/* Generated by Cython 0.25.2 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [
            "/usr/local/include/cpl_conv.h",
            "/usr/local/include/cpl_error.h",
            "/usr/local/include/cpl_string.h",
            "/usr/local/include/cpl_vsi.h",
            "/usr/local/include/gdal.h",
            "/usr/local/include/gdal_alg.h",
            "/usr/local/include/gdalwarper.h",
            "/usr/local/include/ogr_api.h",
            "/usr/local/include/ogr_core.h",
            "/usr/local/include/ogr_geometry.h",
            "/usr/local/include/ogr_spatialref.h",
            "/usr/local/include/ogr_srs_api.h",
            "/var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/numpy/core/include/numpy/arrayobject.h",
            "/var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/numpy/core/include/numpy/ufuncobject.h"
        ],
        "extra_compile_args": [
            "-Wno-unused-parameter",
            "-Wno-unused-function"
        ],
        "include_dirs": [
            "/var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/numpy/core/include",
            "/usr/local/include"
        ],
        "language": "c++",
        "libraries": [
            "gdal"
        ],
        "library_dirs": [
            "/usr/local/lib"
        ]
    },
    "module_name": "rasterio._warp"
}
END: Cython Metadata */

#define PY_SSIZE_T_CLEAN
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03020000)
    #error Cython requires Python 2.6+ or Python 3.2+.
#else
#define CYTHON_ABI "0_25_2"
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x03030000 || (PY_MAJOR_VERSION == 2 && PY_VERSION_HEX >= 0x02070000)
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #include "longintrepr.h"
  #undef SHIFT
  #undef BASE
  #undef MASK
#endif
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_FASTCALL
  #define METH_FASTCALL 0x80
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject **args,
                                              Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None)) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None)) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   PyInt_AsLong
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   PyInt_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? PyMethod_New(func, self) : PyInstanceMethod_New(func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)

#ifndef __cplusplus
  #error "Cython files generated with the C++ option must be compiled with a C++ compiler."
#endif
#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #else
    #define CYTHON_INLINE inline
  #endif
#endif
template<typename T>
void __Pyx_call_destructor(T& x) {
    x.~T();
}
template<typename T>
class __Pyx_FakeReference {
  public:
    __Pyx_FakeReference() : ptr(NULL) { }
    __Pyx_FakeReference(const T& ref) : ptr(const_cast<T*>(&ref)) { }
    T *operator->() { return ptr; }
    T *operator&() { return ptr; }
    operator T&() { return *ptr; }
    template<typename U> bool operator ==(U other) { return *ptr == other; }
    template<typename U> bool operator !=(U other) { return *ptr != other; }
  private:
    T *ptr;
};

#if defined(WIN32) || defined(MS_WINDOWS)
  #define _USE_MATH_DEFINES
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif


#define __PYX_ERR(f_index, lineno, Ln_error) \
{ \
  __pyx_filename = __pyx_f[f_index]; __pyx_lineno = lineno; __pyx_clineno = __LINE__; goto Ln_error; \
}

#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__rasterio___warp
#define __PYX_HAVE_API__rasterio___warp
#include "gdalwarper.h"
#include "ios"
#include "new"
#include "stdexcept"
#include "typeinfo"
#include "ogr_geometry.h"
#include "ogr_spatialref.h"
#include "gdal_alg.h"
#include <string.h>
#include <stdio.h>
#include "cpl_conv.h"
#include "cpl_error.h"
#include "cpl_string.h"
#include "cpl_vsi.h"
#include "ogr_srs_api.h"
#include "gdal.h"
#include "ogr_api.h"
#include "ogr_core.h"
#include <stdlib.h>
#include "numpy/arrayobject.h"
#include "numpy/ufuncobject.h"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#ifdef PYREX_WITHOUT_ASSERTIONS
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT 0
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER) && defined (_M_X64)
    #define __Pyx_sst_abs(value) _abs64(value)
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyObject_AsSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
#if PY_MAJOR_VERSION < 3
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u)
{
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#else
#define __Pyx_Py_UNICODE_strlen Py_UNICODE_strlen
#endif
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
#define __Pyx_PyBool_FromLong(b) ((b) ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False))
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c));
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */

static PyObject *__pyx_m;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;

/* Header.proto */
#if !defined(CYTHON_CCOMPLEX)
  #if defined(__cplusplus)
    #define CYTHON_CCOMPLEX 1
  #elif defined(_Complex_I)
    #define CYTHON_CCOMPLEX 1
  #else
    #define CYTHON_CCOMPLEX 0
  #endif
#endif
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #include <complex>
  #else
    #include <complex.h>
  #endif
#endif
#if CYTHON_CCOMPLEX && !defined(__cplusplus) && defined(__sun__) && defined(__GNUC__)
  #undef _Complex_I
  #define _Complex_I 1.0fj
#endif


static const char *__pyx_f[] = {
  "rasterio/_warp.pyx",
  "__init__.pxd",
  "type.pxd",
  "rasterio/_base.pxd",
  "rasterio/_io.pxd",
  "rasterio/_features.pxd",
};

/* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":725
 * # in Cython to enable them only on the right systems.
 * 
 * ctypedef npy_int8       int8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 */
typedef npy_int8 __pyx_t_5numpy_int8_t;

/* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":726
 * 
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t
 */
typedef npy_int16 __pyx_t_5numpy_int16_t;

/* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":727
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int64      int64_t
 * #ctypedef npy_int96      int96_t
 */
typedef npy_int32 __pyx_t_5numpy_int32_t;

/* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":728
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_int96      int96_t
 * #ctypedef npy_int128     int128_t
 */
typedef npy_int64 __pyx_t_5numpy_int64_t;

/* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":732
 * #ctypedef npy_int128     int128_t
 * 
 * ctypedef npy_uint8      uint8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 */
typedef npy_uint8 __pyx_t_5numpy_uint8_t;

/* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":733
 * 
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t
 */
typedef npy_uint16 __pyx_t_5numpy_uint16_t;

/* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":734
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint64     uint64_t
 * #ctypedef npy_uint96     uint96_t
 */
typedef npy_uint32 __pyx_t_5numpy_uint32_t;

/* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":735
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_uint96     uint96_t
 * #ctypedef npy_uint128    uint128_t
 */
typedef npy_uint64 __pyx_t_5numpy_uint64_t;

/* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":739
 * #ctypedef npy_uint128    uint128_t
 * 
 * ctypedef npy_float32    float32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_float64    float64_t
 * #ctypedef npy_float80    float80_t
 */
typedef npy_float32 __pyx_t_5numpy_float32_t;

/* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":740
 * 
 * ctypedef npy_float32    float32_t
 * ctypedef npy_float64    float64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_float80    float80_t
 * #ctypedef npy_float128   float128_t
 */
typedef npy_float64 __pyx_t_5numpy_float64_t;

/* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":749
 * # The int types are mapped a bit surprising --
 * # numpy.int corresponds to 'l' and numpy.long to 'q'
 * ctypedef npy_long       int_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longlong   long_t
 * ctypedef npy_longlong   longlong_t
 */
typedef npy_long __pyx_t_5numpy_int_t;

/* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":750
 * # numpy.int corresponds to 'l' and numpy.long to 'q'
 * ctypedef npy_long       int_t
 * ctypedef npy_longlong   long_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longlong   longlong_t
 * 
 */
typedef npy_longlong __pyx_t_5numpy_long_t;

/* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":751
 * ctypedef npy_long       int_t
 * ctypedef npy_longlong   long_t
 * ctypedef npy_longlong   longlong_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_ulong      uint_t
 */
typedef npy_longlong __pyx_t_5numpy_longlong_t;

/* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":753
 * ctypedef npy_longlong   longlong_t
 * 
 * ctypedef npy_ulong      uint_t             # <<<<<<<<<<<<<<
 * ctypedef npy_ulonglong  ulong_t
 * ctypedef npy_ulonglong  ulonglong_t
 */
typedef npy_ulong __pyx_t_5numpy_uint_t;

/* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":754
 * 
 * ctypedef npy_ulong      uint_t
 * ctypedef npy_ulonglong  ulong_t             # <<<<<<<<<<<<<<
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 */
typedef npy_ulonglong __pyx_t_5numpy_ulong_t;

/* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":755
 * ctypedef npy_ulong      uint_t
 * ctypedef npy_ulonglong  ulong_t
 * ctypedef npy_ulonglong  ulonglong_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_intp       intp_t
 */
typedef npy_ulonglong __pyx_t_5numpy_ulonglong_t;

/* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":757
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 * ctypedef npy_intp       intp_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uintp      uintp_t
 * 
 */
typedef npy_intp __pyx_t_5numpy_intp_t;

/* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":758
 * 
 * ctypedef npy_intp       intp_t
 * ctypedef npy_uintp      uintp_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_double     float_t
 */
typedef npy_uintp __pyx_t_5numpy_uintp_t;

/* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":760
 * ctypedef npy_uintp      uintp_t
 * 
 * ctypedef npy_double     float_t             # <<<<<<<<<<<<<<
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t
 */
typedef npy_double __pyx_t_5numpy_float_t;

/* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":761
 * 
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longdouble longdouble_t
 * 
 */
typedef npy_double __pyx_t_5numpy_double_t;

/* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":762
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_cfloat      cfloat_t
 */
typedef npy_longdouble __pyx_t_5numpy_longdouble_t;

/* "rasterio/_io.pxd":40
 * 
 * 
 * ctypedef np.uint8_t DTYPE_UBYTE_t             # <<<<<<<<<<<<<<
 * ctypedef np.uint16_t DTYPE_UINT16_t
 * ctypedef np.int16_t DTYPE_INT16_t
 */
typedef __pyx_t_5numpy_uint8_t __pyx_t_8rasterio_3_io_DTYPE_UBYTE_t;

/* "rasterio/_io.pxd":41
 * 
 * ctypedef np.uint8_t DTYPE_UBYTE_t
 * ctypedef np.uint16_t DTYPE_UINT16_t             # <<<<<<<<<<<<<<
 * ctypedef np.int16_t DTYPE_INT16_t
 * ctypedef np.uint32_t DTYPE_UINT32_t
 */
typedef __pyx_t_5numpy_uint16_t __pyx_t_8rasterio_3_io_DTYPE_UINT16_t;

/* "rasterio/_io.pxd":42
 * ctypedef np.uint8_t DTYPE_UBYTE_t
 * ctypedef np.uint16_t DTYPE_UINT16_t
 * ctypedef np.int16_t DTYPE_INT16_t             # <<<<<<<<<<<<<<
 * ctypedef np.uint32_t DTYPE_UINT32_t
 * ctypedef np.int32_t DTYPE_INT32_t
 */
typedef __pyx_t_5numpy_int16_t __pyx_t_8rasterio_3_io_DTYPE_INT16_t;

/* "rasterio/_io.pxd":43
 * ctypedef np.uint16_t DTYPE_UINT16_t
 * ctypedef np.int16_t DTYPE_INT16_t
 * ctypedef np.uint32_t DTYPE_UINT32_t             # <<<<<<<<<<<<<<
 * ctypedef np.int32_t DTYPE_INT32_t
 * ctypedef np.float32_t DTYPE_FLOAT32_t
 */
typedef __pyx_t_5numpy_uint32_t __pyx_t_8rasterio_3_io_DTYPE_UINT32_t;

/* "rasterio/_io.pxd":44
 * ctypedef np.int16_t DTYPE_INT16_t
 * ctypedef np.uint32_t DTYPE_UINT32_t
 * ctypedef np.int32_t DTYPE_INT32_t             # <<<<<<<<<<<<<<
 * ctypedef np.float32_t DTYPE_FLOAT32_t
 * ctypedef np.float64_t DTYPE_FLOAT64_t
 */
typedef __pyx_t_5numpy_int32_t __pyx_t_8rasterio_3_io_DTYPE_INT32_t;

/* "rasterio/_io.pxd":45
 * ctypedef np.uint32_t DTYPE_UINT32_t
 * ctypedef np.int32_t DTYPE_INT32_t
 * ctypedef np.float32_t DTYPE_FLOAT32_t             # <<<<<<<<<<<<<<
 * ctypedef np.float64_t DTYPE_FLOAT64_t
 * 
 */
typedef __pyx_t_5numpy_float32_t __pyx_t_8rasterio_3_io_DTYPE_FLOAT32_t;

/* "rasterio/_io.pxd":46
 * ctypedef np.int32_t DTYPE_INT32_t
 * ctypedef np.float32_t DTYPE_FLOAT32_t
 * ctypedef np.float64_t DTYPE_FLOAT64_t             # <<<<<<<<<<<<<<
 * 
 * 
 */
typedef __pyx_t_5numpy_float64_t __pyx_t_8rasterio_3_io_DTYPE_FLOAT64_t;
/* Declarations.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    typedef ::std::complex< float > __pyx_t_float_complex;
  #else
    typedef float _Complex __pyx_t_float_complex;
  #endif
#else
    typedef struct { float real, imag; } __pyx_t_float_complex;
#endif
static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float, float);

/* Declarations.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    typedef ::std::complex< double > __pyx_t_double_complex;
  #else
    typedef double _Complex __pyx_t_double_complex;
  #endif
#else
    typedef struct { double real, imag; } __pyx_t_double_complex;
#endif
static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double, double);


/*--- Type declarations ---*/
struct __pyx_obj_8rasterio_5_base_DatasetBase;
struct __pyx_obj_8rasterio_3_io_DatasetReaderBase;
struct __pyx_obj_8rasterio_3_io_DatasetWriterBase;
struct __pyx_obj_8rasterio_3_io_BufferedDatasetWriterBase;
struct __pyx_obj_8rasterio_3_io_WarpedVRTReaderBase;
struct __pyx_obj_8rasterio_3_io_InMemoryRaster;
struct __pyx_obj_8rasterio_9_features_GeomBuilder;
struct __pyx_obj_8rasterio_9_features_OGRGeomBuilder;
struct __pyx_obj_8rasterio_9_features_ShapeIterator;
struct __pyx_obj_8rasterio_5_warp_WarpedVRTReaderBase;
struct __pyx_obj_8rasterio_5_warp___pyx_scope_struct___calculate_default_transform;
struct __pyx_obj_8rasterio_5_warp___pyx_scope_struct_1_genexpr;
struct __pyx_obj_8rasterio_5_warp___pyx_scope_struct_2_genexpr;

/* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":764
 * ctypedef npy_longdouble longdouble_t
 * 
 * ctypedef npy_cfloat      cfloat_t             # <<<<<<<<<<<<<<
 * ctypedef npy_cdouble     cdouble_t
 * ctypedef npy_clongdouble clongdouble_t
 */
typedef npy_cfloat __pyx_t_5numpy_cfloat_t;

/* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":765
 * 
 * ctypedef npy_cfloat      cfloat_t
 * ctypedef npy_cdouble     cdouble_t             # <<<<<<<<<<<<<<
 * ctypedef npy_clongdouble clongdouble_t
 * 
 */
typedef npy_cdouble __pyx_t_5numpy_cdouble_t;

/* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":766
 * ctypedef npy_cfloat      cfloat_t
 * ctypedef npy_cdouble     cdouble_t
 * ctypedef npy_clongdouble clongdouble_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_cdouble     complex_t
 */
typedef npy_clongdouble __pyx_t_5numpy_clongdouble_t;

/* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":768
 * ctypedef npy_clongdouble clongdouble_t
 * 
 * ctypedef npy_cdouble     complex_t             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew1(a):
 */
typedef npy_cdouble __pyx_t_5numpy_complex_t;
struct __pyx_opt_args_8rasterio_3_io_io_auto;

/* "rasterio/_io.pxd":51
 * cdef bint in_dtype_range(value, dtype)
 * 
 * cdef int io_auto(image, GDALRasterBandH band, bint write, int resampling=*)             # <<<<<<<<<<<<<<
 */
struct __pyx_opt_args_8rasterio_3_io_io_auto {
  int __pyx_n;
  int resampling;
};

/* "rasterio/_base.pxd":6
 * 
 * 
 * cdef class DatasetBase:             # <<<<<<<<<<<<<<
 * 
 *     cdef GDALDatasetH _hds
 */
struct __pyx_obj_8rasterio_5_base_DatasetBase {
  PyObject_HEAD
  struct __pyx_vtabstruct_8rasterio_5_base_DatasetBase *__pyx_vtab;
  GDALDatasetH _hds;
  PyObject *name;
  PyObject *mode;
  PyObject *options;
  PyObject *width;
  PyObject *height;
  PyObject *shape;
  PyObject *driver;
  PyObject *_count;
  PyObject *_dtypes;
  PyObject *_closed;
  PyObject *_crs;
  PyObject *_crs_wkt;
  PyObject *_transform;
  PyObject *_block_shapes;
  PyObject *_nodatavals;
  PyObject *_units;
  PyObject *_descriptions;
  PyObject *_read;
  PyObject *_gcps;
};


/* "rasterio/_io.pxd":8
 * 
 * 
 * cdef class DatasetReaderBase(DatasetBase):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_8rasterio_3_io_DatasetReaderBase {
  struct __pyx_obj_8rasterio_5_base_DatasetBase __pyx_base;
};


/* "rasterio/_io.pxd":12
 * 
 * 
 * cdef class DatasetWriterBase(DatasetReaderBase):             # <<<<<<<<<<<<<<
 *     cdef readonly object _init_dtype
 *     cdef readonly object _init_nodata
 */
struct __pyx_obj_8rasterio_3_io_DatasetWriterBase {
  struct __pyx_obj_8rasterio_3_io_DatasetReaderBase __pyx_base;
  PyObject *_init_dtype;
  PyObject *_init_nodata;
  PyObject *_options;
};


/* "rasterio/_io.pxd":18
 * 
 * 
 * cdef class BufferedDatasetWriterBase(DatasetWriterBase):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_8rasterio_3_io_BufferedDatasetWriterBase {
  struct __pyx_obj_8rasterio_3_io_DatasetWriterBase __pyx_base;
};


/* "rasterio/_io.pxd":22
 * 
 * 
 * cdef class WarpedVRTReaderBase(DatasetReaderBase):             # <<<<<<<<<<<<<<
 *     cdef GDALDatasetH _hds_source
 *     cdef object dst_crs
 */
struct __pyx_obj_8rasterio_3_io_WarpedVRTReaderBase {
  struct __pyx_obj_8rasterio_3_io_DatasetReaderBase __pyx_base;
  GDALDatasetH _hds_source;
  PyObject *dst_crs;
  PyObject *resampling;
  PyObject *tolerance;
};


/* "rasterio/_io.pxd":29
 * 
 * 
 * cdef class InMemoryRaster:             # <<<<<<<<<<<<<<
 *     cdef GDALDatasetH _hds
 *     cdef double transform[6]
 */
struct __pyx_obj_8rasterio_3_io_InMemoryRaster {
  PyObject_HEAD
  struct __pyx_vtabstruct_8rasterio_3_io_InMemoryRaster *__pyx_vtab;
  GDALDatasetH _hds;
  double transform[6];
  int band_ids[1];
  PyArrayObject *_image;
  PyObject *crs;
};


/* "rasterio/_features.pxd":6
 * 
 * 
 * cdef class GeomBuilder:             # <<<<<<<<<<<<<<
 * 
 *     cdef OGRGeometryH geom
 */
struct __pyx_obj_8rasterio_9_features_GeomBuilder {
  PyObject_HEAD
  struct __pyx_vtabstruct_8rasterio_9_features_GeomBuilder *__pyx_vtab;
  OGRGeometryH geom;
  PyObject *code;
  PyObject *geomtypename;
  PyObject *ndims;
};


/* "rasterio/_features.pxd":24
 * 
 * 
 * cdef class OGRGeomBuilder:             # <<<<<<<<<<<<<<
 * 
 *     cdef OGRGeometryH _createOgrGeometry(self, int geom_type) except NULL
 */
struct __pyx_obj_8rasterio_9_features_OGRGeomBuilder {
  PyObject_HEAD
  struct __pyx_vtabstruct_8rasterio_9_features_OGRGeomBuilder *__pyx_vtab;
};


/* "rasterio/_features.pxd":39
 * 
 * 
 * cdef class ShapeIterator:             # <<<<<<<<<<<<<<
 * 
 *     cdef OGRLayerH layer
 */
struct __pyx_obj_8rasterio_9_features_ShapeIterator {
  PyObject_HEAD
  OGRLayerH layer;
  int fieldtype;
};


/* "rasterio/_warp.pyx":630
 * 
 * 
 * cdef class WarpedVRTReaderBase(DatasetReaderBase):             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, src_dataset, dst_crs=None, resampling=Resampling.nearest,
 */
struct __pyx_obj_8rasterio_5_warp_WarpedVRTReaderBase {
  struct __pyx_obj_8rasterio_3_io_DatasetReaderBase __pyx_base;
};


/* "rasterio/_warp.pyx":557
 * 
 * 
 * def _calculate_default_transform(src_crs, dst_crs, width, height,             # <<<<<<<<<<<<<<
 *                                  left=None, bottom=None, right=None, top=None,
 *                                  gcps=None, **kwargs):
 */
struct __pyx_obj_8rasterio_5_warp___pyx_scope_struct___calculate_default_transform {
  PyObject_HEAD
  PyObject *__pyx_v_bottom;
  PyObject *__pyx_v_left;
  PyObject *__pyx_v_right;
  PyObject *__pyx_v_top;
};


/* "rasterio/_warp.pyx":576
 *     # GDALCreateGenImgProjTransformer().
 * 
 *     if all(x is not None for x in (left, bottom, right, top)):             # <<<<<<<<<<<<<<
 *         transform = from_bounds(left, bottom, right, top, width, height)
 *         transform = transform.to_gdal()
 */
struct __pyx_obj_8rasterio_5_warp___pyx_scope_struct_1_genexpr {
  PyObject_HEAD
  struct __pyx_obj_8rasterio_5_warp___pyx_scope_struct___calculate_default_transform *__pyx_outer_scope;
  PyObject *__pyx_v_x;
};


/* "rasterio/_warp.pyx":579
 *         transform = from_bounds(left, bottom, right, top, width, height)
 *         transform = transform.to_gdal()
 *     elif any(x is not None for x in (left, bottom, right, top)):             # <<<<<<<<<<<<<<
 *         raise ValueError(
 *             "Some, but not all, bounding box parameters were provided.")
 */
struct __pyx_obj_8rasterio_5_warp___pyx_scope_struct_2_genexpr {
  PyObject_HEAD
  struct __pyx_obj_8rasterio_5_warp___pyx_scope_struct___calculate_default_transform *__pyx_outer_scope;
  PyObject *__pyx_v_x;
};



/* "rasterio/_base.pxd":6
 * 
 * 
 * cdef class DatasetBase:             # <<<<<<<<<<<<<<
 * 
 *     cdef GDALDatasetH _hds
 */

struct __pyx_vtabstruct_8rasterio_5_base_DatasetBase {
  GDALDatasetH (*handle)(struct __pyx_obj_8rasterio_5_base_DatasetBase *);
  GDALRasterBandH (*band)(struct __pyx_obj_8rasterio_5_base_DatasetBase *, int);
};
static struct __pyx_vtabstruct_8rasterio_5_base_DatasetBase *__pyx_vtabptr_8rasterio_5_base_DatasetBase;


/* "rasterio/_io.pxd":8
 * 
 * 
 * cdef class DatasetReaderBase(DatasetBase):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

struct __pyx_vtabstruct_8rasterio_3_io_DatasetReaderBase {
  struct __pyx_vtabstruct_8rasterio_5_base_DatasetBase __pyx_base;
};
static struct __pyx_vtabstruct_8rasterio_3_io_DatasetReaderBase *__pyx_vtabptr_8rasterio_3_io_DatasetReaderBase;


/* "rasterio/_io.pxd":12
 * 
 * 
 * cdef class DatasetWriterBase(DatasetReaderBase):             # <<<<<<<<<<<<<<
 *     cdef readonly object _init_dtype
 *     cdef readonly object _init_nodata
 */

struct __pyx_vtabstruct_8rasterio_3_io_DatasetWriterBase {
  struct __pyx_vtabstruct_8rasterio_3_io_DatasetReaderBase __pyx_base;
};
static struct __pyx_vtabstruct_8rasterio_3_io_DatasetWriterBase *__pyx_vtabptr_8rasterio_3_io_DatasetWriterBase;


/* "rasterio/_io.pxd":18
 * 
 * 
 * cdef class BufferedDatasetWriterBase(DatasetWriterBase):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

struct __pyx_vtabstruct_8rasterio_3_io_BufferedDatasetWriterBase {
  struct __pyx_vtabstruct_8rasterio_3_io_DatasetWriterBase __pyx_base;
};
static struct __pyx_vtabstruct_8rasterio_3_io_BufferedDatasetWriterBase *__pyx_vtabptr_8rasterio_3_io_BufferedDatasetWriterBase;


/* "rasterio/_io.pxd":22
 * 
 * 
 * cdef class WarpedVRTReaderBase(DatasetReaderBase):             # <<<<<<<<<<<<<<
 *     cdef GDALDatasetH _hds_source
 *     cdef object dst_crs
 */

struct __pyx_vtabstruct_8rasterio_3_io_WarpedVRTReaderBase {
  struct __pyx_vtabstruct_8rasterio_3_io_DatasetReaderBase __pyx_base;
};
static struct __pyx_vtabstruct_8rasterio_3_io_WarpedVRTReaderBase *__pyx_vtabptr_8rasterio_3_io_WarpedVRTReaderBase;


/* "rasterio/_io.pxd":29
 * 
 * 
 * cdef class InMemoryRaster:             # <<<<<<<<<<<<<<
 *     cdef GDALDatasetH _hds
 *     cdef double transform[6]
 */

struct __pyx_vtabstruct_8rasterio_3_io_InMemoryRaster {
  GDALDatasetH (*handle)(struct __pyx_obj_8rasterio_3_io_InMemoryRaster *);
  GDALRasterBandH (*band)(struct __pyx_obj_8rasterio_3_io_InMemoryRaster *, int);
};
static struct __pyx_vtabstruct_8rasterio_3_io_InMemoryRaster *__pyx_vtabptr_8rasterio_3_io_InMemoryRaster;


/* "rasterio/_features.pxd":6
 * 
 * 
 * cdef class GeomBuilder:             # <<<<<<<<<<<<<<
 * 
 *     cdef OGRGeometryH geom
 */

struct __pyx_vtabstruct_8rasterio_9_features_GeomBuilder {
  PyObject *(*_buildCoords)(struct __pyx_obj_8rasterio_9_features_GeomBuilder *, OGRGeometryH);
  PyObject *(*_buildPoint)(struct __pyx_obj_8rasterio_9_features_GeomBuilder *, int __pyx_skip_dispatch);
  PyObject *(*_buildLineString)(struct __pyx_obj_8rasterio_9_features_GeomBuilder *, int __pyx_skip_dispatch);
  PyObject *(*_buildLinearRing)(struct __pyx_obj_8rasterio_9_features_GeomBuilder *, int __pyx_skip_dispatch);
  PyObject *(*_buildParts)(struct __pyx_obj_8rasterio_9_features_GeomBuilder *, OGRGeometryH);
  PyObject *(*_buildPolygon)(struct __pyx_obj_8rasterio_9_features_GeomBuilder *, int __pyx_skip_dispatch);
  PyObject *(*_buildMultiPoint)(struct __pyx_obj_8rasterio_9_features_GeomBuilder *, int __pyx_skip_dispatch);
  PyObject *(*_buildMultiLineString)(struct __pyx_obj_8rasterio_9_features_GeomBuilder *, int __pyx_skip_dispatch);
  PyObject *(*_buildMultiPolygon)(struct __pyx_obj_8rasterio_9_features_GeomBuilder *, int __pyx_skip_dispatch);
  PyObject *(*build)(struct __pyx_obj_8rasterio_9_features_GeomBuilder *, OGRGeometryH);
};
static struct __pyx_vtabstruct_8rasterio_9_features_GeomBuilder *__pyx_vtabptr_8rasterio_9_features_GeomBuilder;


/* "rasterio/_features.pxd":24
 * 
 * 
 * cdef class OGRGeomBuilder:             # <<<<<<<<<<<<<<
 * 
 *     cdef OGRGeometryH _createOgrGeometry(self, int geom_type) except NULL
 */

struct __pyx_vtabstruct_8rasterio_9_features_OGRGeomBuilder {
  OGRGeometryH (*_createOgrGeometry)(struct __pyx_obj_8rasterio_9_features_OGRGeomBuilder *, int);
  PyObject *(*_addPointToGeometry)(struct __pyx_obj_8rasterio_9_features_OGRGeomBuilder *, OGRGeometryH, PyObject *);
  OGRGeometryH (*_buildPoint)(struct __pyx_obj_8rasterio_9_features_OGRGeomBuilder *, PyObject *);
  OGRGeometryH (*_buildLineString)(struct __pyx_obj_8rasterio_9_features_OGRGeomBuilder *, PyObject *);
  OGRGeometryH (*_buildLinearRing)(struct __pyx_obj_8rasterio_9_features_OGRGeomBuilder *, PyObject *);
  OGRGeometryH (*_buildPolygon)(struct __pyx_obj_8rasterio_9_features_OGRGeomBuilder *, PyObject *);
  OGRGeometryH (*_buildMultiPoint)(struct __pyx_obj_8rasterio_9_features_OGRGeomBuilder *, PyObject *);
  OGRGeometryH (*_buildMultiLineString)(struct __pyx_obj_8rasterio_9_features_OGRGeomBuilder *, PyObject *);
  OGRGeometryH (*_buildMultiPolygon)(struct __pyx_obj_8rasterio_9_features_OGRGeomBuilder *, PyObject *);
  OGRGeometryH (*_buildGeometryCollection)(struct __pyx_obj_8rasterio_9_features_OGRGeomBuilder *, PyObject *);
  OGRGeometryH (*build)(struct __pyx_obj_8rasterio_9_features_OGRGeomBuilder *, PyObject *);
};
static struct __pyx_vtabstruct_8rasterio_9_features_OGRGeomBuilder *__pyx_vtabptr_8rasterio_9_features_OGRGeomBuilder;


/* "rasterio/_warp.pyx":630
 * 
 * 
 * cdef class WarpedVRTReaderBase(DatasetReaderBase):             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, src_dataset, dst_crs=None, resampling=Resampling.nearest,
 */

struct __pyx_vtabstruct_8rasterio_5_warp_WarpedVRTReaderBase {
  struct __pyx_vtabstruct_8rasterio_3_io_DatasetReaderBase __pyx_base;
};
static struct __pyx_vtabstruct_8rasterio_5_warp_WarpedVRTReaderBase *__pyx_vtabptr_8rasterio_5_warp_WarpedVRTReaderBase;

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* ListCompAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_ListComp_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len)) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        Py_SIZE(list) = len+1;
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_ListComp_Append(L,x) PyList_Append(L,x)
#endif

/* GetModuleGlobalName.proto */
static CYTHON_INLINE PyObject *__Pyx_GetModuleGlobalName(PyObject *name);

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, int nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#endif

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = PyThreadState_GET();
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* SwapException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSwap(type, value, tb)  __Pyx__ExceptionSwap(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, long intval, int inplace);
#else
#define __Pyx_PyInt_AddObjC(op1, op2, intval, inplace)\
    (inplace ? PyNumber_InPlaceAdd(op1, op2) : PyNumber_Add(op1, op2))
#endif

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* PyObjectCallNoArg.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);
#else
#define __Pyx_PyObject_CallNoArg(func) __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL)
#endif

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* IterFinish.proto */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* UnpackItemEndCheck.proto */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* py_dict_items.proto */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Items(PyObject* d);

/* UnpackUnboundCMethod.proto */
typedef struct {
    PyObject *type;
    PyObject **method_name;
    PyCFunction func;
    PyObject *method;
    int flag;
} __Pyx_CachedCFunction;

/* CallUnboundCMethod0.proto */
static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CallUnboundCMethod0(cfunc, self)\
    ((likely((cfunc)->func)) ?\
        (likely((cfunc)->flag == METH_NOARGS) ?  (*((cfunc)->func))(self, NULL) :\
         (likely((cfunc)->flag == (METH_VARARGS | METH_KEYWORDS)) ?  ((*(PyCFunctionWithKeywords)(cfunc)->func)(self, __pyx_empty_tuple, NULL)) :\
             ((cfunc)->flag == METH_VARARGS ?  (*((cfunc)->func))(self, __pyx_empty_tuple) :\
              (PY_VERSION_HEX >= 0x030600B1 && (cfunc)->flag == METH_FASTCALL ?  (*(__Pyx_PyCFunctionFast)(cfunc)->func)(self, &PyTuple_GET_ITEM(__pyx_empty_tuple, 0), 0, NULL) :\
                __Pyx__CallUnboundCMethod0(cfunc, self))))) :\
        __Pyx__CallUnboundCMethod0(cfunc, self))
#else
#define __Pyx_CallUnboundCMethod0(cfunc, self)  __Pyx__CallUnboundCMethod0(cfunc, self)
#endif

/* SliceObject.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(
        PyObject* obj, Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** py_start, PyObject** py_stop, PyObject** py_slice,
        int has_cstart, int has_cstop, int wraparound);

/* None.proto */
static CYTHON_INLINE void __Pyx_RaiseClosureNameError(const char *varname);

/* IncludeStringH.proto */
#include <string.h>

/* PyObjectLookupSpecial.proto */
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x02070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_LookupSpecial(PyObject* obj, PyObject* attr_name) {
    PyObject *res;
    PyTypeObject *tp = Py_TYPE(obj);
#if PY_MAJOR_VERSION < 3
    if (unlikely(PyInstance_Check(obj)))
        return __Pyx_PyObject_GetAttrStr(obj, attr_name);
#endif
    res = _PyType_Lookup(tp, attr_name);
    if (likely(res)) {
        descrgetfunc f = Py_TYPE(res)->tp_descr_get;
        if (!f) {
            Py_INCREF(res);
        } else {
            res = f(res, obj, (PyObject *)tp);
        }
    } else {
        PyErr_SetObject(PyExc_AttributeError, attr_name);
    }
    return res;
}
#else
#define __Pyx_PyObject_LookupSpecial(o,n) __Pyx_PyObject_GetAttrStr(o,n)
#endif

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* PySequenceContains.proto */
static CYTHON_INLINE int __Pyx_PySequence_ContainsTF(PyObject* item, PyObject* seq, int eq) {
    int result = PySequence_Contains(seq, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* PyObjectSetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
#define __Pyx_PyObject_DelAttrStr(o,n) __Pyx_PyObject_SetAttrStr(o,n,NULL)
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_setattro))
        return tp->tp_setattro(obj, attr_name, value);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_setattr))
        return tp->tp_setattr(obj, PyString_AS_STRING(attr_name), value);
#endif
    return PyObject_SetAttr(obj, attr_name, value);
}
#else
#define __Pyx_PyObject_DelAttrStr(o,n)   PyObject_DelAttr(o,n)
#define __Pyx_PyObject_SetAttrStr(o,n,v) PyObject_SetAttr(o,n,v)
#endif

/* PyDictContains.proto */
static CYTHON_INLINE int __Pyx_PyDict_ContainsTF(PyObject* item, PyObject* dict, int eq) {
    int result = PyDict_Contains(dict, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* decode_c_string.proto */
static CYTHON_INLINE PyObject* __Pyx_decode_c_string(
         const char* cstring, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors));

/* DictGetItem.proto */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key) {
    PyObject *value;
    value = PyDict_GetItemWithError(d, key);
    if (unlikely(!value)) {
        if (!PyErr_Occurred()) {
            PyObject* args = PyTuple_Pack(1, key);
            if (likely(args))
                PyErr_SetObject(PyExc_KeyError, args);
            Py_XDECREF(args);
        }
        return NULL;
    }
    Py_INCREF(value);
    return value;
}
#else
    #define __Pyx_PyDict_GetItem(d, key) PyObject_GetItem(d, key)
#endif

/* RaiseNoneIterError.proto */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void);

/* CallNextTpDealloc.proto */
static void __Pyx_call_next_tp_dealloc(PyObject* obj, destructor current_tp_dealloc);

/* CallNextTpTraverse.proto */
static int __Pyx_call_next_tp_traverse(PyObject* obj, visitproc v, void *a, traverseproc current_tp_traverse);

/* CallNextTpClear.proto */
static void __Pyx_call_next_tp_clear(PyObject* obj, inquiry current_tp_dealloc);

/* GetVTable.proto */
static void* __Pyx_GetVtable(PyObject *dict);

/* SetVTable.proto */
static int __Pyx_SetVtable(PyObject *dict, void *vtable);

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* CppExceptionConversion.proto */
#ifndef __Pyx_CppExn2PyErr
#include <new>
#include <typeinfo>
#include <stdexcept>
#include <ios>
static void __Pyx_CppExn2PyErr() {
  try {
    if (PyErr_Occurred())
      ; // let the latest Python exn pass through and ignore the current one
    else
      throw;
  } catch (const std::bad_alloc& exn) {
    PyErr_SetString(PyExc_MemoryError, exn.what());
  } catch (const std::bad_cast& exn) {
    PyErr_SetString(PyExc_TypeError, exn.what());
  } catch (const std::bad_typeid& exn) {
    PyErr_SetString(PyExc_TypeError, exn.what());
  } catch (const std::domain_error& exn) {
    PyErr_SetString(PyExc_ValueError, exn.what());
  } catch (const std::invalid_argument& exn) {
    PyErr_SetString(PyExc_ValueError, exn.what());
  } catch (const std::ios_base::failure& exn) {
    PyErr_SetString(PyExc_IOError, exn.what());
  } catch (const std::out_of_range& exn) {
    PyErr_SetString(PyExc_IndexError, exn.what());
  } catch (const std::overflow_error& exn) {
    PyErr_SetString(PyExc_OverflowError, exn.what());
  } catch (const std::range_error& exn) {
    PyErr_SetString(PyExc_ArithmeticError, exn.what());
  } catch (const std::underflow_error& exn) {
    PyErr_SetString(PyExc_ArithmeticError, exn.what());
  } catch (const std::exception& exn) {
    PyErr_SetString(PyExc_RuntimeError, exn.what());
  }
  catch (...)
  {
    PyErr_SetString(PyExc_RuntimeError, "Unknown exception");
  }
}
#endif

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* ForceInitThreads.proto */
#ifndef __PYX_FORCE_INIT_THREADS
  #define __PYX_FORCE_INIT_THREADS 0
#endif

/* RealImag.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #define __Pyx_CREAL(z) ((z).real())
    #define __Pyx_CIMAG(z) ((z).imag())
  #else
    #define __Pyx_CREAL(z) (__real__(z))
    #define __Pyx_CIMAG(z) (__imag__(z))
  #endif
#else
    #define __Pyx_CREAL(z) ((z).real)
    #define __Pyx_CIMAG(z) ((z).imag)
#endif
#if defined(__cplusplus) && CYTHON_CCOMPLEX\
        && (defined(_WIN32) || defined(__clang__) || (defined(__GNUC__) && (__GNUC__ >= 5 || __GNUC__ == 4 && __GNUC_MINOR__ >= 4 )) || __cplusplus >= 201103)
    #define __Pyx_SET_CREAL(z,x) ((z).real(x))
    #define __Pyx_SET_CIMAG(z,y) ((z).imag(y))
#else
    #define __Pyx_SET_CREAL(z,x) __Pyx_CREAL(z) = (x)
    #define __Pyx_SET_CIMAG(z,y) __Pyx_CIMAG(z) = (y)
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX
    #define __Pyx_c_eq_float(a, b)   ((a)==(b))
    #define __Pyx_c_sum_float(a, b)  ((a)+(b))
    #define __Pyx_c_diff_float(a, b) ((a)-(b))
    #define __Pyx_c_prod_float(a, b) ((a)*(b))
    #define __Pyx_c_quot_float(a, b) ((a)/(b))
    #define __Pyx_c_neg_float(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_float(z) ((z)==(float)0)
    #define __Pyx_c_conj_float(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_float(z)     (::std::abs(z))
        #define __Pyx_c_pow_float(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_float(z) ((z)==0)
    #define __Pyx_c_conj_float(z)    (conjf(z))
    #if 1
        #define __Pyx_c_abs_float(z)     (cabsf(z))
        #define __Pyx_c_pow_float(a, b)  (cpowf(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sum_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_diff_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prod_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_neg_float(__pyx_t_float_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_float(__pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conj_float(__pyx_t_float_complex);
    #if 1
        static CYTHON_INLINE float __Pyx_c_abs_float(__pyx_t_float_complex);
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_pow_float(__pyx_t_float_complex, __pyx_t_float_complex);
    #endif
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX
    #define __Pyx_c_eq_double(a, b)   ((a)==(b))
    #define __Pyx_c_sum_double(a, b)  ((a)+(b))
    #define __Pyx_c_diff_double(a, b) ((a)-(b))
    #define __Pyx_c_prod_double(a, b) ((a)*(b))
    #define __Pyx_c_quot_double(a, b) ((a)/(b))
    #define __Pyx_c_neg_double(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_double(z) ((z)==(double)0)
    #define __Pyx_c_conj_double(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_double(z)     (::std::abs(z))
        #define __Pyx_c_pow_double(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_double(z) ((z)==0)
    #define __Pyx_c_conj_double(z)    (conj(z))
    #if 1
        #define __Pyx_c_abs_double(z)     (cabs(z))
        #define __Pyx_c_pow_double(a, b)  (cpow(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg_double(__pyx_t_double_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_double(__pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj_double(__pyx_t_double_complex);
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs_double(__pyx_t_double_complex);
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow_double(__pyx_t_double_complex, __pyx_t_double_complex);
    #endif
#endif

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum__NPY_TYPES(enum NPY_TYPES value);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE GDALDataType __Pyx_PyInt_As_GDALDataType(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE GDALResampleAlg __Pyx_PyInt_As_GDALResampleAlg(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* FetchCommonType.proto */
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type);

/* PyObjectCallMethod1.proto */
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg);

/* CoroutineBase.proto */
typedef PyObject *(*__pyx_coroutine_body_t)(PyObject *, PyObject *);
typedef struct {
    PyObject_HEAD
    __pyx_coroutine_body_t body;
    PyObject *closure;
    PyObject *exc_type;
    PyObject *exc_value;
    PyObject *exc_traceback;
    PyObject *gi_weakreflist;
    PyObject *classobj;
    PyObject *yieldfrom;
    PyObject *gi_name;
    PyObject *gi_qualname;
    PyObject *gi_modulename;
    int resume_label;
    char is_running;
} __pyx_CoroutineObject;
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
    PyTypeObject *type, __pyx_coroutine_body_t body, PyObject *closure,
    PyObject *name, PyObject *qualname, PyObject *module_name);
static int __Pyx_Coroutine_clear(PyObject *self);
#if 1 || PY_VERSION_HEX < 0x030300B0
static int __Pyx_PyGen_FetchStopIterationValue(PyObject **pvalue);
#else
#define __Pyx_PyGen_FetchStopIterationValue(pvalue) PyGen_FetchStopIterationValue(pvalue)
#endif

/* PatchModuleWithCoroutine.proto */
static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code);

/* PatchGeneratorABC.proto */
static int __Pyx_patch_abc(void);

/* Generator.proto */
#define __Pyx_Generator_USED
static PyTypeObject *__pyx_GeneratorType = 0;
#define __Pyx_Generator_CheckExact(obj) (Py_TYPE(obj) == __pyx_GeneratorType)
#define __Pyx_Generator_New(body, closure, name, qualname, module_name)\
    __Pyx__Coroutine_New(__pyx_GeneratorType, body, closure, name, qualname, module_name)
static PyObject *__Pyx_Generator_Next(PyObject *self);
static int __pyx_Generator_init(void);

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* PyIdentifierFromString.proto */
#if !defined(__Pyx_PyIdentifier_FromString)
#if PY_MAJOR_VERSION < 3
  #define __Pyx_PyIdentifier_FromString(s) PyString_FromString(s)
#else
  #define __Pyx_PyIdentifier_FromString(s) PyUnicode_FromString(s)
#endif
#endif

/* ModuleImport.proto */
static PyObject *__Pyx_ImportModule(const char *name);

/* TypeImport.proto */
static PyTypeObject *__Pyx_ImportType(const char *module_name, const char *class_name, size_t size, int strict);

/* FunctionImport.proto */
static int __Pyx_ImportFunction(PyObject *module, const char *funcname, void (**f)(void), const char *sig);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);


/* Module declarations from 'libc.string' */

/* Module declarations from 'libc.stdio' */

/* Module declarations from 'cpython.buffer' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.type' */
static PyTypeObject *__pyx_ptype_7cpython_4type_type = 0;

/* Module declarations from 'cpython' */

/* Module declarations from 'cpython.object' */

/* Module declarations from 'cpython.ref' */

/* Module declarations from 'libc.stdlib' */

/* Module declarations from 'numpy' */

/* Module declarations from 'numpy' */
static PyTypeObject *__pyx_ptype_5numpy_dtype = 0;
static PyTypeObject *__pyx_ptype_5numpy_flatiter = 0;
static PyTypeObject *__pyx_ptype_5numpy_broadcast = 0;
static PyTypeObject *__pyx_ptype_5numpy_ndarray = 0;
static PyTypeObject *__pyx_ptype_5numpy_ufunc = 0;
static CYTHON_INLINE char *__pyx_f_5numpy__util_dtypestring(PyArray_Descr *, char *, char *, int *); /*proto*/

/* Module declarations from 'rasterio._base' */
static PyTypeObject *__pyx_ptype_8rasterio_5_base_DatasetBase = 0;
static char const *(*__pyx_f_8rasterio_5_base_get_driver_name)(GDALDriverH); /*proto*/
static OGRSpatialReferenceH (*__pyx_f_8rasterio_5_base__osr_from_crs)(PyObject *); /*proto*/

/* Module declarations from 'rasterio._err' */
static int (*__pyx_f_8rasterio_4_err_exc_wrap_int)(int); /*proto*/
static void *(*__pyx_f_8rasterio_4_err_exc_wrap_pointer)(void *); /*proto*/

/* Module declarations from 'rasterio._io' */
static PyTypeObject *__pyx_ptype_8rasterio_3_io_DatasetReaderBase = 0;
static PyTypeObject *__pyx_ptype_8rasterio_3_io_DatasetWriterBase = 0;
static PyTypeObject *__pyx_ptype_8rasterio_3_io_BufferedDatasetWriterBase = 0;
static PyTypeObject *__pyx_ptype_8rasterio_3_io_WarpedVRTReaderBase = 0;
static PyTypeObject *__pyx_ptype_8rasterio_3_io_InMemoryRaster = 0;
static int (*__pyx_f_8rasterio_3_io_in_dtype_range)(PyObject *, PyObject *); /*proto*/
static int (*__pyx_f_8rasterio_3_io_io_auto)(PyObject *, GDALRasterBandH, int, struct __pyx_opt_args_8rasterio_3_io_io_auto *__pyx_optional_args); /*proto*/

/* Module declarations from 'rasterio._features' */
static PyTypeObject *__pyx_ptype_8rasterio_9_features_GeomBuilder = 0;
static PyTypeObject *__pyx_ptype_8rasterio_9_features_OGRGeomBuilder = 0;
static PyTypeObject *__pyx_ptype_8rasterio_9_features_ShapeIterator = 0;

/* Module declarations from 'rasterio._warp' */
static PyTypeObject *__pyx_ptype_8rasterio_5_warp_WarpedVRTReaderBase = 0;
static PyTypeObject *__pyx_ptype_8rasterio_5_warp___pyx_scope_struct___calculate_default_transform = 0;
static PyTypeObject *__pyx_ptype_8rasterio_5_warp___pyx_scope_struct_1_genexpr = 0;
static PyTypeObject *__pyx_ptype_8rasterio_5_warp___pyx_scope_struct_2_genexpr = 0;
static GDALWarpOptions *__pyx_f_8rasterio_5_warp_create_warp_options(GDALResampleAlg, PyObject *, PyObject *, int, char const **); /*proto*/
#define __Pyx_MODULE_NAME "rasterio._warp"
int __pyx_module_is_main_rasterio___warp = 0;

/* Implementation of 'rasterio._warp' */
static PyObject *__pyx_builtin_round;
static PyObject *__pyx_builtin_ValueError;
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_enumerate;
static PyObject *__pyx_builtin_super;
static PyObject *__pyx_builtin_RuntimeError;
static PyObject *__pyx_builtin_ImportError;
static const char __pyx_k_i[] = "i";
static const char __pyx_k_x[] = "x";
static const char __pyx_k_y[] = "y";
static const char __pyx_k_z[] = "z";
static const char __pyx_k_GA[] = "GA";
static const char __pyx_k_ds[] = "ds";
static const char __pyx_k_gt[] = "gt";
static const char __pyx_k_np[] = "np";
static const char __pyx_k__11[] = "";
static const char __pyx_k__53[] = "_";
static const char __pyx_k_col[] = "col";
static const char __pyx_k_crs[] = "crs";
static const char __pyx_k_dst[] = "dst";
static const char __pyx_k_err[] = "err";
static const char __pyx_k_img[] = "img";
static const char __pyx_k_key[] = "key";
static const char __pyx_k_log[] = "log";
static const char __pyx_k_obj[] = "obj";
static const char __pyx_k_osr[] = "osr";
static const char __pyx_k_rdr[] = "rdr";
static const char __pyx_k_row[] = "row";
static const char __pyx_k_src[] = "src";
static const char __pyx_k_top[] = "top";
static const char __pyx_k_udr[] = "udr";
static const char __pyx_k_val[] = "val";
static const char __pyx_k_wkt[] = "wkt";
static const char __pyx_k_args[] = "args";
static const char __pyx_k_cols[] = "cols";
static const char __pyx_k_exit[] = "__exit__";
static const char __pyx_k_gcps[] = "gcps";
static const char __pyx_k_geom[] = "geom";
static const char __pyx_k_info[] = "info";
static const char __pyx_k_init[] = "__init__";
static const char __pyx_k_left[] = "left";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_meta[] = "meta";
static const char __pyx_k_name[] = "name";
static const char __pyx_k_none[] = "none";
static const char __pyx_k_part[] = "part";
static const char __pyx_k_rows[] = "rows";
static const char __pyx_k_send[] = "send";
static const char __pyx_k_temp[] = "temp";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_uuid[] = "uuid";
static const char __pyx_k_valb[] = "valb";
static const char __pyx_k_close[] = "close";
static const char __pyx_k_debug[] = "debug";
static const char __pyx_k_dtype[] = "dtype";
static const char __pyx_k_empty[] = "empty";
static const char __pyx_k_enter[] = "__enter__";
static const char __pyx_k_ident[] = "ident";
static const char __pyx_k_items[] = "items";
static const char __pyx_k_numpy[] = "numpy";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_right[] = "right";
static const char __pyx_k_round[] = "round";
static const char __pyx_k_shape[] = "shape";
static const char __pyx_k_super[] = "super";
static const char __pyx_k_throw[] = "throw";
static const char __pyx_k_upper[] = "upper";
static const char __pyx_k_utf_8[] = "utf-8";
static const char __pyx_k_uuid4[] = "uuid4";
static const char __pyx_k_width[] = "width";
static const char __pyx_k_Affine[] = "Affine";
static const char __pyx_k_bottom[] = "bottom";
static const char __pyx_k_driver[] = "driver";
static const char __pyx_k_dstwkt[] = "dstwkt";
static const char __pyx_k_dtypes[] = "dtypes";
static const char __pyx_k_encode[] = "encode";
static const char __pyx_k_errmsg[] = "errmsg";
static const char __pyx_k_extent[] = "extent";
static const char __pyx_k_format[] = "format";
static const char __pyx_k_height[] = "height";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_kwargs[] = "kwargs";
static const char __pyx_k_name_2[] = "__name__";
static const char __pyx_k_nlines[] = "nlines";
static const char __pyx_k_nodata[] = "nodata";
static const char __pyx_k_result[] = "result";
static const char __pyx_k_retval[] = "retval";
static const char __pyx_k_source[] = "source";
static const char __pyx_k_srcwkt[] = "srcwkt";
static const char __pyx_k_NO_DATA[] = "NO_DATA";
static const char __pyx_k_dst_crs[] = "dst_crs";
static const char __pyx_k_factory[] = "factory";
static const char __pyx_k_gcplist[] = "gcplist";
static const char __pyx_k_genexpr[] = "genexpr";
static const char __pyx_k_logging[] = "logging";
static const char __pyx_k_nearest[] = "nearest";
static const char __pyx_k_npixels[] = "npixels";
static const char __pyx_k_oWarper[] = "oWarper";
static const char __pyx_k_options[] = "options";
static const char __pyx_k_reshape[] = "reshape";
static const char __pyx_k_src_crs[] = "src_crs";
static const char __pyx_k_to_gdal[] = "to_gdal";
static const char __pyx_k_CRSError[] = "CRSError";
static const char __pyx_k_dst_bidx[] = "dst_bidx";
static const char __pyx_k_dst_geom[] = "dst_geom";
static const char __pyx_k_rasterio[] = "rasterio";
static const char __pyx_k_read_crs[] = "read_crs";
static const char __pyx_k_src_bidx[] = "src_bidx";
static const char __pyx_k_src_geom[] = "src_geom";
static const char __pyx_k_vsi_path[] = "vsi_path";
static const char __pyx_k_GDALError[] = "GDALError";
static const char __pyx_k_WarpedVRT[] = "WarpedVRT({})";
static const char __pyx_k_dst_width[] = "dst_width";
static const char __pyx_k_dtype_rev[] = "dtype_rev";
static const char __pyx_k_enumerate[] = "enumerate";
static const char __pyx_k_from_gdal[] = "from_gdal";
static const char __pyx_k_getLogger[] = "getLogger";
static const char __pyx_k_init_dest[] = "init_dest";
static const char __pyx_k_precision[] = "precision";
static const char __pyx_k_reproject[] = "_reproject";
static const char __pyx_k_src_count[] = "src_count";
static const char __pyx_k_tolerance[] = "tolerance";
static const char __pyx_k_transform[] = "transform";
static const char __pyx_k_Resampling[] = "Resampling";
static const char __pyx_k_ValueError[] = "ValueError";
static const char __pyx_k_dst_affine[] = "dst_affine";
static const char __pyx_k_dst_height[] = "dst_height";
static const char __pyx_k_dst_nodata[] = "dst_nodata";
static const char __pyx_k_fill_value[] = "fill_value";
static const char __pyx_k_is_ndarray[] = "is_ndarray";
static const char __pyx_k_parse_path[] = "parse_path";
static const char __pyx_k_psWOptions[] = "psWOptions";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_resampling[] = "resampling";
static const char __pyx_k_src_nodata[] = "src_nodata";
static const char __pyx_k_ImportError[] = "ImportError";
static const char __pyx_k_coordinates[] = "coordinates";
static const char __pyx_k_datasetname[] = "datasetname";
static const char __pyx_k_destination[] = "destination";
static const char __pyx_k_dst_dataset[] = "dst_dataset";
static const char __pyx_k_from_bounds[] = "from_bounds";
static const char __pyx_k_num_threads[] = "num_threads";
static const char __pyx_k_src_dataset[] = "src_dataset";
static const char __pyx_k_warp_extras[] = "warp_extras";
static const char __pyx_k_RuntimeError[] = "RuntimeError";
static const char __pyx_k_geotransform[] = "geotransform";
static const char __pyx_k_rasterio_vfs[] = "rasterio.vfs";
static const char __pyx_k_Warp_extras_r[] = "Warp_extras: %r";
static const char __pyx_k_dst_transform[] = "dst_transform";
static const char __pyx_k_hTransformArg[] = "hTransformArg";
static const char __pyx_k_pszWarpThread[] = "pszWarpThread";
static const char __pyx_k_rasterio__err[] = "rasterio._err";
static const char __pyx_k_src_transform[] = "src_transform";
static const char __pyx_k_Invalid_source[] = "Invalid source";
static const char __pyx_k_imgProjOptions[] = "imgProjOptions";
static const char __pyx_k_pfnTransformer[] = "pfnTransformer";
static const char __pyx_k_rasterio__warp[] = "rasterio._warp";
static const char __pyx_k_rasterio_enums[] = "rasterio.enums";
static const char __pyx_k_read_transform[] = "read_transform";
static const char __pyx_k_transform_geom[] = "_transform_geom";
static const char __pyx_k_RasterioIOError[] = "RasterioIOError";
static const char __pyx_k_rasterio_errors[] = "rasterio.errors";
static const char __pyx_k_recursive_round[] = "recursive_round";
static const char __pyx_k_init_dest_nodata[] = "init_dest_nodata";
static const char __pyx_k_rasterio_control[] = "rasterio.control";
static const char __pyx_k_tastes_like_gdal[] = "tastes_like_gdal";
static const char __pyx_k_GroundControlPoint[] = "GroundControlPoint";
static const char __pyx_k_rasterio_transform[] = "rasterio.transform";
static const char __pyx_k_Exported_CRS_to_WKT[] = "Exported CRS to WKT.";
static const char __pyx_k_Invalid_destination[] = "Invalid destination";
static const char __pyx_k_Reprojection_failed[] = "Reprojection failed";
static const char __pyx_k_antimeridian_offset[] = "antimeridian_offset";
static const char __pyx_k_CPLE_AppDefinedError[] = "CPLE_AppDefinedError";
static const char __pyx_k_CPLE_IllegalArgError[] = "CPLE_IllegalArgError";
static const char __pyx_k_CPLE_OpenFailedError[] = "CPLE_OpenFailedError";
static const char __pyx_k_Dataset_r_is_started[] = "Dataset %r is started.";
static const char __pyx_k_antimeridian_cutting[] = "antimeridian_cutting";
static const char __pyx_k_CPLE_NotSupportedError[] = "CPLE_NotSupportedError";
static const char __pyx_k_DriverRegistrationError[] = "DriverRegistrationError";
static const char __pyx_k_Set_transformer_options[] = "Set transformer options";
static const char __pyx_k_Created_temp_source_dataset[] = "Created temp source dataset";
static const char __pyx_k_calculate_default_transform[] = "_calculate_default_transform";
static const char __pyx_k_ndarray_is_not_C_contiguous[] = "ndarray is not C contiguous";
static const char __pyx_k_Setting_NUM_THREADS_option_d[] = "Setting NUM_THREADS option: %d";
static const char __pyx_k_Chunk_and_warp_window_d_d_d_d[] = "Chunk and warp window: %d, %d, %d, %d.";
static const char __pyx_k_Destination_s_shape_is_invalid[] = "Destination's shape is invalid";
static const char __pyx_k_Created_transformer_and_options[] = "Created transformer and options.";
static const char __pyx_k_Failed_to_set_transform_on_temp[] = "Failed to set transform on temp destination dataset.";
static const char __pyx_k_MEM_driver_not_found_Check_that[] = "'MEM' driver not found. Check that this call is contained in a `with rasterio.Env()` or `with rasterio.open()` block.";
static const char __pyx_k_Users_sean_code_frs_wheel_build[] = "/Users/sean/code/frs-wheel-builds/src/rasterio/rasterio/_warp.pyx";
static const char __pyx_k_calculate_default_transform_loc[] = "_calculate_default_transform.<locals>.genexpr";
static const char __pyx_k_numpy_core_multiarray_failed_to[] = "numpy.core.multiarray failed to import";
static const char __pyx_k_unknown_dtype_code_in_numpy_pxd[] = "unknown dtype code in numpy.pxd (%d)";
static const char __pyx_k_CRS_for_temp_destination_dataset[] = "CRS for temp destination dataset: %s.";
static const char __pyx_k_Created_temp_destination_dataset[] = "Created temp destination dataset.";
static const char __pyx_k_Created_transformer_and_warp_out[] = "Created transformer and warp output.";
static const char __pyx_k_Encountered_points_outside_of_va[] = "Encountered points outside of valid dst crs region";
static const char __pyx_k_Failed_to_set_projection_on_temp[] = "Failed to set projection on temp destination dataset.";
static const char __pyx_k_Format_string_allocated_too_shor[] = "Format string allocated too short, see comment in numpy.pxd";
static const char __pyx_k_Non_native_byte_order_not_suppor[] = "Non-native byte order not supported";
static const char __pyx_k_Raster_and_vector_warping_and_re[] = "Raster and vector warping and reprojection.";
static const char __pyx_k_Set_CRS_on_temp_source_dataset_s[] = "Set CRS on temp source dataset: %s";
static const char __pyx_k_Set_transform_on_temp_source_dat[] = "Set transform on temp source dataset: %d";
static const char __pyx_k_Some_but_not_all_bounding_box_pa[] = "Some, but not all, bounding box parameters were provided.";
static const char __pyx_k_Wrote_array_to_temp_output_datas[] = "Wrote array to temp output dataset";
static const char __pyx_k_Wrote_array_to_temp_source_datas[] = "Wrote array to temp source dataset";
static const char __pyx_k_dst_nodata_must_be_in_valid_rang[] = "dst_nodata must be in valid range for destination dtype";
static const char __pyx_k_ndarray_is_not_Fortran_contiguou[] = "ndarray is not Fortran contiguous";
static const char __pyx_k_numpy_core_umath_failed_to_impor[] = "numpy.core.umath failed to import";
static const char __pyx_k_src_nodata_must_be_in_valid_rang[] = "src_nodata must be in valid range for source dtype";
static const char __pyx_k_src_nodata_must_be_provided_beca[] = "src_nodata must be provided because dst_nodata is not None";
static const char __pyx_k_Format_string_allocated_too_shor_2[] = "Format string allocated too short.";
static PyObject *__pyx_n_s_Affine;
static PyObject *__pyx_n_s_CPLE_AppDefinedError;
static PyObject *__pyx_n_s_CPLE_IllegalArgError;
static PyObject *__pyx_n_s_CPLE_NotSupportedError;
static PyObject *__pyx_n_s_CPLE_OpenFailedError;
static PyObject *__pyx_n_s_CRSError;
static PyObject *__pyx_kp_s_CRS_for_temp_destination_dataset;
static PyObject *__pyx_kp_s_Chunk_and_warp_window_d_d_d_d;
static PyObject *__pyx_kp_s_Created_temp_destination_dataset;
static PyObject *__pyx_kp_s_Created_temp_source_dataset;
static PyObject *__pyx_kp_s_Created_transformer_and_options;
static PyObject *__pyx_kp_s_Created_transformer_and_warp_out;
static PyObject *__pyx_kp_s_Dataset_r_is_started;
static PyObject *__pyx_kp_s_Destination_s_shape_is_invalid;
static PyObject *__pyx_n_s_DriverRegistrationError;
static PyObject *__pyx_kp_s_Encountered_points_outside_of_va;
static PyObject *__pyx_kp_s_Exported_CRS_to_WKT;
static PyObject *__pyx_kp_s_Failed_to_set_projection_on_temp;
static PyObject *__pyx_kp_s_Failed_to_set_transform_on_temp;
static PyObject *__pyx_kp_u_Format_string_allocated_too_shor;
static PyObject *__pyx_kp_u_Format_string_allocated_too_shor_2;
static PyObject *__pyx_n_s_GA;
static PyObject *__pyx_n_s_GDALError;
static PyObject *__pyx_n_s_GroundControlPoint;
static PyObject *__pyx_n_s_ImportError;
static PyObject *__pyx_kp_s_Invalid_destination;
static PyObject *__pyx_kp_s_Invalid_source;
static PyObject *__pyx_kp_s_MEM_driver_not_found_Check_that;
static PyObject *__pyx_n_s_NO_DATA;
static PyObject *__pyx_kp_u_Non_native_byte_order_not_suppor;
static PyObject *__pyx_n_s_RasterioIOError;
static PyObject *__pyx_kp_s_Reprojection_failed;
static PyObject *__pyx_n_s_Resampling;
static PyObject *__pyx_n_s_RuntimeError;
static PyObject *__pyx_kp_s_Set_CRS_on_temp_source_dataset_s;
static PyObject *__pyx_kp_s_Set_transform_on_temp_source_dat;
static PyObject *__pyx_kp_s_Set_transformer_options;
static PyObject *__pyx_kp_s_Setting_NUM_THREADS_option_d;
static PyObject *__pyx_kp_s_Some_but_not_all_bounding_box_pa;
static PyObject *__pyx_kp_s_Users_sean_code_frs_wheel_build;
static PyObject *__pyx_n_s_ValueError;
static PyObject *__pyx_kp_s_Warp_extras_r;
static PyObject *__pyx_kp_s_WarpedVRT;
static PyObject *__pyx_kp_s_Wrote_array_to_temp_output_datas;
static PyObject *__pyx_kp_s_Wrote_array_to_temp_source_datas;
static PyObject *__pyx_kp_s__11;
static PyObject *__pyx_n_s__53;
static PyObject *__pyx_n_s_antimeridian_cutting;
static PyObject *__pyx_n_s_antimeridian_offset;
static PyObject *__pyx_n_s_args;
static PyObject *__pyx_n_s_bottom;
static PyObject *__pyx_n_s_calculate_default_transform;
static PyObject *__pyx_n_s_calculate_default_transform_loc;
static PyObject *__pyx_n_s_close;
static PyObject *__pyx_n_s_col;
static PyObject *__pyx_n_s_cols;
static PyObject *__pyx_n_s_coordinates;
static PyObject *__pyx_n_s_crs;
static PyObject *__pyx_n_s_datasetname;
static PyObject *__pyx_n_s_debug;
static PyObject *__pyx_n_s_destination;
static PyObject *__pyx_n_s_driver;
static PyObject *__pyx_n_s_ds;
static PyObject *__pyx_n_s_dst;
static PyObject *__pyx_n_s_dst_affine;
static PyObject *__pyx_n_s_dst_bidx;
static PyObject *__pyx_n_s_dst_crs;
static PyObject *__pyx_n_s_dst_dataset;
static PyObject *__pyx_n_s_dst_geom;
static PyObject *__pyx_n_s_dst_height;
static PyObject *__pyx_n_s_dst_nodata;
static PyObject *__pyx_kp_s_dst_nodata_must_be_in_valid_rang;
static PyObject *__pyx_n_s_dst_transform;
static PyObject *__pyx_n_s_dst_width;
static PyObject *__pyx_n_s_dstwkt;
static PyObject *__pyx_n_s_dtype;
static PyObject *__pyx_n_s_dtype_rev;
static PyObject *__pyx_n_s_dtypes;
static PyObject *__pyx_n_s_empty;
static PyObject *__pyx_n_s_encode;
static PyObject *__pyx_n_s_enter;
static PyObject *__pyx_n_s_enumerate;
static PyObject *__pyx_n_s_err;
static PyObject *__pyx_n_s_errmsg;
static PyObject *__pyx_n_s_exit;
static PyObject *__pyx_n_s_extent;
static PyObject *__pyx_n_s_factory;
static PyObject *__pyx_n_s_fill_value;
static PyObject *__pyx_n_s_format;
static PyObject *__pyx_n_s_from_bounds;
static PyObject *__pyx_n_s_from_gdal;
static PyObject *__pyx_n_s_gcplist;
static PyObject *__pyx_n_s_gcps;
static PyObject *__pyx_n_s_genexpr;
static PyObject *__pyx_n_s_geom;
static PyObject *__pyx_n_s_geotransform;
static PyObject *__pyx_n_s_getLogger;
static PyObject *__pyx_n_s_gt;
static PyObject *__pyx_n_s_hTransformArg;
static PyObject *__pyx_n_s_height;
static PyObject *__pyx_n_s_i;
static PyObject *__pyx_n_s_ident;
static PyObject *__pyx_n_s_img;
static PyObject *__pyx_n_s_imgProjOptions;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_n_s_info;
static PyObject *__pyx_n_s_init;
static PyObject *__pyx_n_s_init_dest;
static PyObject *__pyx_n_s_init_dest_nodata;
static PyObject *__pyx_n_s_is_ndarray;
static PyObject *__pyx_n_s_items;
static PyObject *__pyx_n_s_key;
static PyObject *__pyx_n_s_kwargs;
static PyObject *__pyx_n_s_left;
static PyObject *__pyx_n_s_log;
static PyObject *__pyx_n_s_logging;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_meta;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_n_s_name_2;
static PyObject *__pyx_kp_u_ndarray_is_not_C_contiguous;
static PyObject *__pyx_kp_u_ndarray_is_not_Fortran_contiguou;
static PyObject *__pyx_n_s_nearest;
static PyObject *__pyx_n_s_nlines;
static PyObject *__pyx_n_s_nodata;
static PyObject *__pyx_n_s_none;
static PyObject *__pyx_n_s_np;
static PyObject *__pyx_n_s_npixels;
static PyObject *__pyx_n_s_num_threads;
static PyObject *__pyx_n_s_numpy;
static PyObject *__pyx_kp_s_numpy_core_multiarray_failed_to;
static PyObject *__pyx_kp_s_numpy_core_umath_failed_to_impor;
static PyObject *__pyx_n_s_oWarper;
static PyObject *__pyx_n_s_obj;
static PyObject *__pyx_n_s_options;
static PyObject *__pyx_n_s_osr;
static PyObject *__pyx_n_s_parse_path;
static PyObject *__pyx_n_s_part;
static PyObject *__pyx_n_s_pfnTransformer;
static PyObject *__pyx_n_s_precision;
static PyObject *__pyx_n_s_psWOptions;
static PyObject *__pyx_n_s_pszWarpThread;
static PyObject *__pyx_n_s_pyx_vtable;
static PyObject *__pyx_n_s_range;
static PyObject *__pyx_n_s_rasterio;
static PyObject *__pyx_n_s_rasterio__err;
static PyObject *__pyx_n_s_rasterio__warp;
static PyObject *__pyx_n_s_rasterio_control;
static PyObject *__pyx_n_s_rasterio_enums;
static PyObject *__pyx_n_s_rasterio_errors;
static PyObject *__pyx_n_s_rasterio_transform;
static PyObject *__pyx_n_s_rasterio_vfs;
static PyObject *__pyx_n_s_rdr;
static PyObject *__pyx_n_s_read_crs;
static PyObject *__pyx_n_s_read_transform;
static PyObject *__pyx_n_s_recursive_round;
static PyObject *__pyx_n_s_reproject;
static PyObject *__pyx_n_s_resampling;
static PyObject *__pyx_n_s_reshape;
static PyObject *__pyx_n_s_result;
static PyObject *__pyx_n_s_retval;
static PyObject *__pyx_n_s_right;
static PyObject *__pyx_n_s_round;
static PyObject *__pyx_n_s_row;
static PyObject *__pyx_n_s_rows;
static PyObject *__pyx_n_s_send;
static PyObject *__pyx_n_s_shape;
static PyObject *__pyx_n_s_source;
static PyObject *__pyx_n_s_src;
static PyObject *__pyx_n_s_src_bidx;
static PyObject *__pyx_n_s_src_count;
static PyObject *__pyx_n_s_src_crs;
static PyObject *__pyx_n_s_src_dataset;
static PyObject *__pyx_n_s_src_geom;
static PyObject *__pyx_n_s_src_nodata;
static PyObject *__pyx_kp_s_src_nodata_must_be_in_valid_rang;
static PyObject *__pyx_kp_s_src_nodata_must_be_provided_beca;
static PyObject *__pyx_n_s_src_transform;
static PyObject *__pyx_n_s_srcwkt;
static PyObject *__pyx_n_s_super;
static PyObject *__pyx_n_s_tastes_like_gdal;
static PyObject *__pyx_n_s_temp;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_n_s_throw;
static PyObject *__pyx_n_s_to_gdal;
static PyObject *__pyx_n_s_tolerance;
static PyObject *__pyx_n_s_top;
static PyObject *__pyx_n_s_transform;
static PyObject *__pyx_n_s_transform_geom;
static PyObject *__pyx_n_s_udr;
static PyObject *__pyx_kp_u_unknown_dtype_code_in_numpy_pxd;
static PyObject *__pyx_n_s_upper;
static PyObject *__pyx_kp_s_utf_8;
static PyObject *__pyx_n_s_uuid;
static PyObject *__pyx_n_s_uuid4;
static PyObject *__pyx_n_s_val;
static PyObject *__pyx_n_s_valb;
static PyObject *__pyx_n_s_vsi_path;
static PyObject *__pyx_n_s_warp_extras;
static PyObject *__pyx_n_s_width;
static PyObject *__pyx_n_s_wkt;
static PyObject *__pyx_n_s_x;
static PyObject *__pyx_n_s_y;
static PyObject *__pyx_n_s_z;
static PyObject *__pyx_pf_8rasterio_5_warp_recursive_round(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_val, PyObject *__pyx_v_precision); /* proto */
static PyObject *__pyx_pf_8rasterio_5_warp_2_transform_geom(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_src_crs, PyObject *__pyx_v_dst_crs, PyObject *__pyx_v_geom, PyObject *__pyx_v_antimeridian_cutting, PyObject *__pyx_v_antimeridian_offset, PyObject *__pyx_v_precision); /* proto */
static PyObject *__pyx_pf_8rasterio_5_warp_4_reproject(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_source, PyObject *__pyx_v_destination, PyObject *__pyx_v_src_transform, PyObject *__pyx_v_gcps, PyObject *__pyx_v_src_crs, PyObject *__pyx_v_src_nodata, PyObject *__pyx_v_dst_transform, PyObject *__pyx_v_dst_crs, PyObject *__pyx_v_dst_nodata, PyObject *__pyx_v_resampling, PyObject *__pyx_v_init_dest_nodata, PyObject *__pyx_v_num_threads, PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_8rasterio_5_warp_28_calculate_default_transform_genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_8rasterio_5_warp_28_calculate_default_transform_3genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_8rasterio_5_warp_6_calculate_default_transform(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_src_crs, PyObject *__pyx_v_dst_crs, PyObject *__pyx_v_width, PyObject *__pyx_v_height, PyObject *__pyx_v_left, PyObject *__pyx_v_bottom, PyObject *__pyx_v_right, PyObject *__pyx_v_top, PyObject *__pyx_v_gcps, CYTHON_UNUSED PyObject *__pyx_v_kwargs); /* proto */
static int __pyx_pf_8rasterio_5_warp_19WarpedVRTReaderBase___init__(struct __pyx_obj_8rasterio_5_warp_WarpedVRTReaderBase *__pyx_v_self, PyObject *__pyx_v_src_dataset, PyObject *__pyx_v_dst_crs, PyObject *__pyx_v_resampling, PyObject *__pyx_v_tolerance, PyObject *__pyx_v_src_nodata, PyObject *__pyx_v_dst_nodata, PyObject *__pyx_v_dst_width, PyObject *__pyx_v_dst_height, PyObject *__pyx_v_dst_transform, PyObject *__pyx_v_init_dest_nodata, PyObject *__pyx_v_warp_extras); /* proto */
static PyObject *__pyx_pf_8rasterio_5_warp_19WarpedVRTReaderBase_2start(struct __pyx_obj_8rasterio_5_warp_WarpedVRTReaderBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8rasterio_5_warp_19WarpedVRTReaderBase_4stop(struct __pyx_obj_8rasterio_5_warp_WarpedVRTReaderBase *__pyx_v_self); /* proto */
static int __pyx_pf_5numpy_7ndarray___getbuffer__(PyArrayObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /* proto */
static void __pyx_pf_5numpy_7ndarray_2__releasebuffer__(PyArrayObject *__pyx_v_self, Py_buffer *__pyx_v_info); /* proto */
static PyObject *__pyx_tp_new_8rasterio_5_warp_WarpedVRTReaderBase(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_8rasterio_5_warp___pyx_scope_struct___calculate_default_transform(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_8rasterio_5_warp___pyx_scope_struct_1_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_8rasterio_5_warp___pyx_scope_struct_2_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_items = {0, &__pyx_n_s_items, 0, 0, 0};
static PyObject *__pyx_float_0_125;
static PyObject *__pyx_int_0;
static PyObject *__pyx_int_1;
static PyObject *__pyx_int_neg_2;
static PyObject *__pyx_k__3;
static PyObject *__pyx_k__35;
static PyObject *__pyx_tuple_;
static PyObject *__pyx_tuple__2;
static PyObject *__pyx_tuple__4;
static PyObject *__pyx_tuple__5;
static PyObject *__pyx_tuple__6;
static PyObject *__pyx_tuple__7;
static PyObject *__pyx_tuple__8;
static PyObject *__pyx_tuple__9;
static PyObject *__pyx_slice__30;
static PyObject *__pyx_tuple__10;
static PyObject *__pyx_tuple__12;
static PyObject *__pyx_tuple__13;
static PyObject *__pyx_tuple__14;
static PyObject *__pyx_tuple__15;
static PyObject *__pyx_tuple__16;
static PyObject *__pyx_tuple__17;
static PyObject *__pyx_tuple__18;
static PyObject *__pyx_tuple__19;
static PyObject *__pyx_tuple__20;
static PyObject *__pyx_tuple__21;
static PyObject *__pyx_tuple__22;
static PyObject *__pyx_tuple__23;
static PyObject *__pyx_tuple__24;
static PyObject *__pyx_tuple__25;
static PyObject *__pyx_tuple__26;
static PyObject *__pyx_tuple__27;
static PyObject *__pyx_tuple__28;
static PyObject *__pyx_tuple__29;
static PyObject *__pyx_tuple__31;
static PyObject *__pyx_tuple__32;
static PyObject *__pyx_tuple__33;
static PyObject *__pyx_tuple__34;
static PyObject *__pyx_tuple__36;
static PyObject *__pyx_tuple__37;
static PyObject *__pyx_tuple__38;
static PyObject *__pyx_tuple__39;
static PyObject *__pyx_tuple__40;
static PyObject *__pyx_tuple__41;
static PyObject *__pyx_tuple__42;
static PyObject *__pyx_tuple__43;
static PyObject *__pyx_tuple__44;
static PyObject *__pyx_tuple__45;
static PyObject *__pyx_tuple__46;
static PyObject *__pyx_tuple__47;
static PyObject *__pyx_tuple__48;
static PyObject *__pyx_tuple__49;
static PyObject *__pyx_tuple__51;
static PyObject *__pyx_tuple__54;
static PyObject *__pyx_tuple__56;
static PyObject *__pyx_codeobj__50;
static PyObject *__pyx_codeobj__52;
static PyObject *__pyx_codeobj__55;
static PyObject *__pyx_codeobj__57;

/* "rasterio/_warp.pyx":33
 * 
 * 
 * def recursive_round(val, precision):             # <<<<<<<<<<<<<<
 *     """Recursively round coordinates."""
 *     if isinstance(val, (int, float)):
 */

/* Python wrapper */
static PyObject *__pyx_pw_8rasterio_5_warp_1recursive_round(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_8rasterio_5_warp_recursive_round[] = "Recursively round coordinates.";
static PyMethodDef __pyx_mdef_8rasterio_5_warp_1recursive_round = {"recursive_round", (PyCFunction)__pyx_pw_8rasterio_5_warp_1recursive_round, METH_VARARGS|METH_KEYWORDS, __pyx_doc_8rasterio_5_warp_recursive_round};
static PyObject *__pyx_pw_8rasterio_5_warp_1recursive_round(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_val = 0;
  PyObject *__pyx_v_precision = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("recursive_round (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_val,&__pyx_n_s_precision,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_val)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_precision)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("recursive_round", 1, 2, 2, 1); __PYX_ERR(0, 33, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "recursive_round") < 0)) __PYX_ERR(0, 33, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_val = values[0];
    __pyx_v_precision = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("recursive_round", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 33, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("rasterio._warp.recursive_round", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8rasterio_5_warp_recursive_round(__pyx_self, __pyx_v_val, __pyx_v_precision);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8rasterio_5_warp_recursive_round(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_val, PyObject *__pyx_v_precision) {
  PyObject *__pyx_v_part = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  PyObject *(*__pyx_t_7)(PyObject *);
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  __Pyx_RefNannySetupContext("recursive_round", 0);

  /* "rasterio/_warp.pyx":35
 * def recursive_round(val, precision):
 *     """Recursively round coordinates."""
 *     if isinstance(val, (int, float)):             # <<<<<<<<<<<<<<
 *         return round(val, precision)
 *     else:
 */
  __pyx_t_2 = PyInt_Check(__pyx_v_val); 
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (!__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = PyFloat_Check(__pyx_v_val); 
  __pyx_t_2 = (__pyx_t_3 != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "rasterio/_warp.pyx":36
 *     """Recursively round coordinates."""
 *     if isinstance(val, (int, float)):
 *         return round(val, precision)             # <<<<<<<<<<<<<<
 *     else:
 *         return [recursive_round(part, precision) for part in val]
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 36, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v_val);
    __Pyx_GIVEREF(__pyx_v_val);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_val);
    __Pyx_INCREF(__pyx_v_precision);
    __Pyx_GIVEREF(__pyx_v_precision);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_precision);
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_builtin_round, __pyx_t_4, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 36, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "rasterio/_warp.pyx":35
 * def recursive_round(val, precision):
 *     """Recursively round coordinates."""
 *     if isinstance(val, (int, float)):             # <<<<<<<<<<<<<<
 *         return round(val, precision)
 *     else:
 */
  }

  /* "rasterio/_warp.pyx":38
 *         return round(val, precision)
 *     else:
 *         return [recursive_round(part, precision) for part in val]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_5 = PyList_New(0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 38, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (likely(PyList_CheckExact(__pyx_v_val)) || PyTuple_CheckExact(__pyx_v_val)) {
      __pyx_t_4 = __pyx_v_val; __Pyx_INCREF(__pyx_t_4); __pyx_t_6 = 0;
      __pyx_t_7 = NULL;
    } else {
      __pyx_t_6 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_v_val); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 38, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_7 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 38, __pyx_L1_error)
    }
    for (;;) {
      if (likely(!__pyx_t_7)) {
        if (likely(PyList_CheckExact(__pyx_t_4))) {
          if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_4)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_8 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_6); __Pyx_INCREF(__pyx_t_8); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 38, __pyx_L1_error)
          #else
          __pyx_t_8 = PySequence_ITEM(__pyx_t_4, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 38, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          #endif
        } else {
          if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_8 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_6); __Pyx_INCREF(__pyx_t_8); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 38, __pyx_L1_error)
          #else
          __pyx_t_8 = PySequence_ITEM(__pyx_t_4, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 38, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          #endif
        }
      } else {
        __pyx_t_8 = __pyx_t_7(__pyx_t_4);
        if (unlikely(!__pyx_t_8)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 38, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_8);
      }
      __Pyx_XDECREF_SET(__pyx_v_part, __pyx_t_8);
      __pyx_t_8 = 0;
      __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_recursive_round); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 38, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_10 = NULL;
      __pyx_t_11 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_9))) {
        __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_9);
        if (likely(__pyx_t_10)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
          __Pyx_INCREF(__pyx_t_10);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_9, function);
          __pyx_t_11 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_9)) {
        PyObject *__pyx_temp[3] = {__pyx_t_10, __pyx_v_part, __pyx_v_precision};
        __pyx_t_8 = __Pyx_PyFunction_FastCall(__pyx_t_9, __pyx_temp+1-__pyx_t_11, 2+__pyx_t_11); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 38, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_GOTREF(__pyx_t_8);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_9)) {
        PyObject *__pyx_temp[3] = {__pyx_t_10, __pyx_v_part, __pyx_v_precision};
        __pyx_t_8 = __Pyx_PyCFunction_FastCall(__pyx_t_9, __pyx_temp+1-__pyx_t_11, 2+__pyx_t_11); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 38, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_GOTREF(__pyx_t_8);
      } else
      #endif
      {
        __pyx_t_12 = PyTuple_New(2+__pyx_t_11); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 38, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_12);
        if (__pyx_t_10) {
          __Pyx_GIVEREF(__pyx_t_10); PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_t_10); __pyx_t_10 = NULL;
        }
        __Pyx_INCREF(__pyx_v_part);
        __Pyx_GIVEREF(__pyx_v_part);
        PyTuple_SET_ITEM(__pyx_t_12, 0+__pyx_t_11, __pyx_v_part);
        __Pyx_INCREF(__pyx_v_precision);
        __Pyx_GIVEREF(__pyx_v_precision);
        PyTuple_SET_ITEM(__pyx_t_12, 1+__pyx_t_11, __pyx_v_precision);
        __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_t_12, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 38, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      }
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_5, (PyObject*)__pyx_t_8))) __PYX_ERR(0, 38, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;
  }

  /* "rasterio/_warp.pyx":33
 * 
 * 
 * def recursive_round(val, precision):             # <<<<<<<<<<<<<<
 *     """Recursively round coordinates."""
 *     if isinstance(val, (int, float)):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("rasterio._warp.recursive_round", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_part);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "rasterio/_warp.pyx":41
 * 
 * 
 * def _transform_geom(             # <<<<<<<<<<<<<<
 *         src_crs, dst_crs, geom, antimeridian_cutting, antimeridian_offset,
 *         precision):
 */

/* Python wrapper */
static PyObject *__pyx_pw_8rasterio_5_warp_3_transform_geom(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_8rasterio_5_warp_2_transform_geom[] = "Return a transformed geometry.";
static PyMethodDef __pyx_mdef_8rasterio_5_warp_3_transform_geom = {"_transform_geom", (PyCFunction)__pyx_pw_8rasterio_5_warp_3_transform_geom, METH_VARARGS|METH_KEYWORDS, __pyx_doc_8rasterio_5_warp_2_transform_geom};
static PyObject *__pyx_pw_8rasterio_5_warp_3_transform_geom(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_src_crs = 0;
  PyObject *__pyx_v_dst_crs = 0;
  PyObject *__pyx_v_geom = 0;
  PyObject *__pyx_v_antimeridian_cutting = 0;
  PyObject *__pyx_v_antimeridian_offset = 0;
  PyObject *__pyx_v_precision = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_transform_geom (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_src_crs,&__pyx_n_s_dst_crs,&__pyx_n_s_geom,&__pyx_n_s_antimeridian_cutting,&__pyx_n_s_antimeridian_offset,&__pyx_n_s_precision,0};
    PyObject* values[6] = {0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_src_crs)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dst_crs)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_transform_geom", 1, 6, 6, 1); __PYX_ERR(0, 41, __pyx_L3_error)
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_geom)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_transform_geom", 1, 6, 6, 2); __PYX_ERR(0, 41, __pyx_L3_error)
        }
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_antimeridian_cutting)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_transform_geom", 1, 6, 6, 3); __PYX_ERR(0, 41, __pyx_L3_error)
        }
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_antimeridian_offset)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_transform_geom", 1, 6, 6, 4); __PYX_ERR(0, 41, __pyx_L3_error)
        }
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_precision)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_transform_geom", 1, 6, 6, 5); __PYX_ERR(0, 41, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_transform_geom") < 0)) __PYX_ERR(0, 41, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 6) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
    }
    __pyx_v_src_crs = values[0];
    __pyx_v_dst_crs = values[1];
    __pyx_v_geom = values[2];
    __pyx_v_antimeridian_cutting = values[3];
    __pyx_v_antimeridian_offset = values[4];
    __pyx_v_precision = values[5];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_transform_geom", 1, 6, 6, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 41, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("rasterio._warp._transform_geom", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8rasterio_5_warp_2_transform_geom(__pyx_self, __pyx_v_src_crs, __pyx_v_dst_crs, __pyx_v_geom, __pyx_v_antimeridian_cutting, __pyx_v_antimeridian_offset, __pyx_v_precision);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8rasterio_5_warp_2_transform_geom(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_src_crs, PyObject *__pyx_v_dst_crs, PyObject *__pyx_v_geom, PyObject *__pyx_v_antimeridian_cutting, PyObject *__pyx_v_antimeridian_offset, PyObject *__pyx_v_precision) {
  char **__pyx_v_options;
  OGRSpatialReferenceH __pyx_v_src;
  OGRSpatialReferenceH __pyx_v_dst;
  OGRCoordinateTransformationH __pyx_v_transform;
  OGRGeometryFactory *__pyx_v_factory;
  OGRGeometryH __pyx_v_src_geom;
  OGRGeometryH __pyx_v_dst_geom;
  PyObject *__pyx_v_valb = NULL;
  PyObject *__pyx_v_result = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  OGRSpatialReferenceH __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  void *__pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  char const *__pyx_t_9;
  int __pyx_t_10;
  OGRGeometryFactory *__pyx_t_11;
  OGRGeometryH __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_t_14;
  PyObject *__pyx_t_15 = NULL;
  int __pyx_t_16;
  char const *__pyx_t_17;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  __Pyx_RefNannySetupContext("_transform_geom", 0);

  /* "rasterio/_warp.pyx":45
 *         precision):
 *     """Return a transformed geometry."""
 *     cdef char **options = NULL             # <<<<<<<<<<<<<<
 *     cdef OGRSpatialReferenceH src = NULL
 *     cdef OGRSpatialReferenceH dst = NULL
 */
  __pyx_v_options = NULL;

  /* "rasterio/_warp.pyx":46
 *     """Return a transformed geometry."""
 *     cdef char **options = NULL
 *     cdef OGRSpatialReferenceH src = NULL             # <<<<<<<<<<<<<<
 *     cdef OGRSpatialReferenceH dst = NULL
 *     cdef OGRCoordinateTransformationH transform = NULL
 */
  __pyx_v_src = NULL;

  /* "rasterio/_warp.pyx":47
 *     cdef char **options = NULL
 *     cdef OGRSpatialReferenceH src = NULL
 *     cdef OGRSpatialReferenceH dst = NULL             # <<<<<<<<<<<<<<
 *     cdef OGRCoordinateTransformationH transform = NULL
 *     cdef OGRGeometryFactory *factory = NULL
 */
  __pyx_v_dst = NULL;

  /* "rasterio/_warp.pyx":48
 *     cdef OGRSpatialReferenceH src = NULL
 *     cdef OGRSpatialReferenceH dst = NULL
 *     cdef OGRCoordinateTransformationH transform = NULL             # <<<<<<<<<<<<<<
 *     cdef OGRGeometryFactory *factory = NULL
 *     cdef OGRGeometryH src_geom = NULL
 */
  __pyx_v_transform = NULL;

  /* "rasterio/_warp.pyx":49
 *     cdef OGRSpatialReferenceH dst = NULL
 *     cdef OGRCoordinateTransformationH transform = NULL
 *     cdef OGRGeometryFactory *factory = NULL             # <<<<<<<<<<<<<<
 *     cdef OGRGeometryH src_geom = NULL
 *     cdef OGRGeometryH dst_geom = NULL
 */
  __pyx_v_factory = NULL;

  /* "rasterio/_warp.pyx":50
 *     cdef OGRCoordinateTransformationH transform = NULL
 *     cdef OGRGeometryFactory *factory = NULL
 *     cdef OGRGeometryH src_geom = NULL             # <<<<<<<<<<<<<<
 *     cdef OGRGeometryH dst_geom = NULL
 *     cdef int i
 */
  __pyx_v_src_geom = NULL;

  /* "rasterio/_warp.pyx":51
 *     cdef OGRGeometryFactory *factory = NULL
 *     cdef OGRGeometryH src_geom = NULL
 *     cdef OGRGeometryH dst_geom = NULL             # <<<<<<<<<<<<<<
 *     cdef int i
 * 
 */
  __pyx_v_dst_geom = NULL;

  /* "rasterio/_warp.pyx":54
 *     cdef int i
 * 
 *     src = _osr_from_crs(src_crs)             # <<<<<<<<<<<<<<
 *     dst = _osr_from_crs(dst_crs)
 * 
 */
  __pyx_t_1 = __pyx_f_8rasterio_5_base__osr_from_crs(__pyx_v_src_crs); if (unlikely(__pyx_t_1 == NULL)) __PYX_ERR(0, 54, __pyx_L1_error)
  __pyx_v_src = __pyx_t_1;

  /* "rasterio/_warp.pyx":55
 * 
 *     src = _osr_from_crs(src_crs)
 *     dst = _osr_from_crs(dst_crs)             # <<<<<<<<<<<<<<
 * 
 *     try:
 */
  __pyx_t_1 = __pyx_f_8rasterio_5_base__osr_from_crs(__pyx_v_dst_crs); if (unlikely(__pyx_t_1 == NULL)) __PYX_ERR(0, 55, __pyx_L1_error)
  __pyx_v_dst = __pyx_t_1;

  /* "rasterio/_warp.pyx":57
 *     dst = _osr_from_crs(dst_crs)
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         transform = exc_wrap_pointer(OCTNewCoordinateTransformation(src, dst))
 *     except:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_2, &__pyx_t_3, &__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_4);
    /*try:*/ {

      /* "rasterio/_warp.pyx":58
 * 
 *     try:
 *         transform = exc_wrap_pointer(OCTNewCoordinateTransformation(src, dst))             # <<<<<<<<<<<<<<
 *     except:
 *         OSRRelease(src)
 */
      __pyx_t_5 = __pyx_f_8rasterio_4_err_exc_wrap_pointer(OCTNewCoordinateTransformation(__pyx_v_src, __pyx_v_dst)); if (unlikely(__pyx_t_5 == NULL)) __PYX_ERR(0, 58, __pyx_L3_error)
      __pyx_v_transform = __pyx_t_5;

      /* "rasterio/_warp.pyx":57
 *     dst = _osr_from_crs(dst_crs)
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         transform = exc_wrap_pointer(OCTNewCoordinateTransformation(src, dst))
 *     except:
 */
    }
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    goto __pyx_L10_try_end;
    __pyx_L3_error:;
    __Pyx_PyThreadState_assign

    /* "rasterio/_warp.pyx":59
 *     try:
 *         transform = exc_wrap_pointer(OCTNewCoordinateTransformation(src, dst))
 *     except:             # <<<<<<<<<<<<<<
 *         OSRRelease(src)
 *         OSRRelease(dst)
 */
    /*except:*/ {
      __Pyx_AddTraceback("rasterio._warp._transform_geom", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_6, &__pyx_t_7, &__pyx_t_8) < 0) __PYX_ERR(0, 59, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_GOTREF(__pyx_t_8);

      /* "rasterio/_warp.pyx":60
 *         transform = exc_wrap_pointer(OCTNewCoordinateTransformation(src, dst))
 *     except:
 *         OSRRelease(src)             # <<<<<<<<<<<<<<
 *         OSRRelease(dst)
 *         raise
 */
      OSRRelease(__pyx_v_src);

      /* "rasterio/_warp.pyx":61
 *     except:
 *         OSRRelease(src)
 *         OSRRelease(dst)             # <<<<<<<<<<<<<<
 *         raise
 * 
 */
      OSRRelease(__pyx_v_dst);

      /* "rasterio/_warp.pyx":62
 *         OSRRelease(src)
 *         OSRRelease(dst)
 *         raise             # <<<<<<<<<<<<<<
 * 
 *     # Transform options.
 */
      __Pyx_GIVEREF(__pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ErrRestoreWithState(__pyx_t_6, __pyx_t_7, __pyx_t_8);
      __pyx_t_6 = 0; __pyx_t_7 = 0; __pyx_t_8 = 0; 
      __PYX_ERR(0, 62, __pyx_L5_except_error)
    }
    __pyx_L5_except_error:;

    /* "rasterio/_warp.pyx":57
 *     dst = _osr_from_crs(dst_crs)
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         transform = exc_wrap_pointer(OCTNewCoordinateTransformation(src, dst))
 *     except:
 */
    __Pyx_PyThreadState_assign
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
    goto __pyx_L1_error;
    __pyx_L10_try_end:;
  }

  /* "rasterio/_warp.pyx":65
 * 
 *     # Transform options.
 *     valb = str(antimeridian_offset).encode('utf-8')             # <<<<<<<<<<<<<<
 *     options = CSLSetNameValue(options, "DATELINEOFFSET", <const char *>valb)
 *     if antimeridian_cutting:
 */
  __pyx_t_8 = PyTuple_New(1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_INCREF(__pyx_v_antimeridian_offset);
  __Pyx_GIVEREF(__pyx_v_antimeridian_offset);
  PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_v_antimeridian_offset);
  __pyx_t_7 = __Pyx_PyObject_Call(((PyObject *)(&PyString_Type)), __pyx_t_8, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_encode); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_tuple_, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_v_valb = __pyx_t_7;
  __pyx_t_7 = 0;

  /* "rasterio/_warp.pyx":66
 *     # Transform options.
 *     valb = str(antimeridian_offset).encode('utf-8')
 *     options = CSLSetNameValue(options, "DATELINEOFFSET", <const char *>valb)             # <<<<<<<<<<<<<<
 *     if antimeridian_cutting:
 *         options = CSLSetNameValue(options, "WRAPDATELINE", "YES")
 */
  __pyx_t_9 = __Pyx_PyObject_AsString(__pyx_v_valb); if (unlikely((!__pyx_t_9) && PyErr_Occurred())) __PYX_ERR(0, 66, __pyx_L1_error)
  __pyx_v_options = CSLSetNameValue(__pyx_v_options, ((char *)"DATELINEOFFSET"), ((char const *)__pyx_t_9));

  /* "rasterio/_warp.pyx":67
 *     valb = str(antimeridian_offset).encode('utf-8')
 *     options = CSLSetNameValue(options, "DATELINEOFFSET", <const char *>valb)
 *     if antimeridian_cutting:             # <<<<<<<<<<<<<<
 *         options = CSLSetNameValue(options, "WRAPDATELINE", "YES")
 * 
 */
  __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_v_antimeridian_cutting); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 67, __pyx_L1_error)
  if (__pyx_t_10) {

    /* "rasterio/_warp.pyx":68
 *     options = CSLSetNameValue(options, "DATELINEOFFSET", <const char *>valb)
 *     if antimeridian_cutting:
 *         options = CSLSetNameValue(options, "WRAPDATELINE", "YES")             # <<<<<<<<<<<<<<
 * 
 *     try:
 */
    __pyx_v_options = CSLSetNameValue(__pyx_v_options, ((char *)"WRAPDATELINE"), ((char *)"YES"));

    /* "rasterio/_warp.pyx":67
 *     valb = str(antimeridian_offset).encode('utf-8')
 *     options = CSLSetNameValue(options, "DATELINEOFFSET", <const char *>valb)
 *     if antimeridian_cutting:             # <<<<<<<<<<<<<<
 *         options = CSLSetNameValue(options, "WRAPDATELINE", "YES")
 * 
 */
  }

  /* "rasterio/_warp.pyx":70
 *         options = CSLSetNameValue(options, "WRAPDATELINE", "YES")
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         factory = new OGRGeometryFactory()
 *         src_geom = OGRGeomBuilder().build(geom)
 */
  /*try:*/ {

    /* "rasterio/_warp.pyx":71
 * 
 *     try:
 *         factory = new OGRGeometryFactory()             # <<<<<<<<<<<<<<
 *         src_geom = OGRGeomBuilder().build(geom)
 *         dst_geom = exc_wrap_pointer(
 */
    try {
      __pyx_t_11 = new OGRGeometryFactory();
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 71, __pyx_L15_error)
    }
    __pyx_v_factory = __pyx_t_11;

    /* "rasterio/_warp.pyx":72
 *     try:
 *         factory = new OGRGeometryFactory()
 *         src_geom = OGRGeomBuilder().build(geom)             # <<<<<<<<<<<<<<
 *         dst_geom = exc_wrap_pointer(
 *             factory.transformWithOptions(
 */
    __pyx_t_7 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_8rasterio_9_features_OGRGeomBuilder), __pyx_empty_tuple, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 72, __pyx_L15_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_12 = ((struct __pyx_vtabstruct_8rasterio_9_features_OGRGeomBuilder *)((struct __pyx_obj_8rasterio_9_features_OGRGeomBuilder *)__pyx_t_7)->__pyx_vtab)->build(((struct __pyx_obj_8rasterio_9_features_OGRGeomBuilder *)__pyx_t_7), __pyx_v_geom); if (unlikely(__pyx_t_12 == NULL)) __PYX_ERR(0, 72, __pyx_L15_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_v_src_geom = __pyx_t_12;

    /* "rasterio/_warp.pyx":73
 *         factory = new OGRGeometryFactory()
 *         src_geom = OGRGeomBuilder().build(geom)
 *         dst_geom = exc_wrap_pointer(             # <<<<<<<<<<<<<<
 *             factory.transformWithOptions(
 *                 <const OGRGeometry *>src_geom,
 */
    __pyx_t_5 = __pyx_f_8rasterio_4_err_exc_wrap_pointer(__pyx_v_factory->transformWithOptions(((OGRGeometry const *)__pyx_v_src_geom), ((OGRCoordinateTransformation *)__pyx_v_transform), __pyx_v_options)); if (unlikely(__pyx_t_5 == NULL)) __PYX_ERR(0, 73, __pyx_L15_error)
    __pyx_v_dst_geom = __pyx_t_5;

    /* "rasterio/_warp.pyx":79
 *                 options))
 * 
 *         result = GeomBuilder().build(dst_geom)             # <<<<<<<<<<<<<<
 * 
 *         if precision >= 0:
 */
    __pyx_t_7 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_8rasterio_9_features_GeomBuilder), __pyx_empty_tuple, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 79, __pyx_L15_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = ((struct __pyx_vtabstruct_8rasterio_9_features_GeomBuilder *)((struct __pyx_obj_8rasterio_9_features_GeomBuilder *)__pyx_t_7)->__pyx_vtab)->build(((struct __pyx_obj_8rasterio_9_features_GeomBuilder *)__pyx_t_7), __pyx_v_dst_geom); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 79, __pyx_L15_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_v_result = __pyx_t_8;
    __pyx_t_8 = 0;

    /* "rasterio/_warp.pyx":81
 *         result = GeomBuilder().build(dst_geom)
 * 
 *         if precision >= 0:             # <<<<<<<<<<<<<<
 *             # TODO: Geometry collections.
 *             result['coordinates'] = recursive_round(result['coordinates'],
 */
    __pyx_t_8 = PyObject_RichCompare(__pyx_v_precision, __pyx_int_0, Py_GE); __Pyx_XGOTREF(__pyx_t_8); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 81, __pyx_L15_error)
    __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_8); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 81, __pyx_L15_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (__pyx_t_10) {

      /* "rasterio/_warp.pyx":83
 *         if precision >= 0:
 *             # TODO: Geometry collections.
 *             result['coordinates'] = recursive_round(result['coordinates'],             # <<<<<<<<<<<<<<
 *                                                     precision)
 * 
 */
      __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_recursive_round); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 83, __pyx_L15_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_6 = PyObject_GetItem(__pyx_v_result, __pyx_n_s_coordinates); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 83, __pyx_L15_error)
      __Pyx_GOTREF(__pyx_t_6);

      /* "rasterio/_warp.pyx":84
 *             # TODO: Geometry collections.
 *             result['coordinates'] = recursive_round(result['coordinates'],
 *                                                     precision)             # <<<<<<<<<<<<<<
 * 
 *         return result
 */
      __pyx_t_13 = NULL;
      __pyx_t_14 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_13)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_13);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
          __pyx_t_14 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_7)) {
        PyObject *__pyx_temp[3] = {__pyx_t_13, __pyx_t_6, __pyx_v_precision};
        __pyx_t_8 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_14, 2+__pyx_t_14); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 83, __pyx_L15_error)
        __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
        PyObject *__pyx_temp[3] = {__pyx_t_13, __pyx_t_6, __pyx_v_precision};
        __pyx_t_8 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_14, 2+__pyx_t_14); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 83, __pyx_L15_error)
        __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      {
        __pyx_t_15 = PyTuple_New(2+__pyx_t_14); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 83, __pyx_L15_error)
        __Pyx_GOTREF(__pyx_t_15);
        if (__pyx_t_13) {
          __Pyx_GIVEREF(__pyx_t_13); PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_t_13); __pyx_t_13 = NULL;
        }
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_15, 0+__pyx_t_14, __pyx_t_6);
        __Pyx_INCREF(__pyx_v_precision);
        __Pyx_GIVEREF(__pyx_v_precision);
        PyTuple_SET_ITEM(__pyx_t_15, 1+__pyx_t_14, __pyx_v_precision);
        __pyx_t_6 = 0;
        __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_15, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 83, __pyx_L15_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
      }
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "rasterio/_warp.pyx":83
 *         if precision >= 0:
 *             # TODO: Geometry collections.
 *             result['coordinates'] = recursive_round(result['coordinates'],             # <<<<<<<<<<<<<<
 *                                                     precision)
 * 
 */
      if (unlikely(PyObject_SetItem(__pyx_v_result, __pyx_n_s_coordinates, __pyx_t_8) < 0)) __PYX_ERR(0, 83, __pyx_L15_error)
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "rasterio/_warp.pyx":81
 *         result = GeomBuilder().build(dst_geom)
 * 
 *         if precision >= 0:             # <<<<<<<<<<<<<<
 *             # TODO: Geometry collections.
 *             result['coordinates'] = recursive_round(result['coordinates'],
 */
    }

    /* "rasterio/_warp.pyx":86
 *                                                     precision)
 * 
 *         return result             # <<<<<<<<<<<<<<
 * 
 *     finally:
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_result);
    __pyx_r = __pyx_v_result;
    goto __pyx_L14_return;
  }

  /* "rasterio/_warp.pyx":89
 * 
 *     finally:
 *         del factory             # <<<<<<<<<<<<<<
 *         OGR_G_DestroyGeometry(dst_geom)
 *         OGR_G_DestroyGeometry(src_geom)
 */
  /*finally:*/ {
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __pyx_L15_error:;
      __pyx_t_4 = 0; __pyx_t_3 = 0; __pyx_t_2 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_18, &__pyx_t_19, &__pyx_t_20);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_4, &__pyx_t_3, &__pyx_t_2) < 0)) __Pyx_ErrFetch(&__pyx_t_4, &__pyx_t_3, &__pyx_t_2);
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_2);
      __Pyx_XGOTREF(__pyx_t_18);
      __Pyx_XGOTREF(__pyx_t_19);
      __Pyx_XGOTREF(__pyx_t_20);
      __pyx_t_14 = __pyx_lineno; __pyx_t_16 = __pyx_clineno; __pyx_t_17 = __pyx_filename;
      {
        delete __pyx_v_factory;

        /* "rasterio/_warp.pyx":90
 *     finally:
 *         del factory
 *         OGR_G_DestroyGeometry(dst_geom)             # <<<<<<<<<<<<<<
 *         OGR_G_DestroyGeometry(src_geom)
 *         OCTDestroyCoordinateTransformation(transform)
 */
        OGR_G_DestroyGeometry(__pyx_v_dst_geom);

        /* "rasterio/_warp.pyx":91
 *         del factory
 *         OGR_G_DestroyGeometry(dst_geom)
 *         OGR_G_DestroyGeometry(src_geom)             # <<<<<<<<<<<<<<
 *         OCTDestroyCoordinateTransformation(transform)
 *         if options != NULL:
 */
        OGR_G_DestroyGeometry(__pyx_v_src_geom);

        /* "rasterio/_warp.pyx":92
 *         OGR_G_DestroyGeometry(dst_geom)
 *         OGR_G_DestroyGeometry(src_geom)
 *         OCTDestroyCoordinateTransformation(transform)             # <<<<<<<<<<<<<<
 *         if options != NULL:
 *             CSLDestroy(options)
 */
        OCTDestroyCoordinateTransformation(__pyx_v_transform);

        /* "rasterio/_warp.pyx":93
 *         OGR_G_DestroyGeometry(src_geom)
 *         OCTDestroyCoordinateTransformation(transform)
 *         if options != NULL:             # <<<<<<<<<<<<<<
 *             CSLDestroy(options)
 *         OSRRelease(src)
 */
        __pyx_t_10 = ((__pyx_v_options != NULL) != 0);
        if (__pyx_t_10) {

          /* "rasterio/_warp.pyx":94
 *         OCTDestroyCoordinateTransformation(transform)
 *         if options != NULL:
 *             CSLDestroy(options)             # <<<<<<<<<<<<<<
 *         OSRRelease(src)
 *         OSRRelease(dst)
 */
          CSLDestroy(__pyx_v_options);

          /* "rasterio/_warp.pyx":93
 *         OGR_G_DestroyGeometry(src_geom)
 *         OCTDestroyCoordinateTransformation(transform)
 *         if options != NULL:             # <<<<<<<<<<<<<<
 *             CSLDestroy(options)
 *         OSRRelease(src)
 */
        }

        /* "rasterio/_warp.pyx":95
 *         if options != NULL:
 *             CSLDestroy(options)
 *         OSRRelease(src)             # <<<<<<<<<<<<<<
 *         OSRRelease(dst)
 * 
 */
        OSRRelease(__pyx_v_src);

        /* "rasterio/_warp.pyx":96
 *             CSLDestroy(options)
 *         OSRRelease(src)
 *         OSRRelease(dst)             # <<<<<<<<<<<<<<
 * 
 * 
 */
        OSRRelease(__pyx_v_dst);
      }
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_ExceptionReset(__pyx_t_18, __pyx_t_19, __pyx_t_20);
      }
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_2);
      __Pyx_ErrRestore(__pyx_t_4, __pyx_t_3, __pyx_t_2);
      __pyx_t_4 = 0; __pyx_t_3 = 0; __pyx_t_2 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0;
      __pyx_lineno = __pyx_t_14; __pyx_clineno = __pyx_t_16; __pyx_filename = __pyx_t_17;
      goto __pyx_L1_error;
    }
    __pyx_L14_return: {
      __pyx_t_20 = __pyx_r;
      __pyx_r = 0;

      /* "rasterio/_warp.pyx":89
 * 
 *     finally:
 *         del factory             # <<<<<<<<<<<<<<
 *         OGR_G_DestroyGeometry(dst_geom)
 *         OGR_G_DestroyGeometry(src_geom)
 */
      delete __pyx_v_factory;

      /* "rasterio/_warp.pyx":90
 *     finally:
 *         del factory
 *         OGR_G_DestroyGeometry(dst_geom)             # <<<<<<<<<<<<<<
 *         OGR_G_DestroyGeometry(src_geom)
 *         OCTDestroyCoordinateTransformation(transform)
 */
      OGR_G_DestroyGeometry(__pyx_v_dst_geom);

      /* "rasterio/_warp.pyx":91
 *         del factory
 *         OGR_G_DestroyGeometry(dst_geom)
 *         OGR_G_DestroyGeometry(src_geom)             # <<<<<<<<<<<<<<
 *         OCTDestroyCoordinateTransformation(transform)
 *         if options != NULL:
 */
      OGR_G_DestroyGeometry(__pyx_v_src_geom);

      /* "rasterio/_warp.pyx":92
 *         OGR_G_DestroyGeometry(dst_geom)
 *         OGR_G_DestroyGeometry(src_geom)
 *         OCTDestroyCoordinateTransformation(transform)             # <<<<<<<<<<<<<<
 *         if options != NULL:
 *             CSLDestroy(options)
 */
      OCTDestroyCoordinateTransformation(__pyx_v_transform);

      /* "rasterio/_warp.pyx":93
 *         OGR_G_DestroyGeometry(src_geom)
 *         OCTDestroyCoordinateTransformation(transform)
 *         if options != NULL:             # <<<<<<<<<<<<<<
 *             CSLDestroy(options)
 *         OSRRelease(src)
 */
      __pyx_t_10 = ((__pyx_v_options != NULL) != 0);
      if (__pyx_t_10) {

        /* "rasterio/_warp.pyx":94
 *         OCTDestroyCoordinateTransformation(transform)
 *         if options != NULL:
 *             CSLDestroy(options)             # <<<<<<<<<<<<<<
 *         OSRRelease(src)
 *         OSRRelease(dst)
 */
        CSLDestroy(__pyx_v_options);

        /* "rasterio/_warp.pyx":93
 *         OGR_G_DestroyGeometry(src_geom)
 *         OCTDestroyCoordinateTransformation(transform)
 *         if options != NULL:             # <<<<<<<<<<<<<<
 *             CSLDestroy(options)
 *         OSRRelease(src)
 */
      }

      /* "rasterio/_warp.pyx":95
 *         if options != NULL:
 *             CSLDestroy(options)
 *         OSRRelease(src)             # <<<<<<<<<<<<<<
 *         OSRRelease(dst)
 * 
 */
      OSRRelease(__pyx_v_src);

      /* "rasterio/_warp.pyx":96
 *             CSLDestroy(options)
 *         OSRRelease(src)
 *         OSRRelease(dst)             # <<<<<<<<<<<<<<
 * 
 * 
 */
      OSRRelease(__pyx_v_dst);
      __pyx_r = __pyx_t_20;
      __pyx_t_20 = 0;
      goto __pyx_L0;
    }
  }

  /* "rasterio/_warp.pyx":41
 * 
 * 
 * def _transform_geom(             # <<<<<<<<<<<<<<
 *         src_crs, dst_crs, geom, antimeridian_cutting, antimeridian_offset,
 *         precision):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_AddTraceback("rasterio._warp._transform_geom", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_valb);
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "rasterio/_warp.pyx":99
 * 
 * 
 * cdef GDALWarpOptions * create_warp_options(             # <<<<<<<<<<<<<<
 *         GDALResampleAlg resampling, object src_nodata, object dst_nodata,
 *         int src_count, const char **options) except NULL:
 */

static GDALWarpOptions *__pyx_f_8rasterio_5_warp_create_warp_options(GDALResampleAlg __pyx_v_resampling, PyObject *__pyx_v_src_nodata, PyObject *__pyx_v_dst_nodata, int __pyx_v_src_count, char const **__pyx_v_options) {
  GDALWarpOptions *__pyx_v_psWOptions;
  char **__pyx_v_warp_extras;
  PyObject *__pyx_v_i = NULL;
  GDALWarpOptions *__pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  char **__pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  PyObject *(*__pyx_t_8)(PyObject *);
  double __pyx_t_9;
  Py_ssize_t __pyx_t_10;
  int __pyx_t_11;
  __Pyx_RefNannySetupContext("create_warp_options", 0);
  __Pyx_INCREF(__pyx_v_src_nodata);
  __Pyx_INCREF(__pyx_v_dst_nodata);

  /* "rasterio/_warp.pyx":109
 * 
 *     # We require source nodata if destination nodata is passed.
 *     if src_nodata is None and dst_nodata is not None:             # <<<<<<<<<<<<<<
 *         raise ValueError("src_nodata must be provided because dst_nodata "
 *                          "is not None")
 */
  __pyx_t_2 = (__pyx_v_src_nodata == Py_None);
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = (__pyx_v_dst_nodata != Py_None);
  __pyx_t_2 = (__pyx_t_3 != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "rasterio/_warp.pyx":110
 *     # We require source nodata if destination nodata is passed.
 *     if src_nodata is None and dst_nodata is not None:
 *         raise ValueError("src_nodata must be provided because dst_nodata "             # <<<<<<<<<<<<<<
 *                          "is not None")
 * 
 */
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__2, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 110, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(0, 110, __pyx_L1_error)

    /* "rasterio/_warp.pyx":109
 * 
 *     # We require source nodata if destination nodata is passed.
 *     if src_nodata is None and dst_nodata is not None:             # <<<<<<<<<<<<<<
 *         raise ValueError("src_nodata must be provided because dst_nodata "
 *                          "is not None")
 */
  }

  /* "rasterio/_warp.pyx":113
 *                          "is not None")
 * 
 *     if dst_nodata is None:             # <<<<<<<<<<<<<<
 *         if src_nodata is not None:
 *             dst_nodata = src_nodata
 */
  __pyx_t_1 = (__pyx_v_dst_nodata == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "rasterio/_warp.pyx":114
 * 
 *     if dst_nodata is None:
 *         if src_nodata is not None:             # <<<<<<<<<<<<<<
 *             dst_nodata = src_nodata
 *         else:
 */
    __pyx_t_2 = (__pyx_v_src_nodata != Py_None);
    __pyx_t_1 = (__pyx_t_2 != 0);
    if (__pyx_t_1) {

      /* "rasterio/_warp.pyx":115
 *     if dst_nodata is None:
 *         if src_nodata is not None:
 *             dst_nodata = src_nodata             # <<<<<<<<<<<<<<
 *         else:
 *             dst_nodata = 0
 */
      __Pyx_INCREF(__pyx_v_src_nodata);
      __Pyx_DECREF_SET(__pyx_v_dst_nodata, __pyx_v_src_nodata);

      /* "rasterio/_warp.pyx":114
 * 
 *     if dst_nodata is None:
 *         if src_nodata is not None:             # <<<<<<<<<<<<<<
 *             dst_nodata = src_nodata
 *         else:
 */
      goto __pyx_L7;
    }

    /* "rasterio/_warp.pyx":117
 *             dst_nodata = src_nodata
 *         else:
 *             dst_nodata = 0             # <<<<<<<<<<<<<<
 *             src_nodata = 0
 * 
 */
    /*else*/ {
      __Pyx_INCREF(__pyx_int_0);
      __Pyx_DECREF_SET(__pyx_v_dst_nodata, __pyx_int_0);

      /* "rasterio/_warp.pyx":118
 *         else:
 *             dst_nodata = 0
 *             src_nodata = 0             # <<<<<<<<<<<<<<
 * 
 *     # Convert to float to validate.
 */
      __Pyx_INCREF(__pyx_int_0);
      __Pyx_DECREF_SET(__pyx_v_src_nodata, __pyx_int_0);
    }
    __pyx_L7:;

    /* "rasterio/_warp.pyx":113
 *                          "is not None")
 * 
 *     if dst_nodata is None:             # <<<<<<<<<<<<<<
 *         if src_nodata is not None:
 *             dst_nodata = src_nodata
 */
  }

  /* "rasterio/_warp.pyx":121
 * 
 *     # Convert to float to validate.
 *     dst_nodata = float(dst_nodata)             # <<<<<<<<<<<<<<
 *     src_nodata = float(src_nodata)
 * 
 */
  __pyx_t_4 = __Pyx_PyNumber_Float(__pyx_v_dst_nodata); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 121, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF_SET(__pyx_v_dst_nodata, __pyx_t_4);
  __pyx_t_4 = 0;

  /* "rasterio/_warp.pyx":122
 *     # Convert to float to validate.
 *     dst_nodata = float(dst_nodata)
 *     src_nodata = float(src_nodata)             # <<<<<<<<<<<<<<
 * 
 *     cdef GDALWarpOptions *psWOptions = GDALCreateWarpOptions()
 */
  __pyx_t_4 = __Pyx_PyNumber_Float(__pyx_v_src_nodata); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF_SET(__pyx_v_src_nodata, __pyx_t_4);
  __pyx_t_4 = 0;

  /* "rasterio/_warp.pyx":124
 *     src_nodata = float(src_nodata)
 * 
 *     cdef GDALWarpOptions *psWOptions = GDALCreateWarpOptions()             # <<<<<<<<<<<<<<
 * 
 *     # Note: warp_extras is pointed to different memory locations on every
 */
  __pyx_v_psWOptions = GDALCreateWarpOptions();

  /* "rasterio/_warp.pyx":129
 *     # call to CSLSetNameValue call below, but needs to be set here to
 *     # get the defaults.
 *     cdef char **warp_extras = psWOptions.papszWarpOptions             # <<<<<<<<<<<<<<
 * 
 *     # See http://www.gdal.org/structGDALWarpOptions.html#a0ed77f9917bb96c7a9aabd73d4d06e08
 */
  __pyx_t_5 = __pyx_v_psWOptions->papszWarpOptions;
  __pyx_v_warp_extras = __pyx_t_5;

  /* "rasterio/_warp.pyx":136
 * 
 *     # Use the same default nodata logic as gdalwarp.
 *     warp_extras = CSLSetNameValue(             # <<<<<<<<<<<<<<
 *         warp_extras, "UNIFIED_SRC_NODATA", "YES")
 * 
 */
  __pyx_v_warp_extras = CSLSetNameValue(__pyx_v_warp_extras, ((char *)"UNIFIED_SRC_NODATA"), ((char *)"YES"));

  /* "rasterio/_warp.pyx":139
 *         warp_extras, "UNIFIED_SRC_NODATA", "YES")
 * 
 *     warp_extras = CSLMerge(warp_extras, <char **>options)             # <<<<<<<<<<<<<<
 * 
 *     psWOptions.eResampleAlg = <GDALResampleAlg>resampling
 */
  __pyx_v_warp_extras = CSLMerge(__pyx_v_warp_extras, ((char **)__pyx_v_options));

  /* "rasterio/_warp.pyx":141
 *     warp_extras = CSLMerge(warp_extras, <char **>options)
 * 
 *     psWOptions.eResampleAlg = <GDALResampleAlg>resampling             # <<<<<<<<<<<<<<
 * 
 *     psWOptions.padfSrcNoDataReal = <double*>CPLMalloc(src_count * sizeof(double))
 */
  __pyx_v_psWOptions->eResampleAlg = ((GDALResampleAlg)__pyx_v_resampling);

  /* "rasterio/_warp.pyx":143
 *     psWOptions.eResampleAlg = <GDALResampleAlg>resampling
 * 
 *     psWOptions.padfSrcNoDataReal = <double*>CPLMalloc(src_count * sizeof(double))             # <<<<<<<<<<<<<<
 *     psWOptions.padfSrcNoDataImag = <double*>CPLMalloc(src_count * sizeof(double))
 *     psWOptions.padfDstNoDataReal = <double*>CPLMalloc(src_count * sizeof(double))
 */
  __pyx_v_psWOptions->padfSrcNoDataReal = ((double *)CPLMalloc((__pyx_v_src_count * (sizeof(double)))));

  /* "rasterio/_warp.pyx":144
 * 
 *     psWOptions.padfSrcNoDataReal = <double*>CPLMalloc(src_count * sizeof(double))
 *     psWOptions.padfSrcNoDataImag = <double*>CPLMalloc(src_count * sizeof(double))             # <<<<<<<<<<<<<<
 *     psWOptions.padfDstNoDataReal = <double*>CPLMalloc(src_count * sizeof(double))
 *     psWOptions.padfDstNoDataImag = <double*>CPLMalloc(src_count * sizeof(double))
 */
  __pyx_v_psWOptions->padfSrcNoDataImag = ((double *)CPLMalloc((__pyx_v_src_count * (sizeof(double)))));

  /* "rasterio/_warp.pyx":145
 *     psWOptions.padfSrcNoDataReal = <double*>CPLMalloc(src_count * sizeof(double))
 *     psWOptions.padfSrcNoDataImag = <double*>CPLMalloc(src_count * sizeof(double))
 *     psWOptions.padfDstNoDataReal = <double*>CPLMalloc(src_count * sizeof(double))             # <<<<<<<<<<<<<<
 *     psWOptions.padfDstNoDataImag = <double*>CPLMalloc(src_count * sizeof(double))
 *     for i in range(src_count):
 */
  __pyx_v_psWOptions->padfDstNoDataReal = ((double *)CPLMalloc((__pyx_v_src_count * (sizeof(double)))));

  /* "rasterio/_warp.pyx":146
 *     psWOptions.padfSrcNoDataImag = <double*>CPLMalloc(src_count * sizeof(double))
 *     psWOptions.padfDstNoDataReal = <double*>CPLMalloc(src_count * sizeof(double))
 *     psWOptions.padfDstNoDataImag = <double*>CPLMalloc(src_count * sizeof(double))             # <<<<<<<<<<<<<<
 *     for i in range(src_count):
 *         psWOptions.padfSrcNoDataReal[i] = src_nodata
 */
  __pyx_v_psWOptions->padfDstNoDataImag = ((double *)CPLMalloc((__pyx_v_src_count * (sizeof(double)))));

  /* "rasterio/_warp.pyx":147
 *     psWOptions.padfDstNoDataReal = <double*>CPLMalloc(src_count * sizeof(double))
 *     psWOptions.padfDstNoDataImag = <double*>CPLMalloc(src_count * sizeof(double))
 *     for i in range(src_count):             # <<<<<<<<<<<<<<
 *         psWOptions.padfSrcNoDataReal[i] = src_nodata
 *         psWOptions.padfSrcNoDataImag[i] = 0.0
 */
  __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_src_count); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4);
  __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_range, __pyx_t_6, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (likely(PyList_CheckExact(__pyx_t_4)) || PyTuple_CheckExact(__pyx_t_4)) {
    __pyx_t_6 = __pyx_t_4; __Pyx_INCREF(__pyx_t_6); __pyx_t_7 = 0;
    __pyx_t_8 = NULL;
  } else {
    __pyx_t_7 = -1; __pyx_t_6 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 147, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = Py_TYPE(__pyx_t_6)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 147, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  for (;;) {
    if (likely(!__pyx_t_8)) {
      if (likely(PyList_CheckExact(__pyx_t_6))) {
        if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_6)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_6, __pyx_t_7); __Pyx_INCREF(__pyx_t_4); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 147, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_6, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 147, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_6)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_6, __pyx_t_7); __Pyx_INCREF(__pyx_t_4); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 147, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_6, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 147, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_8(__pyx_t_6);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 147, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "rasterio/_warp.pyx":148
 *     psWOptions.padfDstNoDataImag = <double*>CPLMalloc(src_count * sizeof(double))
 *     for i in range(src_count):
 *         psWOptions.padfSrcNoDataReal[i] = src_nodata             # <<<<<<<<<<<<<<
 *         psWOptions.padfSrcNoDataImag[i] = 0.0
 *         psWOptions.padfDstNoDataReal[i] = dst_nodata
 */
    __pyx_t_9 = __pyx_PyFloat_AsDouble(__pyx_v_src_nodata); if (unlikely((__pyx_t_9 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 148, __pyx_L1_error)
    __pyx_t_10 = __Pyx_PyIndex_AsSsize_t(__pyx_v_i); if (unlikely((__pyx_t_10 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 148, __pyx_L1_error)
    (__pyx_v_psWOptions->padfSrcNoDataReal[__pyx_t_10]) = __pyx_t_9;

    /* "rasterio/_warp.pyx":149
 *     for i in range(src_count):
 *         psWOptions.padfSrcNoDataReal[i] = src_nodata
 *         psWOptions.padfSrcNoDataImag[i] = 0.0             # <<<<<<<<<<<<<<
 *         psWOptions.padfDstNoDataReal[i] = dst_nodata
 *         psWOptions.padfDstNoDataImag[i] = 0.0
 */
    __pyx_t_10 = __Pyx_PyIndex_AsSsize_t(__pyx_v_i); if (unlikely((__pyx_t_10 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 149, __pyx_L1_error)
    (__pyx_v_psWOptions->padfSrcNoDataImag[__pyx_t_10]) = 0.0;

    /* "rasterio/_warp.pyx":150
 *         psWOptions.padfSrcNoDataReal[i] = src_nodata
 *         psWOptions.padfSrcNoDataImag[i] = 0.0
 *         psWOptions.padfDstNoDataReal[i] = dst_nodata             # <<<<<<<<<<<<<<
 *         psWOptions.padfDstNoDataImag[i] = 0.0
 * 
 */
    __pyx_t_9 = __pyx_PyFloat_AsDouble(__pyx_v_dst_nodata); if (unlikely((__pyx_t_9 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 150, __pyx_L1_error)
    __pyx_t_10 = __Pyx_PyIndex_AsSsize_t(__pyx_v_i); if (unlikely((__pyx_t_10 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 150, __pyx_L1_error)
    (__pyx_v_psWOptions->padfDstNoDataReal[__pyx_t_10]) = __pyx_t_9;

    /* "rasterio/_warp.pyx":151
 *         psWOptions.padfSrcNoDataImag[i] = 0.0
 *         psWOptions.padfDstNoDataReal[i] = dst_nodata
 *         psWOptions.padfDstNoDataImag[i] = 0.0             # <<<<<<<<<<<<<<
 * 
 *     # Important: set back into struct or values set above are lost
 */
    __pyx_t_10 = __Pyx_PyIndex_AsSsize_t(__pyx_v_i); if (unlikely((__pyx_t_10 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 151, __pyx_L1_error)
    (__pyx_v_psWOptions->padfDstNoDataImag[__pyx_t_10]) = 0.0;

    /* "rasterio/_warp.pyx":147
 *     psWOptions.padfDstNoDataReal = <double*>CPLMalloc(src_count * sizeof(double))
 *     psWOptions.padfDstNoDataImag = <double*>CPLMalloc(src_count * sizeof(double))
 *     for i in range(src_count):             # <<<<<<<<<<<<<<
 *         psWOptions.padfSrcNoDataReal[i] = src_nodata
 *         psWOptions.padfSrcNoDataImag[i] = 0.0
 */
  }
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "rasterio/_warp.pyx":155
 *     # Important: set back into struct or values set above are lost
 *     # This is because CSLSetNameValue returns a new list each time
 *     psWOptions.papszWarpOptions = warp_extras             # <<<<<<<<<<<<<<
 * 
 *     # Set up band info
 */
  __pyx_v_psWOptions->papszWarpOptions = __pyx_v_warp_extras;

  /* "rasterio/_warp.pyx":158
 * 
 *     # Set up band info
 *     if psWOptions.nBandCount == 0:             # <<<<<<<<<<<<<<
 *         psWOptions.nBandCount = src_count
 * 
 */
  __pyx_t_1 = ((__pyx_v_psWOptions->nBandCount == 0) != 0);
  if (__pyx_t_1) {

    /* "rasterio/_warp.pyx":159
 *     # Set up band info
 *     if psWOptions.nBandCount == 0:
 *         psWOptions.nBandCount = src_count             # <<<<<<<<<<<<<<
 * 
 *         psWOptions.panSrcBands = <int*>CPLMalloc(src_count * sizeof(int))
 */
    __pyx_v_psWOptions->nBandCount = __pyx_v_src_count;

    /* "rasterio/_warp.pyx":161
 *         psWOptions.nBandCount = src_count
 * 
 *         psWOptions.panSrcBands = <int*>CPLMalloc(src_count * sizeof(int))             # <<<<<<<<<<<<<<
 *         psWOptions.panDstBands = <int*>CPLMalloc(src_count * sizeof(int))
 * 
 */
    __pyx_v_psWOptions->panSrcBands = ((int *)CPLMalloc((__pyx_v_src_count * (sizeof(int)))));

    /* "rasterio/_warp.pyx":162
 * 
 *         psWOptions.panSrcBands = <int*>CPLMalloc(src_count * sizeof(int))
 *         psWOptions.panDstBands = <int*>CPLMalloc(src_count * sizeof(int))             # <<<<<<<<<<<<<<
 * 
 *         for i in range(src_count):
 */
    __pyx_v_psWOptions->panDstBands = ((int *)CPLMalloc((__pyx_v_src_count * (sizeof(int)))));

    /* "rasterio/_warp.pyx":164
 *         psWOptions.panDstBands = <int*>CPLMalloc(src_count * sizeof(int))
 * 
 *         for i in range(src_count):             # <<<<<<<<<<<<<<
 *             psWOptions.panSrcBands[i] = i + 1
 *             psWOptions.panDstBands[i] = i + 1
 */
    __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_src_count); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 164, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 164, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_6);
    __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_range, __pyx_t_4, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 164, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (likely(PyList_CheckExact(__pyx_t_6)) || PyTuple_CheckExact(__pyx_t_6)) {
      __pyx_t_4 = __pyx_t_6; __Pyx_INCREF(__pyx_t_4); __pyx_t_7 = 0;
      __pyx_t_8 = NULL;
    } else {
      __pyx_t_7 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 164, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_8 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 164, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    for (;;) {
      if (likely(!__pyx_t_8)) {
        if (likely(PyList_CheckExact(__pyx_t_4))) {
          if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_4)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_6 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_7); __Pyx_INCREF(__pyx_t_6); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 164, __pyx_L1_error)
          #else
          __pyx_t_6 = PySequence_ITEM(__pyx_t_4, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 164, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          #endif
        } else {
          if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_6 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_7); __Pyx_INCREF(__pyx_t_6); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 164, __pyx_L1_error)
          #else
          __pyx_t_6 = PySequence_ITEM(__pyx_t_4, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 164, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          #endif
        }
      } else {
        __pyx_t_6 = __pyx_t_8(__pyx_t_4);
        if (unlikely(!__pyx_t_6)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 164, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_6);
      }
      __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_6);
      __pyx_t_6 = 0;

      /* "rasterio/_warp.pyx":165
 * 
 *         for i in range(src_count):
 *             psWOptions.panSrcBands[i] = i + 1             # <<<<<<<<<<<<<<
 *             psWOptions.panDstBands[i] = i + 1
 * 
 */
      __pyx_t_6 = __Pyx_PyInt_AddObjC(__pyx_v_i, __pyx_int_1, 1, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 165, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_11 = __Pyx_PyInt_As_int(__pyx_t_6); if (unlikely((__pyx_t_11 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 165, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_10 = __Pyx_PyIndex_AsSsize_t(__pyx_v_i); if (unlikely((__pyx_t_10 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 165, __pyx_L1_error)
      (__pyx_v_psWOptions->panSrcBands[__pyx_t_10]) = __pyx_t_11;

      /* "rasterio/_warp.pyx":166
 *         for i in range(src_count):
 *             psWOptions.panSrcBands[i] = i + 1
 *             psWOptions.panDstBands[i] = i + 1             # <<<<<<<<<<<<<<
 * 
 *     return psWOptions
 */
      __pyx_t_6 = __Pyx_PyInt_AddObjC(__pyx_v_i, __pyx_int_1, 1, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 166, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_11 = __Pyx_PyInt_As_int(__pyx_t_6); if (unlikely((__pyx_t_11 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 166, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_10 = __Pyx_PyIndex_AsSsize_t(__pyx_v_i); if (unlikely((__pyx_t_10 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 166, __pyx_L1_error)
      (__pyx_v_psWOptions->panDstBands[__pyx_t_10]) = __pyx_t_11;

      /* "rasterio/_warp.pyx":164
 *         psWOptions.panDstBands = <int*>CPLMalloc(src_count * sizeof(int))
 * 
 *         for i in range(src_count):             # <<<<<<<<<<<<<<
 *             psWOptions.panSrcBands[i] = i + 1
 *             psWOptions.panDstBands[i] = i + 1
 */
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "rasterio/_warp.pyx":158
 * 
 *     # Set up band info
 *     if psWOptions.nBandCount == 0:             # <<<<<<<<<<<<<<
 *         psWOptions.nBandCount = src_count
 * 
 */
  }

  /* "rasterio/_warp.pyx":168
 *             psWOptions.panDstBands[i] = i + 1
 * 
 *     return psWOptions             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_psWOptions;
  goto __pyx_L0;

  /* "rasterio/_warp.pyx":99
 * 
 * 
 * cdef GDALWarpOptions * create_warp_options(             # <<<<<<<<<<<<<<
 *         GDALResampleAlg resampling, object src_nodata, object dst_nodata,
 *         int src_count, const char **options) except NULL:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("rasterio._warp.create_warp_options", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XDECREF(__pyx_v_src_nodata);
  __Pyx_XDECREF(__pyx_v_dst_nodata);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "rasterio/_warp.pyx":171
 * 
 * 
 * def _reproject(             # <<<<<<<<<<<<<<
 *         source, destination,
 *         src_transform=None,
 */

/* Python wrapper */
static PyObject *__pyx_pw_8rasterio_5_warp_5_reproject(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_8rasterio_5_warp_4_reproject[] = "\n    Reproject a source raster to a destination raster.\n\n    If the source and destination are ndarrays, coordinate reference\n    system definitions and affine transformation parameters are required\n    for reprojection.\n\n    If the source and destination are rasterio Bands, shorthand for\n    bands of datasets on disk, the coordinate reference systems and\n    transforms will be read from the appropriate datasets.\n\n    Parameters\n    ------------\n    source: ndarray or rasterio Band\n        Source raster.\n    destination: ndarray or rasterio Band\n        Target raster.\n    src_transform: affine.Affine(), optional\n        Source affine transformation.  Required if source and destination\n        are ndarrays.  Will be derived from source if it is a rasterio Band.\n    gcps: sequence of `GroundControlPoint` instances, optional\n        Ground control points for the source. May be used in place of\n        src_transform.\n    src_crs: dict, optional\n        Source coordinate reference system, in rasterio dict format.\n        Required if source and destination are ndarrays.\n        Will be derived from source if it is a rasterio Band.\n        Example: {'init': 'EPSG:4326'}\n    src_nodata: int or float, optional\n        The source nodata value.  Pixels with this value will not be used\n        for interpolation.  If not set, it will be default to the\n        nodata value of the source image if a masked ndarray or rasterio band,\n        if available.  Must be provided if dst_nodata is not None.\n    dst_transform: affine.Affine(), optional\n        Target affine transformation.  Required if source and destination\n        are ndarrays.  Will be derived from target if it is a rasterio Band.\n    dst_crs: dict, optional\n        Target coordinate reference system.  Required if source and destination\n        are ndarrays.  Will be derived from target if it is a rasterio Band.\n    dst_nodata: int or float, optional\n        The nodata value used to ""initialize the destination; it will remain\n        in all areas not covered by the reprojected source.  Defaults to the\n        nodata value of the destination image (if set), the value of\n        src_nodata, or 0 (gdal default).\n    resampling: int\n        Resampling method to use.  One of the following:\n            Resampling.nearest,\n            Resampling.bilinear,\n            Resampling.cubic,\n            Resampling.cubic_spline,\n            Resampling.lanczos,\n            Resampling.average,\n            Resampling.mode\n    init_dest_nodata: bool\n        Flag to specify initialization of nodata in destination;\n        prevents overwrite of previous warps. Defaults to True.\n    num_threads: int\n        Number of worker threads.\n    kwargs:  dict, optional\n        Additional arguments passed to both the image to image\n        transformer GDALCreateGenImgProjTransformer2() (for example,\n        MAX_GCP_ORDER=2) and to the Warper (for example,\n        INIT_DEST=NO_DATA).\n\n    Returns\n    ---------\n    out: None\n        Output is written to destination.\n    ";
static PyMethodDef __pyx_mdef_8rasterio_5_warp_5_reproject = {"_reproject", (PyCFunction)__pyx_pw_8rasterio_5_warp_5_reproject, METH_VARARGS|METH_KEYWORDS, __pyx_doc_8rasterio_5_warp_4_reproject};
static PyObject *__pyx_pw_8rasterio_5_warp_5_reproject(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_source = 0;
  PyObject *__pyx_v_destination = 0;
  PyObject *__pyx_v_src_transform = 0;
  PyObject *__pyx_v_gcps = 0;
  PyObject *__pyx_v_src_crs = 0;
  PyObject *__pyx_v_src_nodata = 0;
  PyObject *__pyx_v_dst_transform = 0;
  PyObject *__pyx_v_dst_crs = 0;
  PyObject *__pyx_v_dst_nodata = 0;
  PyObject *__pyx_v_resampling = 0;
  PyObject *__pyx_v_init_dest_nodata = 0;
  PyObject *__pyx_v_num_threads = 0;
  PyObject *__pyx_v_kwargs = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_reproject (wrapper)", 0);
  __pyx_v_kwargs = PyDict_New(); if (unlikely(!__pyx_v_kwargs)) return NULL;
  __Pyx_GOTREF(__pyx_v_kwargs);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_source,&__pyx_n_s_destination,&__pyx_n_s_src_transform,&__pyx_n_s_gcps,&__pyx_n_s_src_crs,&__pyx_n_s_src_nodata,&__pyx_n_s_dst_transform,&__pyx_n_s_dst_crs,&__pyx_n_s_dst_nodata,&__pyx_n_s_resampling,&__pyx_n_s_init_dest_nodata,&__pyx_n_s_num_threads,0};
    PyObject* values[12] = {0,0,0,0,0,0,0,0,0,0,0,0};

    /* "rasterio/_warp.pyx":173
 * def _reproject(
 *         source, destination,
 *         src_transform=None,             # <<<<<<<<<<<<<<
 *         gcps=None,
 *         src_crs=None,
 */
    values[2] = ((PyObject *)Py_None);

    /* "rasterio/_warp.pyx":174
 *         source, destination,
 *         src_transform=None,
 *         gcps=None,             # <<<<<<<<<<<<<<
 *         src_crs=None,
 *         src_nodata=None,
 */
    values[3] = ((PyObject *)Py_None);

    /* "rasterio/_warp.pyx":175
 *         src_transform=None,
 *         gcps=None,
 *         src_crs=None,             # <<<<<<<<<<<<<<
 *         src_nodata=None,
 *         dst_transform=None,
 */
    values[4] = ((PyObject *)Py_None);

    /* "rasterio/_warp.pyx":176
 *         gcps=None,
 *         src_crs=None,
 *         src_nodata=None,             # <<<<<<<<<<<<<<
 *         dst_transform=None,
 *         dst_crs=None,
 */
    values[5] = ((PyObject *)Py_None);

    /* "rasterio/_warp.pyx":177
 *         src_crs=None,
 *         src_nodata=None,
 *         dst_transform=None,             # <<<<<<<<<<<<<<
 *         dst_crs=None,
 *         dst_nodata=None,
 */
    values[6] = ((PyObject *)Py_None);

    /* "rasterio/_warp.pyx":178
 *         src_nodata=None,
 *         dst_transform=None,
 *         dst_crs=None,             # <<<<<<<<<<<<<<
 *         dst_nodata=None,
 *         resampling=Resampling.nearest,
 */
    values[7] = ((PyObject *)Py_None);

    /* "rasterio/_warp.pyx":179
 *         dst_transform=None,
 *         dst_crs=None,
 *         dst_nodata=None,             # <<<<<<<<<<<<<<
 *         resampling=Resampling.nearest,
 *         init_dest_nodata=True,
 */
    values[8] = ((PyObject *)Py_None);
    values[9] = __pyx_k__3;

    /* "rasterio/_warp.pyx":181
 *         dst_nodata=None,
 *         resampling=Resampling.nearest,
 *         init_dest_nodata=True,             # <<<<<<<<<<<<<<
 *         num_threads=1,
 *         **kwargs):
 */
    values[10] = ((PyObject *)Py_True);
    values[11] = ((PyObject *)__pyx_int_1);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_source)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_destination)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_reproject", 0, 2, 12, 1); __PYX_ERR(0, 171, __pyx_L3_error)
        }
        case  2:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_src_transform);
          if (value) { values[2] = value; kw_args--; }
        }
        case  3:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_gcps);
          if (value) { values[3] = value; kw_args--; }
        }
        case  4:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_src_crs);
          if (value) { values[4] = value; kw_args--; }
        }
        case  5:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_src_nodata);
          if (value) { values[5] = value; kw_args--; }
        }
        case  6:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dst_transform);
          if (value) { values[6] = value; kw_args--; }
        }
        case  7:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dst_crs);
          if (value) { values[7] = value; kw_args--; }
        }
        case  8:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dst_nodata);
          if (value) { values[8] = value; kw_args--; }
        }
        case  9:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_resampling);
          if (value) { values[9] = value; kw_args--; }
        }
        case 10:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_init_dest_nodata);
          if (value) { values[10] = value; kw_args--; }
        }
        case 11:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_num_threads);
          if (value) { values[11] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, __pyx_v_kwargs, values, pos_args, "_reproject") < 0)) __PYX_ERR(0, 171, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_source = values[0];
    __pyx_v_destination = values[1];
    __pyx_v_src_transform = values[2];
    __pyx_v_gcps = values[3];
    __pyx_v_src_crs = values[4];
    __pyx_v_src_nodata = values[5];
    __pyx_v_dst_transform = values[6];
    __pyx_v_dst_crs = values[7];
    __pyx_v_dst_nodata = values[8];
    __pyx_v_resampling = values[9];
    __pyx_v_init_dest_nodata = values[10];
    __pyx_v_num_threads = values[11];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_reproject", 0, 2, 12, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 171, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_DECREF(__pyx_v_kwargs); __pyx_v_kwargs = 0;
  __Pyx_AddTraceback("rasterio._warp._reproject", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8rasterio_5_warp_4_reproject(__pyx_self, __pyx_v_source, __pyx_v_destination, __pyx_v_src_transform, __pyx_v_gcps, __pyx_v_src_crs, __pyx_v_src_nodata, __pyx_v_dst_transform, __pyx_v_dst_crs, __pyx_v_dst_nodata, __pyx_v_resampling, __pyx_v_init_dest_nodata, __pyx_v_num_threads, __pyx_v_kwargs);

  /* "rasterio/_warp.pyx":171
 * 
 * 
 * def _reproject(             # <<<<<<<<<<<<<<
 *         source, destination,
 *         src_transform=None,
 */

  /* function exit code */
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8rasterio_5_warp_4_reproject(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_source, PyObject *__pyx_v_destination, PyObject *__pyx_v_src_transform, PyObject *__pyx_v_gcps, PyObject *__pyx_v_src_crs, PyObject *__pyx_v_src_nodata, PyObject *__pyx_v_dst_transform, PyObject *__pyx_v_dst_crs, PyObject *__pyx_v_dst_nodata, PyObject *__pyx_v_resampling, PyObject *__pyx_v_init_dest_nodata, PyObject *__pyx_v_num_threads, PyObject *__pyx_v_kwargs) {
  int __pyx_v_retval;
  int __pyx_v_rows;
  int __pyx_v_cols;
  int __pyx_v_src_count;
  GDALDriverH __pyx_v_driver;
  GDALDatasetH __pyx_v_src_dataset;
  GDALDatasetH __pyx_v_dst_dataset;
  double __pyx_v_gt[6];
  char *__pyx_v_srcwkt;
  char *__pyx_v_dstwkt;
  OGRSpatialReferenceH __pyx_v_osr;
  char **__pyx_v_warp_extras;
  CYTHON_UNUSED char const *__pyx_v_pszWarpThread;
  int __pyx_v_i;
  double __pyx_v_tolerance;
  GDAL_GCP *__pyx_v_gcplist;
  void *__pyx_v_hTransformArg;
  GDALTransformerFunc __pyx_v_pfnTransformer;
  GDALWarpOptions *__pyx_v_psWOptions;
  PyObject *__pyx_v_src_bidx = NULL;
  PyObject *__pyx_v_dtype = NULL;
  PyObject *__pyx_v_datasetname = NULL;
  PyObject *__pyx_v_obj = NULL;
  PyObject *__pyx_v_ident = NULL;
  PyObject *__pyx_v_info = NULL;
  PyObject *__pyx_v_rdr = NULL;
  PyObject *__pyx_v_shape = NULL;
  PyObject *__pyx_v_dst_bidx = NULL;
  CYTHON_UNUSED PyObject *__pyx_v__ = NULL;
  PyObject *__pyx_v_udr = NULL;
  char **__pyx_v_imgProjOptions;
  PyObject *__pyx_v_key = NULL;
  PyObject *__pyx_v_val = NULL;
  PyObject *__pyx_v_valb = NULL;
  GDALWarpOperation __pyx_v_oWarper;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  void *__pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  char const *__pyx_t_15;
  GDALDataType __pyx_t_16;
  double __pyx_t_17;
  OGRSpatialReferenceH __pyx_t_18;
  int __pyx_t_19;
  char const *__pyx_t_20;
  PyObject *__pyx_t_21 = NULL;
  PyObject *__pyx_t_22 = NULL;
  PyObject *__pyx_t_23 = NULL;
  PyObject *(*__pyx_t_24)(PyObject *);
  char *__pyx_t_25;
  double __pyx_t_26;
  char const *__pyx_t_27;
  PyObject *(*__pyx_t_28)(PyObject *);
  GDALDatasetH __pyx_t_29;
  char const *__pyx_t_30;
  char const *__pyx_t_31;
  char const *__pyx_t_32;
  char const *__pyx_t_33;
  char const *__pyx_t_34;
  char const *__pyx_t_35;
  char const *__pyx_t_36;
  GDALResampleAlg __pyx_t_37;
  GDALWarpOptions *__pyx_t_38;
  int __pyx_t_39;
  PyObject *__pyx_t_40 = NULL;
  char const *__pyx_t_41;
  __Pyx_RefNannySetupContext("_reproject", 0);
  __Pyx_INCREF(__pyx_v_source);
  __Pyx_INCREF(__pyx_v_destination);
  __Pyx_INCREF(__pyx_v_src_nodata);
  __Pyx_INCREF(__pyx_v_dst_nodata);

  /* "rasterio/_warp.pyx":257
 *     cdef int cols
 *     cdef int src_count
 *     cdef GDALDriverH driver = NULL             # <<<<<<<<<<<<<<
 *     cdef GDALDatasetH src_dataset = NULL
 *     cdef GDALDatasetH dst_dataset = NULL
 */
  __pyx_v_driver = NULL;

  /* "rasterio/_warp.pyx":258
 *     cdef int src_count
 *     cdef GDALDriverH driver = NULL
 *     cdef GDALDatasetH src_dataset = NULL             # <<<<<<<<<<<<<<
 *     cdef GDALDatasetH dst_dataset = NULL
 *     cdef GDALAccess GA
 */
  __pyx_v_src_dataset = NULL;

  /* "rasterio/_warp.pyx":259
 *     cdef GDALDriverH driver = NULL
 *     cdef GDALDatasetH src_dataset = NULL
 *     cdef GDALDatasetH dst_dataset = NULL             # <<<<<<<<<<<<<<
 *     cdef GDALAccess GA
 *     cdef double gt[6]
 */
  __pyx_v_dst_dataset = NULL;

  /* "rasterio/_warp.pyx":262
 *     cdef GDALAccess GA
 *     cdef double gt[6]
 *     cdef char *srcwkt = NULL             # <<<<<<<<<<<<<<
 *     cdef char *dstwkt= NULL
 *     cdef OGRSpatialReferenceH osr = NULL
 */
  __pyx_v_srcwkt = NULL;

  /* "rasterio/_warp.pyx":263
 *     cdef double gt[6]
 *     cdef char *srcwkt = NULL
 *     cdef char *dstwkt= NULL             # <<<<<<<<<<<<<<
 *     cdef OGRSpatialReferenceH osr = NULL
 *     cdef char **warp_extras = NULL
 */
  __pyx_v_dstwkt = NULL;

  /* "rasterio/_warp.pyx":264
 *     cdef char *srcwkt = NULL
 *     cdef char *dstwkt= NULL
 *     cdef OGRSpatialReferenceH osr = NULL             # <<<<<<<<<<<<<<
 *     cdef char **warp_extras = NULL
 *     cdef const char* pszWarpThread = NULL
 */
  __pyx_v_osr = NULL;

  /* "rasterio/_warp.pyx":265
 *     cdef char *dstwkt= NULL
 *     cdef OGRSpatialReferenceH osr = NULL
 *     cdef char **warp_extras = NULL             # <<<<<<<<<<<<<<
 *     cdef const char* pszWarpThread = NULL
 *     cdef int i
 */
  __pyx_v_warp_extras = NULL;

  /* "rasterio/_warp.pyx":266
 *     cdef OGRSpatialReferenceH osr = NULL
 *     cdef char **warp_extras = NULL
 *     cdef const char* pszWarpThread = NULL             # <<<<<<<<<<<<<<
 *     cdef int i
 *     cdef double tolerance = 0.125
 */
  __pyx_v_pszWarpThread = NULL;

  /* "rasterio/_warp.pyx":268
 *     cdef const char* pszWarpThread = NULL
 *     cdef int i
 *     cdef double tolerance = 0.125             # <<<<<<<<<<<<<<
 *     cdef GDAL_GCP *gcplist = NULL
 *     cdef void *hTransformArg = NULL
 */
  __pyx_v_tolerance = 0.125;

  /* "rasterio/_warp.pyx":269
 *     cdef int i
 *     cdef double tolerance = 0.125
 *     cdef GDAL_GCP *gcplist = NULL             # <<<<<<<<<<<<<<
 *     cdef void *hTransformArg = NULL
 *     cdef GDALTransformerFunc pfnTransformer = NULL
 */
  __pyx_v_gcplist = NULL;

  /* "rasterio/_warp.pyx":270
 *     cdef double tolerance = 0.125
 *     cdef GDAL_GCP *gcplist = NULL
 *     cdef void *hTransformArg = NULL             # <<<<<<<<<<<<<<
 *     cdef GDALTransformerFunc pfnTransformer = NULL
 *     cdef GDALWarpOptions *psWOptions = NULL
 */
  __pyx_v_hTransformArg = NULL;

  /* "rasterio/_warp.pyx":271
 *     cdef GDAL_GCP *gcplist = NULL
 *     cdef void *hTransformArg = NULL
 *     cdef GDALTransformerFunc pfnTransformer = NULL             # <<<<<<<<<<<<<<
 *     cdef GDALWarpOptions *psWOptions = NULL
 * 
 */
  __pyx_v_pfnTransformer = NULL;

  /* "rasterio/_warp.pyx":272
 *     cdef void *hTransformArg = NULL
 *     cdef GDALTransformerFunc pfnTransformer = NULL
 *     cdef GDALWarpOptions *psWOptions = NULL             # <<<<<<<<<<<<<<
 * 
 *     # Validate nodata values immediately.
 */
  __pyx_v_psWOptions = NULL;

  /* "rasterio/_warp.pyx":275
 * 
 *     # Validate nodata values immediately.
 *     if src_nodata is not None:             # <<<<<<<<<<<<<<
 *         if not in_dtype_range(src_nodata, source.dtype):
 *             raise ValueError("src_nodata must be in valid range for "
 */
  __pyx_t_1 = (__pyx_v_src_nodata != Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "rasterio/_warp.pyx":276
 *     # Validate nodata values immediately.
 *     if src_nodata is not None:
 *         if not in_dtype_range(src_nodata, source.dtype):             # <<<<<<<<<<<<<<
 *             raise ValueError("src_nodata must be in valid range for "
 *                              "source dtype")
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_source, __pyx_n_s_dtype); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 276, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = ((!(__pyx_f_8rasterio_3_io_in_dtype_range(__pyx_v_src_nodata, __pyx_t_3) != 0)) != 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__pyx_t_2) {

      /* "rasterio/_warp.pyx":277
 *     if src_nodata is not None:
 *         if not in_dtype_range(src_nodata, source.dtype):
 *             raise ValueError("src_nodata must be in valid range for "             # <<<<<<<<<<<<<<
 *                              "source dtype")
 * 
 */
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__4, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 277, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(0, 277, __pyx_L1_error)

      /* "rasterio/_warp.pyx":276
 *     # Validate nodata values immediately.
 *     if src_nodata is not None:
 *         if not in_dtype_range(src_nodata, source.dtype):             # <<<<<<<<<<<<<<
 *             raise ValueError("src_nodata must be in valid range for "
 *                              "source dtype")
 */
    }

    /* "rasterio/_warp.pyx":275
 * 
 *     # Validate nodata values immediately.
 *     if src_nodata is not None:             # <<<<<<<<<<<<<<
 *         if not in_dtype_range(src_nodata, source.dtype):
 *             raise ValueError("src_nodata must be in valid range for "
 */
  }

  /* "rasterio/_warp.pyx":280
 *                              "source dtype")
 * 
 *     if dst_nodata is not None:             # <<<<<<<<<<<<<<
 *         if not in_dtype_range(dst_nodata, destination.dtype):
 *             raise ValueError("dst_nodata must be in valid range for "
 */
  __pyx_t_2 = (__pyx_v_dst_nodata != Py_None);
  __pyx_t_1 = (__pyx_t_2 != 0);
  if (__pyx_t_1) {

    /* "rasterio/_warp.pyx":281
 * 
 *     if dst_nodata is not None:
 *         if not in_dtype_range(dst_nodata, destination.dtype):             # <<<<<<<<<<<<<<
 *             raise ValueError("dst_nodata must be in valid range for "
 *                              "destination dtype")
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_destination, __pyx_n_s_dtype); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 281, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = ((!(__pyx_f_8rasterio_3_io_in_dtype_range(__pyx_v_dst_nodata, __pyx_t_3) != 0)) != 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__pyx_t_1) {

      /* "rasterio/_warp.pyx":282
 *     if dst_nodata is not None:
 *         if not in_dtype_range(dst_nodata, destination.dtype):
 *             raise ValueError("dst_nodata must be in valid range for "             # <<<<<<<<<<<<<<
 *                              "destination dtype")
 * 
 */
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 282, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(0, 282, __pyx_L1_error)

      /* "rasterio/_warp.pyx":281
 * 
 *     if dst_nodata is not None:
 *         if not in_dtype_range(dst_nodata, destination.dtype):             # <<<<<<<<<<<<<<
 *             raise ValueError("dst_nodata must be in valid range for "
 *                              "destination dtype")
 */
    }

    /* "rasterio/_warp.pyx":280
 *                              "source dtype")
 * 
 *     if dst_nodata is not None:             # <<<<<<<<<<<<<<
 *         if not in_dtype_range(dst_nodata, destination.dtype):
 *             raise ValueError("dst_nodata must be in valid range for "
 */
  }

  /* "rasterio/_warp.pyx":288
 *     # We need a src_transform and src_dst in this case. These will
 *     # be copied to the MEM dataset.
 *     if dtypes.is_ndarray(source):             # <<<<<<<<<<<<<<
 *         # Convert 2D single-band arrays to 3D multi-band.
 *         if len(source.shape) == 2:
 */
  __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_dtypes); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 288, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_is_ndarray); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 288, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  if (!__pyx_t_4) {
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_source); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 288, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_source};
      __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 288, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_3);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_source};
      __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 288, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_3);
    } else
    #endif
    {
      __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 288, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;
      __Pyx_INCREF(__pyx_v_source);
      __Pyx_GIVEREF(__pyx_v_source);
      PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_source);
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 288, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 288, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_1) {

    /* "rasterio/_warp.pyx":290
 *     if dtypes.is_ndarray(source):
 *         # Convert 2D single-band arrays to 3D multi-band.
 *         if len(source.shape) == 2:             # <<<<<<<<<<<<<<
 *             source = source.reshape(1, *source.shape)
 *         src_count = source.shape[0]
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_source, __pyx_n_s_shape); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 290, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_7 = PyObject_Length(__pyx_t_3); if (unlikely(__pyx_t_7 == -1)) __PYX_ERR(0, 290, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_1 = ((__pyx_t_7 == 2) != 0);
    if (__pyx_t_1) {

      /* "rasterio/_warp.pyx":291
 *         # Convert 2D single-band arrays to 3D multi-band.
 *         if len(source.shape) == 2:
 *             source = source.reshape(1, *source.shape)             # <<<<<<<<<<<<<<
 *         src_count = source.shape[0]
 *         src_bidx = range(1, src_count + 1)
 */
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_source, __pyx_n_s_reshape); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 291, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_source, __pyx_n_s_shape); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 291, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = PySequence_Tuple(__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 291, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = PyNumber_Add(__pyx_tuple__6, __pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 291, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 291, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF_SET(__pyx_v_source, __pyx_t_6);
      __pyx_t_6 = 0;

      /* "rasterio/_warp.pyx":290
 *     if dtypes.is_ndarray(source):
 *         # Convert 2D single-band arrays to 3D multi-band.
 *         if len(source.shape) == 2:             # <<<<<<<<<<<<<<
 *             source = source.reshape(1, *source.shape)
 *         src_count = source.shape[0]
 */
    }

    /* "rasterio/_warp.pyx":292
 *         if len(source.shape) == 2:
 *             source = source.reshape(1, *source.shape)
 *         src_count = source.shape[0]             # <<<<<<<<<<<<<<
 *         src_bidx = range(1, src_count + 1)
 *         rows = source.shape[1]
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_source, __pyx_n_s_shape); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 292, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_5 = __Pyx_GetItemInt(__pyx_t_6, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 292, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_8 = __Pyx_PyInt_As_int(__pyx_t_5); if (unlikely((__pyx_t_8 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 292, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_src_count = __pyx_t_8;

    /* "rasterio/_warp.pyx":293
 *             source = source.reshape(1, *source.shape)
 *         src_count = source.shape[0]
 *         src_bidx = range(1, src_count + 1)             # <<<<<<<<<<<<<<
 *         rows = source.shape[1]
 *         cols = source.shape[2]
 */
    __pyx_t_5 = __Pyx_PyInt_From_long((__pyx_v_src_count + 1)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 293, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 293, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF(__pyx_int_1);
    __Pyx_GIVEREF(__pyx_int_1);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_int_1);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_5);
    __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_builtin_range, __pyx_t_6, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 293, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_v_src_bidx = __pyx_t_5;
    __pyx_t_5 = 0;

    /* "rasterio/_warp.pyx":294
 *         src_count = source.shape[0]
 *         src_bidx = range(1, src_count + 1)
 *         rows = source.shape[1]             # <<<<<<<<<<<<<<
 *         cols = source.shape[2]
 *         dtype = np.dtype(source.dtype).name
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_source, __pyx_n_s_shape); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 294, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_GetItemInt(__pyx_t_5, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 294, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_8 = __Pyx_PyInt_As_int(__pyx_t_6); if (unlikely((__pyx_t_8 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 294, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_v_rows = __pyx_t_8;

    /* "rasterio/_warp.pyx":295
 *         src_bidx = range(1, src_count + 1)
 *         rows = source.shape[1]
 *         cols = source.shape[2]             # <<<<<<<<<<<<<<
 *         dtype = np.dtype(source.dtype).name
 * 
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_source, __pyx_n_s_shape); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 295, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_5 = __Pyx_GetItemInt(__pyx_t_6, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 295, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_8 = __Pyx_PyInt_As_int(__pyx_t_5); if (unlikely((__pyx_t_8 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 295, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_cols = __pyx_t_8;

    /* "rasterio/_warp.pyx":296
 *         rows = source.shape[1]
 *         cols = source.shape[2]
 *         dtype = np.dtype(source.dtype).name             # <<<<<<<<<<<<<<
 * 
 *         if src_nodata is None and hasattr(source, 'fill_value'):
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_source, __pyx_n_s_dtype); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 296, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 296, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5);
    __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_5numpy_dtype), __pyx_t_6, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 296, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_name); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 296, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_dtype = __pyx_t_6;
    __pyx_t_6 = 0;

    /* "rasterio/_warp.pyx":298
 *         dtype = np.dtype(source.dtype).name
 * 
 *         if src_nodata is None and hasattr(source, 'fill_value'):             # <<<<<<<<<<<<<<
 *             # source is a masked array
 *             src_nodata = source.fill_value
 */
    __pyx_t_2 = (__pyx_v_src_nodata == Py_None);
    __pyx_t_9 = (__pyx_t_2 != 0);
    if (__pyx_t_9) {
    } else {
      __pyx_t_1 = __pyx_t_9;
      goto __pyx_L10_bool_binop_done;
    }
    __pyx_t_9 = PyObject_HasAttr(__pyx_v_source, __pyx_n_s_fill_value); if (unlikely(__pyx_t_9 == -1)) __PYX_ERR(0, 298, __pyx_L1_error)
    __pyx_t_2 = (__pyx_t_9 != 0);
    __pyx_t_1 = __pyx_t_2;
    __pyx_L10_bool_binop_done:;
    if (__pyx_t_1) {

      /* "rasterio/_warp.pyx":300
 *         if src_nodata is None and hasattr(source, 'fill_value'):
 *             # source is a masked array
 *             src_nodata = source.fill_value             # <<<<<<<<<<<<<<
 * 
 *         try:
 */
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_source, __pyx_n_s_fill_value); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 300, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF_SET(__pyx_v_src_nodata, __pyx_t_6);
      __pyx_t_6 = 0;

      /* "rasterio/_warp.pyx":298
 *         dtype = np.dtype(source.dtype).name
 * 
 *         if src_nodata is None and hasattr(source, 'fill_value'):             # <<<<<<<<<<<<<<
 *             # source is a masked array
 *             src_nodata = source.fill_value
 */
    }

    /* "rasterio/_warp.pyx":302
 *             src_nodata = source.fill_value
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             driver = exc_wrap_pointer(GDALGetDriverByName("MEM"))
 *         except:
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_10, &__pyx_t_11, &__pyx_t_12);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_12);
      /*try:*/ {

        /* "rasterio/_warp.pyx":303
 * 
 *         try:
 *             driver = exc_wrap_pointer(GDALGetDriverByName("MEM"))             # <<<<<<<<<<<<<<
 *         except:
 *             raise DriverRegistrationError(
 */
        __pyx_t_13 = __pyx_f_8rasterio_4_err_exc_wrap_pointer(GDALGetDriverByName(((char const *)"MEM"))); if (unlikely(__pyx_t_13 == NULL)) __PYX_ERR(0, 303, __pyx_L12_error)
        __pyx_v_driver = __pyx_t_13;

        /* "rasterio/_warp.pyx":302
 *             src_nodata = source.fill_value
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             driver = exc_wrap_pointer(GDALGetDriverByName("MEM"))
 *         except:
 */
      }
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      goto __pyx_L19_try_end;
      __pyx_L12_error:;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

      /* "rasterio/_warp.pyx":304
 *         try:
 *             driver = exc_wrap_pointer(GDALGetDriverByName("MEM"))
 *         except:             # <<<<<<<<<<<<<<
 *             raise DriverRegistrationError(
 *                 "'MEM' driver not found. Check that this call is contained "
 */
      /*except:*/ {
        __Pyx_AddTraceback("rasterio._warp._reproject", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_6, &__pyx_t_5, &__pyx_t_3) < 0) __PYX_ERR(0, 304, __pyx_L14_except_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GOTREF(__pyx_t_3);

        /* "rasterio/_warp.pyx":305
 *             driver = exc_wrap_pointer(GDALGetDriverByName("MEM"))
 *         except:
 *             raise DriverRegistrationError(             # <<<<<<<<<<<<<<
 *                 "'MEM' driver not found. Check that this call is contained "
 *                 "in a `with rasterio.Env()` or `with rasterio.open()` "
 */
        __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_DriverRegistrationError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 305, __pyx_L14_except_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_14 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_tuple__7, NULL); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 305, __pyx_L14_except_error)
        __Pyx_GOTREF(__pyx_t_14);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_Raise(__pyx_t_14, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
        __PYX_ERR(0, 305, __pyx_L14_except_error)
      }
      __pyx_L14_except_error:;

      /* "rasterio/_warp.pyx":302
 *             src_nodata = source.fill_value
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             driver = exc_wrap_pointer(GDALGetDriverByName("MEM"))
 *         except:
 */
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_11);
      __Pyx_XGIVEREF(__pyx_t_12);
      __Pyx_ExceptionReset(__pyx_t_10, __pyx_t_11, __pyx_t_12);
      goto __pyx_L1_error;
      __pyx_L19_try_end:;
    }

    /* "rasterio/_warp.pyx":310
 *                 "block.")
 * 
 *         datasetname = str(uuid.uuid4()).encode('utf-8')             # <<<<<<<<<<<<<<
 *         src_dataset = exc_wrap_pointer(
 *             GDALCreate(driver, <const char *>datasetname, cols, rows,
 */
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_uuid); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 310, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_uuid4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 310, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
      }
    }
    if (__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 310, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else {
      __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 310, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 310, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3);
    __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_Call(((PyObject *)(&PyString_Type)), __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 310, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_encode); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 310, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_tuple__8, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 310, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_v_datasetname = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "rasterio/_warp.pyx":312
 *         datasetname = str(uuid.uuid4()).encode('utf-8')
 *         src_dataset = exc_wrap_pointer(
 *             GDALCreate(driver, <const char *>datasetname, cols, rows,             # <<<<<<<<<<<<<<
 *                        src_count, dtypes.dtype_rev[dtype], NULL))
 * 
 */
    __pyx_t_15 = __Pyx_PyObject_AsString(__pyx_v_datasetname); if (unlikely((!__pyx_t_15) && PyErr_Occurred())) __PYX_ERR(0, 312, __pyx_L1_error)

    /* "rasterio/_warp.pyx":313
 *         src_dataset = exc_wrap_pointer(
 *             GDALCreate(driver, <const char *>datasetname, cols, rows,
 *                        src_count, dtypes.dtype_rev[dtype], NULL))             # <<<<<<<<<<<<<<
 * 
 *         GDALSetDescription(
 */
    __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_dtypes); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 313, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_dtype_rev); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 313, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyObject_GetItem(__pyx_t_6, __pyx_v_dtype); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 313, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_16 = ((GDALDataType)__Pyx_PyInt_As_GDALDataType(__pyx_t_3)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 313, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "rasterio/_warp.pyx":311
 * 
 *         datasetname = str(uuid.uuid4()).encode('utf-8')
 *         src_dataset = exc_wrap_pointer(             # <<<<<<<<<<<<<<
 *             GDALCreate(driver, <const char *>datasetname, cols, rows,
 *                        src_count, dtypes.dtype_rev[dtype], NULL))
 */
    __pyx_t_13 = __pyx_f_8rasterio_4_err_exc_wrap_pointer(GDALCreate(__pyx_v_driver, ((char const *)__pyx_t_15), __pyx_v_cols, __pyx_v_rows, __pyx_v_src_count, __pyx_t_16, NULL)); if (unlikely(__pyx_t_13 == NULL)) __PYX_ERR(0, 311, __pyx_L1_error)
    __pyx_v_src_dataset = __pyx_t_13;

    /* "rasterio/_warp.pyx":315
 *                        src_count, dtypes.dtype_rev[dtype], NULL))
 * 
 *         GDALSetDescription(             # <<<<<<<<<<<<<<
 *             src_dataset, "Temporary source dataset for _reproject()")
 * 
 */
    GDALSetDescription(__pyx_v_src_dataset, ((char const *)"Temporary source dataset for _reproject()"));

    /* "rasterio/_warp.pyx":318
 *             src_dataset, "Temporary source dataset for _reproject()")
 * 
 *         log.debug("Created temp source dataset")             # <<<<<<<<<<<<<<
 * 
 *         if src_transform:
 */
    __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_log); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 318, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_debug); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 318, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_tuple__9, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 318, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "rasterio/_warp.pyx":320
 *         log.debug("Created temp source dataset")
 * 
 *         if src_transform:             # <<<<<<<<<<<<<<
 *             for i in range(6):
 *                 gt[i] = src_transform[i]
 */
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_src_transform); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 320, __pyx_L1_error)
    if (__pyx_t_1) {

      /* "rasterio/_warp.pyx":321
 * 
 *         if src_transform:
 *             for i in range(6):             # <<<<<<<<<<<<<<
 *                 gt[i] = src_transform[i]
 *             retval = GDALSetGeoTransform(src_dataset, gt)
 */
      for (__pyx_t_8 = 0; __pyx_t_8 < 6; __pyx_t_8+=1) {
        __pyx_v_i = __pyx_t_8;

        /* "rasterio/_warp.pyx":322
 *         if src_transform:
 *             for i in range(6):
 *                 gt[i] = src_transform[i]             # <<<<<<<<<<<<<<
 *             retval = GDALSetGeoTransform(src_dataset, gt)
 * 
 */
        __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_src_transform, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 322, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_17 = __pyx_PyFloat_AsDouble(__pyx_t_3); if (unlikely((__pyx_t_17 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 322, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        (__pyx_v_gt[__pyx_v_i]) = __pyx_t_17;
      }

      /* "rasterio/_warp.pyx":323
 *             for i in range(6):
 *                 gt[i] = src_transform[i]
 *             retval = GDALSetGeoTransform(src_dataset, gt)             # <<<<<<<<<<<<<<
 * 
 *             log.debug("Set transform on temp source dataset: %d", retval)
 */
      __pyx_v_retval = GDALSetGeoTransform(__pyx_v_src_dataset, __pyx_v_gt);

      /* "rasterio/_warp.pyx":325
 *             retval = GDALSetGeoTransform(src_dataset, gt)
 * 
 *             log.debug("Set transform on temp source dataset: %d", retval)             # <<<<<<<<<<<<<<
 * 
 *             try:
 */
      __pyx_t_6 = __Pyx_GetModuleGlobalName(__pyx_n_s_log); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 325, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_debug); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 325, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_retval); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 325, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_14 = NULL;
      __pyx_t_8 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_14)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_14);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
          __pyx_t_8 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_5)) {
        PyObject *__pyx_temp[3] = {__pyx_t_14, __pyx_kp_s_Set_transform_on_temp_source_dat, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 325, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
        PyObject *__pyx_temp[3] = {__pyx_t_14, __pyx_kp_s_Set_transform_on_temp_source_dat, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 325, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      {
        __pyx_t_4 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 325, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        if (__pyx_t_14) {
          __Pyx_GIVEREF(__pyx_t_14); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_14); __pyx_t_14 = NULL;
        }
        __Pyx_INCREF(__pyx_kp_s_Set_transform_on_temp_source_dat);
        __Pyx_GIVEREF(__pyx_kp_s_Set_transform_on_temp_source_dat);
        PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_8, __pyx_kp_s_Set_transform_on_temp_source_dat);
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_8, __pyx_t_6);
        __pyx_t_6 = 0;
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_4, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 325, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      }
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "rasterio/_warp.pyx":327
 *             log.debug("Set transform on temp source dataset: %d", retval)
 * 
 *             try:             # <<<<<<<<<<<<<<
 *                 osr = _osr_from_crs(src_crs)
 *                 OSRExportToWkt(osr, &srcwkt)
 */
      /*try:*/ {

        /* "rasterio/_warp.pyx":328
 * 
 *             try:
 *                 osr = _osr_from_crs(src_crs)             # <<<<<<<<<<<<<<
 *                 OSRExportToWkt(osr, &srcwkt)
 *                 GDALSetProjection(src_dataset, srcwkt)
 */
        __pyx_t_18 = __pyx_f_8rasterio_5_base__osr_from_crs(__pyx_v_src_crs); if (unlikely(__pyx_t_18 == NULL)) __PYX_ERR(0, 328, __pyx_L26_error)
        __pyx_v_osr = __pyx_t_18;

        /* "rasterio/_warp.pyx":329
 *             try:
 *                 osr = _osr_from_crs(src_crs)
 *                 OSRExportToWkt(osr, &srcwkt)             # <<<<<<<<<<<<<<
 *                 GDALSetProjection(src_dataset, srcwkt)
 * 
 */
        OSRExportToWkt(__pyx_v_osr, (&__pyx_v_srcwkt));

        /* "rasterio/_warp.pyx":330
 *                 osr = _osr_from_crs(src_crs)
 *                 OSRExportToWkt(osr, &srcwkt)
 *                 GDALSetProjection(src_dataset, srcwkt)             # <<<<<<<<<<<<<<
 * 
 *                 log.debug("Set CRS on temp source dataset: %s", srcwkt)
 */
        GDALSetProjection(__pyx_v_src_dataset, __pyx_v_srcwkt);

        /* "rasterio/_warp.pyx":332
 *                 GDALSetProjection(src_dataset, srcwkt)
 * 
 *                 log.debug("Set CRS on temp source dataset: %s", srcwkt)             # <<<<<<<<<<<<<<
 * 
 *             finally:
 */
        __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_log); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 332, __pyx_L26_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_debug); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 332, __pyx_L26_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_t_5 = __Pyx_PyBytes_FromString(__pyx_v_srcwkt); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 332, __pyx_L26_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_6 = NULL;
        __pyx_t_8 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_6)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_6);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_8 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_kp_s_Set_CRS_on_temp_source_dataset_s, __pyx_t_5};
          __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 332, __pyx_L26_error)
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_kp_s_Set_CRS_on_temp_source_dataset_s, __pyx_t_5};
          __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 332, __pyx_L26_error)
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        } else
        #endif
        {
          __pyx_t_14 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 332, __pyx_L26_error)
          __Pyx_GOTREF(__pyx_t_14);
          if (__pyx_t_6) {
            __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_t_6); __pyx_t_6 = NULL;
          }
          __Pyx_INCREF(__pyx_kp_s_Set_CRS_on_temp_source_dataset_s);
          __Pyx_GIVEREF(__pyx_kp_s_Set_CRS_on_temp_source_dataset_s);
          PyTuple_SET_ITEM(__pyx_t_14, 0+__pyx_t_8, __pyx_kp_s_Set_CRS_on_temp_source_dataset_s);
          __Pyx_GIVEREF(__pyx_t_5);
          PyTuple_SET_ITEM(__pyx_t_14, 1+__pyx_t_8, __pyx_t_5);
          __pyx_t_5 = 0;
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_14, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 332, __pyx_L26_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }

      /* "rasterio/_warp.pyx":335
 * 
 *             finally:
 *                 CPLFree(srcwkt)             # <<<<<<<<<<<<<<
 *                 OSRRelease(osr)
 *                 osr = NULL
 */
      /*finally:*/ {
        /*normal exit:*/{
          CPLFree(__pyx_v_srcwkt);

          /* "rasterio/_warp.pyx":336
 *             finally:
 *                 CPLFree(srcwkt)
 *                 OSRRelease(osr)             # <<<<<<<<<<<<<<
 *                 osr = NULL
 * 
 */
          OSRRelease(__pyx_v_osr);

          /* "rasterio/_warp.pyx":337
 *                 CPLFree(srcwkt)
 *                 OSRRelease(osr)
 *                 osr = NULL             # <<<<<<<<<<<<<<
 * 
 *         elif gcps:
 */
          __pyx_v_osr = NULL;
          goto __pyx_L27;
        }
        /*exception exit:*/{
          __Pyx_PyThreadState_declare
          __pyx_L26_error:;
          __pyx_t_12 = 0; __pyx_t_11 = 0; __pyx_t_10 = 0; __pyx_t_21 = 0; __pyx_t_22 = 0; __pyx_t_23 = 0;
          __Pyx_PyThreadState_assign
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_21, &__pyx_t_22, &__pyx_t_23);
          if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_12, &__pyx_t_11, &__pyx_t_10) < 0)) __Pyx_ErrFetch(&__pyx_t_12, &__pyx_t_11, &__pyx_t_10);
          __Pyx_XGOTREF(__pyx_t_12);
          __Pyx_XGOTREF(__pyx_t_11);
          __Pyx_XGOTREF(__pyx_t_10);
          __Pyx_XGOTREF(__pyx_t_21);
          __Pyx_XGOTREF(__pyx_t_22);
          __Pyx_XGOTREF(__pyx_t_23);
          __pyx_t_8 = __pyx_lineno; __pyx_t_19 = __pyx_clineno; __pyx_t_20 = __pyx_filename;
          {

            /* "rasterio/_warp.pyx":335
 * 
 *             finally:
 *                 CPLFree(srcwkt)             # <<<<<<<<<<<<<<
 *                 OSRRelease(osr)
 *                 osr = NULL
 */
            CPLFree(__pyx_v_srcwkt);

            /* "rasterio/_warp.pyx":336
 *             finally:
 *                 CPLFree(srcwkt)
 *                 OSRRelease(osr)             # <<<<<<<<<<<<<<
 *                 osr = NULL
 * 
 */
            OSRRelease(__pyx_v_osr);

            /* "rasterio/_warp.pyx":337
 *                 CPLFree(srcwkt)
 *                 OSRRelease(osr)
 *                 osr = NULL             # <<<<<<<<<<<<<<
 * 
 *         elif gcps:
 */
            __pyx_v_osr = NULL;
          }
          __Pyx_PyThreadState_assign
          if (PY_MAJOR_VERSION >= 3) {
            __Pyx_XGIVEREF(__pyx_t_21);
            __Pyx_XGIVEREF(__pyx_t_22);
            __Pyx_XGIVEREF(__pyx_t_23);
            __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_22, __pyx_t_23);
          }
          __Pyx_XGIVEREF(__pyx_t_12);
          __Pyx_XGIVEREF(__pyx_t_11);
          __Pyx_XGIVEREF(__pyx_t_10);
          __Pyx_ErrRestore(__pyx_t_12, __pyx_t_11, __pyx_t_10);
          __pyx_t_12 = 0; __pyx_t_11 = 0; __pyx_t_10 = 0; __pyx_t_21 = 0; __pyx_t_22 = 0; __pyx_t_23 = 0;
          __pyx_lineno = __pyx_t_8; __pyx_clineno = __pyx_t_19; __pyx_filename = __pyx_t_20;
          goto __pyx_L1_error;
        }
        __pyx_L27:;
      }

      /* "rasterio/_warp.pyx":320
 *         log.debug("Created temp source dataset")
 * 
 *         if src_transform:             # <<<<<<<<<<<<<<
 *             for i in range(6):
 *                 gt[i] = src_transform[i]
 */
      goto __pyx_L22;
    }

    /* "rasterio/_warp.pyx":339
 *                 osr = NULL
 * 
 *         elif gcps:             # <<<<<<<<<<<<<<
 *             gcplist = <GDAL_GCP *>CPLMalloc(len(gcps) * sizeof(GDAL_GCP))
 *             try:
 */
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_gcps); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 339, __pyx_L1_error)
    if (__pyx_t_1) {

      /* "rasterio/_warp.pyx":340
 * 
 *         elif gcps:
 *             gcplist = <GDAL_GCP *>CPLMalloc(len(gcps) * sizeof(GDAL_GCP))             # <<<<<<<<<<<<<<
 *             try:
 *                 for i, obj in enumerate(gcps):
 */
      __pyx_t_7 = PyObject_Length(__pyx_v_gcps); if (unlikely(__pyx_t_7 == -1)) __PYX_ERR(0, 340, __pyx_L1_error)
      __pyx_v_gcplist = ((GDAL_GCP *)CPLMalloc((__pyx_t_7 * (sizeof(GDAL_GCP)))));

      /* "rasterio/_warp.pyx":341
 *         elif gcps:
 *             gcplist = <GDAL_GCP *>CPLMalloc(len(gcps) * sizeof(GDAL_GCP))
 *             try:             # <<<<<<<<<<<<<<
 *                 for i, obj in enumerate(gcps):
 *                     ident = str(i).encode('utf-8')
 */
      /*try:*/ {

        /* "rasterio/_warp.pyx":342
 *             gcplist = <GDAL_GCP *>CPLMalloc(len(gcps) * sizeof(GDAL_GCP))
 *             try:
 *                 for i, obj in enumerate(gcps):             # <<<<<<<<<<<<<<
 *                     ident = str(i).encode('utf-8')
 *                     info = "".encode('utf-8')
 */
        __pyx_t_19 = 0;
        if (likely(PyList_CheckExact(__pyx_v_gcps)) || PyTuple_CheckExact(__pyx_v_gcps)) {
          __pyx_t_3 = __pyx_v_gcps; __Pyx_INCREF(__pyx_t_3); __pyx_t_7 = 0;
          __pyx_t_24 = NULL;
        } else {
          __pyx_t_7 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_v_gcps); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 342, __pyx_L31_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_24 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 342, __pyx_L31_error)
        }
        for (;;) {
          if (likely(!__pyx_t_24)) {
            if (likely(PyList_CheckExact(__pyx_t_3))) {
              if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_3)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_4 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_7); __Pyx_INCREF(__pyx_t_4); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 342, __pyx_L31_error)
              #else
              __pyx_t_4 = PySequence_ITEM(__pyx_t_3, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 342, __pyx_L31_error)
              __Pyx_GOTREF(__pyx_t_4);
              #endif
            } else {
              if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_7); __Pyx_INCREF(__pyx_t_4); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 342, __pyx_L31_error)
              #else
              __pyx_t_4 = PySequence_ITEM(__pyx_t_3, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 342, __pyx_L31_error)
              __Pyx_GOTREF(__pyx_t_4);
              #endif
            }
          } else {
            __pyx_t_4 = __pyx_t_24(__pyx_t_3);
            if (unlikely(!__pyx_t_4)) {
              PyObject* exc_type = PyErr_Occurred();
              if (exc_type) {
                if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                else __PYX_ERR(0, 342, __pyx_L31_error)
              }
              break;
            }
            __Pyx_GOTREF(__pyx_t_4);
          }
          __Pyx_XDECREF_SET(__pyx_v_obj, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_v_i = __pyx_t_19;
          __pyx_t_19 = (__pyx_t_19 + 1);

          /* "rasterio/_warp.pyx":343
 *             try:
 *                 for i, obj in enumerate(gcps):
 *                     ident = str(i).encode('utf-8')             # <<<<<<<<<<<<<<
 *                     info = "".encode('utf-8')
 *                     gcplist[i].pszId = ident
 */
          __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 343, __pyx_L31_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_14 = PyTuple_New(1); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 343, __pyx_L31_error)
          __Pyx_GOTREF(__pyx_t_14);
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_4 = __Pyx_PyObject_Call(((PyObject *)(&PyString_Type)), __pyx_t_14, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 343, __pyx_L31_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
          __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_encode); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 343, __pyx_L31_error)
          __Pyx_GOTREF(__pyx_t_14);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_14, __pyx_tuple__10, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 343, __pyx_L31_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
          __Pyx_XDECREF_SET(__pyx_v_ident, __pyx_t_4);
          __pyx_t_4 = 0;

          /* "rasterio/_warp.pyx":344
 *                 for i, obj in enumerate(gcps):
 *                     ident = str(i).encode('utf-8')
 *                     info = "".encode('utf-8')             # <<<<<<<<<<<<<<
 *                     gcplist[i].pszId = ident
 *                     gcplist[i].pszInfo = info
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_kp_s__11, __pyx_n_s_encode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 344, __pyx_L31_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_14 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_tuple__12, NULL); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 344, __pyx_L31_error)
          __Pyx_GOTREF(__pyx_t_14);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_XDECREF_SET(__pyx_v_info, __pyx_t_14);
          __pyx_t_14 = 0;

          /* "rasterio/_warp.pyx":345
 *                     ident = str(i).encode('utf-8')
 *                     info = "".encode('utf-8')
 *                     gcplist[i].pszId = ident             # <<<<<<<<<<<<<<
 *                     gcplist[i].pszInfo = info
 *                     gcplist[i].dfGCPPixel = obj.col
 */
          __pyx_t_25 = __Pyx_PyObject_AsString(__pyx_v_ident); if (unlikely((!__pyx_t_25) && PyErr_Occurred())) __PYX_ERR(0, 345, __pyx_L31_error)
          (__pyx_v_gcplist[__pyx_v_i]).pszId = __pyx_t_25;

          /* "rasterio/_warp.pyx":346
 *                     info = "".encode('utf-8')
 *                     gcplist[i].pszId = ident
 *                     gcplist[i].pszInfo = info             # <<<<<<<<<<<<<<
 *                     gcplist[i].dfGCPPixel = obj.col
 *                     gcplist[i].dfGCPLine = obj.row
 */
          __pyx_t_25 = __Pyx_PyObject_AsString(__pyx_v_info); if (unlikely((!__pyx_t_25) && PyErr_Occurred())) __PYX_ERR(0, 346, __pyx_L31_error)
          (__pyx_v_gcplist[__pyx_v_i]).pszInfo = __pyx_t_25;

          /* "rasterio/_warp.pyx":347
 *                     gcplist[i].pszId = ident
 *                     gcplist[i].pszInfo = info
 *                     gcplist[i].dfGCPPixel = obj.col             # <<<<<<<<<<<<<<
 *                     gcplist[i].dfGCPLine = obj.row
 *                     gcplist[i].dfGCPX = obj.x
 */
          __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_v_obj, __pyx_n_s_col); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 347, __pyx_L31_error)
          __Pyx_GOTREF(__pyx_t_14);
          __pyx_t_17 = __pyx_PyFloat_AsDouble(__pyx_t_14); if (unlikely((__pyx_t_17 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 347, __pyx_L31_error)
          __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
          (__pyx_v_gcplist[__pyx_v_i]).dfGCPPixel = __pyx_t_17;

          /* "rasterio/_warp.pyx":348
 *                     gcplist[i].pszInfo = info
 *                     gcplist[i].dfGCPPixel = obj.col
 *                     gcplist[i].dfGCPLine = obj.row             # <<<<<<<<<<<<<<
 *                     gcplist[i].dfGCPX = obj.x
 *                     gcplist[i].dfGCPY = obj.y
 */
          __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_v_obj, __pyx_n_s_row); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 348, __pyx_L31_error)
          __Pyx_GOTREF(__pyx_t_14);
          __pyx_t_17 = __pyx_PyFloat_AsDouble(__pyx_t_14); if (unlikely((__pyx_t_17 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 348, __pyx_L31_error)
          __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
          (__pyx_v_gcplist[__pyx_v_i]).dfGCPLine = __pyx_t_17;

          /* "rasterio/_warp.pyx":349
 *                     gcplist[i].dfGCPPixel = obj.col
 *                     gcplist[i].dfGCPLine = obj.row
 *                     gcplist[i].dfGCPX = obj.x             # <<<<<<<<<<<<<<
 *                     gcplist[i].dfGCPY = obj.y
 *                     gcplist[i].dfGCPZ = obj.z or 0.0
 */
          __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_v_obj, __pyx_n_s_x); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 349, __pyx_L31_error)
          __Pyx_GOTREF(__pyx_t_14);
          __pyx_t_17 = __pyx_PyFloat_AsDouble(__pyx_t_14); if (unlikely((__pyx_t_17 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 349, __pyx_L31_error)
          __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
          (__pyx_v_gcplist[__pyx_v_i]).dfGCPX = __pyx_t_17;

          /* "rasterio/_warp.pyx":350
 *                     gcplist[i].dfGCPLine = obj.row
 *                     gcplist[i].dfGCPX = obj.x
 *                     gcplist[i].dfGCPY = obj.y             # <<<<<<<<<<<<<<
 *                     gcplist[i].dfGCPZ = obj.z or 0.0
 * 
 */
          __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_v_obj, __pyx_n_s_y); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 350, __pyx_L31_error)
          __Pyx_GOTREF(__pyx_t_14);
          __pyx_t_17 = __pyx_PyFloat_AsDouble(__pyx_t_14); if (unlikely((__pyx_t_17 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 350, __pyx_L31_error)
          __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
          (__pyx_v_gcplist[__pyx_v_i]).dfGCPY = __pyx_t_17;

          /* "rasterio/_warp.pyx":351
 *                     gcplist[i].dfGCPX = obj.x
 *                     gcplist[i].dfGCPY = obj.y
 *                     gcplist[i].dfGCPZ = obj.z or 0.0             # <<<<<<<<<<<<<<
 * 
 *                 GDALSetGCPs(src_dataset, len(gcps), gcplist, srcwkt)
 */
          __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_v_obj, __pyx_n_s_z); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 351, __pyx_L31_error)
          __Pyx_GOTREF(__pyx_t_14);
          __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_14); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 351, __pyx_L31_error)
          if (!__pyx_t_1) {
            __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
          } else {
            __pyx_t_26 = __pyx_PyFloat_AsDouble(__pyx_t_14); if (unlikely((__pyx_t_26 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 351, __pyx_L31_error)
            __pyx_t_17 = __pyx_t_26;
            __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
            goto __pyx_L35_bool_binop_done;
          }
          __pyx_t_17 = 0.0;
          __pyx_L35_bool_binop_done:;
          (__pyx_v_gcplist[__pyx_v_i]).dfGCPZ = __pyx_t_17;

          /* "rasterio/_warp.pyx":342
 *             gcplist = <GDAL_GCP *>CPLMalloc(len(gcps) * sizeof(GDAL_GCP))
 *             try:
 *                 for i, obj in enumerate(gcps):             # <<<<<<<<<<<<<<
 *                     ident = str(i).encode('utf-8')
 *                     info = "".encode('utf-8')
 */
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

        /* "rasterio/_warp.pyx":353
 *                     gcplist[i].dfGCPZ = obj.z or 0.0
 * 
 *                 GDALSetGCPs(src_dataset, len(gcps), gcplist, srcwkt)             # <<<<<<<<<<<<<<
 *             finally:
 *                 CPLFree(gcplist)
 */
        __pyx_t_7 = PyObject_Length(__pyx_v_gcps); if (unlikely(__pyx_t_7 == -1)) __PYX_ERR(0, 353, __pyx_L31_error)
        GDALSetGCPs(__pyx_v_src_dataset, __pyx_t_7, __pyx_v_gcplist, __pyx_v_srcwkt);
      }

      /* "rasterio/_warp.pyx":355
 *                 GDALSetGCPs(src_dataset, len(gcps), gcplist, srcwkt)
 *             finally:
 *                 CPLFree(gcplist)             # <<<<<<<<<<<<<<
 *                 CPLFree(srcwkt)
 * 
 */
      /*finally:*/ {
        /*normal exit:*/{
          CPLFree(__pyx_v_gcplist);

          /* "rasterio/_warp.pyx":356
 *             finally:
 *                 CPLFree(gcplist)
 *                 CPLFree(srcwkt)             # <<<<<<<<<<<<<<
 * 
 *         # Copy arrays to the dataset.
 */
          CPLFree(__pyx_v_srcwkt);
          goto __pyx_L32;
        }
        /*exception exit:*/{
          __Pyx_PyThreadState_declare
          __pyx_L31_error:;
          __pyx_t_23 = 0; __pyx_t_22 = 0; __pyx_t_21 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0;
          __Pyx_PyThreadState_assign
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_10, &__pyx_t_11, &__pyx_t_12);
          if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_23, &__pyx_t_22, &__pyx_t_21) < 0)) __Pyx_ErrFetch(&__pyx_t_23, &__pyx_t_22, &__pyx_t_21);
          __Pyx_XGOTREF(__pyx_t_23);
          __Pyx_XGOTREF(__pyx_t_22);
          __Pyx_XGOTREF(__pyx_t_21);
          __Pyx_XGOTREF(__pyx_t_10);
          __Pyx_XGOTREF(__pyx_t_11);
          __Pyx_XGOTREF(__pyx_t_12);
          __pyx_t_19 = __pyx_lineno; __pyx_t_8 = __pyx_clineno; __pyx_t_27 = __pyx_filename;
          {

            /* "rasterio/_warp.pyx":355
 *                 GDALSetGCPs(src_dataset, len(gcps), gcplist, srcwkt)
 *             finally:
 *                 CPLFree(gcplist)             # <<<<<<<<<<<<<<
 *                 CPLFree(srcwkt)
 * 
 */
            CPLFree(__pyx_v_gcplist);

            /* "rasterio/_warp.pyx":356
 *             finally:
 *                 CPLFree(gcplist)
 *                 CPLFree(srcwkt)             # <<<<<<<<<<<<<<
 * 
 *         # Copy arrays to the dataset.
 */
            CPLFree(__pyx_v_srcwkt);
          }
          __Pyx_PyThreadState_assign
          if (PY_MAJOR_VERSION >= 3) {
            __Pyx_XGIVEREF(__pyx_t_10);
            __Pyx_XGIVEREF(__pyx_t_11);
            __Pyx_XGIVEREF(__pyx_t_12);
            __Pyx_ExceptionReset(__pyx_t_10, __pyx_t_11, __pyx_t_12);
          }
          __Pyx_XGIVEREF(__pyx_t_23);
          __Pyx_XGIVEREF(__pyx_t_22);
          __Pyx_XGIVEREF(__pyx_t_21);
          __Pyx_ErrRestore(__pyx_t_23, __pyx_t_22, __pyx_t_21);
          __pyx_t_23 = 0; __pyx_t_22 = 0; __pyx_t_21 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0;
          __pyx_lineno = __pyx_t_19; __pyx_clineno = __pyx_t_8; __pyx_filename = __pyx_t_27;
          goto __pyx_L1_error;
        }
        __pyx_L32:;
      }

      /* "rasterio/_warp.pyx":339
 *                 osr = NULL
 * 
 *         elif gcps:             # <<<<<<<<<<<<<<
 *             gcplist = <GDAL_GCP *>CPLMalloc(len(gcps) * sizeof(GDAL_GCP))
 *             try:
 */
    }
    __pyx_L22:;

    /* "rasterio/_warp.pyx":359
 * 
 *         # Copy arrays to the dataset.
 *         retval = io_auto(source, src_dataset, 1)             # <<<<<<<<<<<<<<
 *         # TODO: handle errors (by retval).
 * 
 */
    __pyx_v_retval = __pyx_f_8rasterio_3_io_io_auto(__pyx_v_source, __pyx_v_src_dataset, 1, NULL);

    /* "rasterio/_warp.pyx":362
 *         # TODO: handle errors (by retval).
 * 
 *         log.debug("Wrote array to temp source dataset")             # <<<<<<<<<<<<<<
 * 
 *     # If the source is a rasterio MultiBand, no copy necessary.
 */
    __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_log); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 362, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_debug); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 362, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_14, __pyx_tuple__13, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 362, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "rasterio/_warp.pyx":288
 *     # We need a src_transform and src_dst in this case. These will
 *     # be copied to the MEM dataset.
 *     if dtypes.is_ndarray(source):             # <<<<<<<<<<<<<<
 *         # Convert 2D single-band arrays to 3D multi-band.
 *         if len(source.shape) == 2:
 */
    goto __pyx_L7;
  }

  /* "rasterio/_warp.pyx":366
 *     # If the source is a rasterio MultiBand, no copy necessary.
 *     # A MultiBand is a tuple: (dataset, bidx, dtype, shape(2d)).
 *     elif isinstance(source, tuple):             # <<<<<<<<<<<<<<
 *         rdr, src_bidx, dtype, shape = source
 *         if isinstance(src_bidx, int):
 */
  __pyx_t_1 = PyTuple_Check(__pyx_v_source); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "rasterio/_warp.pyx":367
 *     # A MultiBand is a tuple: (dataset, bidx, dtype, shape(2d)).
 *     elif isinstance(source, tuple):
 *         rdr, src_bidx, dtype, shape = source             # <<<<<<<<<<<<<<
 *         if isinstance(src_bidx, int):
 *             src_bidx = [src_bidx]
 */
    if ((likely(PyTuple_CheckExact(__pyx_v_source))) || (PyList_CheckExact(__pyx_v_source))) {
      PyObject* sequence = __pyx_v_source;
      #if !CYTHON_COMPILING_IN_PYPY
      Py_ssize_t size = Py_SIZE(sequence);
      #else
      Py_ssize_t size = PySequence_Size(sequence);
      #endif
      if (unlikely(size != 4)) {
        if (size > 4) __Pyx_RaiseTooManyValuesError(4);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 367, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_14 = PyTuple_GET_ITEM(sequence, 1); 
        __pyx_t_4 = PyTuple_GET_ITEM(sequence, 2); 
        __pyx_t_5 = PyTuple_GET_ITEM(sequence, 3); 
      } else {
        __pyx_t_3 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_14 = PyList_GET_ITEM(sequence, 1); 
        __pyx_t_4 = PyList_GET_ITEM(sequence, 2); 
        __pyx_t_5 = PyList_GET_ITEM(sequence, 3); 
      }
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_14);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      #else
      {
        Py_ssize_t i;
        PyObject** temps[4] = {&__pyx_t_3,&__pyx_t_14,&__pyx_t_4,&__pyx_t_5};
        for (i=0; i < 4; i++) {
          PyObject* item = PySequence_ITEM(sequence, i); if (unlikely(!item)) __PYX_ERR(0, 367, __pyx_L1_error)
          __Pyx_GOTREF(item);
          *(temps[i]) = item;
        }
      }
      #endif
    } else {
      Py_ssize_t index = -1;
      PyObject** temps[4] = {&__pyx_t_3,&__pyx_t_14,&__pyx_t_4,&__pyx_t_5};
      __pyx_t_6 = PyObject_GetIter(__pyx_v_source); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 367, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_28 = Py_TYPE(__pyx_t_6)->tp_iternext;
      for (index=0; index < 4; index++) {
        PyObject* item = __pyx_t_28(__pyx_t_6); if (unlikely(!item)) goto __pyx_L39_unpacking_failed;
        __Pyx_GOTREF(item);
        *(temps[index]) = item;
      }
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_28(__pyx_t_6), 4) < 0) __PYX_ERR(0, 367, __pyx_L1_error)
      __pyx_t_28 = NULL;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      goto __pyx_L40_unpacking_done;
      __pyx_L39_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_28 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 367, __pyx_L1_error)
      __pyx_L40_unpacking_done:;
    }
    __pyx_v_rdr = __pyx_t_3;
    __pyx_t_3 = 0;
    __pyx_v_src_bidx = __pyx_t_14;
    __pyx_t_14 = 0;
    __pyx_v_dtype = __pyx_t_4;
    __pyx_t_4 = 0;
    __pyx_v_shape = __pyx_t_5;
    __pyx_t_5 = 0;

    /* "rasterio/_warp.pyx":368
 *     elif isinstance(source, tuple):
 *         rdr, src_bidx, dtype, shape = source
 *         if isinstance(src_bidx, int):             # <<<<<<<<<<<<<<
 *             src_bidx = [src_bidx]
 *         src_count = len(src_bidx)
 */
    __pyx_t_2 = PyInt_Check(__pyx_v_src_bidx); 
    __pyx_t_1 = (__pyx_t_2 != 0);
    if (__pyx_t_1) {

      /* "rasterio/_warp.pyx":369
 *         rdr, src_bidx, dtype, shape = source
 *         if isinstance(src_bidx, int):
 *             src_bidx = [src_bidx]             # <<<<<<<<<<<<<<
 *         src_count = len(src_bidx)
 *         rows, cols = shape
 */
      __pyx_t_5 = PyList_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 369, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_v_src_bidx);
      __Pyx_GIVEREF(__pyx_v_src_bidx);
      PyList_SET_ITEM(__pyx_t_5, 0, __pyx_v_src_bidx);
      __Pyx_DECREF_SET(__pyx_v_src_bidx, __pyx_t_5);
      __pyx_t_5 = 0;

      /* "rasterio/_warp.pyx":368
 *     elif isinstance(source, tuple):
 *         rdr, src_bidx, dtype, shape = source
 *         if isinstance(src_bidx, int):             # <<<<<<<<<<<<<<
 *             src_bidx = [src_bidx]
 *         src_count = len(src_bidx)
 */
    }

    /* "rasterio/_warp.pyx":370
 *         if isinstance(src_bidx, int):
 *             src_bidx = [src_bidx]
 *         src_count = len(src_bidx)             # <<<<<<<<<<<<<<
 *         rows, cols = shape
 *         src_dataset = (<DatasetReaderBase?>rdr).handle()
 */
    __pyx_t_7 = PyObject_Length(__pyx_v_src_bidx); if (unlikely(__pyx_t_7 == -1)) __PYX_ERR(0, 370, __pyx_L1_error)
    __pyx_v_src_count = __pyx_t_7;

    /* "rasterio/_warp.pyx":371
 *             src_bidx = [src_bidx]
 *         src_count = len(src_bidx)
 *         rows, cols = shape             # <<<<<<<<<<<<<<
 *         src_dataset = (<DatasetReaderBase?>rdr).handle()
 *         if src_nodata is None:
 */
    if ((likely(PyTuple_CheckExact(__pyx_v_shape))) || (PyList_CheckExact(__pyx_v_shape))) {
      PyObject* sequence = __pyx_v_shape;
      #if !CYTHON_COMPILING_IN_PYPY
      Py_ssize_t size = Py_SIZE(sequence);
      #else
      Py_ssize_t size = PySequence_Size(sequence);
      #endif
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 371, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_5 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_4 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_5 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_4 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_4);
      #else
      __pyx_t_5 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 371, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_4 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 371, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      #endif
    } else {
      Py_ssize_t index = -1;
      __pyx_t_14 = PyObject_GetIter(__pyx_v_shape); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 371, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      __pyx_t_28 = Py_TYPE(__pyx_t_14)->tp_iternext;
      index = 0; __pyx_t_5 = __pyx_t_28(__pyx_t_14); if (unlikely(!__pyx_t_5)) goto __pyx_L42_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_5);
      index = 1; __pyx_t_4 = __pyx_t_28(__pyx_t_14); if (unlikely(!__pyx_t_4)) goto __pyx_L42_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_4);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_28(__pyx_t_14), 2) < 0) __PYX_ERR(0, 371, __pyx_L1_error)
      __pyx_t_28 = NULL;
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      goto __pyx_L43_unpacking_done;
      __pyx_L42_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      __pyx_t_28 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 371, __pyx_L1_error)
      __pyx_L43_unpacking_done:;
    }
    __pyx_t_8 = __Pyx_PyInt_As_int(__pyx_t_5); if (unlikely((__pyx_t_8 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 371, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_19 = __Pyx_PyInt_As_int(__pyx_t_4); if (unlikely((__pyx_t_19 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 371, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_rows = __pyx_t_8;
    __pyx_v_cols = __pyx_t_19;

    /* "rasterio/_warp.pyx":372
 *         src_count = len(src_bidx)
 *         rows, cols = shape
 *         src_dataset = (<DatasetReaderBase?>rdr).handle()             # <<<<<<<<<<<<<<
 *         if src_nodata is None:
 *             src_nodata = rdr.nodata
 */
    if (!(likely(__Pyx_TypeTest(__pyx_v_rdr, __pyx_ptype_8rasterio_3_io_DatasetReaderBase)))) __PYX_ERR(0, 372, __pyx_L1_error)
    __pyx_t_29 = ((struct __pyx_vtabstruct_8rasterio_3_io_DatasetReaderBase *)((struct __pyx_obj_8rasterio_3_io_DatasetReaderBase *)__pyx_v_rdr)->__pyx_base.__pyx_vtab)->__pyx_base.handle(((struct __pyx_obj_8rasterio_5_base_DatasetBase *)((struct __pyx_obj_8rasterio_3_io_DatasetReaderBase *)__pyx_v_rdr))); if (unlikely(__pyx_t_29 == NULL)) __PYX_ERR(0, 372, __pyx_L1_error)
    __pyx_v_src_dataset = __pyx_t_29;

    /* "rasterio/_warp.pyx":373
 *         rows, cols = shape
 *         src_dataset = (<DatasetReaderBase?>rdr).handle()
 *         if src_nodata is None:             # <<<<<<<<<<<<<<
 *             src_nodata = rdr.nodata
 *     else:
 */
    __pyx_t_1 = (__pyx_v_src_nodata == Py_None);
    __pyx_t_2 = (__pyx_t_1 != 0);
    if (__pyx_t_2) {

      /* "rasterio/_warp.pyx":374
 *         src_dataset = (<DatasetReaderBase?>rdr).handle()
 *         if src_nodata is None:
 *             src_nodata = rdr.nodata             # <<<<<<<<<<<<<<
 *     else:
 *         raise ValueError("Invalid source")
 */
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_rdr, __pyx_n_s_nodata); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 374, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF_SET(__pyx_v_src_nodata, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "rasterio/_warp.pyx":373
 *         rows, cols = shape
 *         src_dataset = (<DatasetReaderBase?>rdr).handle()
 *         if src_nodata is None:             # <<<<<<<<<<<<<<
 *             src_nodata = rdr.nodata
 *     else:
 */
    }

    /* "rasterio/_warp.pyx":366
 *     # If the source is a rasterio MultiBand, no copy necessary.
 *     # A MultiBand is a tuple: (dataset, bidx, dtype, shape(2d)).
 *     elif isinstance(source, tuple):             # <<<<<<<<<<<<<<
 *         rdr, src_bidx, dtype, shape = source
 *         if isinstance(src_bidx, int):
 */
    goto __pyx_L7;
  }

  /* "rasterio/_warp.pyx":376
 *             src_nodata = rdr.nodata
 *     else:
 *         raise ValueError("Invalid source")             # <<<<<<<<<<<<<<
 * 
 *     # Next, do the same for the destination raster.
 */
  /*else*/ {
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__14, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 376, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(0, 376, __pyx_L1_error)
  }
  __pyx_L7:;

  /* "rasterio/_warp.pyx":379
 * 
 *     # Next, do the same for the destination raster.
 *     if dtypes.is_ndarray(destination):             # <<<<<<<<<<<<<<
 *         if len(destination.shape) == 2:
 *             destination = destination.reshape(1, *destination.shape)
 */
  __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_dtypes); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 379, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_is_ndarray); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 379, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_14))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_14);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_14);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_14, function);
    }
  }
  if (!__pyx_t_5) {
    __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_t_14, __pyx_v_destination); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 379, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_14)) {
      PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_destination};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_14, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 379, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_4);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_14)) {
      PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_destination};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_14, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 379, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_4);
    } else
    #endif
    {
      __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 379, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5); __pyx_t_5 = NULL;
      __Pyx_INCREF(__pyx_v_destination);
      __Pyx_GIVEREF(__pyx_v_destination);
      PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_v_destination);
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_14, __pyx_t_3, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 379, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 379, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__pyx_t_2) {

    /* "rasterio/_warp.pyx":380
 *     # Next, do the same for the destination raster.
 *     if dtypes.is_ndarray(destination):
 *         if len(destination.shape) == 2:             # <<<<<<<<<<<<<<
 *             destination = destination.reshape(1, *destination.shape)
 *             dst_bidx = [1]
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_destination, __pyx_n_s_shape); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 380, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = PyObject_Length(__pyx_t_4); if (unlikely(__pyx_t_7 == -1)) __PYX_ERR(0, 380, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_2 = ((__pyx_t_7 == 2) != 0);
    if (__pyx_t_2) {

      /* "rasterio/_warp.pyx":381
 *     if dtypes.is_ndarray(destination):
 *         if len(destination.shape) == 2:
 *             destination = destination.reshape(1, *destination.shape)             # <<<<<<<<<<<<<<
 *             dst_bidx = [1]
 *         else:
 */
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_destination, __pyx_n_s_reshape); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 381, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_v_destination, __pyx_n_s_shape); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 381, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      __pyx_t_3 = PySequence_Tuple(__pyx_t_14); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 381, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      __pyx_t_14 = PyNumber_Add(__pyx_tuple__15, __pyx_t_3); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 381, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_14, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 381, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      __Pyx_DECREF_SET(__pyx_v_destination, __pyx_t_3);
      __pyx_t_3 = 0;

      /* "rasterio/_warp.pyx":382
 *         if len(destination.shape) == 2:
 *             destination = destination.reshape(1, *destination.shape)
 *             dst_bidx = [1]             # <<<<<<<<<<<<<<
 *         else:
 *             dst_bidx = src_bidx
 */
      __pyx_t_3 = PyList_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 382, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_int_1);
      __Pyx_GIVEREF(__pyx_int_1);
      PyList_SET_ITEM(__pyx_t_3, 0, __pyx_int_1);
      __pyx_v_dst_bidx = __pyx_t_3;
      __pyx_t_3 = 0;

      /* "rasterio/_warp.pyx":380
 *     # Next, do the same for the destination raster.
 *     if dtypes.is_ndarray(destination):
 *         if len(destination.shape) == 2:             # <<<<<<<<<<<<<<
 *             destination = destination.reshape(1, *destination.shape)
 *             dst_bidx = [1]
 */
      goto __pyx_L46;
    }

    /* "rasterio/_warp.pyx":384
 *             dst_bidx = [1]
 *         else:
 *             dst_bidx = src_bidx             # <<<<<<<<<<<<<<
 * 
 *         if destination.shape[0] != src_count:
 */
    /*else*/ {
      __Pyx_INCREF(__pyx_v_src_bidx);
      __pyx_v_dst_bidx = __pyx_v_src_bidx;
    }
    __pyx_L46:;

    /* "rasterio/_warp.pyx":386
 *             dst_bidx = src_bidx
 * 
 *         if destination.shape[0] != src_count:             # <<<<<<<<<<<<<<
 *             raise ValueError("Destination's shape is invalid")
 * 
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_destination, __pyx_n_s_shape); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 386, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_14 = __Pyx_GetItemInt(__pyx_t_3, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 386, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_src_count); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 386, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyObject_RichCompare(__pyx_t_14, __pyx_t_3, Py_NE); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 386, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 386, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (__pyx_t_2) {

      /* "rasterio/_warp.pyx":387
 * 
 *         if destination.shape[0] != src_count:
 *             raise ValueError("Destination's shape is invalid")             # <<<<<<<<<<<<<<
 * 
 *         try:
 */
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__16, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 387, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_Raise(__pyx_t_4, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __PYX_ERR(0, 387, __pyx_L1_error)

      /* "rasterio/_warp.pyx":386
 *             dst_bidx = src_bidx
 * 
 *         if destination.shape[0] != src_count:             # <<<<<<<<<<<<<<
 *             raise ValueError("Destination's shape is invalid")
 * 
 */
    }

    /* "rasterio/_warp.pyx":389
 *             raise ValueError("Destination's shape is invalid")
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             driver = exc_wrap_pointer(GDALGetDriverByName("MEM"))
 *         except:
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_12, &__pyx_t_11, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_12);
      __Pyx_XGOTREF(__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_10);
      /*try:*/ {

        /* "rasterio/_warp.pyx":390
 * 
 *         try:
 *             driver = exc_wrap_pointer(GDALGetDriverByName("MEM"))             # <<<<<<<<<<<<<<
 *         except:
 *             raise DriverRegistrationError(
 */
        __pyx_t_13 = __pyx_f_8rasterio_4_err_exc_wrap_pointer(GDALGetDriverByName(((char const *)"MEM"))); if (unlikely(__pyx_t_13 == NULL)) __PYX_ERR(0, 390, __pyx_L48_error)
        __pyx_v_driver = __pyx_t_13;

        /* "rasterio/_warp.pyx":389
 *             raise ValueError("Destination's shape is invalid")
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             driver = exc_wrap_pointer(GDALGetDriverByName("MEM"))
 *         except:
 */
      }
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L55_try_end;
      __pyx_L48_error:;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "rasterio/_warp.pyx":391
 *         try:
 *             driver = exc_wrap_pointer(GDALGetDriverByName("MEM"))
 *         except:             # <<<<<<<<<<<<<<
 *             raise DriverRegistrationError(
 *                 "'MEM' driver not found. Check that this call is contained "
 */
      /*except:*/ {
        __Pyx_AddTraceback("rasterio._warp._reproject", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_3, &__pyx_t_14) < 0) __PYX_ERR(0, 391, __pyx_L50_except_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_GOTREF(__pyx_t_14);

        /* "rasterio/_warp.pyx":392
 *             driver = exc_wrap_pointer(GDALGetDriverByName("MEM"))
 *         except:
 *             raise DriverRegistrationError(             # <<<<<<<<<<<<<<
 *                 "'MEM' driver not found. Check that this call is contained "
 *                 "in a `with rasterio.Env()` or `with rasterio.open()` "
 */
        __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_DriverRegistrationError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 392, __pyx_L50_except_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_tuple__17, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 392, __pyx_L50_except_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_Raise(__pyx_t_6, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __PYX_ERR(0, 392, __pyx_L50_except_error)
      }
      __pyx_L50_except_error:;

      /* "rasterio/_warp.pyx":389
 *             raise ValueError("Destination's shape is invalid")
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             driver = exc_wrap_pointer(GDALGetDriverByName("MEM"))
 *         except:
 */
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_12);
      __Pyx_XGIVEREF(__pyx_t_11);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_12, __pyx_t_11, __pyx_t_10);
      goto __pyx_L1_error;
      __pyx_L55_try_end:;
    }

    /* "rasterio/_warp.pyx":397
 *                 "block.")
 * 
 *         _, rows, cols = destination.shape             # <<<<<<<<<<<<<<
 * 
 *         datasetname = str(uuid.uuid4()).encode('utf-8')
 */
    __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_v_destination, __pyx_n_s_shape); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 397, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    if ((likely(PyTuple_CheckExact(__pyx_t_14))) || (PyList_CheckExact(__pyx_t_14))) {
      PyObject* sequence = __pyx_t_14;
      #if !CYTHON_COMPILING_IN_PYPY
      Py_ssize_t size = Py_SIZE(sequence);
      #else
      Py_ssize_t size = PySequence_Size(sequence);
      #endif
      if (unlikely(size != 3)) {
        if (size > 3) __Pyx_RaiseTooManyValuesError(3);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 397, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_4 = PyTuple_GET_ITEM(sequence, 1); 
        __pyx_t_6 = PyTuple_GET_ITEM(sequence, 2); 
      } else {
        __pyx_t_3 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_4 = PyList_GET_ITEM(sequence, 1); 
        __pyx_t_6 = PyList_GET_ITEM(sequence, 2); 
      }
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_6);
      #else
      __pyx_t_3 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 397, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 397, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_6 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 397, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      #endif
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_5 = PyObject_GetIter(__pyx_t_14); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 397, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      __pyx_t_28 = Py_TYPE(__pyx_t_5)->tp_iternext;
      index = 0; __pyx_t_3 = __pyx_t_28(__pyx_t_5); if (unlikely(!__pyx_t_3)) goto __pyx_L58_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_3);
      index = 1; __pyx_t_4 = __pyx_t_28(__pyx_t_5); if (unlikely(!__pyx_t_4)) goto __pyx_L58_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_4);
      index = 2; __pyx_t_6 = __pyx_t_28(__pyx_t_5); if (unlikely(!__pyx_t_6)) goto __pyx_L58_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_6);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_28(__pyx_t_5), 3) < 0) __PYX_ERR(0, 397, __pyx_L1_error)
      __pyx_t_28 = NULL;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L59_unpacking_done;
      __pyx_L58_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_28 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 397, __pyx_L1_error)
      __pyx_L59_unpacking_done:;
    }
    __pyx_t_19 = __Pyx_PyInt_As_int(__pyx_t_4); if (unlikely((__pyx_t_19 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 397, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_8 = __Pyx_PyInt_As_int(__pyx_t_6); if (unlikely((__pyx_t_8 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 397, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_v__ = __pyx_t_3;
    __pyx_t_3 = 0;
    __pyx_v_rows = __pyx_t_19;
    __pyx_v_cols = __pyx_t_8;

    /* "rasterio/_warp.pyx":399
 *         _, rows, cols = destination.shape
 * 
 *         datasetname = str(uuid.uuid4()).encode('utf-8')             # <<<<<<<<<<<<<<
 *         dst_dataset = exc_wrap_pointer(
 *             GDALCreate(driver, <const char *>datasetname, cols, rows,
 */
    __pyx_t_6 = __Pyx_GetModuleGlobalName(__pyx_n_s_uuid); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 399, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_uuid4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 399, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (__pyx_t_6) {
      __pyx_t_14 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_6); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 399, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else {
      __pyx_t_14 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 399, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 399, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_14);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_14);
    __pyx_t_14 = 0;
    __pyx_t_14 = __Pyx_PyObject_Call(((PyObject *)(&PyString_Type)), __pyx_t_4, NULL); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 399, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_14, __pyx_n_s_encode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 399, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    __pyx_t_14 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_tuple__18, NULL); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 399, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF_SET(__pyx_v_datasetname, __pyx_t_14);
    __pyx_t_14 = 0;

    /* "rasterio/_warp.pyx":401
 *         datasetname = str(uuid.uuid4()).encode('utf-8')
 *         dst_dataset = exc_wrap_pointer(
 *             GDALCreate(driver, <const char *>datasetname, cols, rows,             # <<<<<<<<<<<<<<
 *                 src_count,
 *                 dtypes.dtype_rev[np.dtype(destination.dtype).name], NULL))
 */
    __pyx_t_30 = __Pyx_PyObject_AsString(__pyx_v_datasetname); if (unlikely((!__pyx_t_30) && PyErr_Occurred())) __PYX_ERR(0, 401, __pyx_L1_error)

    /* "rasterio/_warp.pyx":403
 *             GDALCreate(driver, <const char *>datasetname, cols, rows,
 *                 src_count,
 *                 dtypes.dtype_rev[np.dtype(destination.dtype).name], NULL))             # <<<<<<<<<<<<<<
 * 
 *         GDALSetDescription(
 */
    __pyx_t_14 = __Pyx_GetModuleGlobalName(__pyx_n_s_dtypes); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 403, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_14, __pyx_n_s_dtype_rev); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 403, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_v_destination, __pyx_n_s_dtype); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 403, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 403, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_14);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_14);
    __pyx_t_14 = 0;
    __pyx_t_14 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_5numpy_dtype), __pyx_t_6, NULL); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 403, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_14, __pyx_n_s_name); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 403, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    __pyx_t_14 = PyObject_GetItem(__pyx_t_4, __pyx_t_6); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 403, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_16 = ((GDALDataType)__Pyx_PyInt_As_GDALDataType(__pyx_t_14)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 403, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

    /* "rasterio/_warp.pyx":400
 * 
 *         datasetname = str(uuid.uuid4()).encode('utf-8')
 *         dst_dataset = exc_wrap_pointer(             # <<<<<<<<<<<<<<
 *             GDALCreate(driver, <const char *>datasetname, cols, rows,
 *                 src_count,
 */
    __pyx_t_13 = __pyx_f_8rasterio_4_err_exc_wrap_pointer(GDALCreate(__pyx_v_driver, ((char const *)__pyx_t_30), __pyx_v_cols, __pyx_v_rows, __pyx_v_src_count, __pyx_t_16, NULL)); if (unlikely(__pyx_t_13 == NULL)) __PYX_ERR(0, 400, __pyx_L1_error)
    __pyx_v_dst_dataset = __pyx_t_13;

    /* "rasterio/_warp.pyx":405
 *                 dtypes.dtype_rev[np.dtype(destination.dtype).name], NULL))
 * 
 *         GDALSetDescription(             # <<<<<<<<<<<<<<
 *             dst_dataset, "Temporary destination dataset for _reproject()")
 * 
 */
    GDALSetDescription(__pyx_v_dst_dataset, ((char const *)"Temporary destination dataset for _reproject()"));

    /* "rasterio/_warp.pyx":408
 *             dst_dataset, "Temporary destination dataset for _reproject()")
 * 
 *         log.debug("Created temp destination dataset.")             # <<<<<<<<<<<<<<
 * 
 *         for i in range(6):
 */
    __pyx_t_14 = __Pyx_GetModuleGlobalName(__pyx_n_s_log); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 408, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_14, __pyx_n_s_debug); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 408, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    __pyx_t_14 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_tuple__19, NULL); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 408, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

    /* "rasterio/_warp.pyx":410
 *         log.debug("Created temp destination dataset.")
 * 
 *         for i in range(6):             # <<<<<<<<<<<<<<
 *             gt[i] = dst_transform[i]
 * 
 */
    for (__pyx_t_8 = 0; __pyx_t_8 < 6; __pyx_t_8+=1) {
      __pyx_v_i = __pyx_t_8;

      /* "rasterio/_warp.pyx":411
 * 
 *         for i in range(6):
 *             gt[i] = dst_transform[i]             # <<<<<<<<<<<<<<
 * 
 *         if not GDALError.none == GDALSetGeoTransform(dst_dataset, gt):
 */
      __pyx_t_14 = __Pyx_GetItemInt(__pyx_v_dst_transform, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 411, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      __pyx_t_17 = __pyx_PyFloat_AsDouble(__pyx_t_14); if (unlikely((__pyx_t_17 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 411, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      (__pyx_v_gt[__pyx_v_i]) = __pyx_t_17;
    }

    /* "rasterio/_warp.pyx":413
 *             gt[i] = dst_transform[i]
 * 
 *         if not GDALError.none == GDALSetGeoTransform(dst_dataset, gt):             # <<<<<<<<<<<<<<
 *             raise ValueError(
 *                 "Failed to set transform on temp destination dataset.")
 */
    __pyx_t_14 = __Pyx_GetModuleGlobalName(__pyx_n_s_GDALError); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 413, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_14, __pyx_n_s_none); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 413, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    __pyx_t_14 = __Pyx_PyInt_From_int(GDALSetGeoTransform(__pyx_v_dst_dataset, __pyx_v_gt)); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 413, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __pyx_t_4 = PyObject_RichCompare(__pyx_t_6, __pyx_t_14, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 413, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 413, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_1 = ((!__pyx_t_2) != 0);
    if (__pyx_t_1) {

      /* "rasterio/_warp.pyx":414
 * 
 *         if not GDALError.none == GDALSetGeoTransform(dst_dataset, gt):
 *             raise ValueError(             # <<<<<<<<<<<<<<
 *                 "Failed to set transform on temp destination dataset.")
 * 
 */
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__20, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 414, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_Raise(__pyx_t_4, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __PYX_ERR(0, 414, __pyx_L1_error)

      /* "rasterio/_warp.pyx":413
 *             gt[i] = dst_transform[i]
 * 
 *         if not GDALError.none == GDALSetGeoTransform(dst_dataset, gt):             # <<<<<<<<<<<<<<
 *             raise ValueError(
 *                 "Failed to set transform on temp destination dataset.")
 */
    }

    /* "rasterio/_warp.pyx":417
 *                 "Failed to set transform on temp destination dataset.")
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             osr = _osr_from_crs(dst_crs)
 *             OSRExportToWkt(osr, &dstwkt)
 */
    /*try:*/ {

      /* "rasterio/_warp.pyx":418
 * 
 *         try:
 *             osr = _osr_from_crs(dst_crs)             # <<<<<<<<<<<<<<
 *             OSRExportToWkt(osr, &dstwkt)
 * 
 */
      __pyx_t_18 = __pyx_f_8rasterio_5_base__osr_from_crs(__pyx_v_dst_crs); if (unlikely(__pyx_t_18 == NULL)) __PYX_ERR(0, 418, __pyx_L64_error)
      __pyx_v_osr = __pyx_t_18;

      /* "rasterio/_warp.pyx":419
 *         try:
 *             osr = _osr_from_crs(dst_crs)
 *             OSRExportToWkt(osr, &dstwkt)             # <<<<<<<<<<<<<<
 * 
 *             log.debug("CRS for temp destination dataset: %s.", dstwkt)
 */
      OSRExportToWkt(__pyx_v_osr, (&__pyx_v_dstwkt));

      /* "rasterio/_warp.pyx":421
 *             OSRExportToWkt(osr, &dstwkt)
 * 
 *             log.debug("CRS for temp destination dataset: %s.", dstwkt)             # <<<<<<<<<<<<<<
 * 
 *             if not GDALError.none == GDALSetProjection(
 */
      __pyx_t_14 = __Pyx_GetModuleGlobalName(__pyx_n_s_log); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 421, __pyx_L64_error)
      __Pyx_GOTREF(__pyx_t_14);
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_14, __pyx_n_s_debug); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 421, __pyx_L64_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      __pyx_t_14 = __Pyx_PyBytes_FromString(__pyx_v_dstwkt); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 421, __pyx_L64_error)
      __Pyx_GOTREF(__pyx_t_14);
      __pyx_t_3 = NULL;
      __pyx_t_8 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_6);
        if (likely(__pyx_t_3)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_6, function);
          __pyx_t_8 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_6)) {
        PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_kp_s_CRS_for_temp_destination_dataset, __pyx_t_14};
        __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 421, __pyx_L64_error)
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
        PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_kp_s_CRS_for_temp_destination_dataset, __pyx_t_14};
        __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 421, __pyx_L64_error)
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 421, __pyx_L64_error)
        __Pyx_GOTREF(__pyx_t_5);
        if (__pyx_t_3) {
          __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
        }
        __Pyx_INCREF(__pyx_kp_s_CRS_for_temp_destination_dataset);
        __Pyx_GIVEREF(__pyx_kp_s_CRS_for_temp_destination_dataset);
        PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_8, __pyx_kp_s_CRS_for_temp_destination_dataset);
        __Pyx_GIVEREF(__pyx_t_14);
        PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_8, __pyx_t_14);
        __pyx_t_14 = 0;
        __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_5, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 421, __pyx_L64_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "rasterio/_warp.pyx":423
 *             log.debug("CRS for temp destination dataset: %s.", dstwkt)
 * 
 *             if not GDALError.none == GDALSetProjection(             # <<<<<<<<<<<<<<
 *                     dst_dataset, dstwkt):
 *                 raise ("Failed to set projection on temp destination dataset.")
 */
      __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_GDALError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 423, __pyx_L64_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_none); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 423, __pyx_L64_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "rasterio/_warp.pyx":424
 * 
 *             if not GDALError.none == GDALSetProjection(
 *                     dst_dataset, dstwkt):             # <<<<<<<<<<<<<<
 *                 raise ("Failed to set projection on temp destination dataset.")
 *         finally:
 */
      __pyx_t_4 = __Pyx_PyInt_From_int(GDALSetProjection(__pyx_v_dst_dataset, __pyx_v_dstwkt)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 423, __pyx_L64_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = PyObject_RichCompare(__pyx_t_6, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 423, __pyx_L64_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "rasterio/_warp.pyx":423
 *             log.debug("CRS for temp destination dataset: %s.", dstwkt)
 * 
 *             if not GDALError.none == GDALSetProjection(             # <<<<<<<<<<<<<<
 *                     dst_dataset, dstwkt):
 *                 raise ("Failed to set projection on temp destination dataset.")
 */
      __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 423, __pyx_L64_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_2 = ((!__pyx_t_1) != 0);
      if (__pyx_t_2) {

        /* "rasterio/_warp.pyx":425
 *             if not GDALError.none == GDALSetProjection(
 *                     dst_dataset, dstwkt):
 *                 raise ("Failed to set projection on temp destination dataset.")             # <<<<<<<<<<<<<<
 *         finally:
 *             CPLFree(dstwkt)
 */
        __Pyx_Raise(__pyx_kp_s_Failed_to_set_projection_on_temp, 0, 0, 0);
        __PYX_ERR(0, 425, __pyx_L64_error)

        /* "rasterio/_warp.pyx":423
 *             log.debug("CRS for temp destination dataset: %s.", dstwkt)
 * 
 *             if not GDALError.none == GDALSetProjection(             # <<<<<<<<<<<<<<
 *                     dst_dataset, dstwkt):
 *                 raise ("Failed to set projection on temp destination dataset.")
 */
      }
    }

    /* "rasterio/_warp.pyx":427
 *                 raise ("Failed to set projection on temp destination dataset.")
 *         finally:
 *             CPLFree(dstwkt)             # <<<<<<<<<<<<<<
 *             OSRRelease(osr)
 *             osr = NULL
 */
    /*finally:*/ {
      /*normal exit:*/{
        CPLFree(__pyx_v_dstwkt);

        /* "rasterio/_warp.pyx":428
 *         finally:
 *             CPLFree(dstwkt)
 *             OSRRelease(osr)             # <<<<<<<<<<<<<<
 *             osr = NULL
 * 
 */
        OSRRelease(__pyx_v_osr);

        /* "rasterio/_warp.pyx":429
 *             CPLFree(dstwkt)
 *             OSRRelease(osr)
 *             osr = NULL             # <<<<<<<<<<<<<<
 * 
 *         retval = io_auto(destination, dst_dataset, 1)
 */
        __pyx_v_osr = NULL;
        goto __pyx_L65;
      }
      /*exception exit:*/{
        __Pyx_PyThreadState_declare
        __pyx_L64_error:;
        __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_21 = 0; __pyx_t_22 = 0; __pyx_t_23 = 0;
        __Pyx_PyThreadState_assign
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_21, &__pyx_t_22, &__pyx_t_23);
        if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_10, &__pyx_t_11, &__pyx_t_12) < 0)) __Pyx_ErrFetch(&__pyx_t_10, &__pyx_t_11, &__pyx_t_12);
        __Pyx_XGOTREF(__pyx_t_10);
        __Pyx_XGOTREF(__pyx_t_11);
        __Pyx_XGOTREF(__pyx_t_12);
        __Pyx_XGOTREF(__pyx_t_21);
        __Pyx_XGOTREF(__pyx_t_22);
        __Pyx_XGOTREF(__pyx_t_23);
        __pyx_t_8 = __pyx_lineno; __pyx_t_19 = __pyx_clineno; __pyx_t_31 = __pyx_filename;
        {

          /* "rasterio/_warp.pyx":427
 *                 raise ("Failed to set projection on temp destination dataset.")
 *         finally:
 *             CPLFree(dstwkt)             # <<<<<<<<<<<<<<
 *             OSRRelease(osr)
 *             osr = NULL
 */
          CPLFree(__pyx_v_dstwkt);

          /* "rasterio/_warp.pyx":428
 *         finally:
 *             CPLFree(dstwkt)
 *             OSRRelease(osr)             # <<<<<<<<<<<<<<
 *             osr = NULL
 * 
 */
          OSRRelease(__pyx_v_osr);

          /* "rasterio/_warp.pyx":429
 *             CPLFree(dstwkt)
 *             OSRRelease(osr)
 *             osr = NULL             # <<<<<<<<<<<<<<
 * 
 *         retval = io_auto(destination, dst_dataset, 1)
 */
          __pyx_v_osr = NULL;
        }
        __Pyx_PyThreadState_assign
        if (PY_MAJOR_VERSION >= 3) {
          __Pyx_XGIVEREF(__pyx_t_21);
          __Pyx_XGIVEREF(__pyx_t_22);
          __Pyx_XGIVEREF(__pyx_t_23);
          __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_22, __pyx_t_23);
        }
        __Pyx_XGIVEREF(__pyx_t_10);
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_XGIVEREF(__pyx_t_12);
        __Pyx_ErrRestore(__pyx_t_10, __pyx_t_11, __pyx_t_12);
        __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_21 = 0; __pyx_t_22 = 0; __pyx_t_23 = 0;
        __pyx_lineno = __pyx_t_8; __pyx_clineno = __pyx_t_19; __pyx_filename = __pyx_t_31;
        goto __pyx_L1_error;
      }
      __pyx_L65:;
    }

    /* "rasterio/_warp.pyx":431
 *             osr = NULL
 * 
 *         retval = io_auto(destination, dst_dataset, 1)             # <<<<<<<<<<<<<<
 * 
 *         log.debug("Wrote array to temp output dataset")
 */
    __pyx_v_retval = __pyx_f_8rasterio_3_io_io_auto(__pyx_v_destination, __pyx_v_dst_dataset, 1, NULL);

    /* "rasterio/_warp.pyx":433
 *         retval = io_auto(destination, dst_dataset, 1)
 * 
 *         log.debug("Wrote array to temp output dataset")             # <<<<<<<<<<<<<<
 * 
 *         if dst_nodata is None and hasattr(destination, "fill_value"):
 */
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_log); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 433, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_debug); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 433, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_tuple__21, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 433, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "rasterio/_warp.pyx":435
 *         log.debug("Wrote array to temp output dataset")
 * 
 *         if dst_nodata is None and hasattr(destination, "fill_value"):             # <<<<<<<<<<<<<<
 *             # destination is a masked array
 *             dst_nodata = destination.fill_value
 */
    __pyx_t_1 = (__pyx_v_dst_nodata == Py_None);
    __pyx_t_9 = (__pyx_t_1 != 0);
    if (__pyx_t_9) {
    } else {
      __pyx_t_2 = __pyx_t_9;
      goto __pyx_L70_bool_binop_done;
    }
    __pyx_t_9 = PyObject_HasAttr(__pyx_v_destination, __pyx_n_s_fill_value); if (unlikely(__pyx_t_9 == -1)) __PYX_ERR(0, 435, __pyx_L1_error)
    __pyx_t_1 = (__pyx_t_9 != 0);
    __pyx_t_2 = __pyx_t_1;
    __pyx_L70_bool_binop_done:;
    if (__pyx_t_2) {

      /* "rasterio/_warp.pyx":437
 *         if dst_nodata is None and hasattr(destination, "fill_value"):
 *             # destination is a masked array
 *             dst_nodata = destination.fill_value             # <<<<<<<<<<<<<<
 * 
 *     elif isinstance(destination, tuple):
 */
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_destination, __pyx_n_s_fill_value); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 437, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF_SET(__pyx_v_dst_nodata, __pyx_t_5);
      __pyx_t_5 = 0;

      /* "rasterio/_warp.pyx":435
 *         log.debug("Wrote array to temp output dataset")
 * 
 *         if dst_nodata is None and hasattr(destination, "fill_value"):             # <<<<<<<<<<<<<<
 *             # destination is a masked array
 *             dst_nodata = destination.fill_value
 */
    }

    /* "rasterio/_warp.pyx":379
 * 
 *     # Next, do the same for the destination raster.
 *     if dtypes.is_ndarray(destination):             # <<<<<<<<<<<<<<
 *         if len(destination.shape) == 2:
 *             destination = destination.reshape(1, *destination.shape)
 */
    goto __pyx_L45;
  }

  /* "rasterio/_warp.pyx":439
 *             dst_nodata = destination.fill_value
 * 
 *     elif isinstance(destination, tuple):             # <<<<<<<<<<<<<<
 *         udr, dst_bidx, _, _ = destination
 *         if isinstance(dst_bidx, int):
 */
  __pyx_t_2 = PyTuple_Check(__pyx_v_destination); 
  __pyx_t_1 = (__pyx_t_2 != 0);
  if (__pyx_t_1) {

    /* "rasterio/_warp.pyx":440
 * 
 *     elif isinstance(destination, tuple):
 *         udr, dst_bidx, _, _ = destination             # <<<<<<<<<<<<<<
 *         if isinstance(dst_bidx, int):
 *             dst_bidx = [dst_bidx]
 */
    if ((likely(PyTuple_CheckExact(__pyx_v_destination))) || (PyList_CheckExact(__pyx_v_destination))) {
      PyObject* sequence = __pyx_v_destination;
      #if !CYTHON_COMPILING_IN_PYPY
      Py_ssize_t size = Py_SIZE(sequence);
      #else
      Py_ssize_t size = PySequence_Size(sequence);
      #endif
      if (unlikely(size != 4)) {
        if (size > 4) __Pyx_RaiseTooManyValuesError(4);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 440, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_5 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_4 = PyTuple_GET_ITEM(sequence, 1); 
        __pyx_t_6 = PyTuple_GET_ITEM(sequence, 2); 
        __pyx_t_14 = PyTuple_GET_ITEM(sequence, 3); 
      } else {
        __pyx_t_5 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_4 = PyList_GET_ITEM(sequence, 1); 
        __pyx_t_6 = PyList_GET_ITEM(sequence, 2); 
        __pyx_t_14 = PyList_GET_ITEM(sequence, 3); 
      }
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_14);
      #else
      {
        Py_ssize_t i;
        PyObject** temps[4] = {&__pyx_t_5,&__pyx_t_4,&__pyx_t_6,&__pyx_t_14};
        for (i=0; i < 4; i++) {
          PyObject* item = PySequence_ITEM(sequence, i); if (unlikely(!item)) __PYX_ERR(0, 440, __pyx_L1_error)
          __Pyx_GOTREF(item);
          *(temps[i]) = item;
        }
      }
      #endif
    } else {
      Py_ssize_t index = -1;
      PyObject** temps[4] = {&__pyx_t_5,&__pyx_t_4,&__pyx_t_6,&__pyx_t_14};
      __pyx_t_3 = PyObject_GetIter(__pyx_v_destination); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 440, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_28 = Py_TYPE(__pyx_t_3)->tp_iternext;
      for (index=0; index < 4; index++) {
        PyObject* item = __pyx_t_28(__pyx_t_3); if (unlikely(!item)) goto __pyx_L72_unpacking_failed;
        __Pyx_GOTREF(item);
        *(temps[index]) = item;
      }
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_28(__pyx_t_3), 4) < 0) __PYX_ERR(0, 440, __pyx_L1_error)
      __pyx_t_28 = NULL;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      goto __pyx_L73_unpacking_done;
      __pyx_L72_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_28 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 440, __pyx_L1_error)
      __pyx_L73_unpacking_done:;
    }
    __pyx_v_udr = __pyx_t_5;
    __pyx_t_5 = 0;
    __pyx_v_dst_bidx = __pyx_t_4;
    __pyx_t_4 = 0;
    __pyx_v__ = __pyx_t_6;
    __pyx_t_6 = 0;
    __Pyx_DECREF_SET(__pyx_v__, __pyx_t_14);
    __pyx_t_14 = 0;

    /* "rasterio/_warp.pyx":441
 *     elif isinstance(destination, tuple):
 *         udr, dst_bidx, _, _ = destination
 *         if isinstance(dst_bidx, int):             # <<<<<<<<<<<<<<
 *             dst_bidx = [dst_bidx]
 *         udr = destination.ds
 */
    __pyx_t_1 = PyInt_Check(__pyx_v_dst_bidx); 
    __pyx_t_2 = (__pyx_t_1 != 0);
    if (__pyx_t_2) {

      /* "rasterio/_warp.pyx":442
 *         udr, dst_bidx, _, _ = destination
 *         if isinstance(dst_bidx, int):
 *             dst_bidx = [dst_bidx]             # <<<<<<<<<<<<<<
 *         udr = destination.ds
 *         dst_dataset = (<DatasetReaderBase?>udr).handle()
 */
      __pyx_t_14 = PyList_New(1); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 442, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      __Pyx_INCREF(__pyx_v_dst_bidx);
      __Pyx_GIVEREF(__pyx_v_dst_bidx);
      PyList_SET_ITEM(__pyx_t_14, 0, __pyx_v_dst_bidx);
      __Pyx_DECREF_SET(__pyx_v_dst_bidx, __pyx_t_14);
      __pyx_t_14 = 0;

      /* "rasterio/_warp.pyx":441
 *     elif isinstance(destination, tuple):
 *         udr, dst_bidx, _, _ = destination
 *         if isinstance(dst_bidx, int):             # <<<<<<<<<<<<<<
 *             dst_bidx = [dst_bidx]
 *         udr = destination.ds
 */
    }

    /* "rasterio/_warp.pyx":443
 *         if isinstance(dst_bidx, int):
 *             dst_bidx = [dst_bidx]
 *         udr = destination.ds             # <<<<<<<<<<<<<<
 *         dst_dataset = (<DatasetReaderBase?>udr).handle()
 *         if dst_nodata is None:
 */
    __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_v_destination, __pyx_n_s_ds); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 443, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF_SET(__pyx_v_udr, __pyx_t_14);
    __pyx_t_14 = 0;

    /* "rasterio/_warp.pyx":444
 *             dst_bidx = [dst_bidx]
 *         udr = destination.ds
 *         dst_dataset = (<DatasetReaderBase?>udr).handle()             # <<<<<<<<<<<<<<
 *         if dst_nodata is None:
 *             dst_nodata = udr.nodata
 */
    if (!(likely(__Pyx_TypeTest(__pyx_v_udr, __pyx_ptype_8rasterio_3_io_DatasetReaderBase)))) __PYX_ERR(0, 444, __pyx_L1_error)
    __pyx_t_29 = ((struct __pyx_vtabstruct_8rasterio_3_io_DatasetReaderBase *)((struct __pyx_obj_8rasterio_3_io_DatasetReaderBase *)__pyx_v_udr)->__pyx_base.__pyx_vtab)->__pyx_base.handle(((struct __pyx_obj_8rasterio_5_base_DatasetBase *)((struct __pyx_obj_8rasterio_3_io_DatasetReaderBase *)__pyx_v_udr))); if (unlikely(__pyx_t_29 == NULL)) __PYX_ERR(0, 444, __pyx_L1_error)
    __pyx_v_dst_dataset = __pyx_t_29;

    /* "rasterio/_warp.pyx":445
 *         udr = destination.ds
 *         dst_dataset = (<DatasetReaderBase?>udr).handle()
 *         if dst_nodata is None:             # <<<<<<<<<<<<<<
 *             dst_nodata = udr.nodata
 *     else:
 */
    __pyx_t_2 = (__pyx_v_dst_nodata == Py_None);
    __pyx_t_1 = (__pyx_t_2 != 0);
    if (__pyx_t_1) {

      /* "rasterio/_warp.pyx":446
 *         dst_dataset = (<DatasetReaderBase?>udr).handle()
 *         if dst_nodata is None:
 *             dst_nodata = udr.nodata             # <<<<<<<<<<<<<<
 *     else:
 *         raise ValueError("Invalid destination")
 */
      __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_v_udr, __pyx_n_s_nodata); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 446, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      __Pyx_DECREF_SET(__pyx_v_dst_nodata, __pyx_t_14);
      __pyx_t_14 = 0;

      /* "rasterio/_warp.pyx":445
 *         udr = destination.ds
 *         dst_dataset = (<DatasetReaderBase?>udr).handle()
 *         if dst_nodata is None:             # <<<<<<<<<<<<<<
 *             dst_nodata = udr.nodata
 *     else:
 */
    }

    /* "rasterio/_warp.pyx":439
 *             dst_nodata = destination.fill_value
 * 
 *     elif isinstance(destination, tuple):             # <<<<<<<<<<<<<<
 *         udr, dst_bidx, _, _ = destination
 *         if isinstance(dst_bidx, int):
 */
    goto __pyx_L45;
  }

  /* "rasterio/_warp.pyx":448
 *             dst_nodata = udr.nodata
 *     else:
 *         raise ValueError("Invalid destination")             # <<<<<<<<<<<<<<
 * 
 *     # Set up GDALCreateGenImgProjTransformer2 keyword arguments.
 */
  /*else*/ {
    __pyx_t_14 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__22, NULL); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 448, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_Raise(__pyx_t_14, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    __PYX_ERR(0, 448, __pyx_L1_error)
  }
  __pyx_L45:;

  /* "rasterio/_warp.pyx":451
 * 
 *     # Set up GDALCreateGenImgProjTransformer2 keyword arguments.
 *     cdef char **imgProjOptions = NULL             # <<<<<<<<<<<<<<
 *     CSLSetNameValue(imgProjOptions, "GCPS_OK", "TRUE")
 * 
 */
  __pyx_v_imgProjOptions = NULL;

  /* "rasterio/_warp.pyx":452
 *     # Set up GDALCreateGenImgProjTransformer2 keyword arguments.
 *     cdef char **imgProjOptions = NULL
 *     CSLSetNameValue(imgProjOptions, "GCPS_OK", "TRUE")             # <<<<<<<<<<<<<<
 * 
 *     # See http://www.gdal.org/gdal__alg_8h.html#a94cd172f78dbc41d6f407d662914f2e3
 */
  CSLSetNameValue(__pyx_v_imgProjOptions, ((char *)"GCPS_OK"), ((char *)"TRUE"));

  /* "rasterio/_warp.pyx":459
 *     # image transformer options mapping; unsupported options should be
 *     # okay.
 *     for key, val in kwargs.items():             # <<<<<<<<<<<<<<
 *         key = key.upper().encode('utf-8')
 *         val = str(val).upper().encode('utf-8')
 */
  __pyx_t_14 = __Pyx_PyDict_Items(__pyx_v_kwargs); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 459, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  if (likely(PyList_CheckExact(__pyx_t_14)) || PyTuple_CheckExact(__pyx_t_14)) {
    __pyx_t_6 = __pyx_t_14; __Pyx_INCREF(__pyx_t_6); __pyx_t_7 = 0;
    __pyx_t_24 = NULL;
  } else {
    __pyx_t_7 = -1; __pyx_t_6 = PyObject_GetIter(__pyx_t_14); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 459, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_24 = Py_TYPE(__pyx_t_6)->tp_iternext; if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 459, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  for (;;) {
    if (likely(!__pyx_t_24)) {
      if (likely(PyList_CheckExact(__pyx_t_6))) {
        if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_6)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_14 = PyList_GET_ITEM(__pyx_t_6, __pyx_t_7); __Pyx_INCREF(__pyx_t_14); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 459, __pyx_L1_error)
        #else
        __pyx_t_14 = PySequence_ITEM(__pyx_t_6, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 459, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_14);
        #endif
      } else {
        if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_6)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_14 = PyTuple_GET_ITEM(__pyx_t_6, __pyx_t_7); __Pyx_INCREF(__pyx_t_14); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 459, __pyx_L1_error)
        #else
        __pyx_t_14 = PySequence_ITEM(__pyx_t_6, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 459, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_14);
        #endif
      }
    } else {
      __pyx_t_14 = __pyx_t_24(__pyx_t_6);
      if (unlikely(!__pyx_t_14)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 459, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_14);
    }
    if ((likely(PyTuple_CheckExact(__pyx_t_14))) || (PyList_CheckExact(__pyx_t_14))) {
      PyObject* sequence = __pyx_t_14;
      #if !CYTHON_COMPILING_IN_PYPY
      Py_ssize_t size = Py_SIZE(sequence);
      #else
      Py_ssize_t size = PySequence_Size(sequence);
      #endif
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 459, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_4 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_5 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_4 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_5 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      #else
      __pyx_t_4 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 459, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 459, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      #endif
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_3 = PyObject_GetIter(__pyx_t_14); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 459, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      __pyx_t_28 = Py_TYPE(__pyx_t_3)->tp_iternext;
      index = 0; __pyx_t_4 = __pyx_t_28(__pyx_t_3); if (unlikely(!__pyx_t_4)) goto __pyx_L78_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_4);
      index = 1; __pyx_t_5 = __pyx_t_28(__pyx_t_3); if (unlikely(!__pyx_t_5)) goto __pyx_L78_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_5);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_28(__pyx_t_3), 2) < 0) __PYX_ERR(0, 459, __pyx_L1_error)
      __pyx_t_28 = NULL;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      goto __pyx_L79_unpacking_done;
      __pyx_L78_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_28 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 459, __pyx_L1_error)
      __pyx_L79_unpacking_done:;
    }
    __Pyx_XDECREF_SET(__pyx_v_key, __pyx_t_4);
    __pyx_t_4 = 0;
    __Pyx_XDECREF_SET(__pyx_v_val, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "rasterio/_warp.pyx":460
 *     # okay.
 *     for key, val in kwargs.items():
 *         key = key.upper().encode('utf-8')             # <<<<<<<<<<<<<<
 *         val = str(val).upper().encode('utf-8')
 *         imgProjOptions = CSLSetNameValue(
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_key, __pyx_n_s_upper); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 460, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    if (__pyx_t_4) {
      __pyx_t_14 = __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_4); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 460, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else {
      __pyx_t_14 = __Pyx_PyObject_CallNoArg(__pyx_t_5); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 460, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_14, __pyx_n_s_encode); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 460, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    __pyx_t_14 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_tuple__23, NULL); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 460, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF_SET(__pyx_v_key, __pyx_t_14);
    __pyx_t_14 = 0;

    /* "rasterio/_warp.pyx":461
 *     for key, val in kwargs.items():
 *         key = key.upper().encode('utf-8')
 *         val = str(val).upper().encode('utf-8')             # <<<<<<<<<<<<<<
 *         imgProjOptions = CSLSetNameValue(
 *             imgProjOptions, <const char *>key, <const char *>val)
 */
    __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 461, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_v_val);
    __Pyx_GIVEREF(__pyx_v_val);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_val);
    __pyx_t_4 = __Pyx_PyObject_Call(((PyObject *)(&PyString_Type)), __pyx_t_5, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 461, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_upper); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 461, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    if (__pyx_t_4) {
      __pyx_t_14 = __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_4); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 461, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else {
      __pyx_t_14 = __Pyx_PyObject_CallNoArg(__pyx_t_5); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 461, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_14, __pyx_n_s_encode); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 461, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    __pyx_t_14 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_tuple__24, NULL); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 461, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF_SET(__pyx_v_val, __pyx_t_14);
    __pyx_t_14 = 0;

    /* "rasterio/_warp.pyx":463
 *         val = str(val).upper().encode('utf-8')
 *         imgProjOptions = CSLSetNameValue(
 *             imgProjOptions, <const char *>key, <const char *>val)             # <<<<<<<<<<<<<<
 * 
 *     try:
 */
    __pyx_t_32 = __Pyx_PyObject_AsString(__pyx_v_key); if (unlikely((!__pyx_t_32) && PyErr_Occurred())) __PYX_ERR(0, 463, __pyx_L1_error)
    __pyx_t_33 = __Pyx_PyObject_AsString(__pyx_v_val); if (unlikely((!__pyx_t_33) && PyErr_Occurred())) __PYX_ERR(0, 463, __pyx_L1_error)

    /* "rasterio/_warp.pyx":462
 *         key = key.upper().encode('utf-8')
 *         val = str(val).upper().encode('utf-8')
 *         imgProjOptions = CSLSetNameValue(             # <<<<<<<<<<<<<<
 *             imgProjOptions, <const char *>key, <const char *>val)
 * 
 */
    __pyx_v_imgProjOptions = CSLSetNameValue(__pyx_v_imgProjOptions, ((char const *)__pyx_t_32), ((char const *)__pyx_t_33));

    /* "rasterio/_warp.pyx":459
 *     # image transformer options mapping; unsupported options should be
 *     # okay.
 *     for key, val in kwargs.items():             # <<<<<<<<<<<<<<
 *         key = key.upper().encode('utf-8')
 *         val = str(val).upper().encode('utf-8')
 */
  }
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "rasterio/_warp.pyx":465
 *             imgProjOptions, <const char *>key, <const char *>val)
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         hTransformArg = exc_wrap_pointer(
 *             GDALCreateGenImgProjTransformer2(
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_23, &__pyx_t_22, &__pyx_t_21);
    __Pyx_XGOTREF(__pyx_t_23);
    __Pyx_XGOTREF(__pyx_t_22);
    __Pyx_XGOTREF(__pyx_t_21);
    /*try:*/ {

      /* "rasterio/_warp.pyx":466
 * 
 *     try:
 *         hTransformArg = exc_wrap_pointer(             # <<<<<<<<<<<<<<
 *             GDALCreateGenImgProjTransformer2(
 *                 src_dataset, dst_dataset, imgProjOptions))
 */
      __pyx_t_13 = __pyx_f_8rasterio_4_err_exc_wrap_pointer(GDALCreateGenImgProjTransformer2(__pyx_v_src_dataset, __pyx_v_dst_dataset, __pyx_v_imgProjOptions)); if (unlikely(__pyx_t_13 == NULL)) __PYX_ERR(0, 466, __pyx_L80_error)
      __pyx_v_hTransformArg = __pyx_t_13;

      /* "rasterio/_warp.pyx":469
 *             GDALCreateGenImgProjTransformer2(
 *                 src_dataset, dst_dataset, imgProjOptions))
 *         hTransformArg = exc_wrap_pointer(             # <<<<<<<<<<<<<<
 *             GDALCreateApproxTransformer(
 *                 GDALGenImgProjTransform, hTransformArg, tolerance))
 */
      __pyx_t_13 = __pyx_f_8rasterio_4_err_exc_wrap_pointer(GDALCreateApproxTransformer(GDALGenImgProjTransform, __pyx_v_hTransformArg, __pyx_v_tolerance)); if (unlikely(__pyx_t_13 == NULL)) __PYX_ERR(0, 469, __pyx_L80_error)
      __pyx_v_hTransformArg = __pyx_t_13;

      /* "rasterio/_warp.pyx":472
 *             GDALCreateApproxTransformer(
 *                 GDALGenImgProjTransform, hTransformArg, tolerance))
 *         pfnTransformer = GDALApproxTransform             # <<<<<<<<<<<<<<
 *         GDALApproxTransformerOwnsSubtransformer(hTransformArg, 1)
 * 
 */
      __pyx_v_pfnTransformer = GDALApproxTransform;

      /* "rasterio/_warp.pyx":473
 *                 GDALGenImgProjTransform, hTransformArg, tolerance))
 *         pfnTransformer = GDALApproxTransform
 *         GDALApproxTransformerOwnsSubtransformer(hTransformArg, 1)             # <<<<<<<<<<<<<<
 * 
 *         log.debug("Created transformer and options.")
 */
      GDALApproxTransformerOwnsSubtransformer(__pyx_v_hTransformArg, 1);

      /* "rasterio/_warp.pyx":475
 *         GDALApproxTransformerOwnsSubtransformer(hTransformArg, 1)
 * 
 *         log.debug("Created transformer and options.")             # <<<<<<<<<<<<<<
 * 
 *     except:
 */
      __pyx_t_6 = __Pyx_GetModuleGlobalName(__pyx_n_s_log); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 475, __pyx_L80_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_debug); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 475, __pyx_L80_error)
      __Pyx_GOTREF(__pyx_t_14);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_14, __pyx_tuple__25, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 475, __pyx_L80_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

      /* "rasterio/_warp.pyx":465
 *             imgProjOptions, <const char *>key, <const char *>val)
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         hTransformArg = exc_wrap_pointer(
 *             GDALCreateGenImgProjTransformer2(
 */
    }
    __Pyx_XDECREF(__pyx_t_23); __pyx_t_23 = 0;
    __Pyx_XDECREF(__pyx_t_22); __pyx_t_22 = 0;
    __Pyx_XDECREF(__pyx_t_21); __pyx_t_21 = 0;
    goto __pyx_L87_try_end;
    __pyx_L80_error:;
    __Pyx_PyThreadState_assign
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "rasterio/_warp.pyx":477
 *         log.debug("Created transformer and options.")
 * 
 *     except:             # <<<<<<<<<<<<<<
 *         GDALDestroyApproxTransformer(hTransformArg)
 *         CPLFree(imgProjOptions)
 */
    /*except:*/ {
      __Pyx_AddTraceback("rasterio._warp._reproject", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_6, &__pyx_t_14, &__pyx_t_5) < 0) __PYX_ERR(0, 477, __pyx_L82_except_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_14);
      __Pyx_GOTREF(__pyx_t_5);

      /* "rasterio/_warp.pyx":478
 * 
 *     except:
 *         GDALDestroyApproxTransformer(hTransformArg)             # <<<<<<<<<<<<<<
 *         CPLFree(imgProjOptions)
 *         raise
 */
      GDALDestroyApproxTransformer(__pyx_v_hTransformArg);

      /* "rasterio/_warp.pyx":479
 *     except:
 *         GDALDestroyApproxTransformer(hTransformArg)
 *         CPLFree(imgProjOptions)             # <<<<<<<<<<<<<<
 *         raise
 * 
 */
      CPLFree(__pyx_v_imgProjOptions);

      /* "rasterio/_warp.pyx":480
 *         GDALDestroyApproxTransformer(hTransformArg)
 *         CPLFree(imgProjOptions)
 *         raise             # <<<<<<<<<<<<<<
 * 
 *     # Note: warp_extras is pointed to different memory locations on every
 */
      __Pyx_GIVEREF(__pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_14);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_ErrRestoreWithState(__pyx_t_6, __pyx_t_14, __pyx_t_5);
      __pyx_t_6 = 0; __pyx_t_14 = 0; __pyx_t_5 = 0; 
      __PYX_ERR(0, 480, __pyx_L82_except_error)
    }
    __pyx_L82_except_error:;

    /* "rasterio/_warp.pyx":465
 *             imgProjOptions, <const char *>key, <const char *>val)
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         hTransformArg = exc_wrap_pointer(
 *             GDALCreateGenImgProjTransformer2(
 */
    __Pyx_PyThreadState_assign
    __Pyx_XGIVEREF(__pyx_t_23);
    __Pyx_XGIVEREF(__pyx_t_22);
    __Pyx_XGIVEREF(__pyx_t_21);
    __Pyx_ExceptionReset(__pyx_t_23, __pyx_t_22, __pyx_t_21);
    goto __pyx_L1_error;
    __pyx_L87_try_end:;
  }

  /* "rasterio/_warp.pyx":487
 *     # warp_extras = psWOptions.papszWarpOptions
 * 
 *     valb = str(num_threads).encode('utf-8')             # <<<<<<<<<<<<<<
 *     warp_extras = CSLSetNameValue(warp_extras, "NUM_THREADS", <const char *>valb)
 * 
 */
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 487, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_v_num_threads);
  __Pyx_GIVEREF(__pyx_v_num_threads);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_num_threads);
  __pyx_t_14 = __Pyx_PyObject_Call(((PyObject *)(&PyString_Type)), __pyx_t_5, NULL); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 487, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_14, __pyx_n_s_encode); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 487, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  __pyx_t_14 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_tuple__26, NULL); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 487, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_valb = __pyx_t_14;
  __pyx_t_14 = 0;

  /* "rasterio/_warp.pyx":488
 * 
 *     valb = str(num_threads).encode('utf-8')
 *     warp_extras = CSLSetNameValue(warp_extras, "NUM_THREADS", <const char *>valb)             # <<<<<<<<<<<<<<
 * 
 *     log.debug("Setting NUM_THREADS option: %d", num_threads)
 */
  __pyx_t_34 = __Pyx_PyObject_AsString(__pyx_v_valb); if (unlikely((!__pyx_t_34) && PyErr_Occurred())) __PYX_ERR(0, 488, __pyx_L1_error)
  __pyx_v_warp_extras = CSLSetNameValue(__pyx_v_warp_extras, ((char *)"NUM_THREADS"), ((char const *)__pyx_t_34));

  /* "rasterio/_warp.pyx":490
 *     warp_extras = CSLSetNameValue(warp_extras, "NUM_THREADS", <const char *>valb)
 * 
 *     log.debug("Setting NUM_THREADS option: %d", num_threads)             # <<<<<<<<<<<<<<
 * 
 *     if init_dest_nodata:
 */
  __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_log); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 490, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_debug); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 490, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = NULL;
  __pyx_t_19 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
      __pyx_t_19 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_6)) {
    PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_kp_s_Setting_NUM_THREADS_option_d, __pyx_v_num_threads};
    __pyx_t_14 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_19, 2+__pyx_t_19); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 490, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_14);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
    PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_kp_s_Setting_NUM_THREADS_option_d, __pyx_v_num_threads};
    __pyx_t_14 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_19, 2+__pyx_t_19); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 490, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_14);
  } else
  #endif
  {
    __pyx_t_4 = PyTuple_New(2+__pyx_t_19); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 490, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_kp_s_Setting_NUM_THREADS_option_d);
    __Pyx_GIVEREF(__pyx_kp_s_Setting_NUM_THREADS_option_d);
    PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_19, __pyx_kp_s_Setting_NUM_THREADS_option_d);
    __Pyx_INCREF(__pyx_v_num_threads);
    __Pyx_GIVEREF(__pyx_v_num_threads);
    PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_19, __pyx_v_num_threads);
    __pyx_t_14 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_4, NULL); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 490, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

  /* "rasterio/_warp.pyx":492
 *     log.debug("Setting NUM_THREADS option: %d", num_threads)
 * 
 *     if init_dest_nodata:             # <<<<<<<<<<<<<<
 *         warp_extras = CSLSetNameValue(warp_extras, "INIT_DEST", "NO_DATA")
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_init_dest_nodata); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 492, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "rasterio/_warp.pyx":493
 * 
 *     if init_dest_nodata:
 *         warp_extras = CSLSetNameValue(warp_extras, "INIT_DEST", "NO_DATA")             # <<<<<<<<<<<<<<
 * 
 *     # See http://www.gdal.org/structGDALWarpOptions.html#a0ed77f9917bb96c7a9aabd73d4d06e08
 */
    __pyx_v_warp_extras = CSLSetNameValue(__pyx_v_warp_extras, ((char *)"INIT_DEST"), ((char *)"NO_DATA"));

    /* "rasterio/_warp.pyx":492
 *     log.debug("Setting NUM_THREADS option: %d", num_threads)
 * 
 *     if init_dest_nodata:             # <<<<<<<<<<<<<<
 *         warp_extras = CSLSetNameValue(warp_extras, "INIT_DEST", "NO_DATA")
 * 
 */
  }

  /* "rasterio/_warp.pyx":498
 *     # for a list of supported options. Copying unsupported options
 *     # is fine.
 *     for key, val in kwargs.items():             # <<<<<<<<<<<<<<
 *         key = key.upper().encode('utf-8')
 *         val = str(val).upper().encode('utf-8')
 */
  __pyx_t_14 = __Pyx_PyDict_Items(__pyx_v_kwargs); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 498, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  if (likely(PyList_CheckExact(__pyx_t_14)) || PyTuple_CheckExact(__pyx_t_14)) {
    __pyx_t_6 = __pyx_t_14; __Pyx_INCREF(__pyx_t_6); __pyx_t_7 = 0;
    __pyx_t_24 = NULL;
  } else {
    __pyx_t_7 = -1; __pyx_t_6 = PyObject_GetIter(__pyx_t_14); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 498, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_24 = Py_TYPE(__pyx_t_6)->tp_iternext; if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 498, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  for (;;) {
    if (likely(!__pyx_t_24)) {
      if (likely(PyList_CheckExact(__pyx_t_6))) {
        if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_6)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_14 = PyList_GET_ITEM(__pyx_t_6, __pyx_t_7); __Pyx_INCREF(__pyx_t_14); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 498, __pyx_L1_error)
        #else
        __pyx_t_14 = PySequence_ITEM(__pyx_t_6, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 498, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_14);
        #endif
      } else {
        if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_6)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_14 = PyTuple_GET_ITEM(__pyx_t_6, __pyx_t_7); __Pyx_INCREF(__pyx_t_14); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 498, __pyx_L1_error)
        #else
        __pyx_t_14 = PySequence_ITEM(__pyx_t_6, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 498, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_14);
        #endif
      }
    } else {
      __pyx_t_14 = __pyx_t_24(__pyx_t_6);
      if (unlikely(!__pyx_t_14)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 498, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_14);
    }
    if ((likely(PyTuple_CheckExact(__pyx_t_14))) || (PyList_CheckExact(__pyx_t_14))) {
      PyObject* sequence = __pyx_t_14;
      #if !CYTHON_COMPILING_IN_PYPY
      Py_ssize_t size = Py_SIZE(sequence);
      #else
      Py_ssize_t size = PySequence_Size(sequence);
      #endif
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 498, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_4 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_5 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_4 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_5 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      #else
      __pyx_t_4 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 498, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 498, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      #endif
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_3 = PyObject_GetIter(__pyx_t_14); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 498, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      __pyx_t_28 = Py_TYPE(__pyx_t_3)->tp_iternext;
      index = 0; __pyx_t_4 = __pyx_t_28(__pyx_t_3); if (unlikely(!__pyx_t_4)) goto __pyx_L93_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_4);
      index = 1; __pyx_t_5 = __pyx_t_28(__pyx_t_3); if (unlikely(!__pyx_t_5)) goto __pyx_L93_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_5);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_28(__pyx_t_3), 2) < 0) __PYX_ERR(0, 498, __pyx_L1_error)
      __pyx_t_28 = NULL;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      goto __pyx_L94_unpacking_done;
      __pyx_L93_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_28 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 498, __pyx_L1_error)
      __pyx_L94_unpacking_done:;
    }
    __Pyx_XDECREF_SET(__pyx_v_key, __pyx_t_4);
    __pyx_t_4 = 0;
    __Pyx_XDECREF_SET(__pyx_v_val, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "rasterio/_warp.pyx":499
 *     # is fine.
 *     for key, val in kwargs.items():
 *         key = key.upper().encode('utf-8')             # <<<<<<<<<<<<<<
 *         val = str(val).upper().encode('utf-8')
 *         warp_extras = CSLSetNameValue(
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_key, __pyx_n_s_upper); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 499, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    if (__pyx_t_4) {
      __pyx_t_14 = __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_4); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 499, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else {
      __pyx_t_14 = __Pyx_PyObject_CallNoArg(__pyx_t_5); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 499, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_14, __pyx_n_s_encode); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 499, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    __pyx_t_14 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_tuple__27, NULL); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 499, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF_SET(__pyx_v_key, __pyx_t_14);
    __pyx_t_14 = 0;

    /* "rasterio/_warp.pyx":500
 *     for key, val in kwargs.items():
 *         key = key.upper().encode('utf-8')
 *         val = str(val).upper().encode('utf-8')             # <<<<<<<<<<<<<<
 *         warp_extras = CSLSetNameValue(
 *             warp_extras, <const char *>key, <const char *>val)
 */
    __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 500, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_v_val);
    __Pyx_GIVEREF(__pyx_v_val);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_val);
    __pyx_t_4 = __Pyx_PyObject_Call(((PyObject *)(&PyString_Type)), __pyx_t_5, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 500, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_upper); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 500, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    if (__pyx_t_4) {
      __pyx_t_14 = __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_4); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 500, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else {
      __pyx_t_14 = __Pyx_PyObject_CallNoArg(__pyx_t_5); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 500, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_14, __pyx_n_s_encode); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 500, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    __pyx_t_14 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_tuple__28, NULL); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 500, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF_SET(__pyx_v_val, __pyx_t_14);
    __pyx_t_14 = 0;

    /* "rasterio/_warp.pyx":502
 *         val = str(val).upper().encode('utf-8')
 *         warp_extras = CSLSetNameValue(
 *             warp_extras, <const char *>key, <const char *>val)             # <<<<<<<<<<<<<<
 * 
 *     psWOptions = create_warp_options(
 */
    __pyx_t_35 = __Pyx_PyObject_AsString(__pyx_v_key); if (unlikely((!__pyx_t_35) && PyErr_Occurred())) __PYX_ERR(0, 502, __pyx_L1_error)
    __pyx_t_36 = __Pyx_PyObject_AsString(__pyx_v_val); if (unlikely((!__pyx_t_36) && PyErr_Occurred())) __PYX_ERR(0, 502, __pyx_L1_error)

    /* "rasterio/_warp.pyx":501
 *         key = key.upper().encode('utf-8')
 *         val = str(val).upper().encode('utf-8')
 *         warp_extras = CSLSetNameValue(             # <<<<<<<<<<<<<<
 *             warp_extras, <const char *>key, <const char *>val)
 * 
 */
    __pyx_v_warp_extras = CSLSetNameValue(__pyx_v_warp_extras, ((char const *)__pyx_t_35), ((char const *)__pyx_t_36));

    /* "rasterio/_warp.pyx":498
 *     # for a list of supported options. Copying unsupported options
 *     # is fine.
 *     for key, val in kwargs.items():             # <<<<<<<<<<<<<<
 *         key = key.upper().encode('utf-8')
 *         val = str(val).upper().encode('utf-8')
 */
  }
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "rasterio/_warp.pyx":505
 * 
 *     psWOptions = create_warp_options(
 *         <GDALResampleAlg>resampling, src_nodata,             # <<<<<<<<<<<<<<
 *         dst_nodata, src_count, <const char **>warp_extras)
 * 
 */
  __pyx_t_37 = ((GDALResampleAlg)__Pyx_PyInt_As_GDALResampleAlg(__pyx_v_resampling)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 505, __pyx_L1_error)

  /* "rasterio/_warp.pyx":504
 *             warp_extras, <const char *>key, <const char *>val)
 * 
 *     psWOptions = create_warp_options(             # <<<<<<<<<<<<<<
 *         <GDALResampleAlg>resampling, src_nodata,
 *         dst_nodata, src_count, <const char **>warp_extras)
 */
  __pyx_t_38 = __pyx_f_8rasterio_5_warp_create_warp_options(((GDALResampleAlg)__pyx_t_37), __pyx_v_src_nodata, __pyx_v_dst_nodata, __pyx_v_src_count, ((char const **)__pyx_v_warp_extras)); if (unlikely(__pyx_t_38 == NULL)) __PYX_ERR(0, 504, __pyx_L1_error)
  __pyx_v_psWOptions = __pyx_t_38;

  /* "rasterio/_warp.pyx":508
 *         dst_nodata, src_count, <const char **>warp_extras)
 * 
 *     psWOptions.pfnTransformer = pfnTransformer             # <<<<<<<<<<<<<<
 *     psWOptions.pTransformerArg = hTransformArg
 *     psWOptions.hSrcDS = src_dataset
 */
  __pyx_v_psWOptions->pfnTransformer = __pyx_v_pfnTransformer;

  /* "rasterio/_warp.pyx":509
 * 
 *     psWOptions.pfnTransformer = pfnTransformer
 *     psWOptions.pTransformerArg = hTransformArg             # <<<<<<<<<<<<<<
 *     psWOptions.hSrcDS = src_dataset
 *     psWOptions.hDstDS = dst_dataset
 */
  __pyx_v_psWOptions->pTransformerArg = __pyx_v_hTransformArg;

  /* "rasterio/_warp.pyx":510
 *     psWOptions.pfnTransformer = pfnTransformer
 *     psWOptions.pTransformerArg = hTransformArg
 *     psWOptions.hSrcDS = src_dataset             # <<<<<<<<<<<<<<
 *     psWOptions.hDstDS = dst_dataset
 *     psWOptions.nBandCount = src_count
 */
  __pyx_v_psWOptions->hSrcDS = __pyx_v_src_dataset;

  /* "rasterio/_warp.pyx":511
 *     psWOptions.pTransformerArg = hTransformArg
 *     psWOptions.hSrcDS = src_dataset
 *     psWOptions.hDstDS = dst_dataset             # <<<<<<<<<<<<<<
 *     psWOptions.nBandCount = src_count
 *     psWOptions.panSrcBands = <int *>CPLMalloc(src_count*sizeof(int))
 */
  __pyx_v_psWOptions->hDstDS = __pyx_v_dst_dataset;

  /* "rasterio/_warp.pyx":512
 *     psWOptions.hSrcDS = src_dataset
 *     psWOptions.hDstDS = dst_dataset
 *     psWOptions.nBandCount = src_count             # <<<<<<<<<<<<<<
 *     psWOptions.panSrcBands = <int *>CPLMalloc(src_count*sizeof(int))
 *     psWOptions.panDstBands = <int *>CPLMalloc(src_count*sizeof(int))
 */
  __pyx_v_psWOptions->nBandCount = __pyx_v_src_count;

  /* "rasterio/_warp.pyx":513
 *     psWOptions.hDstDS = dst_dataset
 *     psWOptions.nBandCount = src_count
 *     psWOptions.panSrcBands = <int *>CPLMalloc(src_count*sizeof(int))             # <<<<<<<<<<<<<<
 *     psWOptions.panDstBands = <int *>CPLMalloc(src_count*sizeof(int))
 * 
 */
  __pyx_v_psWOptions->panSrcBands = ((int *)CPLMalloc((__pyx_v_src_count * (sizeof(int)))));

  /* "rasterio/_warp.pyx":514
 *     psWOptions.nBandCount = src_count
 *     psWOptions.panSrcBands = <int *>CPLMalloc(src_count*sizeof(int))
 *     psWOptions.panDstBands = <int *>CPLMalloc(src_count*sizeof(int))             # <<<<<<<<<<<<<<
 * 
 *     for i in range(src_count):
 */
  __pyx_v_psWOptions->panDstBands = ((int *)CPLMalloc((__pyx_v_src_count * (sizeof(int)))));

  /* "rasterio/_warp.pyx":516
 *     psWOptions.panDstBands = <int *>CPLMalloc(src_count*sizeof(int))
 * 
 *     for i in range(src_count):             # <<<<<<<<<<<<<<
 *         psWOptions.panSrcBands[i] = src_bidx[i]
 *         psWOptions.panDstBands[i] = dst_bidx[i]
 */
  __pyx_t_19 = __pyx_v_src_count;
  for (__pyx_t_8 = 0; __pyx_t_8 < __pyx_t_19; __pyx_t_8+=1) {
    __pyx_v_i = __pyx_t_8;

    /* "rasterio/_warp.pyx":517
 * 
 *     for i in range(src_count):
 *         psWOptions.panSrcBands[i] = src_bidx[i]             # <<<<<<<<<<<<<<
 *         psWOptions.panDstBands[i] = dst_bidx[i]
 * 
 */
    __pyx_t_6 = __Pyx_GetItemInt(__pyx_v_src_bidx, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 517, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_39 = __Pyx_PyInt_As_int(__pyx_t_6); if (unlikely((__pyx_t_39 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 517, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    (__pyx_v_psWOptions->panSrcBands[__pyx_v_i]) = __pyx_t_39;

    /* "rasterio/_warp.pyx":518
 *     for i in range(src_count):
 *         psWOptions.panSrcBands[i] = src_bidx[i]
 *         psWOptions.panDstBands[i] = dst_bidx[i]             # <<<<<<<<<<<<<<
 * 
 *     log.debug("Set transformer options")
 */
    __pyx_t_6 = __Pyx_GetItemInt(__pyx_v_dst_bidx, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 518, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_39 = __Pyx_PyInt_As_int(__pyx_t_6); if (unlikely((__pyx_t_39 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 518, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    (__pyx_v_psWOptions->panDstBands[__pyx_v_i]) = __pyx_t_39;
  }

  /* "rasterio/_warp.pyx":520
 *         psWOptions.panDstBands[i] = dst_bidx[i]
 * 
 *     log.debug("Set transformer options")             # <<<<<<<<<<<<<<
 * 
 *     # Now that the transformer and warp options are set up, we init
 */
  __pyx_t_6 = __Pyx_GetModuleGlobalName(__pyx_n_s_log); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 520, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_debug); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 520, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_14, __pyx_tuple__29, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 520, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "rasterio/_warp.pyx":525
 *     # and run the warper.
 *     cdef GDALWarpOperation oWarper
 *     try:             # <<<<<<<<<<<<<<
 *         exc_wrap_int(oWarper.Initialize(psWOptions))
 *         rows, cols = destination.shape[-2:]
 */
  /*try:*/ {

    /* "rasterio/_warp.pyx":526
 *     cdef GDALWarpOperation oWarper
 *     try:
 *         exc_wrap_int(oWarper.Initialize(psWOptions))             # <<<<<<<<<<<<<<
 *         rows, cols = destination.shape[-2:]
 * 
 */
    __pyx_t_19 = __pyx_f_8rasterio_4_err_exc_wrap_int(__pyx_v_oWarper.Initialize(__pyx_v_psWOptions)); if (unlikely(__pyx_t_19 == -1)) __PYX_ERR(0, 526, __pyx_L98_error)

    /* "rasterio/_warp.pyx":527
 *     try:
 *         exc_wrap_int(oWarper.Initialize(psWOptions))
 *         rows, cols = destination.shape[-2:]             # <<<<<<<<<<<<<<
 * 
 *         log.debug(
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_destination, __pyx_n_s_shape); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 527, __pyx_L98_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_14 = __Pyx_PyObject_GetSlice(__pyx_t_6, -2L, 0, NULL, NULL, &__pyx_slice__30, 1, 0, 1); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 527, __pyx_L98_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if ((likely(PyTuple_CheckExact(__pyx_t_14))) || (PyList_CheckExact(__pyx_t_14))) {
      PyObject* sequence = __pyx_t_14;
      #if !CYTHON_COMPILING_IN_PYPY
      Py_ssize_t size = Py_SIZE(sequence);
      #else
      Py_ssize_t size = PySequence_Size(sequence);
      #endif
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 527, __pyx_L98_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_6 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_5 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_6 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_5 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_5);
      #else
      __pyx_t_6 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 527, __pyx_L98_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_5 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 527, __pyx_L98_error)
      __Pyx_GOTREF(__pyx_t_5);
      #endif
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_4 = PyObject_GetIter(__pyx_t_14); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 527, __pyx_L98_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      __pyx_t_28 = Py_TYPE(__pyx_t_4)->tp_iternext;
      index = 0; __pyx_t_6 = __pyx_t_28(__pyx_t_4); if (unlikely(!__pyx_t_6)) goto __pyx_L100_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_6);
      index = 1; __pyx_t_5 = __pyx_t_28(__pyx_t_4); if (unlikely(!__pyx_t_5)) goto __pyx_L100_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_5);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_28(__pyx_t_4), 2) < 0) __PYX_ERR(0, 527, __pyx_L98_error)
      __pyx_t_28 = NULL;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      goto __pyx_L101_unpacking_done;
      __pyx_L100_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_28 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 527, __pyx_L98_error)
      __pyx_L101_unpacking_done:;
    }
    __pyx_t_19 = __Pyx_PyInt_As_int(__pyx_t_6); if (unlikely((__pyx_t_19 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 527, __pyx_L98_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_8 = __Pyx_PyInt_As_int(__pyx_t_5); if (unlikely((__pyx_t_8 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 527, __pyx_L98_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_rows = __pyx_t_19;
    __pyx_v_cols = __pyx_t_8;

    /* "rasterio/_warp.pyx":529
 *         rows, cols = destination.shape[-2:]
 * 
 *         log.debug(             # <<<<<<<<<<<<<<
 *             "Chunk and warp window: %d, %d, %d, %d.",
 *             0, 0, cols, rows)
 */
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_log); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 529, __pyx_L98_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_debug); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 529, __pyx_L98_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "rasterio/_warp.pyx":531
 *         log.debug(
 *             "Chunk and warp window: %d, %d, %d, %d.",
 *             0, 0, cols, rows)             # <<<<<<<<<<<<<<
 * 
 *         if num_threads > 1:
 */
    __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_cols); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 531, __pyx_L98_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_rows); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 531, __pyx_L98_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_6)) {
      PyObject *__pyx_temp[6] = {__pyx_t_3, __pyx_kp_s_Chunk_and_warp_window_d_d_d_d, __pyx_int_0, __pyx_int_0, __pyx_t_5, __pyx_t_4};
      __pyx_t_14 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_8, 5+__pyx_t_8); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 529, __pyx_L98_error)
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_GOTREF(__pyx_t_14);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
      PyObject *__pyx_temp[6] = {__pyx_t_3, __pyx_kp_s_Chunk_and_warp_window_d_d_d_d, __pyx_int_0, __pyx_int_0, __pyx_t_5, __pyx_t_4};
      __pyx_t_14 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_8, 5+__pyx_t_8); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 529, __pyx_L98_error)
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_GOTREF(__pyx_t_14);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else
    #endif
    {
      __pyx_t_40 = PyTuple_New(5+__pyx_t_8); if (unlikely(!__pyx_t_40)) __PYX_ERR(0, 529, __pyx_L98_error)
      __Pyx_GOTREF(__pyx_t_40);
      if (__pyx_t_3) {
        __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_40, 0, __pyx_t_3); __pyx_t_3 = NULL;
      }
      __Pyx_INCREF(__pyx_kp_s_Chunk_and_warp_window_d_d_d_d);
      __Pyx_GIVEREF(__pyx_kp_s_Chunk_and_warp_window_d_d_d_d);
      PyTuple_SET_ITEM(__pyx_t_40, 0+__pyx_t_8, __pyx_kp_s_Chunk_and_warp_window_d_d_d_d);
      __Pyx_INCREF(__pyx_int_0);
      __Pyx_GIVEREF(__pyx_int_0);
      PyTuple_SET_ITEM(__pyx_t_40, 1+__pyx_t_8, __pyx_int_0);
      __Pyx_INCREF(__pyx_int_0);
      __Pyx_GIVEREF(__pyx_int_0);
      PyTuple_SET_ITEM(__pyx_t_40, 2+__pyx_t_8, __pyx_int_0);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_40, 3+__pyx_t_8, __pyx_t_5);
      __Pyx_GIVEREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_40, 4+__pyx_t_8, __pyx_t_4);
      __pyx_t_5 = 0;
      __pyx_t_4 = 0;
      __pyx_t_14 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_40, NULL); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 529, __pyx_L98_error)
      __Pyx_GOTREF(__pyx_t_14);
      __Pyx_DECREF(__pyx_t_40); __pyx_t_40 = 0;
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

    /* "rasterio/_warp.pyx":533
 *             0, 0, cols, rows)
 * 
 *         if num_threads > 1:             # <<<<<<<<<<<<<<
 *             with nogil:
 *                 oWarper.ChunkAndWarpMulti(0, 0, cols, rows)
 */
    __pyx_t_14 = PyObject_RichCompare(__pyx_v_num_threads, __pyx_int_1, Py_GT); __Pyx_XGOTREF(__pyx_t_14); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 533, __pyx_L98_error)
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_14); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 533, __pyx_L98_error)
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    if (__pyx_t_1) {

      /* "rasterio/_warp.pyx":534
 * 
 *         if num_threads > 1:
 *             with nogil:             # <<<<<<<<<<<<<<
 *                 oWarper.ChunkAndWarpMulti(0, 0, cols, rows)
 *         else:
 */
      {
          #ifdef WITH_THREAD
          PyThreadState *_save;
          Py_UNBLOCK_THREADS
          #endif
          /*try:*/ {

            /* "rasterio/_warp.pyx":535
 *         if num_threads > 1:
 *             with nogil:
 *                 oWarper.ChunkAndWarpMulti(0, 0, cols, rows)             # <<<<<<<<<<<<<<
 *         else:
 *             with nogil:
 */
            __pyx_v_oWarper.ChunkAndWarpMulti(0, 0, __pyx_v_cols, __pyx_v_rows);
          }

          /* "rasterio/_warp.pyx":534
 * 
 *         if num_threads > 1:
 *             with nogil:             # <<<<<<<<<<<<<<
 *                 oWarper.ChunkAndWarpMulti(0, 0, cols, rows)
 *         else:
 */
          /*finally:*/ {
            /*normal exit:*/{
              #ifdef WITH_THREAD
              Py_BLOCK_THREADS
              #endif
              goto __pyx_L105;
            }
            __pyx_L105:;
          }
      }

      /* "rasterio/_warp.pyx":533
 *             0, 0, cols, rows)
 * 
 *         if num_threads > 1:             # <<<<<<<<<<<<<<
 *             with nogil:
 *                 oWarper.ChunkAndWarpMulti(0, 0, cols, rows)
 */
      goto __pyx_L102;
    }

    /* "rasterio/_warp.pyx":537
 *                 oWarper.ChunkAndWarpMulti(0, 0, cols, rows)
 *         else:
 *             with nogil:             # <<<<<<<<<<<<<<
 *                 oWarper.ChunkAndWarpImage(0, 0, cols, rows)
 * 
 */
    /*else*/ {
      {
          #ifdef WITH_THREAD
          PyThreadState *_save;
          Py_UNBLOCK_THREADS
          #endif
          /*try:*/ {

            /* "rasterio/_warp.pyx":538
 *         else:
 *             with nogil:
 *                 oWarper.ChunkAndWarpImage(0, 0, cols, rows)             # <<<<<<<<<<<<<<
 * 
 *         if dtypes.is_ndarray(destination):
 */
            __pyx_v_oWarper.ChunkAndWarpImage(0, 0, __pyx_v_cols, __pyx_v_rows);
          }

          /* "rasterio/_warp.pyx":537
 *                 oWarper.ChunkAndWarpMulti(0, 0, cols, rows)
 *         else:
 *             with nogil:             # <<<<<<<<<<<<<<
 *                 oWarper.ChunkAndWarpImage(0, 0, cols, rows)
 * 
 */
          /*finally:*/ {
            /*normal exit:*/{
              #ifdef WITH_THREAD
              Py_BLOCK_THREADS
              #endif
              goto __pyx_L108;
            }
            __pyx_L108:;
          }
      }
    }
    __pyx_L102:;

    /* "rasterio/_warp.pyx":540
 *                 oWarper.ChunkAndWarpImage(0, 0, cols, rows)
 * 
 *         if dtypes.is_ndarray(destination):             # <<<<<<<<<<<<<<
 *             retval = io_auto(destination, dst_dataset, 0)
 *             # TODO: handle errors (by retval).
 */
    __pyx_t_6 = __Pyx_GetModuleGlobalName(__pyx_n_s_dtypes); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 540, __pyx_L98_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_40 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_is_ndarray); if (unlikely(!__pyx_t_40)) __PYX_ERR(0, 540, __pyx_L98_error)
    __Pyx_GOTREF(__pyx_t_40);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_40))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_40);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_40);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_40, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_14 = __Pyx_PyObject_CallOneArg(__pyx_t_40, __pyx_v_destination); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 540, __pyx_L98_error)
      __Pyx_GOTREF(__pyx_t_14);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_40)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_destination};
        __pyx_t_14 = __Pyx_PyFunction_FastCall(__pyx_t_40, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 540, __pyx_L98_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_14);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_40)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_destination};
        __pyx_t_14 = __Pyx_PyCFunction_FastCall(__pyx_t_40, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 540, __pyx_L98_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_14);
      } else
      #endif
      {
        __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 540, __pyx_L98_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_destination);
        __Pyx_GIVEREF(__pyx_v_destination);
        PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_v_destination);
        __pyx_t_14 = __Pyx_PyObject_Call(__pyx_t_40, __pyx_t_4, NULL); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 540, __pyx_L98_error)
        __Pyx_GOTREF(__pyx_t_14);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_40); __pyx_t_40 = 0;
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_14); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 540, __pyx_L98_error)
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    if (__pyx_t_1) {

      /* "rasterio/_warp.pyx":541
 * 
 *         if dtypes.is_ndarray(destination):
 *             retval = io_auto(destination, dst_dataset, 0)             # <<<<<<<<<<<<<<
 *             # TODO: handle errors (by retval).
 * 
 */
      __pyx_v_retval = __pyx_f_8rasterio_3_io_io_auto(__pyx_v_destination, __pyx_v_dst_dataset, 0, NULL);

      /* "rasterio/_warp.pyx":544
 *             # TODO: handle errors (by retval).
 * 
 *             if dst_dataset != NULL:             # <<<<<<<<<<<<<<
 *                 GDALClose(dst_dataset)
 * 
 */
      __pyx_t_1 = ((__pyx_v_dst_dataset != NULL) != 0);
      if (__pyx_t_1) {

        /* "rasterio/_warp.pyx":545
 * 
 *             if dst_dataset != NULL:
 *                 GDALClose(dst_dataset)             # <<<<<<<<<<<<<<
 * 
 *     # Clean up transformer, warp options, and dataset handles.
 */
        GDALClose(__pyx_v_dst_dataset);

        /* "rasterio/_warp.pyx":544
 *             # TODO: handle errors (by retval).
 * 
 *             if dst_dataset != NULL:             # <<<<<<<<<<<<<<
 *                 GDALClose(dst_dataset)
 * 
 */
      }

      /* "rasterio/_warp.pyx":540
 *                 oWarper.ChunkAndWarpImage(0, 0, cols, rows)
 * 
 *         if dtypes.is_ndarray(destination):             # <<<<<<<<<<<<<<
 *             retval = io_auto(destination, dst_dataset, 0)
 *             # TODO: handle errors (by retval).
 */
    }
  }

  /* "rasterio/_warp.pyx":549
 *     # Clean up transformer, warp options, and dataset handles.
 *     finally:
 *         GDALDestroyApproxTransformer(hTransformArg)             # <<<<<<<<<<<<<<
 *         GDALDestroyWarpOptions(psWOptions)
 *         CPLFree(imgProjOptions)
 */
  /*finally:*/ {
    /*normal exit:*/{
      GDALDestroyApproxTransformer(__pyx_v_hTransformArg);

      /* "rasterio/_warp.pyx":550
 *     finally:
 *         GDALDestroyApproxTransformer(hTransformArg)
 *         GDALDestroyWarpOptions(psWOptions)             # <<<<<<<<<<<<<<
 *         CPLFree(imgProjOptions)
 *         if dtypes.is_ndarray(source):
 */
      GDALDestroyWarpOptions(__pyx_v_psWOptions);

      /* "rasterio/_warp.pyx":551
 *         GDALDestroyApproxTransformer(hTransformArg)
 *         GDALDestroyWarpOptions(psWOptions)
 *         CPLFree(imgProjOptions)             # <<<<<<<<<<<<<<
 *         if dtypes.is_ndarray(source):
 *             if src_dataset != NULL:
 */
      CPLFree(__pyx_v_imgProjOptions);

      /* "rasterio/_warp.pyx":552
 *         GDALDestroyWarpOptions(psWOptions)
 *         CPLFree(imgProjOptions)
 *         if dtypes.is_ndarray(source):             # <<<<<<<<<<<<<<
 *             if src_dataset != NULL:
 *                 GDALClose(src_dataset)
 */
      __pyx_t_40 = __Pyx_GetModuleGlobalName(__pyx_n_s_dtypes); if (unlikely(!__pyx_t_40)) __PYX_ERR(0, 552, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_40);
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_40, __pyx_n_s_is_ndarray); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 552, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_40); __pyx_t_40 = 0;
      __pyx_t_40 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_40 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_40)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_40);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      if (!__pyx_t_40) {
        __pyx_t_14 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_source); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 552, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_14);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[2] = {__pyx_t_40, __pyx_v_source};
          __pyx_t_14 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 552, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_40); __pyx_t_40 = 0;
          __Pyx_GOTREF(__pyx_t_14);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[2] = {__pyx_t_40, __pyx_v_source};
          __pyx_t_14 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 552, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_40); __pyx_t_40 = 0;
          __Pyx_GOTREF(__pyx_t_14);
        } else
        #endif
        {
          __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 552, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_GIVEREF(__pyx_t_40); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_40); __pyx_t_40 = NULL;
          __Pyx_INCREF(__pyx_v_source);
          __Pyx_GIVEREF(__pyx_v_source);
          PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_source);
          __pyx_t_14 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 552, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_14);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_14); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 552, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      if (__pyx_t_1) {

        /* "rasterio/_warp.pyx":553
 *         CPLFree(imgProjOptions)
 *         if dtypes.is_ndarray(source):
 *             if src_dataset != NULL:             # <<<<<<<<<<<<<<
 *                 GDALClose(src_dataset)
 * 
 */
        __pyx_t_1 = ((__pyx_v_src_dataset != NULL) != 0);
        if (__pyx_t_1) {

          /* "rasterio/_warp.pyx":554
 *         if dtypes.is_ndarray(source):
 *             if src_dataset != NULL:
 *                 GDALClose(src_dataset)             # <<<<<<<<<<<<<<
 * 
 * 
 */
          GDALClose(__pyx_v_src_dataset);

          /* "rasterio/_warp.pyx":553
 *         CPLFree(imgProjOptions)
 *         if dtypes.is_ndarray(source):
 *             if src_dataset != NULL:             # <<<<<<<<<<<<<<
 *                 GDALClose(src_dataset)
 * 
 */
        }

        /* "rasterio/_warp.pyx":552
 *         GDALDestroyWarpOptions(psWOptions)
 *         CPLFree(imgProjOptions)
 *         if dtypes.is_ndarray(source):             # <<<<<<<<<<<<<<
 *             if src_dataset != NULL:
 *                 GDALClose(src_dataset)
 */
      }
      goto __pyx_L99;
    }
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __pyx_L98_error:;
      __pyx_t_21 = 0; __pyx_t_22 = 0; __pyx_t_23 = 0; __pyx_t_12 = 0; __pyx_t_11 = 0; __pyx_t_10 = 0;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_40); __pyx_t_40 = 0;
      __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_12, &__pyx_t_11, &__pyx_t_10);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_21, &__pyx_t_22, &__pyx_t_23) < 0)) __Pyx_ErrFetch(&__pyx_t_21, &__pyx_t_22, &__pyx_t_23);
      __Pyx_XGOTREF(__pyx_t_21);
      __Pyx_XGOTREF(__pyx_t_22);
      __Pyx_XGOTREF(__pyx_t_23);
      __Pyx_XGOTREF(__pyx_t_12);
      __Pyx_XGOTREF(__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_10);
      __pyx_t_8 = __pyx_lineno; __pyx_t_19 = __pyx_clineno; __pyx_t_41 = __pyx_filename;
      {

        /* "rasterio/_warp.pyx":549
 *     # Clean up transformer, warp options, and dataset handles.
 *     finally:
 *         GDALDestroyApproxTransformer(hTransformArg)             # <<<<<<<<<<<<<<
 *         GDALDestroyWarpOptions(psWOptions)
 *         CPLFree(imgProjOptions)
 */
        GDALDestroyApproxTransformer(__pyx_v_hTransformArg);

        /* "rasterio/_warp.pyx":550
 *     finally:
 *         GDALDestroyApproxTransformer(hTransformArg)
 *         GDALDestroyWarpOptions(psWOptions)             # <<<<<<<<<<<<<<
 *         CPLFree(imgProjOptions)
 *         if dtypes.is_ndarray(source):
 */
        GDALDestroyWarpOptions(__pyx_v_psWOptions);

        /* "rasterio/_warp.pyx":551
 *         GDALDestroyApproxTransformer(hTransformArg)
 *         GDALDestroyWarpOptions(psWOptions)
 *         CPLFree(imgProjOptions)             # <<<<<<<<<<<<<<
 *         if dtypes.is_ndarray(source):
 *             if src_dataset != NULL:
 */
        CPLFree(__pyx_v_imgProjOptions);

        /* "rasterio/_warp.pyx":552
 *         GDALDestroyWarpOptions(psWOptions)
 *         CPLFree(imgProjOptions)
 *         if dtypes.is_ndarray(source):             # <<<<<<<<<<<<<<
 *             if src_dataset != NULL:
 *                 GDALClose(src_dataset)
 */
        __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_dtypes); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 552, __pyx_L114_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_is_ndarray); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 552, __pyx_L114_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_6);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_6, function);
          }
        }
        if (!__pyx_t_4) {
          __pyx_t_14 = __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_v_source); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 552, __pyx_L114_error)
          __Pyx_GOTREF(__pyx_t_14);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_6)) {
            PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_source};
            __pyx_t_14 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 552, __pyx_L114_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_14);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
            PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_source};
            __pyx_t_14 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 552, __pyx_L114_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_14);
          } else
          #endif
          {
            __pyx_t_40 = PyTuple_New(1+1); if (unlikely(!__pyx_t_40)) __PYX_ERR(0, 552, __pyx_L114_error)
            __Pyx_GOTREF(__pyx_t_40);
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_40, 0, __pyx_t_4); __pyx_t_4 = NULL;
            __Pyx_INCREF(__pyx_v_source);
            __Pyx_GIVEREF(__pyx_v_source);
            PyTuple_SET_ITEM(__pyx_t_40, 0+1, __pyx_v_source);
            __pyx_t_14 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_40, NULL); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 552, __pyx_L114_error)
            __Pyx_GOTREF(__pyx_t_14);
            __Pyx_DECREF(__pyx_t_40); __pyx_t_40 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_14); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 552, __pyx_L114_error)
        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
        if (__pyx_t_1) {

          /* "rasterio/_warp.pyx":553
 *         CPLFree(imgProjOptions)
 *         if dtypes.is_ndarray(source):
 *             if src_dataset != NULL:             # <<<<<<<<<<<<<<
 *                 GDALClose(src_dataset)
 * 
 */
          __pyx_t_1 = ((__pyx_v_src_dataset != NULL) != 0);
          if (__pyx_t_1) {

            /* "rasterio/_warp.pyx":554
 *         if dtypes.is_ndarray(source):
 *             if src_dataset != NULL:
 *                 GDALClose(src_dataset)             # <<<<<<<<<<<<<<
 * 
 * 
 */
            GDALClose(__pyx_v_src_dataset);

            /* "rasterio/_warp.pyx":553
 *         CPLFree(imgProjOptions)
 *         if dtypes.is_ndarray(source):
 *             if src_dataset != NULL:             # <<<<<<<<<<<<<<
 *                 GDALClose(src_dataset)
 * 
 */
          }

          /* "rasterio/_warp.pyx":552
 *         GDALDestroyWarpOptions(psWOptions)
 *         CPLFree(imgProjOptions)
 *         if dtypes.is_ndarray(source):             # <<<<<<<<<<<<<<
 *             if src_dataset != NULL:
 *                 GDALClose(src_dataset)
 */
        }
      }
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_12);
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_XGIVEREF(__pyx_t_10);
        __Pyx_ExceptionReset(__pyx_t_12, __pyx_t_11, __pyx_t_10);
      }
      __Pyx_XGIVEREF(__pyx_t_21);
      __Pyx_XGIVEREF(__pyx_t_22);
      __Pyx_XGIVEREF(__pyx_t_23);
      __Pyx_ErrRestore(__pyx_t_21, __pyx_t_22, __pyx_t_23);
      __pyx_t_21 = 0; __pyx_t_22 = 0; __pyx_t_23 = 0; __pyx_t_12 = 0; __pyx_t_11 = 0; __pyx_t_10 = 0;
      __pyx_lineno = __pyx_t_8; __pyx_clineno = __pyx_t_19; __pyx_filename = __pyx_t_41;
      goto __pyx_L1_error;
      __pyx_L114_error:;
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_12);
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_XGIVEREF(__pyx_t_10);
        __Pyx_ExceptionReset(__pyx_t_12, __pyx_t_11, __pyx_t_10);
      }
      __Pyx_XDECREF(__pyx_t_21); __pyx_t_21 = 0;
      __Pyx_XDECREF(__pyx_t_22); __pyx_t_22 = 0;
      __Pyx_XDECREF(__pyx_t_23); __pyx_t_23 = 0;
      __pyx_t_12 = 0; __pyx_t_11 = 0; __pyx_t_10 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L99:;
  }

  /* "rasterio/_warp.pyx":171
 * 
 * 
 * def _reproject(             # <<<<<<<<<<<<<<
 *         source, destination,
 *         src_transform=None,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_40);
  __Pyx_AddTraceback("rasterio._warp._reproject", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_src_bidx);
  __Pyx_XDECREF(__pyx_v_dtype);
  __Pyx_XDECREF(__pyx_v_datasetname);
  __Pyx_XDECREF(__pyx_v_obj);
  __Pyx_XDECREF(__pyx_v_ident);
  __Pyx_XDECREF(__pyx_v_info);
  __Pyx_XDECREF(__pyx_v_rdr);
  __Pyx_XDECREF(__pyx_v_shape);
  __Pyx_XDECREF(__pyx_v_dst_bidx);
  __Pyx_XDECREF(__pyx_v__);
  __Pyx_XDECREF(__pyx_v_udr);
  __Pyx_XDECREF(__pyx_v_key);
  __Pyx_XDECREF(__pyx_v_val);
  __Pyx_XDECREF(__pyx_v_valb);
  __Pyx_XDECREF(__pyx_v_source);
  __Pyx_XDECREF(__pyx_v_destination);
  __Pyx_XDECREF(__pyx_v_src_nodata);
  __Pyx_XDECREF(__pyx_v_dst_nodata);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "rasterio/_warp.pyx":557
 * 
 * 
 * def _calculate_default_transform(src_crs, dst_crs, width, height,             # <<<<<<<<<<<<<<
 *                                  left=None, bottom=None, right=None, top=None,
 *                                  gcps=None, **kwargs):
 */

/* Python wrapper */
static PyObject *__pyx_pw_8rasterio_5_warp_7_calculate_default_transform(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_8rasterio_5_warp_6_calculate_default_transform[] = "Wraps GDAL's algorithm.";
static PyMethodDef __pyx_mdef_8rasterio_5_warp_7_calculate_default_transform = {"_calculate_default_transform", (PyCFunction)__pyx_pw_8rasterio_5_warp_7_calculate_default_transform, METH_VARARGS|METH_KEYWORDS, __pyx_doc_8rasterio_5_warp_6_calculate_default_transform};
static PyObject *__pyx_pw_8rasterio_5_warp_7_calculate_default_transform(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_src_crs = 0;
  PyObject *__pyx_v_dst_crs = 0;
  PyObject *__pyx_v_width = 0;
  PyObject *__pyx_v_height = 0;
  PyObject *__pyx_v_left = 0;
  PyObject *__pyx_v_bottom = 0;
  PyObject *__pyx_v_right = 0;
  PyObject *__pyx_v_top = 0;
  PyObject *__pyx_v_gcps = 0;
  CYTHON_UNUSED PyObject *__pyx_v_kwargs = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_calculate_default_transform (wrapper)", 0);
  __pyx_v_kwargs = PyDict_New(); if (unlikely(!__pyx_v_kwargs)) return NULL;
  __Pyx_GOTREF(__pyx_v_kwargs);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_src_crs,&__pyx_n_s_dst_crs,&__pyx_n_s_width,&__pyx_n_s_height,&__pyx_n_s_left,&__pyx_n_s_bottom,&__pyx_n_s_right,&__pyx_n_s_top,&__pyx_n_s_gcps,0};
    PyObject* values[9] = {0,0,0,0,0,0,0,0,0};

    /* "rasterio/_warp.pyx":558
 * 
 * def _calculate_default_transform(src_crs, dst_crs, width, height,
 *                                  left=None, bottom=None, right=None, top=None,             # <<<<<<<<<<<<<<
 *                                  gcps=None, **kwargs):
 *     """Wraps GDAL's algorithm."""
 */
    values[4] = ((PyObject *)Py_None);
    values[5] = ((PyObject *)Py_None);
    values[6] = ((PyObject *)Py_None);
    values[7] = ((PyObject *)Py_None);

    /* "rasterio/_warp.pyx":559
 * def _calculate_default_transform(src_crs, dst_crs, width, height,
 *                                  left=None, bottom=None, right=None, top=None,
 *                                  gcps=None, **kwargs):             # <<<<<<<<<<<<<<
 *     """Wraps GDAL's algorithm."""
 *     cdef void *hTransformArg = NULL
 */
    values[8] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_src_crs)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dst_crs)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_calculate_default_transform", 0, 4, 9, 1); __PYX_ERR(0, 557, __pyx_L3_error)
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_width)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_calculate_default_transform", 0, 4, 9, 2); __PYX_ERR(0, 557, __pyx_L3_error)
        }
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_height)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_calculate_default_transform", 0, 4, 9, 3); __PYX_ERR(0, 557, __pyx_L3_error)
        }
        case  4:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_left);
          if (value) { values[4] = value; kw_args--; }
        }
        case  5:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_bottom);
          if (value) { values[5] = value; kw_args--; }
        }
        case  6:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_right);
          if (value) { values[6] = value; kw_args--; }
        }
        case  7:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_top);
          if (value) { values[7] = value; kw_args--; }
        }
        case  8:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_gcps);
          if (value) { values[8] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, __pyx_v_kwargs, values, pos_args, "_calculate_default_transform") < 0)) __PYX_ERR(0, 557, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_src_crs = values[0];
    __pyx_v_dst_crs = values[1];
    __pyx_v_width = values[2];
    __pyx_v_height = values[3];
    __pyx_v_left = values[4];
    __pyx_v_bottom = values[5];
    __pyx_v_right = values[6];
    __pyx_v_top = values[7];
    __pyx_v_gcps = values[8];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_calculate_default_transform", 0, 4, 9, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 557, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_DECREF(__pyx_v_kwargs); __pyx_v_kwargs = 0;
  __Pyx_AddTraceback("rasterio._warp._calculate_default_transform", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8rasterio_5_warp_6_calculate_default_transform(__pyx_self, __pyx_v_src_crs, __pyx_v_dst_crs, __pyx_v_width, __pyx_v_height, __pyx_v_left, __pyx_v_bottom, __pyx_v_right, __pyx_v_top, __pyx_v_gcps, __pyx_v_kwargs);

  /* "rasterio/_warp.pyx":557
 * 
 * 
 * def _calculate_default_transform(src_crs, dst_crs, width, height,             # <<<<<<<<<<<<<<
 *                                  left=None, bottom=None, right=None, top=None,
 *                                  gcps=None, **kwargs):
 */

  /* function exit code */
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_8rasterio_5_warp_28_calculate_default_transform_2generator(__pyx_CoroutineObject *__pyx_generator, PyObject *__pyx_sent_value); /* proto */

/* "rasterio/_warp.pyx":576
 *     # GDALCreateGenImgProjTransformer().
 * 
 *     if all(x is not None for x in (left, bottom, right, top)):             # <<<<<<<<<<<<<<
 *         transform = from_bounds(left, bottom, right, top, width, height)
 *         transform = transform.to_gdal()
 */

static PyObject *__pyx_pf_8rasterio_5_warp_28_calculate_default_transform_genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_8rasterio_5_warp___pyx_scope_struct_1_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_8rasterio_5_warp___pyx_scope_struct_1_genexpr *)__pyx_tp_new_8rasterio_5_warp___pyx_scope_struct_1_genexpr(__pyx_ptype_8rasterio_5_warp___pyx_scope_struct_1_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_8rasterio_5_warp___pyx_scope_struct_1_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 576, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_8rasterio_5_warp___pyx_scope_struct___calculate_default_transform *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_8rasterio_5_warp_28_calculate_default_transform_2generator, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_calculate_default_transform_loc, __pyx_n_s_rasterio__warp); if (unlikely(!gen)) __PYX_ERR(0, 576, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("rasterio._warp._calculate_default_transform.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_8rasterio_5_warp_28_calculate_default_transform_2generator(__pyx_CoroutineObject *__pyx_generator, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_8rasterio_5_warp___pyx_scope_struct_1_genexpr *__pyx_cur_scope = ((struct __pyx_obj_8rasterio_5_warp___pyx_scope_struct_1_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("None", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 576, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_left)) { __Pyx_RaiseClosureNameError("left"); __PYX_ERR(0, 576, __pyx_L1_error) }
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_bottom)) { __Pyx_RaiseClosureNameError("bottom"); __PYX_ERR(0, 576, __pyx_L1_error) }
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_right)) { __Pyx_RaiseClosureNameError("right"); __PYX_ERR(0, 576, __pyx_L1_error) }
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_top)) { __Pyx_RaiseClosureNameError("top"); __PYX_ERR(0, 576, __pyx_L1_error) }
  __pyx_t_1 = PyTuple_New(4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 576, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_left);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_left);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_cur_scope->__pyx_outer_scope->__pyx_v_left);
  __Pyx_INCREF(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_bottom);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_bottom);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_cur_scope->__pyx_outer_scope->__pyx_v_bottom);
  __Pyx_INCREF(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_right);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_right);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_cur_scope->__pyx_outer_scope->__pyx_v_right);
  __Pyx_INCREF(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_top);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_top);
  PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_cur_scope->__pyx_outer_scope->__pyx_v_top);
  __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (__pyx_t_3 >= 4) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 576, __pyx_L1_error)
    #else
    __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 576, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    #endif
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_x);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_x, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_4 = (__pyx_cur_scope->__pyx_v_x != Py_None);
    __pyx_t_5 = ((!(__pyx_t_4 != 0)) != 0);
    if (__pyx_t_5) {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(Py_False);
      __pyx_r = Py_False;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      goto __pyx_L0;
    }
  }
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_True);
    __pyx_r = Py_True;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    goto __pyx_L0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_8rasterio_5_warp_28_calculate_default_transform_5generator1(__pyx_CoroutineObject *__pyx_generator, PyObject *__pyx_sent_value); /* proto */

/* "rasterio/_warp.pyx":579
 *         transform = from_bounds(left, bottom, right, top, width, height)
 *         transform = transform.to_gdal()
 *     elif any(x is not None for x in (left, bottom, right, top)):             # <<<<<<<<<<<<<<
 *         raise ValueError(
 *             "Some, but not all, bounding box parameters were provided.")
 */

static PyObject *__pyx_pf_8rasterio_5_warp_28_calculate_default_transform_3genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_8rasterio_5_warp___pyx_scope_struct_2_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_8rasterio_5_warp___pyx_scope_struct_2_genexpr *)__pyx_tp_new_8rasterio_5_warp___pyx_scope_struct_2_genexpr(__pyx_ptype_8rasterio_5_warp___pyx_scope_struct_2_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_8rasterio_5_warp___pyx_scope_struct_2_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 579, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_8rasterio_5_warp___pyx_scope_struct___calculate_default_transform *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_8rasterio_5_warp_28_calculate_default_transform_5generator1, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_calculate_default_transform_loc, __pyx_n_s_rasterio__warp); if (unlikely(!gen)) __PYX_ERR(0, 579, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("rasterio._warp._calculate_default_transform.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_8rasterio_5_warp_28_calculate_default_transform_5generator1(__pyx_CoroutineObject *__pyx_generator, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_8rasterio_5_warp___pyx_scope_struct_2_genexpr *__pyx_cur_scope = ((struct __pyx_obj_8rasterio_5_warp___pyx_scope_struct_2_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("None", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 579, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_left)) { __Pyx_RaiseClosureNameError("left"); __PYX_ERR(0, 579, __pyx_L1_error) }
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_bottom)) { __Pyx_RaiseClosureNameError("bottom"); __PYX_ERR(0, 579, __pyx_L1_error) }
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_right)) { __Pyx_RaiseClosureNameError("right"); __PYX_ERR(0, 579, __pyx_L1_error) }
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_top)) { __Pyx_RaiseClosureNameError("top"); __PYX_ERR(0, 579, __pyx_L1_error) }
  __pyx_t_1 = PyTuple_New(4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 579, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_left);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_left);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_cur_scope->__pyx_outer_scope->__pyx_v_left);
  __Pyx_INCREF(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_bottom);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_bottom);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_cur_scope->__pyx_outer_scope->__pyx_v_bottom);
  __Pyx_INCREF(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_right);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_right);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_cur_scope->__pyx_outer_scope->__pyx_v_right);
  __Pyx_INCREF(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_top);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_top);
  PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_cur_scope->__pyx_outer_scope->__pyx_v_top);
  __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (__pyx_t_3 >= 4) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 579, __pyx_L1_error)
    #else
    __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 579, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    #endif
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_x);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_x, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_4 = (__pyx_cur_scope->__pyx_v_x != Py_None);
    __pyx_t_5 = (__pyx_t_4 != 0);
    if (__pyx_t_5) {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(Py_True);
      __pyx_r = Py_True;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      goto __pyx_L0;
    }
  }
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_False);
    __pyx_r = Py_False;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    goto __pyx_L0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "rasterio/_warp.pyx":557
 * 
 * 
 * def _calculate_default_transform(src_crs, dst_crs, width, height,             # <<<<<<<<<<<<<<
 *                                  left=None, bottom=None, right=None, top=None,
 *                                  gcps=None, **kwargs):
 */

static PyObject *__pyx_pf_8rasterio_5_warp_6_calculate_default_transform(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_src_crs, PyObject *__pyx_v_dst_crs, PyObject *__pyx_v_width, PyObject *__pyx_v_height, PyObject *__pyx_v_left, PyObject *__pyx_v_bottom, PyObject *__pyx_v_right, PyObject *__pyx_v_top, PyObject *__pyx_v_gcps, CYTHON_UNUSED PyObject *__pyx_v_kwargs) {
  struct __pyx_obj_8rasterio_5_warp___pyx_scope_struct___calculate_default_transform *__pyx_cur_scope;
  void *__pyx_v_hTransformArg;
  int __pyx_v_npixels;
  int __pyx_v_nlines;
  double __pyx_v_extent[4];
  double __pyx_v_geotransform[6];
  OGRSpatialReferenceH __pyx_v_osr;
  char *__pyx_v_wkt;
  struct __pyx_obj_8rasterio_3_io_InMemoryRaster *__pyx_v_temp = 0;
  PyObject *__pyx_v_transform = NULL;
  CYTHON_UNUSED PyObject *__pyx_v_img = NULL;
  PyObject *__pyx_v_err = NULL;
  PyObject *__pyx_v_dst_affine = NULL;
  int __pyx_v_dst_width;
  int __pyx_v_dst_height;
  long __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1[4];
  double __pyx_t_2[6];
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  OGRSpatialReferenceH __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  void *__pyx_t_18;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  int __pyx_t_22;
  int __pyx_t_23;
  char const *__pyx_t_24;
  PyObject *__pyx_t_25 = NULL;
  PyObject *__pyx_t_26 = NULL;
  PyObject *__pyx_t_27 = NULL;
  long __pyx_t_28;
  __Pyx_RefNannySetupContext("_calculate_default_transform", 0);
  __pyx_cur_scope = (struct __pyx_obj_8rasterio_5_warp___pyx_scope_struct___calculate_default_transform *)__pyx_tp_new_8rasterio_5_warp___pyx_scope_struct___calculate_default_transform(__pyx_ptype_8rasterio_5_warp___pyx_scope_struct___calculate_default_transform, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_8rasterio_5_warp___pyx_scope_struct___calculate_default_transform *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 557, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_left = __pyx_v_left;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_left);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_left);
  __pyx_cur_scope->__pyx_v_bottom = __pyx_v_bottom;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_bottom);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_bottom);
  __pyx_cur_scope->__pyx_v_right = __pyx_v_right;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_right);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_right);
  __pyx_cur_scope->__pyx_v_top = __pyx_v_top;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_top);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_top);

  /* "rasterio/_warp.pyx":561
 *                                  gcps=None, **kwargs):
 *     """Wraps GDAL's algorithm."""
 *     cdef void *hTransformArg = NULL             # <<<<<<<<<<<<<<
 *     cdef int npixels = 0
 *     cdef int nlines = 0
 */
  __pyx_v_hTransformArg = NULL;

  /* "rasterio/_warp.pyx":562
 *     """Wraps GDAL's algorithm."""
 *     cdef void *hTransformArg = NULL
 *     cdef int npixels = 0             # <<<<<<<<<<<<<<
 *     cdef int nlines = 0
 *     cdef double extent[4]
 */
  __pyx_v_npixels = 0;

  /* "rasterio/_warp.pyx":563
 *     cdef void *hTransformArg = NULL
 *     cdef int npixels = 0
 *     cdef int nlines = 0             # <<<<<<<<<<<<<<
 *     cdef double extent[4]
 *     cdef double geotransform[6]
 */
  __pyx_v_nlines = 0;

  /* "rasterio/_warp.pyx":566
 *     cdef double extent[4]
 *     cdef double geotransform[6]
 *     cdef OGRSpatialReferenceH osr = NULL             # <<<<<<<<<<<<<<
 *     cdef char *wkt = NULL
 *     cdef InMemoryRaster temp = None
 */
  __pyx_v_osr = NULL;

  /* "rasterio/_warp.pyx":567
 *     cdef double geotransform[6]
 *     cdef OGRSpatialReferenceH osr = NULL
 *     cdef char *wkt = NULL             # <<<<<<<<<<<<<<
 *     cdef InMemoryRaster temp = None
 * 
 */
  __pyx_v_wkt = NULL;

  /* "rasterio/_warp.pyx":568
 *     cdef OGRSpatialReferenceH osr = NULL
 *     cdef char *wkt = NULL
 *     cdef InMemoryRaster temp = None             # <<<<<<<<<<<<<<
 * 
 *     extent[:] = [0.0, 0.0, 0.0, 0.0]
 */
  __Pyx_INCREF(Py_None);
  __pyx_v_temp = ((struct __pyx_obj_8rasterio_3_io_InMemoryRaster *)Py_None);

  /* "rasterio/_warp.pyx":570
 *     cdef InMemoryRaster temp = None
 * 
 *     extent[:] = [0.0, 0.0, 0.0, 0.0]             # <<<<<<<<<<<<<<
 *     geotransform[:] = [0.0, 0.0, 0.0, 0.0, 0.0, 0.0]
 * 
 */
  __pyx_t_1[0] = 0.0;
  __pyx_t_1[1] = 0.0;
  __pyx_t_1[2] = 0.0;
  __pyx_t_1[3] = 0.0;
  memcpy(&(__pyx_v_extent[0]), __pyx_t_1, sizeof(__pyx_v_extent[0]) * (4));

  /* "rasterio/_warp.pyx":571
 * 
 *     extent[:] = [0.0, 0.0, 0.0, 0.0]
 *     geotransform[:] = [0.0, 0.0, 0.0, 0.0, 0.0, 0.0]             # <<<<<<<<<<<<<<
 * 
 *     # Make an in-memory raster dataset we can pass to
 */
  __pyx_t_2[0] = 0.0;
  __pyx_t_2[1] = 0.0;
  __pyx_t_2[2] = 0.0;
  __pyx_t_2[3] = 0.0;
  __pyx_t_2[4] = 0.0;
  __pyx_t_2[5] = 0.0;
  memcpy(&(__pyx_v_geotransform[0]), __pyx_t_2, sizeof(__pyx_v_geotransform[0]) * (6));

  /* "rasterio/_warp.pyx":576
 *     # GDALCreateGenImgProjTransformer().
 * 
 *     if all(x is not None for x in (left, bottom, right, top)):             # <<<<<<<<<<<<<<
 *         transform = from_bounds(left, bottom, right, top, width, height)
 *         transform = transform.to_gdal()
 */
  __pyx_t_3 = __pyx_pf_8rasterio_5_warp_28_calculate_default_transform_genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 576, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_Generator_Next(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 576, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 576, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__pyx_t_5) {

    /* "rasterio/_warp.pyx":577
 * 
 *     if all(x is not None for x in (left, bottom, right, top)):
 *         transform = from_bounds(left, bottom, right, top, width, height)             # <<<<<<<<<<<<<<
 *         transform = transform.to_gdal()
 *     elif any(x is not None for x in (left, bottom, right, top)):
 */
    __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_from_bounds); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 577, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_6 = NULL;
    __pyx_t_7 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_7 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[7] = {__pyx_t_6, __pyx_cur_scope->__pyx_v_left, __pyx_cur_scope->__pyx_v_bottom, __pyx_cur_scope->__pyx_v_right, __pyx_cur_scope->__pyx_v_top, __pyx_v_width, __pyx_v_height};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_7, 6+__pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 577, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_4);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[7] = {__pyx_t_6, __pyx_cur_scope->__pyx_v_left, __pyx_cur_scope->__pyx_v_bottom, __pyx_cur_scope->__pyx_v_right, __pyx_cur_scope->__pyx_v_top, __pyx_v_width, __pyx_v_height};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_7, 6+__pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 577, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_4);
    } else
    #endif
    {
      __pyx_t_8 = PyTuple_New(6+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 577, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (__pyx_t_6) {
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;
      }
      __Pyx_INCREF(__pyx_cur_scope->__pyx_v_left);
      __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_left);
      PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_cur_scope->__pyx_v_left);
      __Pyx_INCREF(__pyx_cur_scope->__pyx_v_bottom);
      __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_bottom);
      PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_cur_scope->__pyx_v_bottom);
      __Pyx_INCREF(__pyx_cur_scope->__pyx_v_right);
      __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_right);
      PyTuple_SET_ITEM(__pyx_t_8, 2+__pyx_t_7, __pyx_cur_scope->__pyx_v_right);
      __Pyx_INCREF(__pyx_cur_scope->__pyx_v_top);
      __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_top);
      PyTuple_SET_ITEM(__pyx_t_8, 3+__pyx_t_7, __pyx_cur_scope->__pyx_v_top);
      __Pyx_INCREF(__pyx_v_width);
      __Pyx_GIVEREF(__pyx_v_width);
      PyTuple_SET_ITEM(__pyx_t_8, 4+__pyx_t_7, __pyx_v_width);
      __Pyx_INCREF(__pyx_v_height);
      __Pyx_GIVEREF(__pyx_v_height);
      PyTuple_SET_ITEM(__pyx_t_8, 5+__pyx_t_7, __pyx_v_height);
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_8, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 577, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_transform = __pyx_t_4;
    __pyx_t_4 = 0;

    /* "rasterio/_warp.pyx":578
 *     if all(x is not None for x in (left, bottom, right, top)):
 *         transform = from_bounds(left, bottom, right, top, width, height)
 *         transform = transform.to_gdal()             # <<<<<<<<<<<<<<
 *     elif any(x is not None for x in (left, bottom, right, top)):
 *         raise ValueError(
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_transform, __pyx_n_s_to_gdal); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 578, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    if (__pyx_t_8) {
      __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_8); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 578, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    } else {
      __pyx_t_4 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 578, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF_SET(__pyx_v_transform, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "rasterio/_warp.pyx":576
 *     # GDALCreateGenImgProjTransformer().
 * 
 *     if all(x is not None for x in (left, bottom, right, top)):             # <<<<<<<<<<<<<<
 *         transform = from_bounds(left, bottom, right, top, width, height)
 *         transform = transform.to_gdal()
 */
    goto __pyx_L3;
  }

  /* "rasterio/_warp.pyx":579
 *         transform = from_bounds(left, bottom, right, top, width, height)
 *         transform = transform.to_gdal()
 *     elif any(x is not None for x in (left, bottom, right, top)):             # <<<<<<<<<<<<<<
 *         raise ValueError(
 *             "Some, but not all, bounding box parameters were provided.")
 */
  __pyx_t_4 = __pyx_pf_8rasterio_5_warp_28_calculate_default_transform_3genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 579, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __Pyx_Generator_Next(__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 579, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 579, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_5) {

    /* "rasterio/_warp.pyx":580
 *         transform = transform.to_gdal()
 *     elif any(x is not None for x in (left, bottom, right, top)):
 *         raise ValueError(             # <<<<<<<<<<<<<<
 *             "Some, but not all, bounding box parameters were provided.")
 *     else:
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__31, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 580, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 580, __pyx_L1_error)

    /* "rasterio/_warp.pyx":579
 *         transform = from_bounds(left, bottom, right, top, width, height)
 *         transform = transform.to_gdal()
 *     elif any(x is not None for x in (left, bottom, right, top)):             # <<<<<<<<<<<<<<
 *         raise ValueError(
 *             "Some, but not all, bounding box parameters were provided.")
 */
  }

  /* "rasterio/_warp.pyx":583
 *             "Some, but not all, bounding box parameters were provided.")
 *     else:
 *         transform = None             # <<<<<<<<<<<<<<
 *     img = np.empty((height, width))
 * 
 */
  /*else*/ {
    __Pyx_INCREF(Py_None);
    __pyx_v_transform = Py_None;
  }
  __pyx_L3:;

  /* "rasterio/_warp.pyx":584
 *     else:
 *         transform = None
 *     img = np.empty((height, width))             # <<<<<<<<<<<<<<
 * 
 *     osr = _osr_from_crs(dst_crs)
 */
  __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 584, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_empty); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 584, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 584, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_v_height);
  __Pyx_GIVEREF(__pyx_v_height);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_height);
  __Pyx_INCREF(__pyx_v_width);
  __Pyx_GIVEREF(__pyx_v_width);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_width);
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_8))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_8);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_8, function);
    }
  }
  if (!__pyx_t_6) {
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 584, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_3);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_8)) {
      PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_t_4};
      __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_8, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 584, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_8)) {
      PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_t_4};
      __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_8, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 584, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else
    #endif
    {
      __pyx_t_9 = PyTuple_New(1+1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 584, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_6); __pyx_t_6 = NULL;
      __Pyx_GIVEREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_9, 0+1, __pyx_t_4);
      __pyx_t_4 = 0;
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_9, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 584, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_v_img = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "rasterio/_warp.pyx":586
 *     img = np.empty((height, width))
 * 
 *     osr = _osr_from_crs(dst_crs)             # <<<<<<<<<<<<<<
 *     OSRExportToWkt(osr, &wkt)
 *     OSRRelease(osr)
 */
  __pyx_t_10 = __pyx_f_8rasterio_5_base__osr_from_crs(__pyx_v_dst_crs); if (unlikely(__pyx_t_10 == NULL)) __PYX_ERR(0, 586, __pyx_L1_error)
  __pyx_v_osr = __pyx_t_10;

  /* "rasterio/_warp.pyx":587
 * 
 *     osr = _osr_from_crs(dst_crs)
 *     OSRExportToWkt(osr, &wkt)             # <<<<<<<<<<<<<<
 *     OSRRelease(osr)
 *     osr = NULL
 */
  OSRExportToWkt(__pyx_v_osr, (&__pyx_v_wkt));

  /* "rasterio/_warp.pyx":588
 *     osr = _osr_from_crs(dst_crs)
 *     OSRExportToWkt(osr, &wkt)
 *     OSRRelease(osr)             # <<<<<<<<<<<<<<
 *     osr = NULL
 * 
 */
  OSRRelease(__pyx_v_osr);

  /* "rasterio/_warp.pyx":589
 *     OSRExportToWkt(osr, &wkt)
 *     OSRRelease(osr)
 *     osr = NULL             # <<<<<<<<<<<<<<
 * 
 *     with InMemoryRaster(width=width, height=height, transform=transform,
 */
  __pyx_v_osr = NULL;

  /* "rasterio/_warp.pyx":591
 *     osr = NULL
 * 
 *     with InMemoryRaster(width=width, height=height, transform=transform,             # <<<<<<<<<<<<<<
 *                         gcps=gcps, crs=src_crs) as temp:
 *         try:
 */
  /*with:*/ {
    __pyx_t_3 = PyDict_New(); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 591, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_width, __pyx_v_width) < 0) __PYX_ERR(0, 591, __pyx_L1_error)
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_height, __pyx_v_height) < 0) __PYX_ERR(0, 591, __pyx_L1_error)
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_transform, __pyx_v_transform) < 0) __PYX_ERR(0, 591, __pyx_L1_error)

    /* "rasterio/_warp.pyx":592
 * 
 *     with InMemoryRaster(width=width, height=height, transform=transform,
 *                         gcps=gcps, crs=src_crs) as temp:             # <<<<<<<<<<<<<<
 *         try:
 *             hTransformArg = exc_wrap_pointer(
 */
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_gcps, __pyx_v_gcps) < 0) __PYX_ERR(0, 591, __pyx_L1_error)
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_crs, __pyx_v_src_crs) < 0) __PYX_ERR(0, 591, __pyx_L1_error)

    /* "rasterio/_warp.pyx":591
 *     osr = NULL
 * 
 *     with InMemoryRaster(width=width, height=height, transform=transform,             # <<<<<<<<<<<<<<
 *                         gcps=gcps, crs=src_crs) as temp:
 *         try:
 */
    __pyx_t_8 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_8rasterio_3_io_InMemoryRaster), __pyx_empty_tuple, __pyx_t_3); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 591, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_11 = __Pyx_PyObject_LookupSpecial(__pyx_t_8, __pyx_n_s_exit); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 591, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_9 = __Pyx_PyObject_LookupSpecial(__pyx_t_8, __pyx_n_s_enter); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 591, __pyx_L4_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_9);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_9, function);
      }
    }
    if (__pyx_t_4) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 591, __pyx_L4_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else {
      __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_9); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 591, __pyx_L4_error)
    }
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_9 = __pyx_t_3;
    __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    /*try:*/ {
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_12, &__pyx_t_13, &__pyx_t_14);
        __Pyx_XGOTREF(__pyx_t_12);
        __Pyx_XGOTREF(__pyx_t_13);
        __Pyx_XGOTREF(__pyx_t_14);
        /*try:*/ {
          if (!(likely(((__pyx_t_9) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_9, __pyx_ptype_8rasterio_3_io_InMemoryRaster))))) __PYX_ERR(0, 591, __pyx_L8_error)
          __Pyx_DECREF_SET(__pyx_v_temp, ((struct __pyx_obj_8rasterio_3_io_InMemoryRaster *)__pyx_t_9));
          __pyx_t_9 = 0;

          /* "rasterio/_warp.pyx":593
 *     with InMemoryRaster(width=width, height=height, transform=transform,
 *                         gcps=gcps, crs=src_crs) as temp:
 *         try:             # <<<<<<<<<<<<<<
 *             hTransformArg = exc_wrap_pointer(
 *                 GDALCreateGenImgProjTransformer(
 */
          /*try:*/ {
            {
              __Pyx_PyThreadState_declare
              __Pyx_PyThreadState_assign
              __Pyx_ExceptionSave(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17);
              __Pyx_XGOTREF(__pyx_t_15);
              __Pyx_XGOTREF(__pyx_t_16);
              __Pyx_XGOTREF(__pyx_t_17);
              /*try:*/ {

                /* "rasterio/_warp.pyx":594
 *                         gcps=gcps, crs=src_crs) as temp:
 *         try:
 *             hTransformArg = exc_wrap_pointer(             # <<<<<<<<<<<<<<
 *                 GDALCreateGenImgProjTransformer(
 *                     temp._hds, NULL, NULL, wkt, 1, 1000.0,0))
 */
                __pyx_t_18 = __pyx_f_8rasterio_4_err_exc_wrap_pointer(GDALCreateGenImgProjTransformer(__pyx_v_temp->_hds, NULL, NULL, __pyx_v_wkt, 1, 1000.0, 0)); if (unlikely(__pyx_t_18 == NULL)) __PYX_ERR(0, 594, __pyx_L21_error)
                __pyx_v_hTransformArg = __pyx_t_18;

                /* "rasterio/_warp.pyx":597
 *                 GDALCreateGenImgProjTransformer(
 *                     temp._hds, NULL, NULL, wkt, 1, 1000.0,0))
 *             exc_wrap_int(             # <<<<<<<<<<<<<<
 *                 GDALSuggestedWarpOutput2(
 *                     temp._hds, GDALGenImgProjTransform, hTransformArg,
 */
                __pyx_t_7 = __pyx_f_8rasterio_4_err_exc_wrap_int(GDALSuggestedWarpOutput2(__pyx_v_temp->_hds, GDALGenImgProjTransform, __pyx_v_hTransformArg, __pyx_v_geotransform, (&__pyx_v_npixels), (&__pyx_v_nlines), __pyx_v_extent, 0)); if (unlikely(__pyx_t_7 == -1)) __PYX_ERR(0, 597, __pyx_L21_error)

                /* "rasterio/_warp.pyx":602
 *                     geotransform, &npixels, &nlines, extent, 0))
 * 
 *             log.debug("Created transformer and warp output.")             # <<<<<<<<<<<<<<
 * 
 *         except CPLE_NotSupportedError as err:
 */
                __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_log); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 602, __pyx_L21_error)
                __Pyx_GOTREF(__pyx_t_9);
                __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_debug); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 602, __pyx_L21_error)
                __Pyx_GOTREF(__pyx_t_8);
                __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
                __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_tuple__32, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 602, __pyx_L21_error)
                __Pyx_GOTREF(__pyx_t_9);
                __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
                __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

                /* "rasterio/_warp.pyx":593
 *     with InMemoryRaster(width=width, height=height, transform=transform,
 *                         gcps=gcps, crs=src_crs) as temp:
 *         try:             # <<<<<<<<<<<<<<
 *             hTransformArg = exc_wrap_pointer(
 *                 GDALCreateGenImgProjTransformer(
 */
              }
              __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
              __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
              __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
              goto __pyx_L28_try_end;
              __pyx_L21_error:;
              __Pyx_PyThreadState_assign
              __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
              __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
              __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

              /* "rasterio/_warp.pyx":604
 *             log.debug("Created transformer and warp output.")
 * 
 *         except CPLE_NotSupportedError as err:             # <<<<<<<<<<<<<<
 *             raise CRSError(err.errmsg)
 * 
 */
              __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_CPLE_NotSupportedError); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 604, __pyx_L23_except_error)
              __Pyx_GOTREF(__pyx_t_9);
              __pyx_t_7 = __Pyx_PyErr_ExceptionMatches(__pyx_t_9);
              __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
              if (__pyx_t_7) {
                __Pyx_AddTraceback("rasterio._warp._calculate_default_transform", __pyx_clineno, __pyx_lineno, __pyx_filename);
                if (__Pyx_GetException(&__pyx_t_9, &__pyx_t_8, &__pyx_t_3) < 0) __PYX_ERR(0, 604, __pyx_L23_except_error)
                __Pyx_GOTREF(__pyx_t_9);
                __Pyx_GOTREF(__pyx_t_8);
                __Pyx_GOTREF(__pyx_t_3);
                __Pyx_INCREF(__pyx_t_8);
                __pyx_v_err = __pyx_t_8;

                /* "rasterio/_warp.pyx":605
 * 
 *         except CPLE_NotSupportedError as err:
 *             raise CRSError(err.errmsg)             # <<<<<<<<<<<<<<
 * 
 *         except CPLE_AppDefinedError as err:
 */
                __pyx_t_6 = __Pyx_GetModuleGlobalName(__pyx_n_s_CRSError); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 605, __pyx_L23_except_error)
                __Pyx_GOTREF(__pyx_t_6);
                __pyx_t_19 = __Pyx_PyObject_GetAttrStr(__pyx_v_err, __pyx_n_s_errmsg); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 605, __pyx_L23_except_error)
                __Pyx_GOTREF(__pyx_t_19);
                __pyx_t_20 = NULL;
                if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
                  __pyx_t_20 = PyMethod_GET_SELF(__pyx_t_6);
                  if (likely(__pyx_t_20)) {
                    PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
                    __Pyx_INCREF(__pyx_t_20);
                    __Pyx_INCREF(function);
                    __Pyx_DECREF_SET(__pyx_t_6, function);
                  }
                }
                if (!__pyx_t_20) {
                  __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_19); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 605, __pyx_L23_except_error)
                  __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
                  __Pyx_GOTREF(__pyx_t_4);
                } else {
                  #if CYTHON_FAST_PYCALL
                  if (PyFunction_Check(__pyx_t_6)) {
                    PyObject *__pyx_temp[2] = {__pyx_t_20, __pyx_t_19};
                    __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 605, __pyx_L23_except_error)
                    __Pyx_XDECREF(__pyx_t_20); __pyx_t_20 = 0;
                    __Pyx_GOTREF(__pyx_t_4);
                    __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
                  } else
                  #endif
                  #if CYTHON_FAST_PYCCALL
                  if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
                    PyObject *__pyx_temp[2] = {__pyx_t_20, __pyx_t_19};
                    __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 605, __pyx_L23_except_error)
                    __Pyx_XDECREF(__pyx_t_20); __pyx_t_20 = 0;
                    __Pyx_GOTREF(__pyx_t_4);
                    __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
                  } else
                  #endif
                  {
                    __pyx_t_21 = PyTuple_New(1+1); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 605, __pyx_L23_except_error)
                    __Pyx_GOTREF(__pyx_t_21);
                    __Pyx_GIVEREF(__pyx_t_20); PyTuple_SET_ITEM(__pyx_t_21, 0, __pyx_t_20); __pyx_t_20 = NULL;
                    __Pyx_GIVEREF(__pyx_t_19);
                    PyTuple_SET_ITEM(__pyx_t_21, 0+1, __pyx_t_19);
                    __pyx_t_19 = 0;
                    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_21, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 605, __pyx_L23_except_error)
                    __Pyx_GOTREF(__pyx_t_4);
                    __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
                  }
                }
                __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
                __Pyx_Raise(__pyx_t_4, 0, 0, 0);
                __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
                __PYX_ERR(0, 605, __pyx_L23_except_error)
              }

              /* "rasterio/_warp.pyx":607
 *             raise CRSError(err.errmsg)
 * 
 *         except CPLE_AppDefinedError as err:             # <<<<<<<<<<<<<<
 *             if "Reprojection failed" in str(err):
 *                 # This "exception" should be treated as a debug msg, not error
 */
              __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_CPLE_AppDefinedError); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 607, __pyx_L23_except_error)
              __Pyx_GOTREF(__pyx_t_3);
              __pyx_t_7 = __Pyx_PyErr_ExceptionMatches(__pyx_t_3);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              if (__pyx_t_7) {
                __Pyx_AddTraceback("rasterio._warp._calculate_default_transform", __pyx_clineno, __pyx_lineno, __pyx_filename);
                if (__Pyx_GetException(&__pyx_t_3, &__pyx_t_8, &__pyx_t_9) < 0) __PYX_ERR(0, 607, __pyx_L23_except_error)
                __Pyx_GOTREF(__pyx_t_3);
                __Pyx_GOTREF(__pyx_t_8);
                __Pyx_GOTREF(__pyx_t_9);
                __Pyx_INCREF(__pyx_t_8);
                __pyx_v_err = __pyx_t_8;

                /* "rasterio/_warp.pyx":608
 * 
 *         except CPLE_AppDefinedError as err:
 *             if "Reprojection failed" in str(err):             # <<<<<<<<<<<<<<
 *                 # This "exception" should be treated as a debug msg, not error
 *                 # "Reprojection failed, err = -14, further errors will be
 */
                __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 608, __pyx_L23_except_error)
                __Pyx_GOTREF(__pyx_t_4);
                __Pyx_INCREF(__pyx_v_err);
                __Pyx_GIVEREF(__pyx_v_err);
                PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_err);
                __pyx_t_6 = __Pyx_PyObject_Call(((PyObject *)(&PyString_Type)), __pyx_t_4, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 608, __pyx_L23_except_error)
                __Pyx_GOTREF(__pyx_t_6);
                __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
                __pyx_t_5 = (__Pyx_PySequence_ContainsTF(__pyx_kp_s_Reprojection_failed, __pyx_t_6, Py_EQ)); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 608, __pyx_L23_except_error)
                __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
                __pyx_t_22 = (__pyx_t_5 != 0);
                if (__pyx_t_22) {

                  /* "rasterio/_warp.pyx":612
 *                 # "Reprojection failed, err = -14, further errors will be
 *                 # suppressed on the transform object."
 *                 log.info("Encountered points outside of valid dst crs region")             # <<<<<<<<<<<<<<
 *                 pass
 *             else:
 */
                  __pyx_t_6 = __Pyx_GetModuleGlobalName(__pyx_n_s_log); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 612, __pyx_L23_except_error)
                  __Pyx_GOTREF(__pyx_t_6);
                  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_info); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 612, __pyx_L23_except_error)
                  __Pyx_GOTREF(__pyx_t_4);
                  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
                  __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_tuple__33, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 612, __pyx_L23_except_error)
                  __Pyx_GOTREF(__pyx_t_6);
                  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
                  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

                  /* "rasterio/_warp.pyx":608
 * 
 *         except CPLE_AppDefinedError as err:
 *             if "Reprojection failed" in str(err):             # <<<<<<<<<<<<<<
 *                 # This "exception" should be treated as a debug msg, not error
 *                 # "Reprojection failed, err = -14, further errors will be
 */
                  goto __pyx_L33;
                }

                /* "rasterio/_warp.pyx":615
 *                 pass
 *             else:
 *                 raise err             # <<<<<<<<<<<<<<
 *         finally:
 *             if wkt != NULL:
 */
                /*else*/ {
                  __Pyx_Raise(__pyx_v_err, 0, 0, 0);
                  __PYX_ERR(0, 615, __pyx_L23_except_error)
                }
                __pyx_L33:;
                __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
                __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
                __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
                goto __pyx_L22_exception_handled;
              }
              goto __pyx_L23_except_error;
              __pyx_L23_except_error:;

              /* "rasterio/_warp.pyx":593
 *     with InMemoryRaster(width=width, height=height, transform=transform,
 *                         gcps=gcps, crs=src_crs) as temp:
 *         try:             # <<<<<<<<<<<<<<
 *             hTransformArg = exc_wrap_pointer(
 *                 GDALCreateGenImgProjTransformer(
 */
              __Pyx_PyThreadState_assign
              __Pyx_XGIVEREF(__pyx_t_15);
              __Pyx_XGIVEREF(__pyx_t_16);
              __Pyx_XGIVEREF(__pyx_t_17);
              __Pyx_ExceptionReset(__pyx_t_15, __pyx_t_16, __pyx_t_17);
              goto __pyx_L19_error;
              __pyx_L22_exception_handled:;
              __Pyx_PyThreadState_assign
              __Pyx_XGIVEREF(__pyx_t_15);
              __Pyx_XGIVEREF(__pyx_t_16);
              __Pyx_XGIVEREF(__pyx_t_17);
              __Pyx_ExceptionReset(__pyx_t_15, __pyx_t_16, __pyx_t_17);
              __pyx_L28_try_end:;
            }
          }

          /* "rasterio/_warp.pyx":617
 *                 raise err
 *         finally:
 *             if wkt != NULL:             # <<<<<<<<<<<<<<
 *                 CPLFree(wkt)
 *             if hTransformArg != NULL:
 */
          /*finally:*/ {
            /*normal exit:*/{
              __pyx_t_22 = ((__pyx_v_wkt != NULL) != 0);
              if (__pyx_t_22) {

                /* "rasterio/_warp.pyx":618
 *         finally:
 *             if wkt != NULL:
 *                 CPLFree(wkt)             # <<<<<<<<<<<<<<
 *             if hTransformArg != NULL:
 *                 GDALDestroyGenImgProjTransformer(hTransformArg)
 */
                CPLFree(__pyx_v_wkt);

                /* "rasterio/_warp.pyx":617
 *                 raise err
 *         finally:
 *             if wkt != NULL:             # <<<<<<<<<<<<<<
 *                 CPLFree(wkt)
 *             if hTransformArg != NULL:
 */
              }

              /* "rasterio/_warp.pyx":619
 *             if wkt != NULL:
 *                 CPLFree(wkt)
 *             if hTransformArg != NULL:             # <<<<<<<<<<<<<<
 *                 GDALDestroyGenImgProjTransformer(hTransformArg)
 * 
 */
              __pyx_t_22 = ((__pyx_v_hTransformArg != NULL) != 0);
              if (__pyx_t_22) {

                /* "rasterio/_warp.pyx":620
 *                 CPLFree(wkt)
 *             if hTransformArg != NULL:
 *                 GDALDestroyGenImgProjTransformer(hTransformArg)             # <<<<<<<<<<<<<<
 * 
 *     # Convert those modified arguments to Python values.
 */
                GDALDestroyGenImgProjTransformer(__pyx_v_hTransformArg);

                /* "rasterio/_warp.pyx":619
 *             if wkt != NULL:
 *                 CPLFree(wkt)
 *             if hTransformArg != NULL:             # <<<<<<<<<<<<<<
 *                 GDALDestroyGenImgProjTransformer(hTransformArg)
 * 
 */
              }
              goto __pyx_L20;
            }
            /*exception exit:*/{
              __Pyx_PyThreadState_declare
              __pyx_L19_error:;
              __pyx_t_17 = 0; __pyx_t_16 = 0; __pyx_t_15 = 0; __pyx_t_25 = 0; __pyx_t_26 = 0; __pyx_t_27 = 0;
              __Pyx_PyThreadState_assign
              __Pyx_XDECREF(__pyx_t_20); __pyx_t_20 = 0;
              __Pyx_XDECREF(__pyx_t_19); __pyx_t_19 = 0;
              __Pyx_XDECREF(__pyx_t_21); __pyx_t_21 = 0;
              __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
              __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
              __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
              if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_25, &__pyx_t_26, &__pyx_t_27);
              if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_17, &__pyx_t_16, &__pyx_t_15) < 0)) __Pyx_ErrFetch(&__pyx_t_17, &__pyx_t_16, &__pyx_t_15);
              __Pyx_XGOTREF(__pyx_t_17);
              __Pyx_XGOTREF(__pyx_t_16);
              __Pyx_XGOTREF(__pyx_t_15);
              __Pyx_XGOTREF(__pyx_t_25);
              __Pyx_XGOTREF(__pyx_t_26);
              __Pyx_XGOTREF(__pyx_t_27);
              __pyx_t_7 = __pyx_lineno; __pyx_t_23 = __pyx_clineno; __pyx_t_24 = __pyx_filename;
              {

                /* "rasterio/_warp.pyx":617
 *                 raise err
 *         finally:
 *             if wkt != NULL:             # <<<<<<<<<<<<<<
 *                 CPLFree(wkt)
 *             if hTransformArg != NULL:
 */
                __pyx_t_22 = ((__pyx_v_wkt != NULL) != 0);
                if (__pyx_t_22) {

                  /* "rasterio/_warp.pyx":618
 *         finally:
 *             if wkt != NULL:
 *                 CPLFree(wkt)             # <<<<<<<<<<<<<<
 *             if hTransformArg != NULL:
 *                 GDALDestroyGenImgProjTransformer(hTransformArg)
 */
                  CPLFree(__pyx_v_wkt);

                  /* "rasterio/_warp.pyx":617
 *                 raise err
 *         finally:
 *             if wkt != NULL:             # <<<<<<<<<<<<<<
 *                 CPLFree(wkt)
 *             if hTransformArg != NULL:
 */
                }

                /* "rasterio/_warp.pyx":619
 *             if wkt != NULL:
 *                 CPLFree(wkt)
 *             if hTransformArg != NULL:             # <<<<<<<<<<<<<<
 *                 GDALDestroyGenImgProjTransformer(hTransformArg)
 * 
 */
                __pyx_t_22 = ((__pyx_v_hTransformArg != NULL) != 0);
                if (__pyx_t_22) {

                  /* "rasterio/_warp.pyx":620
 *                 CPLFree(wkt)
 *             if hTransformArg != NULL:
 *                 GDALDestroyGenImgProjTransformer(hTransformArg)             # <<<<<<<<<<<<<<
 * 
 *     # Convert those modified arguments to Python values.
 */
                  GDALDestroyGenImgProjTransformer(__pyx_v_hTransformArg);

                  /* "rasterio/_warp.pyx":619
 *             if wkt != NULL:
 *                 CPLFree(wkt)
 *             if hTransformArg != NULL:             # <<<<<<<<<<<<<<
 *                 GDALDestroyGenImgProjTransformer(hTransformArg)
 * 
 */
                }
              }
              __Pyx_PyThreadState_assign
              if (PY_MAJOR_VERSION >= 3) {
                __Pyx_XGIVEREF(__pyx_t_25);
                __Pyx_XGIVEREF(__pyx_t_26);
                __Pyx_XGIVEREF(__pyx_t_27);
                __Pyx_ExceptionReset(__pyx_t_25, __pyx_t_26, __pyx_t_27);
              }
              __Pyx_XGIVEREF(__pyx_t_17);
              __Pyx_XGIVEREF(__pyx_t_16);
              __Pyx_XGIVEREF(__pyx_t_15);
              __Pyx_ErrRestore(__pyx_t_17, __pyx_t_16, __pyx_t_15);
              __pyx_t_17 = 0; __pyx_t_16 = 0; __pyx_t_15 = 0; __pyx_t_25 = 0; __pyx_t_26 = 0; __pyx_t_27 = 0;
              __pyx_lineno = __pyx_t_7; __pyx_clineno = __pyx_t_23; __pyx_filename = __pyx_t_24;
              goto __pyx_L8_error;
            }
            __pyx_L20:;
          }

          /* "rasterio/_warp.pyx":591
 *     osr = NULL
 * 
 *     with InMemoryRaster(width=width, height=height, transform=transform,             # <<<<<<<<<<<<<<
 *                         gcps=gcps, crs=src_crs) as temp:
 *         try:
 */
        }
        __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
        __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
        __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
        goto __pyx_L15_try_end;
        __pyx_L8_error:;
        __Pyx_PyThreadState_assign
        __Pyx_XDECREF(__pyx_t_20); __pyx_t_20 = 0;
        __Pyx_XDECREF(__pyx_t_19); __pyx_t_19 = 0;
        __Pyx_XDECREF(__pyx_t_21); __pyx_t_21 = 0;
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        /*except:*/ {
          __Pyx_AddTraceback("rasterio._warp._calculate_default_transform", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_9, &__pyx_t_8, &__pyx_t_3) < 0) __PYX_ERR(0, 591, __pyx_L10_except_error)
          __Pyx_GOTREF(__pyx_t_9);
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_6 = PyTuple_Pack(3, __pyx_t_9, __pyx_t_8, __pyx_t_3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 591, __pyx_L10_except_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_27 = __Pyx_PyObject_Call(__pyx_t_11, __pyx_t_6, NULL);
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          if (unlikely(!__pyx_t_27)) __PYX_ERR(0, 591, __pyx_L10_except_error)
          __Pyx_GOTREF(__pyx_t_27);
          __pyx_t_22 = __Pyx_PyObject_IsTrue(__pyx_t_27);
          __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;
          if (__pyx_t_22 < 0) __PYX_ERR(0, 591, __pyx_L10_except_error)
          __pyx_t_5 = ((!(__pyx_t_22 != 0)) != 0);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_9);
            __Pyx_GIVEREF(__pyx_t_8);
            __Pyx_XGIVEREF(__pyx_t_3);
            __Pyx_ErrRestoreWithState(__pyx_t_9, __pyx_t_8, __pyx_t_3);
            __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_3 = 0; 
            __PYX_ERR(0, 591, __pyx_L10_except_error)
          }
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          goto __pyx_L9_exception_handled;
        }
        __pyx_L10_except_error:;
        __Pyx_PyThreadState_assign
        __Pyx_XGIVEREF(__pyx_t_12);
        __Pyx_XGIVEREF(__pyx_t_13);
        __Pyx_XGIVEREF(__pyx_t_14);
        __Pyx_ExceptionReset(__pyx_t_12, __pyx_t_13, __pyx_t_14);
        goto __pyx_L1_error;
        __pyx_L9_exception_handled:;
        __Pyx_PyThreadState_assign
        __Pyx_XGIVEREF(__pyx_t_12);
        __Pyx_XGIVEREF(__pyx_t_13);
        __Pyx_XGIVEREF(__pyx_t_14);
        __Pyx_ExceptionReset(__pyx_t_12, __pyx_t_13, __pyx_t_14);
        __pyx_L15_try_end:;
      }
    }
    /*finally:*/ {
      /*normal exit:*/{
        if (__pyx_t_11) {
          __pyx_t_14 = __Pyx_PyObject_Call(__pyx_t_11, __pyx_tuple__34, NULL);
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 591, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_14);
          __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
        }
        goto __pyx_L7;
      }
      __pyx_L7:;
    }
    goto __pyx_L45;
    __pyx_L4_error:;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    goto __pyx_L1_error;
    __pyx_L45:;
  }

  /* "rasterio/_warp.pyx":623
 * 
 *     # Convert those modified arguments to Python values.
 *     dst_affine = Affine.from_gdal(*[geotransform[i] for i in range(6)])             # <<<<<<<<<<<<<<
 *     dst_width = npixels
 *     dst_height = nlines
 */
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_Affine); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 623, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_from_gdal); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 623, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 623, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  for (__pyx_t_28 = 0; __pyx_t_28 < 6; __pyx_t_28+=1) {
    __pyx_v_i = __pyx_t_28;
    __pyx_t_9 = PyFloat_FromDouble((__pyx_v_geotransform[__pyx_v_i])); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 623, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_3, (PyObject*)__pyx_t_9))) __PYX_ERR(0, 623, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  }
  __pyx_t_9 = PySequence_Tuple(__pyx_t_3); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 623, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_9, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 623, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_v_dst_affine = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "rasterio/_warp.pyx":624
 *     # Convert those modified arguments to Python values.
 *     dst_affine = Affine.from_gdal(*[geotransform[i] for i in range(6)])
 *     dst_width = npixels             # <<<<<<<<<<<<<<
 *     dst_height = nlines
 * 
 */
  __pyx_v_dst_width = __pyx_v_npixels;

  /* "rasterio/_warp.pyx":625
 *     dst_affine = Affine.from_gdal(*[geotransform[i] for i in range(6)])
 *     dst_width = npixels
 *     dst_height = nlines             # <<<<<<<<<<<<<<
 * 
 *     return dst_affine, dst_width, dst_height
 */
  __pyx_v_dst_height = __pyx_v_nlines;

  /* "rasterio/_warp.pyx":627
 *     dst_height = nlines
 * 
 *     return dst_affine, dst_width, dst_height             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_dst_width); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 627, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_9 = __Pyx_PyInt_From_int(__pyx_v_dst_height); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 627, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_8 = PyTuple_New(3); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 627, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_INCREF(__pyx_v_dst_affine);
  __Pyx_GIVEREF(__pyx_v_dst_affine);
  PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_v_dst_affine);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_8, 2, __pyx_t_9);
  __pyx_t_3 = 0;
  __pyx_t_9 = 0;
  __pyx_r = __pyx_t_8;
  __pyx_t_8 = 0;
  goto __pyx_L0;

  /* "rasterio/_warp.pyx":557
 * 
 * 
 * def _calculate_default_transform(src_crs, dst_crs, width, height,             # <<<<<<<<<<<<<<
 *                                  left=None, bottom=None, right=None, top=None,
 *                                  gcps=None, **kwargs):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_19);
  __Pyx_XDECREF(__pyx_t_20);
  __Pyx_XDECREF(__pyx_t_21);
  __Pyx_AddTraceback("rasterio._warp._calculate_default_transform", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_temp);
  __Pyx_XDECREF(__pyx_v_transform);
  __Pyx_XDECREF(__pyx_v_img);
  __Pyx_XDECREF(__pyx_v_err);
  __Pyx_XDECREF(__pyx_v_dst_affine);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "rasterio/_warp.pyx":632
 * cdef class WarpedVRTReaderBase(DatasetReaderBase):
 * 
 *     def __init__(self, src_dataset, dst_crs=None, resampling=Resampling.nearest,             # <<<<<<<<<<<<<<
 *                  tolerance=0.125, src_nodata=None, dst_nodata=None,
 *                  dst_width=None, dst_height=None, dst_transform=None,
 */

/* Python wrapper */
static int __pyx_pw_8rasterio_5_warp_19WarpedVRTReaderBase_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_8rasterio_5_warp_19WarpedVRTReaderBase_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_src_dataset = 0;
  PyObject *__pyx_v_dst_crs = 0;
  PyObject *__pyx_v_resampling = 0;
  PyObject *__pyx_v_tolerance = 0;
  PyObject *__pyx_v_src_nodata = 0;
  PyObject *__pyx_v_dst_nodata = 0;
  PyObject *__pyx_v_dst_width = 0;
  PyObject *__pyx_v_dst_height = 0;
  PyObject *__pyx_v_dst_transform = 0;
  PyObject *__pyx_v_init_dest_nodata = 0;
  PyObject *__pyx_v_warp_extras = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  __pyx_v_warp_extras = PyDict_New(); if (unlikely(!__pyx_v_warp_extras)) return -1;
  __Pyx_GOTREF(__pyx_v_warp_extras);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_src_dataset,&__pyx_n_s_dst_crs,&__pyx_n_s_resampling,&__pyx_n_s_tolerance,&__pyx_n_s_src_nodata,&__pyx_n_s_dst_nodata,&__pyx_n_s_dst_width,&__pyx_n_s_dst_height,&__pyx_n_s_dst_transform,&__pyx_n_s_init_dest_nodata,0};
    PyObject* values[10] = {0,0,0,0,0,0,0,0,0,0};
    values[1] = ((PyObject *)Py_None);
    values[2] = __pyx_k__35;
    values[3] = ((PyObject *)__pyx_float_0_125);

    /* "rasterio/_warp.pyx":633
 * 
 *     def __init__(self, src_dataset, dst_crs=None, resampling=Resampling.nearest,
 *                  tolerance=0.125, src_nodata=None, dst_nodata=None,             # <<<<<<<<<<<<<<
 *                  dst_width=None, dst_height=None, dst_transform=None,
 *                  init_dest_nodata=True, **warp_extras):
 */
    values[4] = ((PyObject *)Py_None);
    values[5] = ((PyObject *)Py_None);

    /* "rasterio/_warp.pyx":634
 *     def __init__(self, src_dataset, dst_crs=None, resampling=Resampling.nearest,
 *                  tolerance=0.125, src_nodata=None, dst_nodata=None,
 *                  dst_width=None, dst_height=None, dst_transform=None,             # <<<<<<<<<<<<<<
 *                  init_dest_nodata=True, **warp_extras):
 *         # kwargs become warp options.
 */
    values[6] = ((PyObject *)Py_None);
    values[7] = ((PyObject *)Py_None);
    values[8] = ((PyObject *)Py_None);

    /* "rasterio/_warp.pyx":635
 *                  tolerance=0.125, src_nodata=None, dst_nodata=None,
 *                  dst_width=None, dst_height=None, dst_transform=None,
 *                  init_dest_nodata=True, **warp_extras):             # <<<<<<<<<<<<<<
 *         # kwargs become warp options.
 *         super(WarpedVRTReaderBase, self).__init__(self)
 */
    values[9] = ((PyObject *)Py_True);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_src_dataset)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dst_crs);
          if (value) { values[1] = value; kw_args--; }
        }
        case  2:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_resampling);
          if (value) { values[2] = value; kw_args--; }
        }
        case  3:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_tolerance);
          if (value) { values[3] = value; kw_args--; }
        }
        case  4:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_src_nodata);
          if (value) { values[4] = value; kw_args--; }
        }
        case  5:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dst_nodata);
          if (value) { values[5] = value; kw_args--; }
        }
        case  6:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dst_width);
          if (value) { values[6] = value; kw_args--; }
        }
        case  7:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dst_height);
          if (value) { values[7] = value; kw_args--; }
        }
        case  8:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dst_transform);
          if (value) { values[8] = value; kw_args--; }
        }
        case  9:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_init_dest_nodata);
          if (value) { values[9] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, __pyx_v_warp_extras, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 632, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_src_dataset = values[0];
    __pyx_v_dst_crs = values[1];
    __pyx_v_resampling = values[2];
    __pyx_v_tolerance = values[3];
    __pyx_v_src_nodata = values[4];
    __pyx_v_dst_nodata = values[5];
    __pyx_v_dst_width = values[6];
    __pyx_v_dst_height = values[7];
    __pyx_v_dst_transform = values[8];
    __pyx_v_init_dest_nodata = values[9];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 1, 10, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 632, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_DECREF(__pyx_v_warp_extras); __pyx_v_warp_extras = 0;
  __Pyx_AddTraceback("rasterio._warp.WarpedVRTReaderBase.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8rasterio_5_warp_19WarpedVRTReaderBase___init__(((struct __pyx_obj_8rasterio_5_warp_WarpedVRTReaderBase *)__pyx_v_self), __pyx_v_src_dataset, __pyx_v_dst_crs, __pyx_v_resampling, __pyx_v_tolerance, __pyx_v_src_nodata, __pyx_v_dst_nodata, __pyx_v_dst_width, __pyx_v_dst_height, __pyx_v_dst_transform, __pyx_v_init_dest_nodata, __pyx_v_warp_extras);

  /* "rasterio/_warp.pyx":632
 * cdef class WarpedVRTReaderBase(DatasetReaderBase):
 * 
 *     def __init__(self, src_dataset, dst_crs=None, resampling=Resampling.nearest,             # <<<<<<<<<<<<<<
 *                  tolerance=0.125, src_nodata=None, dst_nodata=None,
 *                  dst_width=None, dst_height=None, dst_transform=None,
 */

  /* function exit code */
  __Pyx_XDECREF(__pyx_v_warp_extras);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_8rasterio_5_warp_19WarpedVRTReaderBase___init__(struct __pyx_obj_8rasterio_5_warp_WarpedVRTReaderBase *__pyx_v_self, PyObject *__pyx_v_src_dataset, PyObject *__pyx_v_dst_crs, PyObject *__pyx_v_resampling, PyObject *__pyx_v_tolerance, PyObject *__pyx_v_src_nodata, PyObject *__pyx_v_dst_nodata, PyObject *__pyx_v_dst_width, PyObject *__pyx_v_dst_height, PyObject *__pyx_v_dst_transform, PyObject *__pyx_v_init_dest_nodata, PyObject *__pyx_v_warp_extras) {
  GDALDriverH __pyx_v_driver;
  GDALDatasetH __pyx_v_hds;
  GDALDatasetH __pyx_v_hds_warped;
  CYTHON_UNUSED char const *__pyx_v_cypath;
  char *__pyx_v_dst_crs_wkt;
  OGRSpatialReferenceH __pyx_v_osr;
  char **__pyx_v_c_warp_extras;
  GDALWarpOptions *__pyx_v_psWOptions;
  float __pyx_v_c_tolerance;
  GDALResampleAlg __pyx_v_c_resampling;
  int __pyx_v_c_width;
  int __pyx_v_c_height;
  double __pyx_v_gt[6];
  void *__pyx_v_hTransformArg;
  PyObject *__pyx_v_t = NULL;
  long __pyx_v_i;
  PyObject *__pyx_v_key = NULL;
  PyObject *__pyx_v_val = NULL;
  PyObject *__pyx_v_err = NULL;
  CYTHON_UNUSED PyObject *__pyx_v__ = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  float __pyx_t_9;
  GDALResampleAlg __pyx_t_10;
  int __pyx_t_11;
  int __pyx_t_12;
  OGRSpatialReferenceH __pyx_t_13;
  char const *__pyx_t_14;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  GDALDatasetH __pyx_t_21;
  void *__pyx_t_22;
  long __pyx_t_23;
  double __pyx_t_24;
  Py_ssize_t __pyx_t_25;
  PyObject *(*__pyx_t_26)(PyObject *);
  PyObject *(*__pyx_t_27)(PyObject *);
  char const *__pyx_t_28;
  char const *__pyx_t_29;
  GDALWarpOptions *__pyx_t_30;
  PyObject *__pyx_t_31 = NULL;
  PyObject *__pyx_t_32 = NULL;
  PyObject *__pyx_t_33 = NULL;
  char const *__pyx_t_34;
  char const *__pyx_t_35;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "rasterio/_warp.pyx":637
 *                  init_dest_nodata=True, **warp_extras):
 *         # kwargs become warp options.
 *         super(WarpedVRTReaderBase, self).__init__(self)             # <<<<<<<<<<<<<<
 *         self.src_dataset = src_dataset
 *         self.name = "WarpedVRT({})".format(src_dataset.name)
 */
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 637, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(((PyObject *)__pyx_ptype_8rasterio_5_warp_WarpedVRTReaderBase));
  __Pyx_GIVEREF(((PyObject *)__pyx_ptype_8rasterio_5_warp_WarpedVRTReaderBase));
  PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_ptype_8rasterio_5_warp_WarpedVRTReaderBase));
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
  PyTuple_SET_ITEM(__pyx_t_2, 1, ((PyObject *)__pyx_v_self));
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_2, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 637, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_init); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 637, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  if (!__pyx_t_3) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, ((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 637, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[2] = {__pyx_t_3, ((PyObject *)__pyx_v_self)};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 637, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[2] = {__pyx_t_3, ((PyObject *)__pyx_v_self)};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 637, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 637, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3); __pyx_t_3 = NULL;
      __Pyx_INCREF(((PyObject *)__pyx_v_self));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
      PyTuple_SET_ITEM(__pyx_t_4, 0+1, ((PyObject *)__pyx_v_self));
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 637, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "rasterio/_warp.pyx":638
 *         # kwargs become warp options.
 *         super(WarpedVRTReaderBase, self).__init__(self)
 *         self.src_dataset = src_dataset             # <<<<<<<<<<<<<<
 *         self.name = "WarpedVRT({})".format(src_dataset.name)
 *         self.dst_crs = dst_crs
 */
  if (__Pyx_PyObject_SetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_src_dataset, __pyx_v_src_dataset) < 0) __PYX_ERR(0, 638, __pyx_L1_error)

  /* "rasterio/_warp.pyx":639
 *         super(WarpedVRTReaderBase, self).__init__(self)
 *         self.src_dataset = src_dataset
 *         self.name = "WarpedVRT({})".format(src_dataset.name)             # <<<<<<<<<<<<<<
 *         self.dst_crs = dst_crs
 *         self.resampling = resampling
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_kp_s_WarpedVRT, __pyx_n_s_format); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 639, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_src_dataset, __pyx_n_s_name); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 639, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  if (!__pyx_t_3) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 639, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_t_4};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 639, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_t_4};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 639, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else
    #endif
    {
      __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 639, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
      __Pyx_GIVEREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_t_4);
      __pyx_t_4 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 639, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->__pyx_base.__pyx_base.name);
  __Pyx_DECREF(__pyx_v_self->__pyx_base.__pyx_base.name);
  __pyx_v_self->__pyx_base.__pyx_base.name = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "rasterio/_warp.pyx":640
 *         self.src_dataset = src_dataset
 *         self.name = "WarpedVRT({})".format(src_dataset.name)
 *         self.dst_crs = dst_crs             # <<<<<<<<<<<<<<
 *         self.resampling = resampling
 *         self.tolerance = tolerance
 */
  if (__Pyx_PyObject_SetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dst_crs, __pyx_v_dst_crs) < 0) __PYX_ERR(0, 640, __pyx_L1_error)

  /* "rasterio/_warp.pyx":641
 *         self.name = "WarpedVRT({})".format(src_dataset.name)
 *         self.dst_crs = dst_crs
 *         self.resampling = resampling             # <<<<<<<<<<<<<<
 *         self.tolerance = tolerance
 *         self.src_nodata = src_nodata
 */
  if (__Pyx_PyObject_SetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_resampling, __pyx_v_resampling) < 0) __PYX_ERR(0, 641, __pyx_L1_error)

  /* "rasterio/_warp.pyx":642
 *         self.dst_crs = dst_crs
 *         self.resampling = resampling
 *         self.tolerance = tolerance             # <<<<<<<<<<<<<<
 *         self.src_nodata = src_nodata
 *         self.dst_nodata = dst_nodata
 */
  if (__Pyx_PyObject_SetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_tolerance, __pyx_v_tolerance) < 0) __PYX_ERR(0, 642, __pyx_L1_error)

  /* "rasterio/_warp.pyx":643
 *         self.resampling = resampling
 *         self.tolerance = tolerance
 *         self.src_nodata = src_nodata             # <<<<<<<<<<<<<<
 *         self.dst_nodata = dst_nodata
 *         self.dst_width = dst_width
 */
  if (__Pyx_PyObject_SetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_src_nodata, __pyx_v_src_nodata) < 0) __PYX_ERR(0, 643, __pyx_L1_error)

  /* "rasterio/_warp.pyx":644
 *         self.tolerance = tolerance
 *         self.src_nodata = src_nodata
 *         self.dst_nodata = dst_nodata             # <<<<<<<<<<<<<<
 *         self.dst_width = dst_width
 *         self.dst_height = dst_height
 */
  if (__Pyx_PyObject_SetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dst_nodata, __pyx_v_dst_nodata) < 0) __PYX_ERR(0, 644, __pyx_L1_error)

  /* "rasterio/_warp.pyx":645
 *         self.src_nodata = src_nodata
 *         self.dst_nodata = dst_nodata
 *         self.dst_width = dst_width             # <<<<<<<<<<<<<<
 *         self.dst_height = dst_height
 *         self.dst_transform = dst_transform
 */
  if (__Pyx_PyObject_SetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dst_width, __pyx_v_dst_width) < 0) __PYX_ERR(0, 645, __pyx_L1_error)

  /* "rasterio/_warp.pyx":646
 *         self.dst_nodata = dst_nodata
 *         self.dst_width = dst_width
 *         self.dst_height = dst_height             # <<<<<<<<<<<<<<
 *         self.dst_transform = dst_transform
 *         self.warp_extras = warp_extras.copy()
 */
  if (__Pyx_PyObject_SetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dst_height, __pyx_v_dst_height) < 0) __PYX_ERR(0, 646, __pyx_L1_error)

  /* "rasterio/_warp.pyx":647
 *         self.dst_width = dst_width
 *         self.dst_height = dst_height
 *         self.dst_transform = dst_transform             # <<<<<<<<<<<<<<
 *         self.warp_extras = warp_extras.copy()
 *         if init_dest_nodata is True and 'init_dest' not in warp_extras:
 */
  if (__Pyx_PyObject_SetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dst_transform, __pyx_v_dst_transform) < 0) __PYX_ERR(0, 647, __pyx_L1_error)

  /* "rasterio/_warp.pyx":648
 *         self.dst_height = dst_height
 *         self.dst_transform = dst_transform
 *         self.warp_extras = warp_extras.copy()             # <<<<<<<<<<<<<<
 *         if init_dest_nodata is True and 'init_dest' not in warp_extras:
 *             self.warp_extras['init_dest'] = 'NO_DATA'
 */
  __pyx_t_1 = PyDict_Copy(__pyx_v_warp_extras); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 648, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PyObject_SetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_warp_extras, __pyx_t_1) < 0) __PYX_ERR(0, 648, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "rasterio/_warp.pyx":649
 *         self.dst_transform = dst_transform
 *         self.warp_extras = warp_extras.copy()
 *         if init_dest_nodata is True and 'init_dest' not in warp_extras:             # <<<<<<<<<<<<<<
 *             self.warp_extras['init_dest'] = 'NO_DATA'
 * 
 */
  __pyx_t_7 = (__pyx_v_init_dest_nodata == Py_True);
  __pyx_t_8 = (__pyx_t_7 != 0);
  if (__pyx_t_8) {
  } else {
    __pyx_t_6 = __pyx_t_8;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_8 = (__Pyx_PyDict_ContainsTF(__pyx_n_s_init_dest, __pyx_v_warp_extras, Py_NE)); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 649, __pyx_L1_error)
  __pyx_t_7 = (__pyx_t_8 != 0);
  __pyx_t_6 = __pyx_t_7;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_6) {

    /* "rasterio/_warp.pyx":650
 *         self.warp_extras = warp_extras.copy()
 *         if init_dest_nodata is True and 'init_dest' not in warp_extras:
 *             self.warp_extras['init_dest'] = 'NO_DATA'             # <<<<<<<<<<<<<<
 * 
 *         cdef GDALDriverH driver = NULL
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_warp_extras); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 650, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (unlikely(PyObject_SetItem(__pyx_t_1, __pyx_n_s_init_dest, __pyx_n_s_NO_DATA) < 0)) __PYX_ERR(0, 650, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "rasterio/_warp.pyx":649
 *         self.dst_transform = dst_transform
 *         self.warp_extras = warp_extras.copy()
 *         if init_dest_nodata is True and 'init_dest' not in warp_extras:             # <<<<<<<<<<<<<<
 *             self.warp_extras['init_dest'] = 'NO_DATA'
 * 
 */
  }

  /* "rasterio/_warp.pyx":652
 *             self.warp_extras['init_dest'] = 'NO_DATA'
 * 
 *         cdef GDALDriverH driver = NULL             # <<<<<<<<<<<<<<
 *         cdef GDALDatasetH hds = NULL
 *         cdef GDALDatasetH hds_warped = NULL
 */
  __pyx_v_driver = NULL;

  /* "rasterio/_warp.pyx":653
 * 
 *         cdef GDALDriverH driver = NULL
 *         cdef GDALDatasetH hds = NULL             # <<<<<<<<<<<<<<
 *         cdef GDALDatasetH hds_warped = NULL
 *         cdef const char *cypath = NULL
 */
  __pyx_v_hds = NULL;

  /* "rasterio/_warp.pyx":654
 *         cdef GDALDriverH driver = NULL
 *         cdef GDALDatasetH hds = NULL
 *         cdef GDALDatasetH hds_warped = NULL             # <<<<<<<<<<<<<<
 *         cdef const char *cypath = NULL
 *         cdef char *dst_crs_wkt = NULL
 */
  __pyx_v_hds_warped = NULL;

  /* "rasterio/_warp.pyx":655
 *         cdef GDALDatasetH hds = NULL
 *         cdef GDALDatasetH hds_warped = NULL
 *         cdef const char *cypath = NULL             # <<<<<<<<<<<<<<
 *         cdef char *dst_crs_wkt = NULL
 *         cdef OGRSpatialReferenceH osr = NULL
 */
  __pyx_v_cypath = NULL;

  /* "rasterio/_warp.pyx":656
 *         cdef GDALDatasetH hds_warped = NULL
 *         cdef const char *cypath = NULL
 *         cdef char *dst_crs_wkt = NULL             # <<<<<<<<<<<<<<
 *         cdef OGRSpatialReferenceH osr = NULL
 *         cdef char **c_warp_extras = NULL
 */
  __pyx_v_dst_crs_wkt = NULL;

  /* "rasterio/_warp.pyx":657
 *         cdef const char *cypath = NULL
 *         cdef char *dst_crs_wkt = NULL
 *         cdef OGRSpatialReferenceH osr = NULL             # <<<<<<<<<<<<<<
 *         cdef char **c_warp_extras = NULL
 *         cdef GDALWarpOptions *psWOptions = NULL
 */
  __pyx_v_osr = NULL;

  /* "rasterio/_warp.pyx":658
 *         cdef char *dst_crs_wkt = NULL
 *         cdef OGRSpatialReferenceH osr = NULL
 *         cdef char **c_warp_extras = NULL             # <<<<<<<<<<<<<<
 *         cdef GDALWarpOptions *psWOptions = NULL
 *         cdef float c_tolerance = tolerance
 */
  __pyx_v_c_warp_extras = NULL;

  /* "rasterio/_warp.pyx":659
 *         cdef OGRSpatialReferenceH osr = NULL
 *         cdef char **c_warp_extras = NULL
 *         cdef GDALWarpOptions *psWOptions = NULL             # <<<<<<<<<<<<<<
 *         cdef float c_tolerance = tolerance
 *         cdef GDALResampleAlg c_resampling = resampling
 */
  __pyx_v_psWOptions = NULL;

  /* "rasterio/_warp.pyx":660
 *         cdef char **c_warp_extras = NULL
 *         cdef GDALWarpOptions *psWOptions = NULL
 *         cdef float c_tolerance = tolerance             # <<<<<<<<<<<<<<
 *         cdef GDALResampleAlg c_resampling = resampling
 *         cdef int c_width = dst_width or 0
 */
  __pyx_t_9 = __pyx_PyFloat_AsFloat(__pyx_v_tolerance); if (unlikely((__pyx_t_9 == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 660, __pyx_L1_error)
  __pyx_v_c_tolerance = __pyx_t_9;

  /* "rasterio/_warp.pyx":661
 *         cdef GDALWarpOptions *psWOptions = NULL
 *         cdef float c_tolerance = tolerance
 *         cdef GDALResampleAlg c_resampling = resampling             # <<<<<<<<<<<<<<
 *         cdef int c_width = dst_width or 0
 *         cdef int c_height = dst_height or 0
 */
  __pyx_t_10 = ((GDALResampleAlg)__Pyx_PyInt_As_GDALResampleAlg(__pyx_v_resampling)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 661, __pyx_L1_error)
  __pyx_v_c_resampling = __pyx_t_10;

  /* "rasterio/_warp.pyx":662
 *         cdef float c_tolerance = tolerance
 *         cdef GDALResampleAlg c_resampling = resampling
 *         cdef int c_width = dst_width or 0             # <<<<<<<<<<<<<<
 *         cdef int c_height = dst_height or 0
 *         cdef double gt[6]
 */
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_v_dst_width); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 662, __pyx_L1_error)
  if (!__pyx_t_6) {
  } else {
    __pyx_t_12 = __Pyx_PyInt_As_int(__pyx_v_dst_width); if (unlikely((__pyx_t_12 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 662, __pyx_L1_error)
    __pyx_t_11 = __pyx_t_12;
    goto __pyx_L6_bool_binop_done;
  }
  __pyx_t_11 = 0;
  __pyx_L6_bool_binop_done:;
  __pyx_v_c_width = __pyx_t_11;

  /* "rasterio/_warp.pyx":663
 *         cdef GDALResampleAlg c_resampling = resampling
 *         cdef int c_width = dst_width or 0
 *         cdef int c_height = dst_height or 0             # <<<<<<<<<<<<<<
 *         cdef double gt[6]
 *         cdef void *hTransformArg = NULL
 */
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_v_dst_height); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 663, __pyx_L1_error)
  if (!__pyx_t_6) {
  } else {
    __pyx_t_12 = __Pyx_PyInt_As_int(__pyx_v_dst_height); if (unlikely((__pyx_t_12 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 663, __pyx_L1_error)
    __pyx_t_11 = __pyx_t_12;
    goto __pyx_L8_bool_binop_done;
  }
  __pyx_t_11 = 0;
  __pyx_L8_bool_binop_done:;
  __pyx_v_c_height = __pyx_t_11;

  /* "rasterio/_warp.pyx":665
 *         cdef int c_height = dst_height or 0
 *         cdef double gt[6]
 *         cdef void *hTransformArg = NULL             # <<<<<<<<<<<<<<
 * 
 *         # Convert destination CRS to a C WKT string.
 */
  __pyx_v_hTransformArg = NULL;

  /* "rasterio/_warp.pyx":668
 * 
 *         # Convert destination CRS to a C WKT string.
 *         try:             # <<<<<<<<<<<<<<
 *             osr = _osr_from_crs(self.dst_crs)
 *             OSRExportToWkt(osr, &dst_crs_wkt)
 */
  /*try:*/ {

    /* "rasterio/_warp.pyx":669
 *         # Convert destination CRS to a C WKT string.
 *         try:
 *             osr = _osr_from_crs(self.dst_crs)             # <<<<<<<<<<<<<<
 *             OSRExportToWkt(osr, &dst_crs_wkt)
 *         finally:
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dst_crs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 669, __pyx_L11_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_13 = __pyx_f_8rasterio_5_base__osr_from_crs(__pyx_t_1); if (unlikely(__pyx_t_13 == NULL)) __PYX_ERR(0, 669, __pyx_L11_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_osr = __pyx_t_13;

    /* "rasterio/_warp.pyx":670
 *         try:
 *             osr = _osr_from_crs(self.dst_crs)
 *             OSRExportToWkt(osr, &dst_crs_wkt)             # <<<<<<<<<<<<<<
 *         finally:
 *             OSRRelease(osr)
 */
    OSRExportToWkt(__pyx_v_osr, (&__pyx_v_dst_crs_wkt));
  }

  /* "rasterio/_warp.pyx":672
 *             OSRExportToWkt(osr, &dst_crs_wkt)
 *         finally:
 *             OSRRelease(osr)             # <<<<<<<<<<<<<<
 *             osr = NULL
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      OSRRelease(__pyx_v_osr);

      /* "rasterio/_warp.pyx":673
 *         finally:
 *             OSRRelease(osr)
 *             osr = NULL             # <<<<<<<<<<<<<<
 * 
 *         log.debug("Exported CRS to WKT.")
 */
      __pyx_v_osr = NULL;
      goto __pyx_L12;
    }
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __pyx_L11_error:;
      __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_18, &__pyx_t_19, &__pyx_t_20);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17) < 0)) __Pyx_ErrFetch(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17);
      __Pyx_XGOTREF(__pyx_t_15);
      __Pyx_XGOTREF(__pyx_t_16);
      __Pyx_XGOTREF(__pyx_t_17);
      __Pyx_XGOTREF(__pyx_t_18);
      __Pyx_XGOTREF(__pyx_t_19);
      __Pyx_XGOTREF(__pyx_t_20);
      __pyx_t_11 = __pyx_lineno; __pyx_t_12 = __pyx_clineno; __pyx_t_14 = __pyx_filename;
      {

        /* "rasterio/_warp.pyx":672
 *             OSRExportToWkt(osr, &dst_crs_wkt)
 *         finally:
 *             OSRRelease(osr)             # <<<<<<<<<<<<<<
 *             osr = NULL
 * 
 */
        OSRRelease(__pyx_v_osr);

        /* "rasterio/_warp.pyx":673
 *         finally:
 *             OSRRelease(osr)
 *             osr = NULL             # <<<<<<<<<<<<<<
 * 
 *         log.debug("Exported CRS to WKT.")
 */
        __pyx_v_osr = NULL;
      }
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_ExceptionReset(__pyx_t_18, __pyx_t_19, __pyx_t_20);
      }
      __Pyx_XGIVEREF(__pyx_t_15);
      __Pyx_XGIVEREF(__pyx_t_16);
      __Pyx_XGIVEREF(__pyx_t_17);
      __Pyx_ErrRestore(__pyx_t_15, __pyx_t_16, __pyx_t_17);
      __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0;
      __pyx_lineno = __pyx_t_11; __pyx_clineno = __pyx_t_12; __pyx_filename = __pyx_t_14;
      goto __pyx_L1_error;
    }
    __pyx_L12:;
  }

  /* "rasterio/_warp.pyx":675
 *             osr = NULL
 * 
 *         log.debug("Exported CRS to WKT.")             # <<<<<<<<<<<<<<
 * 
 *         hds = (<DatasetReaderBase?>self.src_dataset).handle()
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_log); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 675, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_debug); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 675, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_tuple__36, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 675, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "rasterio/_warp.pyx":677
 *         log.debug("Exported CRS to WKT.")
 * 
 *         hds = (<DatasetReaderBase?>self.src_dataset).handle()             # <<<<<<<<<<<<<<
 *         hds = exc_wrap_pointer(hds)
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_src_dataset); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 677, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_8rasterio_3_io_DatasetReaderBase)))) __PYX_ERR(0, 677, __pyx_L1_error)
  __pyx_t_21 = ((struct __pyx_vtabstruct_8rasterio_3_io_DatasetReaderBase *)((struct __pyx_obj_8rasterio_3_io_DatasetReaderBase *)__pyx_t_1)->__pyx_base.__pyx_vtab)->__pyx_base.handle(((struct __pyx_obj_8rasterio_5_base_DatasetBase *)((struct __pyx_obj_8rasterio_3_io_DatasetReaderBase *)__pyx_t_1))); if (unlikely(__pyx_t_21 == NULL)) __PYX_ERR(0, 677, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_hds = __pyx_t_21;

  /* "rasterio/_warp.pyx":678
 * 
 *         hds = (<DatasetReaderBase?>self.src_dataset).handle()
 *         hds = exc_wrap_pointer(hds)             # <<<<<<<<<<<<<<
 * 
 *         if dst_transform:
 */
  __pyx_t_22 = __pyx_f_8rasterio_4_err_exc_wrap_pointer(__pyx_v_hds); if (unlikely(__pyx_t_22 == NULL)) __PYX_ERR(0, 678, __pyx_L1_error)
  __pyx_v_hds = __pyx_t_22;

  /* "rasterio/_warp.pyx":680
 *         hds = exc_wrap_pointer(hds)
 * 
 *         if dst_transform:             # <<<<<<<<<<<<<<
 *             t = dst_transform.to_gdal()
 *             for i in range(6):
 */
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_v_dst_transform); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 680, __pyx_L1_error)
  if (__pyx_t_6) {

    /* "rasterio/_warp.pyx":681
 * 
 *         if dst_transform:
 *             t = dst_transform.to_gdal()             # <<<<<<<<<<<<<<
 *             for i in range(6):
 *                 gt[i] = t[i]
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_dst_transform, __pyx_n_s_to_gdal); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 681, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    if (__pyx_t_5) {
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 681, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else {
      __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 681, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_t = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "rasterio/_warp.pyx":682
 *         if dst_transform:
 *             t = dst_transform.to_gdal()
 *             for i in range(6):             # <<<<<<<<<<<<<<
 *                 gt[i] = t[i]
 * 
 */
    for (__pyx_t_23 = 0; __pyx_t_23 < 6; __pyx_t_23+=1) {
      __pyx_v_i = __pyx_t_23;

      /* "rasterio/_warp.pyx":683
 *             t = dst_transform.to_gdal()
 *             for i in range(6):
 *                 gt[i] = t[i]             # <<<<<<<<<<<<<<
 * 
 *         log.debug("Warp_extras: %r", self.warp_extras)
 */
      __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_t, __pyx_v_i, long, 1, __Pyx_PyInt_From_long, 0, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 683, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_24 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_24 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 683, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      (__pyx_v_gt[__pyx_v_i]) = __pyx_t_24;
    }

    /* "rasterio/_warp.pyx":680
 *         hds = exc_wrap_pointer(hds)
 * 
 *         if dst_transform:             # <<<<<<<<<<<<<<
 *             t = dst_transform.to_gdal()
 *             for i in range(6):
 */
  }

  /* "rasterio/_warp.pyx":685
 *                 gt[i] = t[i]
 * 
 *         log.debug("Warp_extras: %r", self.warp_extras)             # <<<<<<<<<<<<<<
 * 
 *         for key, val in self.warp_extras.items():
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_log); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 685, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_debug); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 685, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_warp_extras); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 685, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  __pyx_t_12 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
      __pyx_t_12 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_5)) {
    PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_kp_s_Warp_extras_r, __pyx_t_2};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_12, 2+__pyx_t_12); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 685, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
    PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_kp_s_Warp_extras_r, __pyx_t_2};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_12, 2+__pyx_t_12); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 685, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else
  #endif
  {
    __pyx_t_3 = PyTuple_New(2+__pyx_t_12); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 685, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_kp_s_Warp_extras_r);
    __Pyx_GIVEREF(__pyx_kp_s_Warp_extras_r);
    PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_12, __pyx_kp_s_Warp_extras_r);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_12, __pyx_t_2);
    __pyx_t_2 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 685, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "rasterio/_warp.pyx":687
 *         log.debug("Warp_extras: %r", self.warp_extras)
 * 
 *         for key, val in self.warp_extras.items():             # <<<<<<<<<<<<<<
 *             key = key.upper().encode('utf-8')
 *             val = str(val).upper().encode('utf-8')
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_warp_extras); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 687, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_items); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 687, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (__pyx_t_5) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 687, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  } else {
    __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 687, __pyx_L1_error)
  }
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_3 = __pyx_t_1; __Pyx_INCREF(__pyx_t_3); __pyx_t_25 = 0;
    __pyx_t_26 = NULL;
  } else {
    __pyx_t_25 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 687, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_26 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_26)) __PYX_ERR(0, 687, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_26)) {
      if (likely(PyList_CheckExact(__pyx_t_3))) {
        if (__pyx_t_25 >= PyList_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_25); __Pyx_INCREF(__pyx_t_1); __pyx_t_25++; if (unlikely(0 < 0)) __PYX_ERR(0, 687, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_3, __pyx_t_25); __pyx_t_25++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 687, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_25 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_25); __Pyx_INCREF(__pyx_t_1); __pyx_t_25++; if (unlikely(0 < 0)) __PYX_ERR(0, 687, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_3, __pyx_t_25); __pyx_t_25++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 687, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_26(__pyx_t_3);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 687, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
      PyObject* sequence = __pyx_t_1;
      #if !CYTHON_COMPILING_IN_PYPY
      Py_ssize_t size = Py_SIZE(sequence);
      #else
      Py_ssize_t size = PySequence_Size(sequence);
      #endif
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 687, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_5 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_2 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_5 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_2 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_2);
      #else
      __pyx_t_5 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 687, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_2 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 687, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_4 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 687, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_27 = Py_TYPE(__pyx_t_4)->tp_iternext;
      index = 0; __pyx_t_5 = __pyx_t_27(__pyx_t_4); if (unlikely(!__pyx_t_5)) goto __pyx_L20_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_5);
      index = 1; __pyx_t_2 = __pyx_t_27(__pyx_t_4); if (unlikely(!__pyx_t_2)) goto __pyx_L20_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_2);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_27(__pyx_t_4), 2) < 0) __PYX_ERR(0, 687, __pyx_L1_error)
      __pyx_t_27 = NULL;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      goto __pyx_L21_unpacking_done;
      __pyx_L20_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_27 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 687, __pyx_L1_error)
      __pyx_L21_unpacking_done:;
    }
    __Pyx_XDECREF_SET(__pyx_v_key, __pyx_t_5);
    __pyx_t_5 = 0;
    __Pyx_XDECREF_SET(__pyx_v_val, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "rasterio/_warp.pyx":688
 * 
 *         for key, val in self.warp_extras.items():
 *             key = key.upper().encode('utf-8')             # <<<<<<<<<<<<<<
 *             val = str(val).upper().encode('utf-8')
 *             c_warp_extras = CSLSetNameValue(
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_key, __pyx_n_s_upper); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 688, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    if (__pyx_t_5) {
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 688, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else {
      __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 688, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_encode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 688, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_tuple__37, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 688, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF_SET(__pyx_v_key, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "rasterio/_warp.pyx":689
 *         for key, val in self.warp_extras.items():
 *             key = key.upper().encode('utf-8')
 *             val = str(val).upper().encode('utf-8')             # <<<<<<<<<<<<<<
 *             c_warp_extras = CSLSetNameValue(
 *                 c_warp_extras, <const char *>key, <const char *>val)
 */
    __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 689, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v_val);
    __Pyx_GIVEREF(__pyx_v_val);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_val);
    __pyx_t_5 = __Pyx_PyObject_Call(((PyObject *)(&PyString_Type)), __pyx_t_2, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 689, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_upper); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 689, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    if (__pyx_t_5) {
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 689, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else {
      __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 689, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_encode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 689, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_tuple__38, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 689, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF_SET(__pyx_v_val, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "rasterio/_warp.pyx":691
 *             val = str(val).upper().encode('utf-8')
 *             c_warp_extras = CSLSetNameValue(
 *                 c_warp_extras, <const char *>key, <const char *>val)             # <<<<<<<<<<<<<<
 * 
 *         psWOptions = create_warp_options(
 */
    __pyx_t_28 = __Pyx_PyObject_AsString(__pyx_v_key); if (unlikely((!__pyx_t_28) && PyErr_Occurred())) __PYX_ERR(0, 691, __pyx_L1_error)
    __pyx_t_29 = __Pyx_PyObject_AsString(__pyx_v_val); if (unlikely((!__pyx_t_29) && PyErr_Occurred())) __PYX_ERR(0, 691, __pyx_L1_error)

    /* "rasterio/_warp.pyx":690
 *             key = key.upper().encode('utf-8')
 *             val = str(val).upper().encode('utf-8')
 *             c_warp_extras = CSLSetNameValue(             # <<<<<<<<<<<<<<
 *                 c_warp_extras, <const char *>key, <const char *>val)
 * 
 */
    __pyx_v_c_warp_extras = CSLSetNameValue(__pyx_v_c_warp_extras, ((char const *)__pyx_t_28), ((char const *)__pyx_t_29));

    /* "rasterio/_warp.pyx":687
 *         log.debug("Warp_extras: %r", self.warp_extras)
 * 
 *         for key, val in self.warp_extras.items():             # <<<<<<<<<<<<<<
 *             key = key.upper().encode('utf-8')
 *             val = str(val).upper().encode('utf-8')
 */
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "rasterio/_warp.pyx":694
 * 
 *         psWOptions = create_warp_options(
 *             <GDALResampleAlg>c_resampling, self.src_nodata,             # <<<<<<<<<<<<<<
 *             self.dst_nodata, GDALGetRasterCount(hds), <const char **>c_warp_extras)
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_src_nodata); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 694, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "rasterio/_warp.pyx":695
 *         psWOptions = create_warp_options(
 *             <GDALResampleAlg>c_resampling, self.src_nodata,
 *             self.dst_nodata, GDALGetRasterCount(hds), <const char **>c_warp_extras)             # <<<<<<<<<<<<<<
 * 
 *         try:
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dst_nodata); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 695, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "rasterio/_warp.pyx":693
 *                 c_warp_extras, <const char *>key, <const char *>val)
 * 
 *         psWOptions = create_warp_options(             # <<<<<<<<<<<<<<
 *             <GDALResampleAlg>c_resampling, self.src_nodata,
 *             self.dst_nodata, GDALGetRasterCount(hds), <const char **>c_warp_extras)
 */
  __pyx_t_30 = __pyx_f_8rasterio_5_warp_create_warp_options(((GDALResampleAlg)__pyx_v_c_resampling), __pyx_t_3, __pyx_t_1, GDALGetRasterCount(__pyx_v_hds), ((char const **)__pyx_v_c_warp_extras)); if (unlikely(__pyx_t_30 == NULL)) __PYX_ERR(0, 693, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_psWOptions = __pyx_t_30;

  /* "rasterio/_warp.pyx":697
 *             self.dst_nodata, GDALGetRasterCount(hds), <const char **>c_warp_extras)
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             if dst_width and dst_height and dst_transform:
 *                 # set up transform args (otherwise handled in
 */
  /*try:*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_20, &__pyx_t_19, &__pyx_t_18);
      __Pyx_XGOTREF(__pyx_t_20);
      __Pyx_XGOTREF(__pyx_t_19);
      __Pyx_XGOTREF(__pyx_t_18);
      /*try:*/ {

        /* "rasterio/_warp.pyx":698
 * 
 *         try:
 *             if dst_width and dst_height and dst_transform:             # <<<<<<<<<<<<<<
 *                 # set up transform args (otherwise handled in
 *                 # GDALAutoCreateWarpedVRT)
 */
        __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_v_dst_width); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 698, __pyx_L25_error)
        if (__pyx_t_7) {
        } else {
          __pyx_t_6 = __pyx_t_7;
          goto __pyx_L34_bool_binop_done;
        }
        __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_v_dst_height); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 698, __pyx_L25_error)
        if (__pyx_t_7) {
        } else {
          __pyx_t_6 = __pyx_t_7;
          goto __pyx_L34_bool_binop_done;
        }
        __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_v_dst_transform); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 698, __pyx_L25_error)
        __pyx_t_6 = __pyx_t_7;
        __pyx_L34_bool_binop_done:;
        if (__pyx_t_6) {

          /* "rasterio/_warp.pyx":701
 *                 # set up transform args (otherwise handled in
 *                 # GDALAutoCreateWarpedVRT)
 *                 try:             # <<<<<<<<<<<<<<
 *                     hTransformArg = exc_wrap_pointer(
 *                         GDALCreateGenImgProjTransformer(
 */
          {
            __Pyx_PyThreadState_declare
            __Pyx_PyThreadState_assign
            __Pyx_ExceptionSave(&__pyx_t_17, &__pyx_t_16, &__pyx_t_15);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_15);
            /*try:*/ {

              /* "rasterio/_warp.pyx":702
 *                 # GDALAutoCreateWarpedVRT)
 *                 try:
 *                     hTransformArg = exc_wrap_pointer(             # <<<<<<<<<<<<<<
 *                         GDALCreateGenImgProjTransformer(
 *                             hds, NULL, NULL, dst_crs_wkt, True, 1.0, 0))
 */
              __pyx_t_22 = __pyx_f_8rasterio_4_err_exc_wrap_pointer(GDALCreateGenImgProjTransformer(__pyx_v_hds, NULL, NULL, __pyx_v_dst_crs_wkt, 1, 1.0, 0)); if (unlikely(__pyx_t_22 == NULL)) __PYX_ERR(0, 702, __pyx_L37_error)
              __pyx_v_hTransformArg = __pyx_t_22;

              /* "rasterio/_warp.pyx":706
 *                             hds, NULL, NULL, dst_crs_wkt, True, 1.0, 0))
 * 
 *                     GDALSetGenImgProjTransformerDstGeoTransform(             # <<<<<<<<<<<<<<
 *                         hTransformArg, gt)
 * 
 */
              GDALSetGenImgProjTransformerDstGeoTransform(__pyx_v_hTransformArg, __pyx_v_gt);

              /* "rasterio/_warp.pyx":709
 *                         hTransformArg, gt)
 * 
 *                     if c_tolerance > 0.0:             # <<<<<<<<<<<<<<
 *                         hTransformArg = exc_wrap_pointer(
 *                             GDALCreateApproxTransformer(
 */
              __pyx_t_6 = ((__pyx_v_c_tolerance > 0.0) != 0);
              if (__pyx_t_6) {

                /* "rasterio/_warp.pyx":710
 * 
 *                     if c_tolerance > 0.0:
 *                         hTransformArg = exc_wrap_pointer(             # <<<<<<<<<<<<<<
 *                             GDALCreateApproxTransformer(
 *                                 GDALGenImgProjTransform,
 */
                __pyx_t_22 = __pyx_f_8rasterio_4_err_exc_wrap_pointer(GDALCreateApproxTransformer(GDALGenImgProjTransform, __pyx_v_hTransformArg, __pyx_v_c_tolerance)); if (unlikely(__pyx_t_22 == NULL)) __PYX_ERR(0, 710, __pyx_L37_error)
                __pyx_v_hTransformArg = __pyx_t_22;

                /* "rasterio/_warp.pyx":716
 *                                 c_tolerance))
 * 
 *                         psWOptions.pfnTransformer = GDALApproxTransform             # <<<<<<<<<<<<<<
 * 
 *                         GDALApproxTransformerOwnsSubtransformer(
 */
                __pyx_v_psWOptions->pfnTransformer = GDALApproxTransform;

                /* "rasterio/_warp.pyx":718
 *                         psWOptions.pfnTransformer = GDALApproxTransform
 * 
 *                         GDALApproxTransformerOwnsSubtransformer(             # <<<<<<<<<<<<<<
 *                             hTransformArg, 1)
 * 
 */
                GDALApproxTransformerOwnsSubtransformer(__pyx_v_hTransformArg, 1);

                /* "rasterio/_warp.pyx":709
 *                         hTransformArg, gt)
 * 
 *                     if c_tolerance > 0.0:             # <<<<<<<<<<<<<<
 *                         hTransformArg = exc_wrap_pointer(
 *                             GDALCreateApproxTransformer(
 */
              }

              /* "rasterio/_warp.pyx":721
 *                             hTransformArg, 1)
 * 
 *                     log.debug("Created transformer and options.")             # <<<<<<<<<<<<<<
 *                     psWOptions.pTransformerArg = hTransformArg
 *                 except Exception:
 */
              __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_log); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 721, __pyx_L37_error)
              __Pyx_GOTREF(__pyx_t_1);
              __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_debug); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 721, __pyx_L37_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__39, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 721, __pyx_L37_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

              /* "rasterio/_warp.pyx":722
 * 
 *                     log.debug("Created transformer and options.")
 *                     psWOptions.pTransformerArg = hTransformArg             # <<<<<<<<<<<<<<
 *                 except Exception:
 *                     GDALDestroyApproxTransformer(hTransformArg)
 */
              __pyx_v_psWOptions->pTransformerArg = __pyx_v_hTransformArg;

              /* "rasterio/_warp.pyx":701
 *                 # set up transform args (otherwise handled in
 *                 # GDALAutoCreateWarpedVRT)
 *                 try:             # <<<<<<<<<<<<<<
 *                     hTransformArg = exc_wrap_pointer(
 *                         GDALCreateGenImgProjTransformer(
 */
            }
            __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
            __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
            __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
            goto __pyx_L44_try_end;
            __pyx_L37_error:;
            __Pyx_PyThreadState_assign
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

            /* "rasterio/_warp.pyx":723
 *                     log.debug("Created transformer and options.")
 *                     psWOptions.pTransformerArg = hTransformArg
 *                 except Exception:             # <<<<<<<<<<<<<<
 *                     GDALDestroyApproxTransformer(hTransformArg)
 *                     raise
 */
            __pyx_t_12 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
            if (__pyx_t_12) {
              __Pyx_AddTraceback("rasterio._warp.WarpedVRTReaderBase.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
              if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_3, &__pyx_t_2) < 0) __PYX_ERR(0, 723, __pyx_L39_except_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_GOTREF(__pyx_t_2);

              /* "rasterio/_warp.pyx":724
 *                     psWOptions.pTransformerArg = hTransformArg
 *                 except Exception:
 *                     GDALDestroyApproxTransformer(hTransformArg)             # <<<<<<<<<<<<<<
 *                     raise
 * 
 */
              GDALDestroyApproxTransformer(__pyx_v_hTransformArg);

              /* "rasterio/_warp.pyx":725
 *                 except Exception:
 *                     GDALDestroyApproxTransformer(hTransformArg)
 *                     raise             # <<<<<<<<<<<<<<
 * 
 *                 psWOptions.hSrcDS = hds
 */
              __Pyx_GIVEREF(__pyx_t_1);
              __Pyx_GIVEREF(__pyx_t_3);
              __Pyx_XGIVEREF(__pyx_t_2);
              __Pyx_ErrRestoreWithState(__pyx_t_1, __pyx_t_3, __pyx_t_2);
              __pyx_t_1 = 0; __pyx_t_3 = 0; __pyx_t_2 = 0; 
              __PYX_ERR(0, 725, __pyx_L39_except_error)
            }
            goto __pyx_L39_except_error;
            __pyx_L39_except_error:;

            /* "rasterio/_warp.pyx":701
 *                 # set up transform args (otherwise handled in
 *                 # GDALAutoCreateWarpedVRT)
 *                 try:             # <<<<<<<<<<<<<<
 *                     hTransformArg = exc_wrap_pointer(
 *                         GDALCreateGenImgProjTransformer(
 */
            __Pyx_PyThreadState_assign
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_15);
            __Pyx_ExceptionReset(__pyx_t_17, __pyx_t_16, __pyx_t_15);
            goto __pyx_L25_error;
            __pyx_L44_try_end:;
          }

          /* "rasterio/_warp.pyx":727
 *                     raise
 * 
 *                 psWOptions.hSrcDS = hds             # <<<<<<<<<<<<<<
 * 
 *                 with nogil:
 */
          __pyx_v_psWOptions->hSrcDS = __pyx_v_hds;

          /* "rasterio/_warp.pyx":729
 *                 psWOptions.hSrcDS = hds
 * 
 *                 with nogil:             # <<<<<<<<<<<<<<
 *                     hds_warped = GDALCreateWarpedVRT(
 *                         hds, c_width, c_height, gt, psWOptions)
 */
          {
              #ifdef WITH_THREAD
              PyThreadState *_save;
              Py_UNBLOCK_THREADS
              #endif
              /*try:*/ {

                /* "rasterio/_warp.pyx":730
 * 
 *                 with nogil:
 *                     hds_warped = GDALCreateWarpedVRT(             # <<<<<<<<<<<<<<
 *                         hds, c_width, c_height, gt, psWOptions)
 *                 self._hds = exc_wrap_pointer(hds_warped)
 */
                __pyx_v_hds_warped = GDALCreateWarpedVRT(__pyx_v_hds, __pyx_v_c_width, __pyx_v_c_height, __pyx_v_gt, __pyx_v_psWOptions);
              }

              /* "rasterio/_warp.pyx":729
 *                 psWOptions.hSrcDS = hds
 * 
 *                 with nogil:             # <<<<<<<<<<<<<<
 *                     hds_warped = GDALCreateWarpedVRT(
 *                         hds, c_width, c_height, gt, psWOptions)
 */
              /*finally:*/ {
                /*normal exit:*/{
                  #ifdef WITH_THREAD
                  Py_BLOCK_THREADS
                  #endif
                  goto __pyx_L52;
                }
                __pyx_L52:;
              }
          }

          /* "rasterio/_warp.pyx":732
 *                     hds_warped = GDALCreateWarpedVRT(
 *                         hds, c_width, c_height, gt, psWOptions)
 *                 self._hds = exc_wrap_pointer(hds_warped)             # <<<<<<<<<<<<<<
 *             else:
 *                 with nogil:
 */
          __pyx_t_22 = __pyx_f_8rasterio_4_err_exc_wrap_pointer(__pyx_v_hds_warped); if (unlikely(__pyx_t_22 == NULL)) __PYX_ERR(0, 732, __pyx_L25_error)
          __pyx_v_self->__pyx_base.__pyx_base._hds = __pyx_t_22;

          /* "rasterio/_warp.pyx":698
 * 
 *         try:
 *             if dst_width and dst_height and dst_transform:             # <<<<<<<<<<<<<<
 *                 # set up transform args (otherwise handled in
 *                 # GDALAutoCreateWarpedVRT)
 */
          goto __pyx_L33;
        }

        /* "rasterio/_warp.pyx":734
 *                 self._hds = exc_wrap_pointer(hds_warped)
 *             else:
 *                 with nogil:             # <<<<<<<<<<<<<<
 *                     hds_warped = GDALAutoCreateWarpedVRT(
 *                         hds, NULL, dst_crs_wkt, c_resampling,
 */
        /*else*/ {
          {
              #ifdef WITH_THREAD
              PyThreadState *_save;
              Py_UNBLOCK_THREADS
              #endif
              /*try:*/ {

                /* "rasterio/_warp.pyx":735
 *             else:
 *                 with nogil:
 *                     hds_warped = GDALAutoCreateWarpedVRT(             # <<<<<<<<<<<<<<
 *                         hds, NULL, dst_crs_wkt, c_resampling,
 *                         c_tolerance, psWOptions)
 */
                __pyx_v_hds_warped = GDALAutoCreateWarpedVRT(__pyx_v_hds, NULL, __pyx_v_dst_crs_wkt, __pyx_v_c_resampling, __pyx_v_c_tolerance, __pyx_v_psWOptions);
              }

              /* "rasterio/_warp.pyx":734
 *                 self._hds = exc_wrap_pointer(hds_warped)
 *             else:
 *                 with nogil:             # <<<<<<<<<<<<<<
 *                     hds_warped = GDALAutoCreateWarpedVRT(
 *                         hds, NULL, dst_crs_wkt, c_resampling,
 */
              /*finally:*/ {
                /*normal exit:*/{
                  #ifdef WITH_THREAD
                  Py_BLOCK_THREADS
                  #endif
                  goto __pyx_L57;
                }
                __pyx_L57:;
              }
          }

          /* "rasterio/_warp.pyx":738
 *                         hds, NULL, dst_crs_wkt, c_resampling,
 *                         c_tolerance, psWOptions)
 *                 self._hds = exc_wrap_pointer(hds_warped)             # <<<<<<<<<<<<<<
 *         except CPLE_OpenFailedError as err:
 *             raise RasterioIOError(err.errmsg)
 */
          __pyx_t_22 = __pyx_f_8rasterio_4_err_exc_wrap_pointer(__pyx_v_hds_warped); if (unlikely(__pyx_t_22 == NULL)) __PYX_ERR(0, 738, __pyx_L25_error)
          __pyx_v_self->__pyx_base.__pyx_base._hds = __pyx_t_22;
        }
        __pyx_L33:;

        /* "rasterio/_warp.pyx":697
 *             self.dst_nodata, GDALGetRasterCount(hds), <const char **>c_warp_extras)
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             if dst_width and dst_height and dst_transform:
 *                 # set up transform args (otherwise handled in
 */
      }
      __Pyx_XDECREF(__pyx_t_20); __pyx_t_20 = 0;
      __Pyx_XDECREF(__pyx_t_19); __pyx_t_19 = 0;
      __Pyx_XDECREF(__pyx_t_18); __pyx_t_18 = 0;
      goto __pyx_L32_try_end;
      __pyx_L25_error:;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "rasterio/_warp.pyx":739
 *                         c_tolerance, psWOptions)
 *                 self._hds = exc_wrap_pointer(hds_warped)
 *         except CPLE_OpenFailedError as err:             # <<<<<<<<<<<<<<
 *             raise RasterioIOError(err.errmsg)
 *         finally:
 */
      __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_CPLE_OpenFailedError); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 739, __pyx_L27_except_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_12 = __Pyx_PyErr_ExceptionMatches(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (__pyx_t_12) {
        __Pyx_AddTraceback("rasterio._warp.WarpedVRTReaderBase.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_2, &__pyx_t_3, &__pyx_t_1) < 0) __PYX_ERR(0, 739, __pyx_L27_except_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_3);
        __pyx_v_err = __pyx_t_3;

        /* "rasterio/_warp.pyx":740
 *                 self._hds = exc_wrap_pointer(hds_warped)
 *         except CPLE_OpenFailedError as err:
 *             raise RasterioIOError(err.errmsg)             # <<<<<<<<<<<<<<
 *         finally:
 *             CPLFree(dst_crs_wkt)
 */
        __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_RasterioIOError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 740, __pyx_L27_except_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_31 = __Pyx_PyObject_GetAttrStr(__pyx_v_err, __pyx_n_s_errmsg); if (unlikely(!__pyx_t_31)) __PYX_ERR(0, 740, __pyx_L27_except_error)
        __Pyx_GOTREF(__pyx_t_31);
        __pyx_t_32 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_32 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_32)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_32);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        if (!__pyx_t_32) {
          __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_31); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 740, __pyx_L27_except_error)
          __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;
          __Pyx_GOTREF(__pyx_t_5);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[2] = {__pyx_t_32, __pyx_t_31};
            __pyx_t_5 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 740, __pyx_L27_except_error)
            __Pyx_XDECREF(__pyx_t_32); __pyx_t_32 = 0;
            __Pyx_GOTREF(__pyx_t_5);
            __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[2] = {__pyx_t_32, __pyx_t_31};
            __pyx_t_5 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 740, __pyx_L27_except_error)
            __Pyx_XDECREF(__pyx_t_32); __pyx_t_32 = 0;
            __Pyx_GOTREF(__pyx_t_5);
            __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;
          } else
          #endif
          {
            __pyx_t_33 = PyTuple_New(1+1); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 740, __pyx_L27_except_error)
            __Pyx_GOTREF(__pyx_t_33);
            __Pyx_GIVEREF(__pyx_t_32); PyTuple_SET_ITEM(__pyx_t_33, 0, __pyx_t_32); __pyx_t_32 = NULL;
            __Pyx_GIVEREF(__pyx_t_31);
            PyTuple_SET_ITEM(__pyx_t_33, 0+1, __pyx_t_31);
            __pyx_t_31 = 0;
            __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_33, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 740, __pyx_L27_except_error)
            __Pyx_GOTREF(__pyx_t_5);
            __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_Raise(__pyx_t_5, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __PYX_ERR(0, 740, __pyx_L27_except_error)
      }
      goto __pyx_L27_except_error;
      __pyx_L27_except_error:;

      /* "rasterio/_warp.pyx":697
 *             self.dst_nodata, GDALGetRasterCount(hds), <const char **>c_warp_extras)
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             if dst_width and dst_height and dst_transform:
 *                 # set up transform args (otherwise handled in
 */
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_20);
      __Pyx_XGIVEREF(__pyx_t_19);
      __Pyx_XGIVEREF(__pyx_t_18);
      __Pyx_ExceptionReset(__pyx_t_20, __pyx_t_19, __pyx_t_18);
      goto __pyx_L23_error;
      __pyx_L32_try_end:;
    }
  }

  /* "rasterio/_warp.pyx":742
 *             raise RasterioIOError(err.errmsg)
 *         finally:
 *             CPLFree(dst_crs_wkt)             # <<<<<<<<<<<<<<
 *             CSLDestroy(c_warp_extras)
 *             GDALDestroyWarpOptions(psWOptions)
 */
  /*finally:*/ {
    /*normal exit:*/{
      CPLFree(__pyx_v_dst_crs_wkt);

      /* "rasterio/_warp.pyx":743
 *         finally:
 *             CPLFree(dst_crs_wkt)
 *             CSLDestroy(c_warp_extras)             # <<<<<<<<<<<<<<
 *             GDALDestroyWarpOptions(psWOptions)
 * 
 */
      CSLDestroy(__pyx_v_c_warp_extras);

      /* "rasterio/_warp.pyx":744
 *             CPLFree(dst_crs_wkt)
 *             CSLDestroy(c_warp_extras)
 *             GDALDestroyWarpOptions(psWOptions)             # <<<<<<<<<<<<<<
 * 
 * 
 */
      GDALDestroyWarpOptions(__pyx_v_psWOptions);
      goto __pyx_L24;
    }
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __pyx_L23_error:;
      __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_32); __pyx_t_32 = 0;
      __Pyx_XDECREF(__pyx_t_31); __pyx_t_31 = 0;
      __Pyx_XDECREF(__pyx_t_33); __pyx_t_33 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_18, &__pyx_t_19, &__pyx_t_20) < 0)) __Pyx_ErrFetch(&__pyx_t_18, &__pyx_t_19, &__pyx_t_20);
      __Pyx_XGOTREF(__pyx_t_18);
      __Pyx_XGOTREF(__pyx_t_19);
      __Pyx_XGOTREF(__pyx_t_20);
      __Pyx_XGOTREF(__pyx_t_15);
      __Pyx_XGOTREF(__pyx_t_16);
      __Pyx_XGOTREF(__pyx_t_17);
      __pyx_t_12 = __pyx_lineno; __pyx_t_11 = __pyx_clineno; __pyx_t_34 = __pyx_filename;
      {

        /* "rasterio/_warp.pyx":742
 *             raise RasterioIOError(err.errmsg)
 *         finally:
 *             CPLFree(dst_crs_wkt)             # <<<<<<<<<<<<<<
 *             CSLDestroy(c_warp_extras)
 *             GDALDestroyWarpOptions(psWOptions)
 */
        CPLFree(__pyx_v_dst_crs_wkt);

        /* "rasterio/_warp.pyx":743
 *         finally:
 *             CPLFree(dst_crs_wkt)
 *             CSLDestroy(c_warp_extras)             # <<<<<<<<<<<<<<
 *             GDALDestroyWarpOptions(psWOptions)
 * 
 */
        CSLDestroy(__pyx_v_c_warp_extras);

        /* "rasterio/_warp.pyx":744
 *             CPLFree(dst_crs_wkt)
 *             CSLDestroy(c_warp_extras)
 *             GDALDestroyWarpOptions(psWOptions)             # <<<<<<<<<<<<<<
 * 
 * 
 */
        GDALDestroyWarpOptions(__pyx_v_psWOptions);
      }
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_15);
        __Pyx_XGIVEREF(__pyx_t_16);
        __Pyx_XGIVEREF(__pyx_t_17);
        __Pyx_ExceptionReset(__pyx_t_15, __pyx_t_16, __pyx_t_17);
      }
      __Pyx_XGIVEREF(__pyx_t_18);
      __Pyx_XGIVEREF(__pyx_t_19);
      __Pyx_XGIVEREF(__pyx_t_20);
      __Pyx_ErrRestore(__pyx_t_18, __pyx_t_19, __pyx_t_20);
      __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0;
      __pyx_lineno = __pyx_t_12; __pyx_clineno = __pyx_t_11; __pyx_filename = __pyx_t_34;
      goto __pyx_L1_error;
    }
    __pyx_L24:;
  }

  /* "rasterio/_warp.pyx":747
 * 
 * 
 *         driver = GDALGetDatasetDriver(self._hds)             # <<<<<<<<<<<<<<
 *         self.driver = get_driver_name(driver).decode('utf-8')
 * 
 */
  __pyx_v_driver = GDALGetDatasetDriver(__pyx_v_self->__pyx_base.__pyx_base._hds);

  /* "rasterio/_warp.pyx":748
 * 
 *         driver = GDALGetDatasetDriver(self._hds)
 *         self.driver = get_driver_name(driver).decode('utf-8')             # <<<<<<<<<<<<<<
 * 
 *         self._count = GDALGetRasterCount(self._hds)
 */
  __pyx_t_35 = __pyx_f_8rasterio_5_base_get_driver_name(__pyx_v_driver);
  __pyx_t_1 = __Pyx_decode_c_string(__pyx_t_35, 0, strlen(__pyx_t_35), NULL, NULL, PyUnicode_DecodeUTF8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 748, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->__pyx_base.__pyx_base.driver);
  __Pyx_DECREF(__pyx_v_self->__pyx_base.__pyx_base.driver);
  __pyx_v_self->__pyx_base.__pyx_base.driver = __pyx_t_1;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "rasterio/_warp.pyx":750
 *         self.driver = get_driver_name(driver).decode('utf-8')
 * 
 *         self._count = GDALGetRasterCount(self._hds)             # <<<<<<<<<<<<<<
 *         self.width = GDALGetRasterXSize(self._hds)
 *         self.height = GDALGetRasterYSize(self._hds)
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(GDALGetRasterCount(__pyx_v_self->__pyx_base.__pyx_base._hds)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 750, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->__pyx_base.__pyx_base._count);
  __Pyx_DECREF(__pyx_v_self->__pyx_base.__pyx_base._count);
  __pyx_v_self->__pyx_base.__pyx_base._count = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "rasterio/_warp.pyx":751
 * 
 *         self._count = GDALGetRasterCount(self._hds)
 *         self.width = GDALGetRasterXSize(self._hds)             # <<<<<<<<<<<<<<
 *         self.height = GDALGetRasterYSize(self._hds)
 *         self.shape = (self.height, self.width)
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(GDALGetRasterXSize(__pyx_v_self->__pyx_base.__pyx_base._hds)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 751, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->__pyx_base.__pyx_base.width);
  __Pyx_DECREF(__pyx_v_self->__pyx_base.__pyx_base.width);
  __pyx_v_self->__pyx_base.__pyx_base.width = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "rasterio/_warp.pyx":752
 *         self._count = GDALGetRasterCount(self._hds)
 *         self.width = GDALGetRasterXSize(self._hds)
 *         self.height = GDALGetRasterYSize(self._hds)             # <<<<<<<<<<<<<<
 *         self.shape = (self.height, self.width)
 * 
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(GDALGetRasterYSize(__pyx_v_self->__pyx_base.__pyx_base._hds)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 752, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->__pyx_base.__pyx_base.height);
  __Pyx_DECREF(__pyx_v_self->__pyx_base.__pyx_base.height);
  __pyx_v_self->__pyx_base.__pyx_base.height = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "rasterio/_warp.pyx":753
 *         self.width = GDALGetRasterXSize(self._hds)
 *         self.height = GDALGetRasterYSize(self._hds)
 *         self.shape = (self.height, self.width)             # <<<<<<<<<<<<<<
 * 
 *         self._transform = self.read_transform()
 */
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 753, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.__pyx_base.height);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.__pyx_base.height);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->__pyx_base.__pyx_base.height);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.__pyx_base.width);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.__pyx_base.width);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_self->__pyx_base.__pyx_base.width);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->__pyx_base.__pyx_base.shape);
  __Pyx_DECREF(__pyx_v_self->__pyx_base.__pyx_base.shape);
  __pyx_v_self->__pyx_base.__pyx_base.shape = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "rasterio/_warp.pyx":755
 *         self.shape = (self.height, self.width)
 * 
 *         self._transform = self.read_transform()             # <<<<<<<<<<<<<<
 *         self._crs = self.read_crs()
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_read_transform); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 755, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (__pyx_t_2) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 755, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else {
    __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 755, __pyx_L1_error)
  }
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->__pyx_base.__pyx_base._transform);
  __Pyx_DECREF(__pyx_v_self->__pyx_base.__pyx_base._transform);
  __pyx_v_self->__pyx_base.__pyx_base._transform = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "rasterio/_warp.pyx":756
 * 
 *         self._transform = self.read_transform()
 *         self._crs = self.read_crs()             # <<<<<<<<<<<<<<
 * 
 *         # touch self.meta
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_read_crs); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 756, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (__pyx_t_2) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 756, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else {
    __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 756, __pyx_L1_error)
  }
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->__pyx_base.__pyx_base._crs);
  __Pyx_DECREF(__pyx_v_self->__pyx_base.__pyx_base._crs);
  __pyx_v_self->__pyx_base.__pyx_base._crs = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "rasterio/_warp.pyx":759
 * 
 *         # touch self.meta
 *         _ = self.meta             # <<<<<<<<<<<<<<
 * 
 *         self._closed = False
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_meta); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 759, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__ = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "rasterio/_warp.pyx":761
 *         _ = self.meta
 * 
 *         self._closed = False             # <<<<<<<<<<<<<<
 * 
 *     def start(self):
 */
  __Pyx_INCREF(Py_False);
  __Pyx_GIVEREF(Py_False);
  __Pyx_GOTREF(__pyx_v_self->__pyx_base.__pyx_base._closed);
  __Pyx_DECREF(__pyx_v_self->__pyx_base.__pyx_base._closed);
  __pyx_v_self->__pyx_base.__pyx_base._closed = Py_False;

  /* "rasterio/_warp.pyx":632
 * cdef class WarpedVRTReaderBase(DatasetReaderBase):
 * 
 *     def __init__(self, src_dataset, dst_crs=None, resampling=Resampling.nearest,             # <<<<<<<<<<<<<<
 *                  tolerance=0.125, src_nodata=None, dst_nodata=None,
 *                  dst_width=None, dst_height=None, dst_transform=None,
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_31);
  __Pyx_XDECREF(__pyx_t_32);
  __Pyx_XDECREF(__pyx_t_33);
  __Pyx_AddTraceback("rasterio._warp.WarpedVRTReaderBase.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_t);
  __Pyx_XDECREF(__pyx_v_key);
  __Pyx_XDECREF(__pyx_v_val);
  __Pyx_XDECREF(__pyx_v_err);
  __Pyx_XDECREF(__pyx_v__);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "rasterio/_warp.pyx":763
 *         self._closed = False
 * 
 *     def start(self):             # <<<<<<<<<<<<<<
 *         """Starts the VRT's life cycle."""
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_8rasterio_5_warp_19WarpedVRTReaderBase_3start(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_8rasterio_5_warp_19WarpedVRTReaderBase_2start[] = "Starts the VRT's life cycle.";
static PyObject *__pyx_pw_8rasterio_5_warp_19WarpedVRTReaderBase_3start(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("start (wrapper)", 0);
  __pyx_r = __pyx_pf_8rasterio_5_warp_19WarpedVRTReaderBase_2start(((struct __pyx_obj_8rasterio_5_warp_WarpedVRTReaderBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8rasterio_5_warp_19WarpedVRTReaderBase_2start(struct __pyx_obj_8rasterio_5_warp_WarpedVRTReaderBase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("start", 0);

  /* "rasterio/_warp.pyx":766
 *         """Starts the VRT's life cycle."""
 * 
 *         log.debug("Dataset %r is started.", self)             # <<<<<<<<<<<<<<
 * 
 *     def stop(self):
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_log); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 766, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_debug); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 766, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_kp_s_Dataset_r_is_started, ((PyObject *)__pyx_v_self)};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 766, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_kp_s_Dataset_r_is_started, ((PyObject *)__pyx_v_self)};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 766, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 766, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_kp_s_Dataset_r_is_started);
    __Pyx_GIVEREF(__pyx_kp_s_Dataset_r_is_started);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_kp_s_Dataset_r_is_started);
    __Pyx_INCREF(((PyObject *)__pyx_v_self));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, ((PyObject *)__pyx_v_self));
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 766, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "rasterio/_warp.pyx":763
 *         self._closed = False
 * 
 *     def start(self):             # <<<<<<<<<<<<<<
 *         """Starts the VRT's life cycle."""
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("rasterio._warp.WarpedVRTReaderBase.start", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "rasterio/_warp.pyx":768
 *         log.debug("Dataset %r is started.", self)
 * 
 *     def stop(self):             # <<<<<<<<<<<<<<
 *         """Ends the VRT's life cycle"""
 *         if self._hds != NULL:
 */

/* Python wrapper */
static PyObject *__pyx_pw_8rasterio_5_warp_19WarpedVRTReaderBase_5stop(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_8rasterio_5_warp_19WarpedVRTReaderBase_4stop[] = "Ends the VRT's life cycle";
static PyObject *__pyx_pw_8rasterio_5_warp_19WarpedVRTReaderBase_5stop(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("stop (wrapper)", 0);
  __pyx_r = __pyx_pf_8rasterio_5_warp_19WarpedVRTReaderBase_4stop(((struct __pyx_obj_8rasterio_5_warp_WarpedVRTReaderBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8rasterio_5_warp_19WarpedVRTReaderBase_4stop(struct __pyx_obj_8rasterio_5_warp_WarpedVRTReaderBase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("stop", 0);

  /* "rasterio/_warp.pyx":770
 *     def stop(self):
 *         """Ends the VRT's life cycle"""
 *         if self._hds != NULL:             # <<<<<<<<<<<<<<
 *             GDALClose(self._hds)
 *         self._hds = NULL
 */
  __pyx_t_1 = ((__pyx_v_self->__pyx_base.__pyx_base._hds != NULL) != 0);
  if (__pyx_t_1) {

    /* "rasterio/_warp.pyx":771
 *         """Ends the VRT's life cycle"""
 *         if self._hds != NULL:
 *             GDALClose(self._hds)             # <<<<<<<<<<<<<<
 *         self._hds = NULL
 */
    GDALClose(__pyx_v_self->__pyx_base.__pyx_base._hds);

    /* "rasterio/_warp.pyx":770
 *     def stop(self):
 *         """Ends the VRT's life cycle"""
 *         if self._hds != NULL:             # <<<<<<<<<<<<<<
 *             GDALClose(self._hds)
 *         self._hds = NULL
 */
  }

  /* "rasterio/_warp.pyx":772
 *         if self._hds != NULL:
 *             GDALClose(self._hds)
 *         self._hds = NULL             # <<<<<<<<<<<<<<
 */
  __pyx_v_self->__pyx_base.__pyx_base._hds = NULL;

  /* "rasterio/_warp.pyx":768
 *         log.debug("Dataset %r is started.", self)
 * 
 *     def stop(self):             # <<<<<<<<<<<<<<
 *         """Ends the VRT's life cycle"""
 *         if self._hds != NULL:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":197
 *         # experimental exception made for __getbuffer__ and __releasebuffer__
 *         # -- the details of this may change.
 *         def __getbuffer__(ndarray self, Py_buffer* info, int flags):             # <<<<<<<<<<<<<<
 *             # This implementation of getbuffer is geared towards Cython
 *             # requirements, and does not yet fullfill the PEP.
 */

/* Python wrapper */
static CYTHON_UNUSED int __pyx_pw_5numpy_7ndarray_1__getbuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /*proto*/
static CYTHON_UNUSED int __pyx_pw_5numpy_7ndarray_1__getbuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getbuffer__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5numpy_7ndarray___getbuffer__(((PyArrayObject *)__pyx_v_self), ((Py_buffer *)__pyx_v_info), ((int)__pyx_v_flags));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5numpy_7ndarray___getbuffer__(PyArrayObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  int __pyx_v_copy_shape;
  int __pyx_v_i;
  int __pyx_v_ndim;
  int __pyx_v_endian_detector;
  int __pyx_v_little_endian;
  int __pyx_v_t;
  char *__pyx_v_f;
  PyArray_Descr *__pyx_v_descr = 0;
  int __pyx_v_offset;
  int __pyx_v_hasfields;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  char *__pyx_t_7;
  __Pyx_RefNannySetupContext("__getbuffer__", 0);
  if (__pyx_v_info != NULL) {
    __pyx_v_info->obj = Py_None; __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(__pyx_v_info->obj);
  }

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":203
 *             # of flags
 * 
 *             if info == NULL: return             # <<<<<<<<<<<<<<
 * 
 *             cdef int copy_shape, i, ndim
 */
  __pyx_t_1 = ((__pyx_v_info == NULL) != 0);
  if (__pyx_t_1) {
    __pyx_r = 0;
    goto __pyx_L0;
  }

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":206
 * 
 *             cdef int copy_shape, i, ndim
 *             cdef int endian_detector = 1             # <<<<<<<<<<<<<<
 *             cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)
 * 
 */
  __pyx_v_endian_detector = 1;

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":207
 *             cdef int copy_shape, i, ndim
 *             cdef int endian_detector = 1
 *             cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)             # <<<<<<<<<<<<<<
 * 
 *             ndim = PyArray_NDIM(self)
 */
  __pyx_v_little_endian = ((((char *)(&__pyx_v_endian_detector))[0]) != 0);

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":209
 *             cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)
 * 
 *             ndim = PyArray_NDIM(self)             # <<<<<<<<<<<<<<
 * 
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 */
  __pyx_v_ndim = PyArray_NDIM(__pyx_v_self);

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":211
 *             ndim = PyArray_NDIM(self)
 * 
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):             # <<<<<<<<<<<<<<
 *                 copy_shape = 1
 *             else:
 */
  __pyx_t_1 = (((sizeof(npy_intp)) != (sizeof(Py_ssize_t))) != 0);
  if (__pyx_t_1) {

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":212
 * 
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 *                 copy_shape = 1             # <<<<<<<<<<<<<<
 *             else:
 *                 copy_shape = 0
 */
    __pyx_v_copy_shape = 1;

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":211
 *             ndim = PyArray_NDIM(self)
 * 
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):             # <<<<<<<<<<<<<<
 *                 copy_shape = 1
 *             else:
 */
    goto __pyx_L4;
  }

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":214
 *                 copy_shape = 1
 *             else:
 *                 copy_shape = 0             # <<<<<<<<<<<<<<
 * 
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)
 */
  /*else*/ {
    __pyx_v_copy_shape = 0;
  }
  __pyx_L4:;

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":216
 *                 copy_shape = 0
 * 
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_C_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not C contiguous")
 */
  __pyx_t_2 = (((__pyx_v_flags & PyBUF_C_CONTIGUOUS) == PyBUF_C_CONTIGUOUS) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L6_bool_binop_done;
  }

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":217
 * 
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_C_CONTIGUOUS)):             # <<<<<<<<<<<<<<
 *                 raise ValueError(u"ndarray is not C contiguous")
 * 
 */
  __pyx_t_2 = ((!(PyArray_CHKFLAGS(__pyx_v_self, NPY_C_CONTIGUOUS) != 0)) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L6_bool_binop_done:;

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":216
 *                 copy_shape = 0
 * 
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_C_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not C contiguous")
 */
  if (__pyx_t_1) {

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":218
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_C_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not C contiguous")             # <<<<<<<<<<<<<<
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__40, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 218, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 218, __pyx_L1_error)

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":216
 *                 copy_shape = 0
 * 
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_C_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not C contiguous")
 */
  }

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":220
 *                 raise ValueError(u"ndarray is not C contiguous")
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_F_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not Fortran contiguous")
 */
  __pyx_t_2 = (((__pyx_v_flags & PyBUF_F_CONTIGUOUS) == PyBUF_F_CONTIGUOUS) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L9_bool_binop_done;
  }

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":221
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_F_CONTIGUOUS)):             # <<<<<<<<<<<<<<
 *                 raise ValueError(u"ndarray is not Fortran contiguous")
 * 
 */
  __pyx_t_2 = ((!(PyArray_CHKFLAGS(__pyx_v_self, NPY_F_CONTIGUOUS) != 0)) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L9_bool_binop_done:;

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":220
 *                 raise ValueError(u"ndarray is not C contiguous")
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_F_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not Fortran contiguous")
 */
  if (__pyx_t_1) {

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":222
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_F_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not Fortran contiguous")             # <<<<<<<<<<<<<<
 * 
 *             info.buf = PyArray_DATA(self)
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__41, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 222, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 222, __pyx_L1_error)

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":220
 *                 raise ValueError(u"ndarray is not C contiguous")
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_F_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not Fortran contiguous")
 */
  }

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":224
 *                 raise ValueError(u"ndarray is not Fortran contiguous")
 * 
 *             info.buf = PyArray_DATA(self)             # <<<<<<<<<<<<<<
 *             info.ndim = ndim
 *             if copy_shape:
 */
  __pyx_v_info->buf = PyArray_DATA(__pyx_v_self);

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":225
 * 
 *             info.buf = PyArray_DATA(self)
 *             info.ndim = ndim             # <<<<<<<<<<<<<<
 *             if copy_shape:
 *                 # Allocate new buffer for strides and shape info.
 */
  __pyx_v_info->ndim = __pyx_v_ndim;

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":226
 *             info.buf = PyArray_DATA(self)
 *             info.ndim = ndim
 *             if copy_shape:             # <<<<<<<<<<<<<<
 *                 # Allocate new buffer for strides and shape info.
 *                 # This is allocated as one block, strides first.
 */
  __pyx_t_1 = (__pyx_v_copy_shape != 0);
  if (__pyx_t_1) {

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":229
 *                 # Allocate new buffer for strides and shape info.
 *                 # This is allocated as one block, strides first.
 *                 info.strides = <Py_ssize_t*>stdlib.malloc(sizeof(Py_ssize_t) * <size_t>ndim * 2)             # <<<<<<<<<<<<<<
 *                 info.shape = info.strides + ndim
 *                 for i in range(ndim):
 */
    __pyx_v_info->strides = ((Py_ssize_t *)malloc((((sizeof(Py_ssize_t)) * ((size_t)__pyx_v_ndim)) * 2)));

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":230
 *                 # This is allocated as one block, strides first.
 *                 info.strides = <Py_ssize_t*>stdlib.malloc(sizeof(Py_ssize_t) * <size_t>ndim * 2)
 *                 info.shape = info.strides + ndim             # <<<<<<<<<<<<<<
 *                 for i in range(ndim):
 *                     info.strides[i] = PyArray_STRIDES(self)[i]
 */
    __pyx_v_info->shape = (__pyx_v_info->strides + __pyx_v_ndim);

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":231
 *                 info.strides = <Py_ssize_t*>stdlib.malloc(sizeof(Py_ssize_t) * <size_t>ndim * 2)
 *                 info.shape = info.strides + ndim
 *                 for i in range(ndim):             # <<<<<<<<<<<<<<
 *                     info.strides[i] = PyArray_STRIDES(self)[i]
 *                     info.shape[i] = PyArray_DIMS(self)[i]
 */
    __pyx_t_4 = __pyx_v_ndim;
    for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
      __pyx_v_i = __pyx_t_5;

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":232
 *                 info.shape = info.strides + ndim
 *                 for i in range(ndim):
 *                     info.strides[i] = PyArray_STRIDES(self)[i]             # <<<<<<<<<<<<<<
 *                     info.shape[i] = PyArray_DIMS(self)[i]
 *             else:
 */
      (__pyx_v_info->strides[__pyx_v_i]) = (PyArray_STRIDES(__pyx_v_self)[__pyx_v_i]);

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":233
 *                 for i in range(ndim):
 *                     info.strides[i] = PyArray_STRIDES(self)[i]
 *                     info.shape[i] = PyArray_DIMS(self)[i]             # <<<<<<<<<<<<<<
 *             else:
 *                 info.strides = <Py_ssize_t*>PyArray_STRIDES(self)
 */
      (__pyx_v_info->shape[__pyx_v_i]) = (PyArray_DIMS(__pyx_v_self)[__pyx_v_i]);
    }

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":226
 *             info.buf = PyArray_DATA(self)
 *             info.ndim = ndim
 *             if copy_shape:             # <<<<<<<<<<<<<<
 *                 # Allocate new buffer for strides and shape info.
 *                 # This is allocated as one block, strides first.
 */
    goto __pyx_L11;
  }

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":235
 *                     info.shape[i] = PyArray_DIMS(self)[i]
 *             else:
 *                 info.strides = <Py_ssize_t*>PyArray_STRIDES(self)             # <<<<<<<<<<<<<<
 *                 info.shape = <Py_ssize_t*>PyArray_DIMS(self)
 *             info.suboffsets = NULL
 */
  /*else*/ {
    __pyx_v_info->strides = ((Py_ssize_t *)PyArray_STRIDES(__pyx_v_self));

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":236
 *             else:
 *                 info.strides = <Py_ssize_t*>PyArray_STRIDES(self)
 *                 info.shape = <Py_ssize_t*>PyArray_DIMS(self)             # <<<<<<<<<<<<<<
 *             info.suboffsets = NULL
 *             info.itemsize = PyArray_ITEMSIZE(self)
 */
    __pyx_v_info->shape = ((Py_ssize_t *)PyArray_DIMS(__pyx_v_self));
  }
  __pyx_L11:;

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":237
 *                 info.strides = <Py_ssize_t*>PyArray_STRIDES(self)
 *                 info.shape = <Py_ssize_t*>PyArray_DIMS(self)
 *             info.suboffsets = NULL             # <<<<<<<<<<<<<<
 *             info.itemsize = PyArray_ITEMSIZE(self)
 *             info.readonly = not PyArray_ISWRITEABLE(self)
 */
  __pyx_v_info->suboffsets = NULL;

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":238
 *                 info.shape = <Py_ssize_t*>PyArray_DIMS(self)
 *             info.suboffsets = NULL
 *             info.itemsize = PyArray_ITEMSIZE(self)             # <<<<<<<<<<<<<<
 *             info.readonly = not PyArray_ISWRITEABLE(self)
 * 
 */
  __pyx_v_info->itemsize = PyArray_ITEMSIZE(__pyx_v_self);

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":239
 *             info.suboffsets = NULL
 *             info.itemsize = PyArray_ITEMSIZE(self)
 *             info.readonly = not PyArray_ISWRITEABLE(self)             # <<<<<<<<<<<<<<
 * 
 *             cdef int t
 */
  __pyx_v_info->readonly = (!(PyArray_ISWRITEABLE(__pyx_v_self) != 0));

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":242
 * 
 *             cdef int t
 *             cdef char* f = NULL             # <<<<<<<<<<<<<<
 *             cdef dtype descr = self.descr
 *             cdef int offset
 */
  __pyx_v_f = NULL;

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":243
 *             cdef int t
 *             cdef char* f = NULL
 *             cdef dtype descr = self.descr             # <<<<<<<<<<<<<<
 *             cdef int offset
 * 
 */
  __pyx_t_3 = ((PyObject *)__pyx_v_self->descr);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_v_descr = ((PyArray_Descr *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":246
 *             cdef int offset
 * 
 *             cdef bint hasfields = PyDataType_HASFIELDS(descr)             # <<<<<<<<<<<<<<
 * 
 *             if not hasfields and not copy_shape:
 */
  __pyx_v_hasfields = PyDataType_HASFIELDS(__pyx_v_descr);

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":248
 *             cdef bint hasfields = PyDataType_HASFIELDS(descr)
 * 
 *             if not hasfields and not copy_shape:             # <<<<<<<<<<<<<<
 *                 # do not call releasebuffer
 *                 info.obj = None
 */
  __pyx_t_2 = ((!(__pyx_v_hasfields != 0)) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L15_bool_binop_done;
  }
  __pyx_t_2 = ((!(__pyx_v_copy_shape != 0)) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L15_bool_binop_done:;
  if (__pyx_t_1) {

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":250
 *             if not hasfields and not copy_shape:
 *                 # do not call releasebuffer
 *                 info.obj = None             # <<<<<<<<<<<<<<
 *             else:
 *                 # need to call releasebuffer
 */
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj);
    __pyx_v_info->obj = Py_None;

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":248
 *             cdef bint hasfields = PyDataType_HASFIELDS(descr)
 * 
 *             if not hasfields and not copy_shape:             # <<<<<<<<<<<<<<
 *                 # do not call releasebuffer
 *                 info.obj = None
 */
    goto __pyx_L14;
  }

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":253
 *             else:
 *                 # need to call releasebuffer
 *                 info.obj = self             # <<<<<<<<<<<<<<
 * 
 *             if not hasfields:
 */
  /*else*/ {
    __Pyx_INCREF(((PyObject *)__pyx_v_self));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj);
    __pyx_v_info->obj = ((PyObject *)__pyx_v_self);
  }
  __pyx_L14:;

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":255
 *                 info.obj = self
 * 
 *             if not hasfields:             # <<<<<<<<<<<<<<
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or
 */
  __pyx_t_1 = ((!(__pyx_v_hasfields != 0)) != 0);
  if (__pyx_t_1) {

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":256
 * 
 *             if not hasfields:
 *                 t = descr.type_num             # <<<<<<<<<<<<<<
 *                 if ((descr.byteorder == c'>' and little_endian) or
 *                     (descr.byteorder == c'<' and not little_endian)):
 */
    __pyx_t_4 = __pyx_v_descr->type_num;
    __pyx_v_t = __pyx_t_4;

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":257
 *             if not hasfields:
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")
 */
    __pyx_t_2 = ((__pyx_v_descr->byteorder == '>') != 0);
    if (!__pyx_t_2) {
      goto __pyx_L20_next_or;
    } else {
    }
    __pyx_t_2 = (__pyx_v_little_endian != 0);
    if (!__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L19_bool_binop_done;
    }
    __pyx_L20_next_or:;

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":258
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or
 *                     (descr.byteorder == c'<' and not little_endian)):             # <<<<<<<<<<<<<<
 *                     raise ValueError(u"Non-native byte order not supported")
 *                 if   t == NPY_BYTE:        f = "b"
 */
    __pyx_t_2 = ((__pyx_v_descr->byteorder == '<') != 0);
    if (__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L19_bool_binop_done;
    }
    __pyx_t_2 = ((!(__pyx_v_little_endian != 0)) != 0);
    __pyx_t_1 = __pyx_t_2;
    __pyx_L19_bool_binop_done:;

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":257
 *             if not hasfields:
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")
 */
    if (__pyx_t_1) {

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":259
 *                 if ((descr.byteorder == c'>' and little_endian) or
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")             # <<<<<<<<<<<<<<
 *                 if   t == NPY_BYTE:        f = "b"
 *                 elif t == NPY_UBYTE:       f = "B"
 */
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__42, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 259, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(1, 259, __pyx_L1_error)

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":257
 *             if not hasfields:
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")
 */
    }

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":260
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")
 *                 if   t == NPY_BYTE:        f = "b"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_UBYTE:       f = "B"
 *                 elif t == NPY_SHORT:       f = "h"
 */
    switch (__pyx_v_t) {
      case NPY_BYTE:
      __pyx_v_f = ((char *)"b");
      break;

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":261
 *                     raise ValueError(u"Non-native byte order not supported")
 *                 if   t == NPY_BYTE:        f = "b"
 *                 elif t == NPY_UBYTE:       f = "B"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_SHORT:       f = "h"
 *                 elif t == NPY_USHORT:      f = "H"
 */
      case NPY_UBYTE:
      __pyx_v_f = ((char *)"B");
      break;

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":262
 *                 if   t == NPY_BYTE:        f = "b"
 *                 elif t == NPY_UBYTE:       f = "B"
 *                 elif t == NPY_SHORT:       f = "h"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_USHORT:      f = "H"
 *                 elif t == NPY_INT:         f = "i"
 */
      case NPY_SHORT:
      __pyx_v_f = ((char *)"h");
      break;

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":263
 *                 elif t == NPY_UBYTE:       f = "B"
 *                 elif t == NPY_SHORT:       f = "h"
 *                 elif t == NPY_USHORT:      f = "H"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_INT:         f = "i"
 *                 elif t == NPY_UINT:        f = "I"
 */
      case NPY_USHORT:
      __pyx_v_f = ((char *)"H");
      break;

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":264
 *                 elif t == NPY_SHORT:       f = "h"
 *                 elif t == NPY_USHORT:      f = "H"
 *                 elif t == NPY_INT:         f = "i"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_UINT:        f = "I"
 *                 elif t == NPY_LONG:        f = "l"
 */
      case NPY_INT:
      __pyx_v_f = ((char *)"i");
      break;

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":265
 *                 elif t == NPY_USHORT:      f = "H"
 *                 elif t == NPY_INT:         f = "i"
 *                 elif t == NPY_UINT:        f = "I"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_LONG:        f = "l"
 *                 elif t == NPY_ULONG:       f = "L"
 */
      case NPY_UINT:
      __pyx_v_f = ((char *)"I");
      break;

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":266
 *                 elif t == NPY_INT:         f = "i"
 *                 elif t == NPY_UINT:        f = "I"
 *                 elif t == NPY_LONG:        f = "l"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_ULONG:       f = "L"
 *                 elif t == NPY_LONGLONG:    f = "q"
 */
      case NPY_LONG:
      __pyx_v_f = ((char *)"l");
      break;

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":267
 *                 elif t == NPY_UINT:        f = "I"
 *                 elif t == NPY_LONG:        f = "l"
 *                 elif t == NPY_ULONG:       f = "L"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_LONGLONG:    f = "q"
 *                 elif t == NPY_ULONGLONG:   f = "Q"
 */
      case NPY_ULONG:
      __pyx_v_f = ((char *)"L");
      break;

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":268
 *                 elif t == NPY_LONG:        f = "l"
 *                 elif t == NPY_ULONG:       f = "L"
 *                 elif t == NPY_LONGLONG:    f = "q"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_ULONGLONG:   f = "Q"
 *                 elif t == NPY_FLOAT:       f = "f"
 */
      case NPY_LONGLONG:
      __pyx_v_f = ((char *)"q");
      break;

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":269
 *                 elif t == NPY_ULONG:       f = "L"
 *                 elif t == NPY_LONGLONG:    f = "q"
 *                 elif t == NPY_ULONGLONG:   f = "Q"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_FLOAT:       f = "f"
 *                 elif t == NPY_DOUBLE:      f = "d"
 */
      case NPY_ULONGLONG:
      __pyx_v_f = ((char *)"Q");
      break;

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":270
 *                 elif t == NPY_LONGLONG:    f = "q"
 *                 elif t == NPY_ULONGLONG:   f = "Q"
 *                 elif t == NPY_FLOAT:       f = "f"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_DOUBLE:      f = "d"
 *                 elif t == NPY_LONGDOUBLE:  f = "g"
 */
      case NPY_FLOAT:
      __pyx_v_f = ((char *)"f");
      break;

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":271
 *                 elif t == NPY_ULONGLONG:   f = "Q"
 *                 elif t == NPY_FLOAT:       f = "f"
 *                 elif t == NPY_DOUBLE:      f = "d"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_LONGDOUBLE:  f = "g"
 *                 elif t == NPY_CFLOAT:      f = "Zf"
 */
      case NPY_DOUBLE:
      __pyx_v_f = ((char *)"d");
      break;

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":272
 *                 elif t == NPY_FLOAT:       f = "f"
 *                 elif t == NPY_DOUBLE:      f = "d"
 *                 elif t == NPY_LONGDOUBLE:  f = "g"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_CFLOAT:      f = "Zf"
 *                 elif t == NPY_CDOUBLE:     f = "Zd"
 */
      case NPY_LONGDOUBLE:
      __pyx_v_f = ((char *)"g");
      break;

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":273
 *                 elif t == NPY_DOUBLE:      f = "d"
 *                 elif t == NPY_LONGDOUBLE:  f = "g"
 *                 elif t == NPY_CFLOAT:      f = "Zf"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_CDOUBLE:     f = "Zd"
 *                 elif t == NPY_CLONGDOUBLE: f = "Zg"
 */
      case NPY_CFLOAT:
      __pyx_v_f = ((char *)"Zf");
      break;

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":274
 *                 elif t == NPY_LONGDOUBLE:  f = "g"
 *                 elif t == NPY_CFLOAT:      f = "Zf"
 *                 elif t == NPY_CDOUBLE:     f = "Zd"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_CLONGDOUBLE: f = "Zg"
 *                 elif t == NPY_OBJECT:      f = "O"
 */
      case NPY_CDOUBLE:
      __pyx_v_f = ((char *)"Zd");
      break;

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":275
 *                 elif t == NPY_CFLOAT:      f = "Zf"
 *                 elif t == NPY_CDOUBLE:     f = "Zd"
 *                 elif t == NPY_CLONGDOUBLE: f = "Zg"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_OBJECT:      f = "O"
 *                 else:
 */
      case NPY_CLONGDOUBLE:
      __pyx_v_f = ((char *)"Zg");
      break;

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":276
 *                 elif t == NPY_CDOUBLE:     f = "Zd"
 *                 elif t == NPY_CLONGDOUBLE: f = "Zg"
 *                 elif t == NPY_OBJECT:      f = "O"             # <<<<<<<<<<<<<<
 *                 else:
 *                     raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 */
      case NPY_OBJECT:
      __pyx_v_f = ((char *)"O");
      break;
      default:

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":278
 *                 elif t == NPY_OBJECT:      f = "O"
 *                 else:
 *                     raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)             # <<<<<<<<<<<<<<
 *                 info.format = f
 *                 return
 */
      __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_t); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 278, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_6 = PyUnicode_Format(__pyx_kp_u_unknown_dtype_code_in_numpy_pxd, __pyx_t_3); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 278, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 278, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_6);
      __pyx_t_6 = 0;
      __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_t_3, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 278, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_Raise(__pyx_t_6, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __PYX_ERR(1, 278, __pyx_L1_error)
      break;
    }

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":279
 *                 else:
 *                     raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 *                 info.format = f             # <<<<<<<<<<<<<<
 *                 return
 *             else:
 */
    __pyx_v_info->format = __pyx_v_f;

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":280
 *                     raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 *                 info.format = f
 *                 return             # <<<<<<<<<<<<<<
 *             else:
 *                 info.format = <char*>stdlib.malloc(_buffer_format_string_len)
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":255
 *                 info.obj = self
 * 
 *             if not hasfields:             # <<<<<<<<<<<<<<
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or
 */
  }

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":282
 *                 return
 *             else:
 *                 info.format = <char*>stdlib.malloc(_buffer_format_string_len)             # <<<<<<<<<<<<<<
 *                 info.format[0] = c'^' # Native data types, manual alignment
 *                 offset = 0
 */
  /*else*/ {
    __pyx_v_info->format = ((char *)malloc(0xFF));

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":283
 *             else:
 *                 info.format = <char*>stdlib.malloc(_buffer_format_string_len)
 *                 info.format[0] = c'^' # Native data types, manual alignment             # <<<<<<<<<<<<<<
 *                 offset = 0
 *                 f = _util_dtypestring(descr, info.format + 1,
 */
    (__pyx_v_info->format[0]) = '^';

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":284
 *                 info.format = <char*>stdlib.malloc(_buffer_format_string_len)
 *                 info.format[0] = c'^' # Native data types, manual alignment
 *                 offset = 0             # <<<<<<<<<<<<<<
 *                 f = _util_dtypestring(descr, info.format + 1,
 *                                       info.format + _buffer_format_string_len,
 */
    __pyx_v_offset = 0;

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":285
 *                 info.format[0] = c'^' # Native data types, manual alignment
 *                 offset = 0
 *                 f = _util_dtypestring(descr, info.format + 1,             # <<<<<<<<<<<<<<
 *                                       info.format + _buffer_format_string_len,
 *                                       &offset)
 */
    __pyx_t_7 = __pyx_f_5numpy__util_dtypestring(__pyx_v_descr, (__pyx_v_info->format + 1), (__pyx_v_info->format + 0xFF), (&__pyx_v_offset)); if (unlikely(__pyx_t_7 == NULL)) __PYX_ERR(1, 285, __pyx_L1_error)
    __pyx_v_f = __pyx_t_7;

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":288
 *                                       info.format + _buffer_format_string_len,
 *                                       &offset)
 *                 f[0] = c'\0' # Terminate format string             # <<<<<<<<<<<<<<
 * 
 *         def __releasebuffer__(ndarray self, Py_buffer* info):
 */
    (__pyx_v_f[0]) = '\x00';
  }

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":197
 *         # experimental exception made for __getbuffer__ and __releasebuffer__
 *         # -- the details of this may change.
 *         def __getbuffer__(ndarray self, Py_buffer* info, int flags):             # <<<<<<<<<<<<<<
 *             # This implementation of getbuffer is geared towards Cython
 *             # requirements, and does not yet fullfill the PEP.
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("numpy.ndarray.__getbuffer__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  if (__pyx_v_info != NULL && __pyx_v_info->obj != NULL) {
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = NULL;
  }
  goto __pyx_L2;
  __pyx_L0:;
  if (__pyx_v_info != NULL && __pyx_v_info->obj == Py_None) {
    __Pyx_GOTREF(Py_None);
    __Pyx_DECREF(Py_None); __pyx_v_info->obj = NULL;
  }
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_descr);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":290
 *                 f[0] = c'\0' # Terminate format string
 * 
 *         def __releasebuffer__(ndarray self, Py_buffer* info):             # <<<<<<<<<<<<<<
 *             if PyArray_HASFIELDS(self):
 *                 stdlib.free(info.format)
 */

/* Python wrapper */
static CYTHON_UNUSED void __pyx_pw_5numpy_7ndarray_3__releasebuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info); /*proto*/
static CYTHON_UNUSED void __pyx_pw_5numpy_7ndarray_3__releasebuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__releasebuffer__ (wrapper)", 0);
  __pyx_pf_5numpy_7ndarray_2__releasebuffer__(((PyArrayObject *)__pyx_v_self), ((Py_buffer *)__pyx_v_info));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_5numpy_7ndarray_2__releasebuffer__(PyArrayObject *__pyx_v_self, Py_buffer *__pyx_v_info) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__releasebuffer__", 0);

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":291
 * 
 *         def __releasebuffer__(ndarray self, Py_buffer* info):
 *             if PyArray_HASFIELDS(self):             # <<<<<<<<<<<<<<
 *                 stdlib.free(info.format)
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 */
  __pyx_t_1 = (PyArray_HASFIELDS(__pyx_v_self) != 0);
  if (__pyx_t_1) {

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":292
 *         def __releasebuffer__(ndarray self, Py_buffer* info):
 *             if PyArray_HASFIELDS(self):
 *                 stdlib.free(info.format)             # <<<<<<<<<<<<<<
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 *                 stdlib.free(info.strides)
 */
    free(__pyx_v_info->format);

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":291
 * 
 *         def __releasebuffer__(ndarray self, Py_buffer* info):
 *             if PyArray_HASFIELDS(self):             # <<<<<<<<<<<<<<
 *                 stdlib.free(info.format)
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 */
  }

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":293
 *             if PyArray_HASFIELDS(self):
 *                 stdlib.free(info.format)
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):             # <<<<<<<<<<<<<<
 *                 stdlib.free(info.strides)
 *                 # info.shape was stored after info.strides in the same block
 */
  __pyx_t_1 = (((sizeof(npy_intp)) != (sizeof(Py_ssize_t))) != 0);
  if (__pyx_t_1) {

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":294
 *                 stdlib.free(info.format)
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 *                 stdlib.free(info.strides)             # <<<<<<<<<<<<<<
 *                 # info.shape was stored after info.strides in the same block
 * 
 */
    free(__pyx_v_info->strides);

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":293
 *             if PyArray_HASFIELDS(self):
 *                 stdlib.free(info.format)
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):             # <<<<<<<<<<<<<<
 *                 stdlib.free(info.strides)
 *                 # info.shape was stored after info.strides in the same block
 */
  }

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":290
 *                 f[0] = c'\0' # Terminate format string
 * 
 *         def __releasebuffer__(ndarray self, Py_buffer* info):             # <<<<<<<<<<<<<<
 *             if PyArray_HASFIELDS(self):
 *                 stdlib.free(info.format)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":770
 * ctypedef npy_cdouble     complex_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew1(PyObject *__pyx_v_a) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew1", 0);

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":771
 * 
 * cdef inline object PyArray_MultiIterNew1(a):
 *     return PyArray_MultiIterNew(1, <void*>a)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(1, ((void *)__pyx_v_a)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 771, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":770
 * ctypedef npy_cdouble     complex_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew1", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":773
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew2(PyObject *__pyx_v_a, PyObject *__pyx_v_b) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew2", 0);

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":774
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(2, ((void *)__pyx_v_a), ((void *)__pyx_v_b)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 774, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":773
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew2", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":776
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew3(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew3", 0);

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":777
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(3, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 777, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":776
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew3", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":779
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew4(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew4", 0);

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":780
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(4, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 780, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":779
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew4", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":782
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew5(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d, PyObject *__pyx_v_e) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew5", 0);

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":783
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)             # <<<<<<<<<<<<<<
 * 
 * cdef inline char* _util_dtypestring(dtype descr, char* f, char* end, int* offset) except NULL:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(5, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d), ((void *)__pyx_v_e)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 783, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":782
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew5", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":785
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline char* _util_dtypestring(dtype descr, char* f, char* end, int* offset) except NULL:             # <<<<<<<<<<<<<<
 *     # Recursive utility function used in __getbuffer__ to get format
 *     # string. The new location in the format string is returned.
 */

static CYTHON_INLINE char *__pyx_f_5numpy__util_dtypestring(PyArray_Descr *__pyx_v_descr, char *__pyx_v_f, char *__pyx_v_end, int *__pyx_v_offset) {
  PyArray_Descr *__pyx_v_child = 0;
  int __pyx_v_endian_detector;
  int __pyx_v_little_endian;
  PyObject *__pyx_v_fields = 0;
  PyObject *__pyx_v_childname = NULL;
  PyObject *__pyx_v_new_offset = NULL;
  PyObject *__pyx_v_t = NULL;
  char *__pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  long __pyx_t_8;
  char *__pyx_t_9;
  __Pyx_RefNannySetupContext("_util_dtypestring", 0);

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":790
 * 
 *     cdef dtype child
 *     cdef int endian_detector = 1             # <<<<<<<<<<<<<<
 *     cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)
 *     cdef tuple fields
 */
  __pyx_v_endian_detector = 1;

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":791
 *     cdef dtype child
 *     cdef int endian_detector = 1
 *     cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)             # <<<<<<<<<<<<<<
 *     cdef tuple fields
 * 
 */
  __pyx_v_little_endian = ((((char *)(&__pyx_v_endian_detector))[0]) != 0);

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":794
 *     cdef tuple fields
 * 
 *     for childname in descr.names:             # <<<<<<<<<<<<<<
 *         fields = descr.fields[childname]
 *         child, new_offset = fields
 */
  if (unlikely(__pyx_v_descr->names == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(1, 794, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_v_descr->names; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
  for (;;) {
    if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(1, 794, __pyx_L1_error)
    #else
    __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 794, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_childname, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":795
 * 
 *     for childname in descr.names:
 *         fields = descr.fields[childname]             # <<<<<<<<<<<<<<
 *         child, new_offset = fields
 * 
 */
    if (unlikely(__pyx_v_descr->fields == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 795, __pyx_L1_error)
    }
    __pyx_t_3 = __Pyx_PyDict_GetItem(__pyx_v_descr->fields, __pyx_v_childname); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 795, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (!(likely(PyTuple_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_t_3)->tp_name), 0))) __PYX_ERR(1, 795, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_fields, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":796
 *     for childname in descr.names:
 *         fields = descr.fields[childname]
 *         child, new_offset = fields             # <<<<<<<<<<<<<<
 * 
 *         if (end - f) - <int>(new_offset - offset[0]) < 15:
 */
    if (likely(__pyx_v_fields != Py_None)) {
      PyObject* sequence = __pyx_v_fields;
      #if !CYTHON_COMPILING_IN_PYPY
      Py_ssize_t size = Py_SIZE(sequence);
      #else
      Py_ssize_t size = PySequence_Size(sequence);
      #endif
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(1, 796, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_4 = PyTuple_GET_ITEM(sequence, 1); 
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      #else
      __pyx_t_3 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 796, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 796, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      #endif
    } else {
      __Pyx_RaiseNoneNotIterableError(); __PYX_ERR(1, 796, __pyx_L1_error)
    }
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5numpy_dtype))))) __PYX_ERR(1, 796, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_child, ((PyArray_Descr *)__pyx_t_3));
    __pyx_t_3 = 0;
    __Pyx_XDECREF_SET(__pyx_v_new_offset, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":798
 *         child, new_offset = fields
 * 
 *         if (end - f) - <int>(new_offset - offset[0]) < 15:             # <<<<<<<<<<<<<<
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")
 * 
 */
    __pyx_t_4 = __Pyx_PyInt_From_int((__pyx_v_offset[0])); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 798, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = PyNumber_Subtract(__pyx_v_new_offset, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 798, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 798, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_6 = ((((__pyx_v_end - __pyx_v_f) - ((int)__pyx_t_5)) < 15) != 0);
    if (__pyx_t_6) {

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":799
 * 
 *         if (end - f) - <int>(new_offset - offset[0]) < 15:
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")             # <<<<<<<<<<<<<<
 * 
 *         if ((child.byteorder == c'>' and little_endian) or
 */
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__43, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 799, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(1, 799, __pyx_L1_error)

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":798
 *         child, new_offset = fields
 * 
 *         if (end - f) - <int>(new_offset - offset[0]) < 15:             # <<<<<<<<<<<<<<
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")
 * 
 */
    }

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":801
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")
 * 
 *         if ((child.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *             (child.byteorder == c'<' and not little_endian)):
 *             raise ValueError(u"Non-native byte order not supported")
 */
    __pyx_t_7 = ((__pyx_v_child->byteorder == '>') != 0);
    if (!__pyx_t_7) {
      goto __pyx_L8_next_or;
    } else {
    }
    __pyx_t_7 = (__pyx_v_little_endian != 0);
    if (!__pyx_t_7) {
    } else {
      __pyx_t_6 = __pyx_t_7;
      goto __pyx_L7_bool_binop_done;
    }
    __pyx_L8_next_or:;

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":802
 * 
 *         if ((child.byteorder == c'>' and little_endian) or
 *             (child.byteorder == c'<' and not little_endian)):             # <<<<<<<<<<<<<<
 *             raise ValueError(u"Non-native byte order not supported")
 *             # One could encode it in the format string and have Cython
 */
    __pyx_t_7 = ((__pyx_v_child->byteorder == '<') != 0);
    if (__pyx_t_7) {
    } else {
      __pyx_t_6 = __pyx_t_7;
      goto __pyx_L7_bool_binop_done;
    }
    __pyx_t_7 = ((!(__pyx_v_little_endian != 0)) != 0);
    __pyx_t_6 = __pyx_t_7;
    __pyx_L7_bool_binop_done:;

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":801
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")
 * 
 *         if ((child.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *             (child.byteorder == c'<' and not little_endian)):
 *             raise ValueError(u"Non-native byte order not supported")
 */
    if (__pyx_t_6) {

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":803
 *         if ((child.byteorder == c'>' and little_endian) or
 *             (child.byteorder == c'<' and not little_endian)):
 *             raise ValueError(u"Non-native byte order not supported")             # <<<<<<<<<<<<<<
 *             # One could encode it in the format string and have Cython
 *             # complain instead, BUT: < and > in format strings also imply
 */
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__44, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 803, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(1, 803, __pyx_L1_error)

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":801
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")
 * 
 *         if ((child.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *             (child.byteorder == c'<' and not little_endian)):
 *             raise ValueError(u"Non-native byte order not supported")
 */
    }

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":813
 * 
 *         # Output padding bytes
 *         while offset[0] < new_offset:             # <<<<<<<<<<<<<<
 *             f[0] = 120 # "x"; pad byte
 *             f += 1
 */
    while (1) {
      __pyx_t_3 = __Pyx_PyInt_From_int((__pyx_v_offset[0])); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 813, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_t_3, __pyx_v_new_offset, Py_LT); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 813, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(1, 813, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (!__pyx_t_6) break;

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":814
 *         # Output padding bytes
 *         while offset[0] < new_offset:
 *             f[0] = 120 # "x"; pad byte             # <<<<<<<<<<<<<<
 *             f += 1
 *             offset[0] += 1
 */
      (__pyx_v_f[0]) = 0x78;

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":815
 *         while offset[0] < new_offset:
 *             f[0] = 120 # "x"; pad byte
 *             f += 1             # <<<<<<<<<<<<<<
 *             offset[0] += 1
 * 
 */
      __pyx_v_f = (__pyx_v_f + 1);

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":816
 *             f[0] = 120 # "x"; pad byte
 *             f += 1
 *             offset[0] += 1             # <<<<<<<<<<<<<<
 * 
 *         offset[0] += child.itemsize
 */
      __pyx_t_8 = 0;
      (__pyx_v_offset[__pyx_t_8]) = ((__pyx_v_offset[__pyx_t_8]) + 1);
    }

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":818
 *             offset[0] += 1
 * 
 *         offset[0] += child.itemsize             # <<<<<<<<<<<<<<
 * 
 *         if not PyDataType_HASFIELDS(child):
 */
    __pyx_t_8 = 0;
    (__pyx_v_offset[__pyx_t_8]) = ((__pyx_v_offset[__pyx_t_8]) + __pyx_v_child->elsize);

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":820
 *         offset[0] += child.itemsize
 * 
 *         if not PyDataType_HASFIELDS(child):             # <<<<<<<<<<<<<<
 *             t = child.type_num
 *             if end - f < 5:
 */
    __pyx_t_6 = ((!(PyDataType_HASFIELDS(__pyx_v_child) != 0)) != 0);
    if (__pyx_t_6) {

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":821
 * 
 *         if not PyDataType_HASFIELDS(child):
 *             t = child.type_num             # <<<<<<<<<<<<<<
 *             if end - f < 5:
 *                 raise RuntimeError(u"Format string allocated too short.")
 */
      __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_child->type_num); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 821, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_XDECREF_SET(__pyx_v_t, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":822
 *         if not PyDataType_HASFIELDS(child):
 *             t = child.type_num
 *             if end - f < 5:             # <<<<<<<<<<<<<<
 *                 raise RuntimeError(u"Format string allocated too short.")
 * 
 */
      __pyx_t_6 = (((__pyx_v_end - __pyx_v_f) < 5) != 0);
      if (__pyx_t_6) {

        /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":823
 *             t = child.type_num
 *             if end - f < 5:
 *                 raise RuntimeError(u"Format string allocated too short.")             # <<<<<<<<<<<<<<
 * 
 *             # Until ticket #99 is fixed, use integers to avoid warnings
 */
        __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__45, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 823, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_Raise(__pyx_t_4, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __PYX_ERR(1, 823, __pyx_L1_error)

        /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":822
 *         if not PyDataType_HASFIELDS(child):
 *             t = child.type_num
 *             if end - f < 5:             # <<<<<<<<<<<<<<
 *                 raise RuntimeError(u"Format string allocated too short.")
 * 
 */
      }

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":826
 * 
 *             # Until ticket #99 is fixed, use integers to avoid warnings
 *             if   t == NPY_BYTE:        f[0] =  98 #"b"             # <<<<<<<<<<<<<<
 *             elif t == NPY_UBYTE:       f[0] =  66 #"B"
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_BYTE); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 826, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 826, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(1, 826, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 98;
        goto __pyx_L15;
      }

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":827
 *             # Until ticket #99 is fixed, use integers to avoid warnings
 *             if   t == NPY_BYTE:        f[0] =  98 #"b"
 *             elif t == NPY_UBYTE:       f[0] =  66 #"B"             # <<<<<<<<<<<<<<
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_UBYTE); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 827, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 827, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(1, 827, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 66;
        goto __pyx_L15;
      }

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":828
 *             if   t == NPY_BYTE:        f[0] =  98 #"b"
 *             elif t == NPY_UBYTE:       f[0] =  66 #"B"
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"             # <<<<<<<<<<<<<<
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"
 *             elif t == NPY_INT:         f[0] = 105 #"i"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_SHORT); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 828, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 828, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(1, 828, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x68;
        goto __pyx_L15;
      }

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":829
 *             elif t == NPY_UBYTE:       f[0] =  66 #"B"
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"             # <<<<<<<<<<<<<<
 *             elif t == NPY_INT:         f[0] = 105 #"i"
 *             elif t == NPY_UINT:        f[0] =  73 #"I"
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_USHORT); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 829, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 829, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(1, 829, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 72;
        goto __pyx_L15;
      }

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":830
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"
 *             elif t == NPY_INT:         f[0] = 105 #"i"             # <<<<<<<<<<<<<<
 *             elif t == NPY_UINT:        f[0] =  73 #"I"
 *             elif t == NPY_LONG:        f[0] = 108 #"l"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_INT); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 830, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 830, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(1, 830, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x69;
        goto __pyx_L15;
      }

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":831
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"
 *             elif t == NPY_INT:         f[0] = 105 #"i"
 *             elif t == NPY_UINT:        f[0] =  73 #"I"             # <<<<<<<<<<<<<<
 *             elif t == NPY_LONG:        f[0] = 108 #"l"
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_UINT); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 831, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 831, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(1, 831, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 73;
        goto __pyx_L15;
      }

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":832
 *             elif t == NPY_INT:         f[0] = 105 #"i"
 *             elif t == NPY_UINT:        f[0] =  73 #"I"
 *             elif t == NPY_LONG:        f[0] = 108 #"l"             # <<<<<<<<<<<<<<
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_LONG); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 832, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 832, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(1, 832, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x6C;
        goto __pyx_L15;
      }

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":833
 *             elif t == NPY_UINT:        f[0] =  73 #"I"
 *             elif t == NPY_LONG:        f[0] = 108 #"l"
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"             # <<<<<<<<<<<<<<
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_ULONG); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 833, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 833, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(1, 833, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 76;
        goto __pyx_L15;
      }

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":834
 *             elif t == NPY_LONG:        f[0] = 108 #"l"
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"             # <<<<<<<<<<<<<<
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_LONGLONG); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 834, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 834, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(1, 834, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x71;
        goto __pyx_L15;
      }

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":835
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"             # <<<<<<<<<<<<<<
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_ULONGLONG); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 835, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 835, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(1, 835, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 81;
        goto __pyx_L15;
      }

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":836
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"             # <<<<<<<<<<<<<<
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_FLOAT); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 836, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 836, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(1, 836, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x66;
        goto __pyx_L15;
      }

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":837
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"             # <<<<<<<<<<<<<<
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_DOUBLE); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 837, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 837, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(1, 837, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x64;
        goto __pyx_L15;
      }

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":838
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"             # <<<<<<<<<<<<<<
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_LONGDOUBLE); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 838, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 838, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(1, 838, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x67;
        goto __pyx_L15;
      }

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":839
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf             # <<<<<<<<<<<<<<
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd
 *             elif t == NPY_CLONGDOUBLE: f[0] = 90; f[1] = 103; f += 1 # Zg
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_CFLOAT); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 839, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 839, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(1, 839, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 90;
        (__pyx_v_f[1]) = 0x66;
        __pyx_v_f = (__pyx_v_f + 1);
        goto __pyx_L15;
      }

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":840
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd             # <<<<<<<<<<<<<<
 *             elif t == NPY_CLONGDOUBLE: f[0] = 90; f[1] = 103; f += 1 # Zg
 *             elif t == NPY_OBJECT:      f[0] = 79 #"O"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_CDOUBLE); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 840, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 840, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(1, 840, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 90;
        (__pyx_v_f[1]) = 0x64;
        __pyx_v_f = (__pyx_v_f + 1);
        goto __pyx_L15;
      }

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":841
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd
 *             elif t == NPY_CLONGDOUBLE: f[0] = 90; f[1] = 103; f += 1 # Zg             # <<<<<<<<<<<<<<
 *             elif t == NPY_OBJECT:      f[0] = 79 #"O"
 *             else:
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_CLONGDOUBLE); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 841, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 841, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(1, 841, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 90;
        (__pyx_v_f[1]) = 0x67;
        __pyx_v_f = (__pyx_v_f + 1);
        goto __pyx_L15;
      }

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":842
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd
 *             elif t == NPY_CLONGDOUBLE: f[0] = 90; f[1] = 103; f += 1 # Zg
 *             elif t == NPY_OBJECT:      f[0] = 79 #"O"             # <<<<<<<<<<<<<<
 *             else:
 *                 raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_OBJECT); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 842, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 842, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(1, 842, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 79;
        goto __pyx_L15;
      }

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":844
 *             elif t == NPY_OBJECT:      f[0] = 79 #"O"
 *             else:
 *                 raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)             # <<<<<<<<<<<<<<
 *             f += 1
 *         else:
 */
      /*else*/ {
        __pyx_t_3 = PyUnicode_Format(__pyx_kp_u_unknown_dtype_code_in_numpy_pxd, __pyx_v_t); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 844, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 844, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_GIVEREF(__pyx_t_3);
        PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3);
        __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_t_4, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 844, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_Raise(__pyx_t_3, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __PYX_ERR(1, 844, __pyx_L1_error)
      }
      __pyx_L15:;

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":845
 *             else:
 *                 raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 *             f += 1             # <<<<<<<<<<<<<<
 *         else:
 *             # Cython ignores struct boundary information ("T{...}"),
 */
      __pyx_v_f = (__pyx_v_f + 1);

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":820
 *         offset[0] += child.itemsize
 * 
 *         if not PyDataType_HASFIELDS(child):             # <<<<<<<<<<<<<<
 *             t = child.type_num
 *             if end - f < 5:
 */
      goto __pyx_L13;
    }

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":849
 *             # Cython ignores struct boundary information ("T{...}"),
 *             # so don't output it
 *             f = _util_dtypestring(child, f, end, offset)             # <<<<<<<<<<<<<<
 *     return f
 * 
 */
    /*else*/ {
      __pyx_t_9 = __pyx_f_5numpy__util_dtypestring(__pyx_v_child, __pyx_v_f, __pyx_v_end, __pyx_v_offset); if (unlikely(__pyx_t_9 == NULL)) __PYX_ERR(1, 849, __pyx_L1_error)
      __pyx_v_f = __pyx_t_9;
    }
    __pyx_L13:;

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":794
 *     cdef tuple fields
 * 
 *     for childname in descr.names:             # <<<<<<<<<<<<<<
 *         fields = descr.fields[childname]
 *         child, new_offset = fields
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":850
 *             # so don't output it
 *             f = _util_dtypestring(child, f, end, offset)
 *     return f             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_f;
  goto __pyx_L0;

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":785
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline char* _util_dtypestring(dtype descr, char* f, char* end, int* offset) except NULL:             # <<<<<<<<<<<<<<
 *     # Recursive utility function used in __getbuffer__ to get format
 *     # string. The new location in the format string is returned.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("numpy._util_dtypestring", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_child);
  __Pyx_XDECREF(__pyx_v_fields);
  __Pyx_XDECREF(__pyx_v_childname);
  __Pyx_XDECREF(__pyx_v_new_offset);
  __Pyx_XDECREF(__pyx_v_t);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":966
 * 
 * 
 * cdef inline void set_array_base(ndarray arr, object base):             # <<<<<<<<<<<<<<
 *      cdef PyObject* baseptr
 *      if base is None:
 */

static CYTHON_INLINE void __pyx_f_5numpy_set_array_base(PyArrayObject *__pyx_v_arr, PyObject *__pyx_v_base) {
  PyObject *__pyx_v_baseptr;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  __Pyx_RefNannySetupContext("set_array_base", 0);

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":968
 * cdef inline void set_array_base(ndarray arr, object base):
 *      cdef PyObject* baseptr
 *      if base is None:             # <<<<<<<<<<<<<<
 *          baseptr = NULL
 *      else:
 */
  __pyx_t_1 = (__pyx_v_base == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":969
 *      cdef PyObject* baseptr
 *      if base is None:
 *          baseptr = NULL             # <<<<<<<<<<<<<<
 *      else:
 *          Py_INCREF(base) # important to do this before decref below!
 */
    __pyx_v_baseptr = NULL;

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":968
 * cdef inline void set_array_base(ndarray arr, object base):
 *      cdef PyObject* baseptr
 *      if base is None:             # <<<<<<<<<<<<<<
 *          baseptr = NULL
 *      else:
 */
    goto __pyx_L3;
  }

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":971
 *          baseptr = NULL
 *      else:
 *          Py_INCREF(base) # important to do this before decref below!             # <<<<<<<<<<<<<<
 *          baseptr = <PyObject*>base
 *      Py_XDECREF(arr.base)
 */
  /*else*/ {
    Py_INCREF(__pyx_v_base);

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":972
 *      else:
 *          Py_INCREF(base) # important to do this before decref below!
 *          baseptr = <PyObject*>base             # <<<<<<<<<<<<<<
 *      Py_XDECREF(arr.base)
 *      arr.base = baseptr
 */
    __pyx_v_baseptr = ((PyObject *)__pyx_v_base);
  }
  __pyx_L3:;

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":973
 *          Py_INCREF(base) # important to do this before decref below!
 *          baseptr = <PyObject*>base
 *      Py_XDECREF(arr.base)             # <<<<<<<<<<<<<<
 *      arr.base = baseptr
 * 
 */
  Py_XDECREF(__pyx_v_arr->base);

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":974
 *          baseptr = <PyObject*>base
 *      Py_XDECREF(arr.base)
 *      arr.base = baseptr             # <<<<<<<<<<<<<<
 * 
 * cdef inline object get_array_base(ndarray arr):
 */
  __pyx_v_arr->base = __pyx_v_baseptr;

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":966
 * 
 * 
 * cdef inline void set_array_base(ndarray arr, object base):             # <<<<<<<<<<<<<<
 *      cdef PyObject* baseptr
 *      if base is None:
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":976
 *      arr.base = baseptr
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     if arr.base is NULL:
 *         return None
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_get_array_base(PyArrayObject *__pyx_v_arr) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("get_array_base", 0);

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":977
 * 
 * cdef inline object get_array_base(ndarray arr):
 *     if arr.base is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     else:
 */
  __pyx_t_1 = ((__pyx_v_arr->base == NULL) != 0);
  if (__pyx_t_1) {

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":978
 * cdef inline object get_array_base(ndarray arr):
 *     if arr.base is NULL:
 *         return None             # <<<<<<<<<<<<<<
 *     else:
 *         return <object>arr.base
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_None);
    __pyx_r = Py_None;
    goto __pyx_L0;

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":977
 * 
 * cdef inline object get_array_base(ndarray arr):
 *     if arr.base is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     else:
 */
  }

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":980
 *         return None
 *     else:
 *         return <object>arr.base             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(((PyObject *)__pyx_v_arr->base));
    __pyx_r = ((PyObject *)__pyx_v_arr->base);
    goto __pyx_L0;
  }

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":976
 *      arr.base = baseptr
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     if arr.base is NULL:
 *         return None
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":985
 * # Versions of the import_* functions which are more suitable for
 * # Cython code.
 * cdef inline int import_array() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_array()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_array(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("import_array", 0);

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":986
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_array()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":987
 * cdef inline int import_array() except -1:
 *     try:
 *         _import_array()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")
 */
      __pyx_t_4 = _import_array(); if (unlikely(__pyx_t_4 == -1)) __PYX_ERR(1, 987, __pyx_L3_error)

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":986
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_array()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L10_try_end;
    __pyx_L3_error:;
    __Pyx_PyThreadState_assign

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":988
 *     try:
 *         _import_array()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_array", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(1, 988, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":989
 *         _import_array()
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_umath() except -1:
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__46, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 989, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(1, 989, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":986
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_array()
 *     except Exception:
 */
    __Pyx_PyThreadState_assign
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L10_try_end:;
  }

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":985
 * # Versions of the import_* functions which are more suitable for
 * # Cython code.
 * cdef inline int import_array() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_array()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_array", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":991
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 * cdef inline int import_umath() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_umath(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("import_umath", 0);

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":992
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":993
 * cdef inline int import_umath() except -1:
 *     try:
 *         _import_umath()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")
 */
      __pyx_t_4 = _import_umath(); if (unlikely(__pyx_t_4 == -1)) __PYX_ERR(1, 993, __pyx_L3_error)

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":992
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L10_try_end;
    __pyx_L3_error:;
    __Pyx_PyThreadState_assign

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":994
 *     try:
 *         _import_umath()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_umath", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(1, 994, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":995
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_ufunc() except -1:
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__47, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 995, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(1, 995, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":992
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    __Pyx_PyThreadState_assign
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L10_try_end:;
  }

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":991
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 * cdef inline int import_umath() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_umath", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":997
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_ufunc(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("import_ufunc", 0);

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":998
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":999
 * cdef inline int import_ufunc() except -1:
 *     try:
 *         _import_umath()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")
 */
      __pyx_t_4 = _import_umath(); if (unlikely(__pyx_t_4 == -1)) __PYX_ERR(1, 999, __pyx_L3_error)

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":998
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L10_try_end;
    __pyx_L3_error:;
    __Pyx_PyThreadState_assign

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":1000
 *     try:
 *         _import_umath()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.umath failed to import")
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_ufunc", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(1, 1000, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":1001
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__48, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 1001, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(1, 1001, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":998
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    __Pyx_PyThreadState_assign
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L10_try_end:;
  }

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":997
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_ufunc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static struct __pyx_vtabstruct_8rasterio_5_warp_WarpedVRTReaderBase __pyx_vtable_8rasterio_5_warp_WarpedVRTReaderBase;

static PyObject *__pyx_tp_new_8rasterio_5_warp_WarpedVRTReaderBase(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_8rasterio_5_warp_WarpedVRTReaderBase *p;
  PyObject *o = __pyx_ptype_8rasterio_3_io_DatasetReaderBase->tp_new(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_8rasterio_5_warp_WarpedVRTReaderBase *)o);
  p->__pyx_base.__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_8rasterio_5_base_DatasetBase*)__pyx_vtabptr_8rasterio_5_warp_WarpedVRTReaderBase;
  return o;
}

static void __pyx_tp_dealloc_8rasterio_5_warp_WarpedVRTReaderBase(PyObject *o) {
  #if PY_VERSION_HEX >= 0x030400a1
  if (unlikely(Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  PyObject_GC_Track(o);
  if (likely(__pyx_ptype_8rasterio_3_io_DatasetReaderBase)) __pyx_ptype_8rasterio_3_io_DatasetReaderBase->tp_dealloc(o); else __Pyx_call_next_tp_dealloc(o, __pyx_tp_dealloc_8rasterio_5_warp_WarpedVRTReaderBase);
}

static int __pyx_tp_traverse_8rasterio_5_warp_WarpedVRTReaderBase(PyObject *o, visitproc v, void *a) {
  int e;
  e = ((likely(__pyx_ptype_8rasterio_3_io_DatasetReaderBase)) ? ((__pyx_ptype_8rasterio_3_io_DatasetReaderBase->tp_traverse) ? __pyx_ptype_8rasterio_3_io_DatasetReaderBase->tp_traverse(o, v, a) : 0) : __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_8rasterio_5_warp_WarpedVRTReaderBase)); if (e) return e;
  return 0;
}

static int __pyx_tp_clear_8rasterio_5_warp_WarpedVRTReaderBase(PyObject *o) {
  if (likely(__pyx_ptype_8rasterio_3_io_DatasetReaderBase)) { if (__pyx_ptype_8rasterio_3_io_DatasetReaderBase->tp_clear) __pyx_ptype_8rasterio_3_io_DatasetReaderBase->tp_clear(o); } else __Pyx_call_next_tp_clear(o, __pyx_tp_clear_8rasterio_5_warp_WarpedVRTReaderBase);
  return 0;
}

static PyMethodDef __pyx_methods_8rasterio_5_warp_WarpedVRTReaderBase[] = {
  {"start", (PyCFunction)__pyx_pw_8rasterio_5_warp_19WarpedVRTReaderBase_3start, METH_NOARGS, __pyx_doc_8rasterio_5_warp_19WarpedVRTReaderBase_2start},
  {"stop", (PyCFunction)__pyx_pw_8rasterio_5_warp_19WarpedVRTReaderBase_5stop, METH_NOARGS, __pyx_doc_8rasterio_5_warp_19WarpedVRTReaderBase_4stop},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_8rasterio_5_warp_WarpedVRTReaderBase = {
  PyVarObject_HEAD_INIT(0, 0)
  "rasterio._warp.WarpedVRTReaderBase", /*tp_name*/
  sizeof(struct __pyx_obj_8rasterio_5_warp_WarpedVRTReaderBase), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_8rasterio_5_warp_WarpedVRTReaderBase, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_8rasterio_5_warp_WarpedVRTReaderBase, /*tp_traverse*/
  __pyx_tp_clear_8rasterio_5_warp_WarpedVRTReaderBase, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_8rasterio_5_warp_WarpedVRTReaderBase, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_8rasterio_5_warp_19WarpedVRTReaderBase_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_8rasterio_5_warp_WarpedVRTReaderBase, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static struct __pyx_obj_8rasterio_5_warp___pyx_scope_struct___calculate_default_transform *__pyx_freelist_8rasterio_5_warp___pyx_scope_struct___calculate_default_transform[8];
static int __pyx_freecount_8rasterio_5_warp___pyx_scope_struct___calculate_default_transform = 0;

static PyObject *__pyx_tp_new_8rasterio_5_warp___pyx_scope_struct___calculate_default_transform(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_8rasterio_5_warp___pyx_scope_struct___calculate_default_transform > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_8rasterio_5_warp___pyx_scope_struct___calculate_default_transform)))) {
    o = (PyObject*)__pyx_freelist_8rasterio_5_warp___pyx_scope_struct___calculate_default_transform[--__pyx_freecount_8rasterio_5_warp___pyx_scope_struct___calculate_default_transform];
    memset(o, 0, sizeof(struct __pyx_obj_8rasterio_5_warp___pyx_scope_struct___calculate_default_transform));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_8rasterio_5_warp___pyx_scope_struct___calculate_default_transform(PyObject *o) {
  struct __pyx_obj_8rasterio_5_warp___pyx_scope_struct___calculate_default_transform *p = (struct __pyx_obj_8rasterio_5_warp___pyx_scope_struct___calculate_default_transform *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_bottom);
  Py_CLEAR(p->__pyx_v_left);
  Py_CLEAR(p->__pyx_v_right);
  Py_CLEAR(p->__pyx_v_top);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_8rasterio_5_warp___pyx_scope_struct___calculate_default_transform < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_8rasterio_5_warp___pyx_scope_struct___calculate_default_transform)))) {
    __pyx_freelist_8rasterio_5_warp___pyx_scope_struct___calculate_default_transform[__pyx_freecount_8rasterio_5_warp___pyx_scope_struct___calculate_default_transform++] = ((struct __pyx_obj_8rasterio_5_warp___pyx_scope_struct___calculate_default_transform *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_8rasterio_5_warp___pyx_scope_struct___calculate_default_transform(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_8rasterio_5_warp___pyx_scope_struct___calculate_default_transform *p = (struct __pyx_obj_8rasterio_5_warp___pyx_scope_struct___calculate_default_transform *)o;
  if (p->__pyx_v_bottom) {
    e = (*v)(p->__pyx_v_bottom, a); if (e) return e;
  }
  if (p->__pyx_v_left) {
    e = (*v)(p->__pyx_v_left, a); if (e) return e;
  }
  if (p->__pyx_v_right) {
    e = (*v)(p->__pyx_v_right, a); if (e) return e;
  }
  if (p->__pyx_v_top) {
    e = (*v)(p->__pyx_v_top, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_8rasterio_5_warp___pyx_scope_struct___calculate_default_transform(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_8rasterio_5_warp___pyx_scope_struct___calculate_default_transform *p = (struct __pyx_obj_8rasterio_5_warp___pyx_scope_struct___calculate_default_transform *)o;
  tmp = ((PyObject*)p->__pyx_v_bottom);
  p->__pyx_v_bottom = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_left);
  p->__pyx_v_left = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_right);
  p->__pyx_v_right = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_top);
  p->__pyx_v_top = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_8rasterio_5_warp___pyx_scope_struct___calculate_default_transform = {
  PyVarObject_HEAD_INIT(0, 0)
  "rasterio._warp.__pyx_scope_struct___calculate_default_transform", /*tp_name*/
  sizeof(struct __pyx_obj_8rasterio_5_warp___pyx_scope_struct___calculate_default_transform), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_8rasterio_5_warp___pyx_scope_struct___calculate_default_transform, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_8rasterio_5_warp___pyx_scope_struct___calculate_default_transform, /*tp_traverse*/
  __pyx_tp_clear_8rasterio_5_warp___pyx_scope_struct___calculate_default_transform, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_8rasterio_5_warp___pyx_scope_struct___calculate_default_transform, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static struct __pyx_obj_8rasterio_5_warp___pyx_scope_struct_1_genexpr *__pyx_freelist_8rasterio_5_warp___pyx_scope_struct_1_genexpr[8];
static int __pyx_freecount_8rasterio_5_warp___pyx_scope_struct_1_genexpr = 0;

static PyObject *__pyx_tp_new_8rasterio_5_warp___pyx_scope_struct_1_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_8rasterio_5_warp___pyx_scope_struct_1_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_8rasterio_5_warp___pyx_scope_struct_1_genexpr)))) {
    o = (PyObject*)__pyx_freelist_8rasterio_5_warp___pyx_scope_struct_1_genexpr[--__pyx_freecount_8rasterio_5_warp___pyx_scope_struct_1_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_8rasterio_5_warp___pyx_scope_struct_1_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_8rasterio_5_warp___pyx_scope_struct_1_genexpr(PyObject *o) {
  struct __pyx_obj_8rasterio_5_warp___pyx_scope_struct_1_genexpr *p = (struct __pyx_obj_8rasterio_5_warp___pyx_scope_struct_1_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_x);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_8rasterio_5_warp___pyx_scope_struct_1_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_8rasterio_5_warp___pyx_scope_struct_1_genexpr)))) {
    __pyx_freelist_8rasterio_5_warp___pyx_scope_struct_1_genexpr[__pyx_freecount_8rasterio_5_warp___pyx_scope_struct_1_genexpr++] = ((struct __pyx_obj_8rasterio_5_warp___pyx_scope_struct_1_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_8rasterio_5_warp___pyx_scope_struct_1_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_8rasterio_5_warp___pyx_scope_struct_1_genexpr *p = (struct __pyx_obj_8rasterio_5_warp___pyx_scope_struct_1_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject*)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_x) {
    e = (*v)(p->__pyx_v_x, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_8rasterio_5_warp___pyx_scope_struct_1_genexpr(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_8rasterio_5_warp___pyx_scope_struct_1_genexpr *p = (struct __pyx_obj_8rasterio_5_warp___pyx_scope_struct_1_genexpr *)o;
  tmp = ((PyObject*)p->__pyx_outer_scope);
  p->__pyx_outer_scope = ((struct __pyx_obj_8rasterio_5_warp___pyx_scope_struct___calculate_default_transform *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_x);
  p->__pyx_v_x = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_8rasterio_5_warp___pyx_scope_struct_1_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "rasterio._warp.__pyx_scope_struct_1_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_8rasterio_5_warp___pyx_scope_struct_1_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_8rasterio_5_warp___pyx_scope_struct_1_genexpr, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_8rasterio_5_warp___pyx_scope_struct_1_genexpr, /*tp_traverse*/
  __pyx_tp_clear_8rasterio_5_warp___pyx_scope_struct_1_genexpr, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_8rasterio_5_warp___pyx_scope_struct_1_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static struct __pyx_obj_8rasterio_5_warp___pyx_scope_struct_2_genexpr *__pyx_freelist_8rasterio_5_warp___pyx_scope_struct_2_genexpr[8];
static int __pyx_freecount_8rasterio_5_warp___pyx_scope_struct_2_genexpr = 0;

static PyObject *__pyx_tp_new_8rasterio_5_warp___pyx_scope_struct_2_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_8rasterio_5_warp___pyx_scope_struct_2_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_8rasterio_5_warp___pyx_scope_struct_2_genexpr)))) {
    o = (PyObject*)__pyx_freelist_8rasterio_5_warp___pyx_scope_struct_2_genexpr[--__pyx_freecount_8rasterio_5_warp___pyx_scope_struct_2_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_8rasterio_5_warp___pyx_scope_struct_2_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_8rasterio_5_warp___pyx_scope_struct_2_genexpr(PyObject *o) {
  struct __pyx_obj_8rasterio_5_warp___pyx_scope_struct_2_genexpr *p = (struct __pyx_obj_8rasterio_5_warp___pyx_scope_struct_2_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_x);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_8rasterio_5_warp___pyx_scope_struct_2_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_8rasterio_5_warp___pyx_scope_struct_2_genexpr)))) {
    __pyx_freelist_8rasterio_5_warp___pyx_scope_struct_2_genexpr[__pyx_freecount_8rasterio_5_warp___pyx_scope_struct_2_genexpr++] = ((struct __pyx_obj_8rasterio_5_warp___pyx_scope_struct_2_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_8rasterio_5_warp___pyx_scope_struct_2_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_8rasterio_5_warp___pyx_scope_struct_2_genexpr *p = (struct __pyx_obj_8rasterio_5_warp___pyx_scope_struct_2_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject*)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_x) {
    e = (*v)(p->__pyx_v_x, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_8rasterio_5_warp___pyx_scope_struct_2_genexpr(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_8rasterio_5_warp___pyx_scope_struct_2_genexpr *p = (struct __pyx_obj_8rasterio_5_warp___pyx_scope_struct_2_genexpr *)o;
  tmp = ((PyObject*)p->__pyx_outer_scope);
  p->__pyx_outer_scope = ((struct __pyx_obj_8rasterio_5_warp___pyx_scope_struct___calculate_default_transform *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_x);
  p->__pyx_v_x = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_8rasterio_5_warp___pyx_scope_struct_2_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "rasterio._warp.__pyx_scope_struct_2_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_8rasterio_5_warp___pyx_scope_struct_2_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_8rasterio_5_warp___pyx_scope_struct_2_genexpr, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_8rasterio_5_warp___pyx_scope_struct_2_genexpr, /*tp_traverse*/
  __pyx_tp_clear_8rasterio_5_warp___pyx_scope_struct_2_genexpr, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_8rasterio_5_warp___pyx_scope_struct_2_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
static struct PyModuleDef __pyx_moduledef = {
  #if PY_VERSION_HEX < 0x03020000
    { PyObject_HEAD_INIT(NULL) NULL, 0, NULL },
  #else
    PyModuleDef_HEAD_INIT,
  #endif
    "_warp",
    __pyx_k_Raster_and_vector_warping_and_re, /* m_doc */
    -1, /* m_size */
    __pyx_methods /* m_methods */,
    NULL, /* m_reload */
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_n_s_Affine, __pyx_k_Affine, sizeof(__pyx_k_Affine), 0, 0, 1, 1},
  {&__pyx_n_s_CPLE_AppDefinedError, __pyx_k_CPLE_AppDefinedError, sizeof(__pyx_k_CPLE_AppDefinedError), 0, 0, 1, 1},
  {&__pyx_n_s_CPLE_IllegalArgError, __pyx_k_CPLE_IllegalArgError, sizeof(__pyx_k_CPLE_IllegalArgError), 0, 0, 1, 1},
  {&__pyx_n_s_CPLE_NotSupportedError, __pyx_k_CPLE_NotSupportedError, sizeof(__pyx_k_CPLE_NotSupportedError), 0, 0, 1, 1},
  {&__pyx_n_s_CPLE_OpenFailedError, __pyx_k_CPLE_OpenFailedError, sizeof(__pyx_k_CPLE_OpenFailedError), 0, 0, 1, 1},
  {&__pyx_n_s_CRSError, __pyx_k_CRSError, sizeof(__pyx_k_CRSError), 0, 0, 1, 1},
  {&__pyx_kp_s_CRS_for_temp_destination_dataset, __pyx_k_CRS_for_temp_destination_dataset, sizeof(__pyx_k_CRS_for_temp_destination_dataset), 0, 0, 1, 0},
  {&__pyx_kp_s_Chunk_and_warp_window_d_d_d_d, __pyx_k_Chunk_and_warp_window_d_d_d_d, sizeof(__pyx_k_Chunk_and_warp_window_d_d_d_d), 0, 0, 1, 0},
  {&__pyx_kp_s_Created_temp_destination_dataset, __pyx_k_Created_temp_destination_dataset, sizeof(__pyx_k_Created_temp_destination_dataset), 0, 0, 1, 0},
  {&__pyx_kp_s_Created_temp_source_dataset, __pyx_k_Created_temp_source_dataset, sizeof(__pyx_k_Created_temp_source_dataset), 0, 0, 1, 0},
  {&__pyx_kp_s_Created_transformer_and_options, __pyx_k_Created_transformer_and_options, sizeof(__pyx_k_Created_transformer_and_options), 0, 0, 1, 0},
  {&__pyx_kp_s_Created_transformer_and_warp_out, __pyx_k_Created_transformer_and_warp_out, sizeof(__pyx_k_Created_transformer_and_warp_out), 0, 0, 1, 0},
  {&__pyx_kp_s_Dataset_r_is_started, __pyx_k_Dataset_r_is_started, sizeof(__pyx_k_Dataset_r_is_started), 0, 0, 1, 0},
  {&__pyx_kp_s_Destination_s_shape_is_invalid, __pyx_k_Destination_s_shape_is_invalid, sizeof(__pyx_k_Destination_s_shape_is_invalid), 0, 0, 1, 0},
  {&__pyx_n_s_DriverRegistrationError, __pyx_k_DriverRegistrationError, sizeof(__pyx_k_DriverRegistrationError), 0, 0, 1, 1},
  {&__pyx_kp_s_Encountered_points_outside_of_va, __pyx_k_Encountered_points_outside_of_va, sizeof(__pyx_k_Encountered_points_outside_of_va), 0, 0, 1, 0},
  {&__pyx_kp_s_Exported_CRS_to_WKT, __pyx_k_Exported_CRS_to_WKT, sizeof(__pyx_k_Exported_CRS_to_WKT), 0, 0, 1, 0},
  {&__pyx_kp_s_Failed_to_set_projection_on_temp, __pyx_k_Failed_to_set_projection_on_temp, sizeof(__pyx_k_Failed_to_set_projection_on_temp), 0, 0, 1, 0},
  {&__pyx_kp_s_Failed_to_set_transform_on_temp, __pyx_k_Failed_to_set_transform_on_temp, sizeof(__pyx_k_Failed_to_set_transform_on_temp), 0, 0, 1, 0},
  {&__pyx_kp_u_Format_string_allocated_too_shor, __pyx_k_Format_string_allocated_too_shor, sizeof(__pyx_k_Format_string_allocated_too_shor), 0, 1, 0, 0},
  {&__pyx_kp_u_Format_string_allocated_too_shor_2, __pyx_k_Format_string_allocated_too_shor_2, sizeof(__pyx_k_Format_string_allocated_too_shor_2), 0, 1, 0, 0},
  {&__pyx_n_s_GA, __pyx_k_GA, sizeof(__pyx_k_GA), 0, 0, 1, 1},
  {&__pyx_n_s_GDALError, __pyx_k_GDALError, sizeof(__pyx_k_GDALError), 0, 0, 1, 1},
  {&__pyx_n_s_GroundControlPoint, __pyx_k_GroundControlPoint, sizeof(__pyx_k_GroundControlPoint), 0, 0, 1, 1},
  {&__pyx_n_s_ImportError, __pyx_k_ImportError, sizeof(__pyx_k_ImportError), 0, 0, 1, 1},
  {&__pyx_kp_s_Invalid_destination, __pyx_k_Invalid_destination, sizeof(__pyx_k_Invalid_destination), 0, 0, 1, 0},
  {&__pyx_kp_s_Invalid_source, __pyx_k_Invalid_source, sizeof(__pyx_k_Invalid_source), 0, 0, 1, 0},
  {&__pyx_kp_s_MEM_driver_not_found_Check_that, __pyx_k_MEM_driver_not_found_Check_that, sizeof(__pyx_k_MEM_driver_not_found_Check_that), 0, 0, 1, 0},
  {&__pyx_n_s_NO_DATA, __pyx_k_NO_DATA, sizeof(__pyx_k_NO_DATA), 0, 0, 1, 1},
  {&__pyx_kp_u_Non_native_byte_order_not_suppor, __pyx_k_Non_native_byte_order_not_suppor, sizeof(__pyx_k_Non_native_byte_order_not_suppor), 0, 1, 0, 0},
  {&__pyx_n_s_RasterioIOError, __pyx_k_RasterioIOError, sizeof(__pyx_k_RasterioIOError), 0, 0, 1, 1},
  {&__pyx_kp_s_Reprojection_failed, __pyx_k_Reprojection_failed, sizeof(__pyx_k_Reprojection_failed), 0, 0, 1, 0},
  {&__pyx_n_s_Resampling, __pyx_k_Resampling, sizeof(__pyx_k_Resampling), 0, 0, 1, 1},
  {&__pyx_n_s_RuntimeError, __pyx_k_RuntimeError, sizeof(__pyx_k_RuntimeError), 0, 0, 1, 1},
  {&__pyx_kp_s_Set_CRS_on_temp_source_dataset_s, __pyx_k_Set_CRS_on_temp_source_dataset_s, sizeof(__pyx_k_Set_CRS_on_temp_source_dataset_s), 0, 0, 1, 0},
  {&__pyx_kp_s_Set_transform_on_temp_source_dat, __pyx_k_Set_transform_on_temp_source_dat, sizeof(__pyx_k_Set_transform_on_temp_source_dat), 0, 0, 1, 0},
  {&__pyx_kp_s_Set_transformer_options, __pyx_k_Set_transformer_options, sizeof(__pyx_k_Set_transformer_options), 0, 0, 1, 0},
  {&__pyx_kp_s_Setting_NUM_THREADS_option_d, __pyx_k_Setting_NUM_THREADS_option_d, sizeof(__pyx_k_Setting_NUM_THREADS_option_d), 0, 0, 1, 0},
  {&__pyx_kp_s_Some_but_not_all_bounding_box_pa, __pyx_k_Some_but_not_all_bounding_box_pa, sizeof(__pyx_k_Some_but_not_all_bounding_box_pa), 0, 0, 1, 0},
  {&__pyx_kp_s_Users_sean_code_frs_wheel_build, __pyx_k_Users_sean_code_frs_wheel_build, sizeof(__pyx_k_Users_sean_code_frs_wheel_build), 0, 0, 1, 0},
  {&__pyx_n_s_ValueError, __pyx_k_ValueError, sizeof(__pyx_k_ValueError), 0, 0, 1, 1},
  {&__pyx_kp_s_Warp_extras_r, __pyx_k_Warp_extras_r, sizeof(__pyx_k_Warp_extras_r), 0, 0, 1, 0},
  {&__pyx_kp_s_WarpedVRT, __pyx_k_WarpedVRT, sizeof(__pyx_k_WarpedVRT), 0, 0, 1, 0},
  {&__pyx_kp_s_Wrote_array_to_temp_output_datas, __pyx_k_Wrote_array_to_temp_output_datas, sizeof(__pyx_k_Wrote_array_to_temp_output_datas), 0, 0, 1, 0},
  {&__pyx_kp_s_Wrote_array_to_temp_source_datas, __pyx_k_Wrote_array_to_temp_source_datas, sizeof(__pyx_k_Wrote_array_to_temp_source_datas), 0, 0, 1, 0},
  {&__pyx_kp_s__11, __pyx_k__11, sizeof(__pyx_k__11), 0, 0, 1, 0},
  {&__pyx_n_s__53, __pyx_k__53, sizeof(__pyx_k__53), 0, 0, 1, 1},
  {&__pyx_n_s_antimeridian_cutting, __pyx_k_antimeridian_cutting, sizeof(__pyx_k_antimeridian_cutting), 0, 0, 1, 1},
  {&__pyx_n_s_antimeridian_offset, __pyx_k_antimeridian_offset, sizeof(__pyx_k_antimeridian_offset), 0, 0, 1, 1},
  {&__pyx_n_s_args, __pyx_k_args, sizeof(__pyx_k_args), 0, 0, 1, 1},
  {&__pyx_n_s_bottom, __pyx_k_bottom, sizeof(__pyx_k_bottom), 0, 0, 1, 1},
  {&__pyx_n_s_calculate_default_transform, __pyx_k_calculate_default_transform, sizeof(__pyx_k_calculate_default_transform), 0, 0, 1, 1},
  {&__pyx_n_s_calculate_default_transform_loc, __pyx_k_calculate_default_transform_loc, sizeof(__pyx_k_calculate_default_transform_loc), 0, 0, 1, 1},
  {&__pyx_n_s_close, __pyx_k_close, sizeof(__pyx_k_close), 0, 0, 1, 1},
  {&__pyx_n_s_col, __pyx_k_col, sizeof(__pyx_k_col), 0, 0, 1, 1},
  {&__pyx_n_s_cols, __pyx_k_cols, sizeof(__pyx_k_cols), 0, 0, 1, 1},
  {&__pyx_n_s_coordinates, __pyx_k_coordinates, sizeof(__pyx_k_coordinates), 0, 0, 1, 1},
  {&__pyx_n_s_crs, __pyx_k_crs, sizeof(__pyx_k_crs), 0, 0, 1, 1},
  {&__pyx_n_s_datasetname, __pyx_k_datasetname, sizeof(__pyx_k_datasetname), 0, 0, 1, 1},
  {&__pyx_n_s_debug, __pyx_k_debug, sizeof(__pyx_k_debug), 0, 0, 1, 1},
  {&__pyx_n_s_destination, __pyx_k_destination, sizeof(__pyx_k_destination), 0, 0, 1, 1},
  {&__pyx_n_s_driver, __pyx_k_driver, sizeof(__pyx_k_driver), 0, 0, 1, 1},
  {&__pyx_n_s_ds, __pyx_k_ds, sizeof(__pyx_k_ds), 0, 0, 1, 1},
  {&__pyx_n_s_dst, __pyx_k_dst, sizeof(__pyx_k_dst), 0, 0, 1, 1},
  {&__pyx_n_s_dst_affine, __pyx_k_dst_affine, sizeof(__pyx_k_dst_affine), 0, 0, 1, 1},
  {&__pyx_n_s_dst_bidx, __pyx_k_dst_bidx, sizeof(__pyx_k_dst_bidx), 0, 0, 1, 1},
  {&__pyx_n_s_dst_crs, __pyx_k_dst_crs, sizeof(__pyx_k_dst_crs), 0, 0, 1, 1},
  {&__pyx_n_s_dst_dataset, __pyx_k_dst_dataset, sizeof(__pyx_k_dst_dataset), 0, 0, 1, 1},
  {&__pyx_n_s_dst_geom, __pyx_k_dst_geom, sizeof(__pyx_k_dst_geom), 0, 0, 1, 1},
  {&__pyx_n_s_dst_height, __pyx_k_dst_height, sizeof(__pyx_k_dst_height), 0, 0, 1, 1},
  {&__pyx_n_s_dst_nodata, __pyx_k_dst_nodata, sizeof(__pyx_k_dst_nodata), 0, 0, 1, 1},
  {&__pyx_kp_s_dst_nodata_must_be_in_valid_rang, __pyx_k_dst_nodata_must_be_in_valid_rang, sizeof(__pyx_k_dst_nodata_must_be_in_valid_rang), 0, 0, 1, 0},
  {&__pyx_n_s_dst_transform, __pyx_k_dst_transform, sizeof(__pyx_k_dst_transform), 0, 0, 1, 1},
  {&__pyx_n_s_dst_width, __pyx_k_dst_width, sizeof(__pyx_k_dst_width), 0, 0, 1, 1},
  {&__pyx_n_s_dstwkt, __pyx_k_dstwkt, sizeof(__pyx_k_dstwkt), 0, 0, 1, 1},
  {&__pyx_n_s_dtype, __pyx_k_dtype, sizeof(__pyx_k_dtype), 0, 0, 1, 1},
  {&__pyx_n_s_dtype_rev, __pyx_k_dtype_rev, sizeof(__pyx_k_dtype_rev), 0, 0, 1, 1},
  {&__pyx_n_s_dtypes, __pyx_k_dtypes, sizeof(__pyx_k_dtypes), 0, 0, 1, 1},
  {&__pyx_n_s_empty, __pyx_k_empty, sizeof(__pyx_k_empty), 0, 0, 1, 1},
  {&__pyx_n_s_encode, __pyx_k_encode, sizeof(__pyx_k_encode), 0, 0, 1, 1},
  {&__pyx_n_s_enter, __pyx_k_enter, sizeof(__pyx_k_enter), 0, 0, 1, 1},
  {&__pyx_n_s_enumerate, __pyx_k_enumerate, sizeof(__pyx_k_enumerate), 0, 0, 1, 1},
  {&__pyx_n_s_err, __pyx_k_err, sizeof(__pyx_k_err), 0, 0, 1, 1},
  {&__pyx_n_s_errmsg, __pyx_k_errmsg, sizeof(__pyx_k_errmsg), 0, 0, 1, 1},
  {&__pyx_n_s_exit, __pyx_k_exit, sizeof(__pyx_k_exit), 0, 0, 1, 1},
  {&__pyx_n_s_extent, __pyx_k_extent, sizeof(__pyx_k_extent), 0, 0, 1, 1},
  {&__pyx_n_s_factory, __pyx_k_factory, sizeof(__pyx_k_factory), 0, 0, 1, 1},
  {&__pyx_n_s_fill_value, __pyx_k_fill_value, sizeof(__pyx_k_fill_value), 0, 0, 1, 1},
  {&__pyx_n_s_format, __pyx_k_format, sizeof(__pyx_k_format), 0, 0, 1, 1},
  {&__pyx_n_s_from_bounds, __pyx_k_from_bounds, sizeof(__pyx_k_from_bounds), 0, 0, 1, 1},
  {&__pyx_n_s_from_gdal, __pyx_k_from_gdal, sizeof(__pyx_k_from_gdal), 0, 0, 1, 1},
  {&__pyx_n_s_gcplist, __pyx_k_gcplist, sizeof(__pyx_k_gcplist), 0, 0, 1, 1},
  {&__pyx_n_s_gcps, __pyx_k_gcps, sizeof(__pyx_k_gcps), 0, 0, 1, 1},
  {&__pyx_n_s_genexpr, __pyx_k_genexpr, sizeof(__pyx_k_genexpr), 0, 0, 1, 1},
  {&__pyx_n_s_geom, __pyx_k_geom, sizeof(__pyx_k_geom), 0, 0, 1, 1},
  {&__pyx_n_s_geotransform, __pyx_k_geotransform, sizeof(__pyx_k_geotransform), 0, 0, 1, 1},
  {&__pyx_n_s_getLogger, __pyx_k_getLogger, sizeof(__pyx_k_getLogger), 0, 0, 1, 1},
  {&__pyx_n_s_gt, __pyx_k_gt, sizeof(__pyx_k_gt), 0, 0, 1, 1},
  {&__pyx_n_s_hTransformArg, __pyx_k_hTransformArg, sizeof(__pyx_k_hTransformArg), 0, 0, 1, 1},
  {&__pyx_n_s_height, __pyx_k_height, sizeof(__pyx_k_height), 0, 0, 1, 1},
  {&__pyx_n_s_i, __pyx_k_i, sizeof(__pyx_k_i), 0, 0, 1, 1},
  {&__pyx_n_s_ident, __pyx_k_ident, sizeof(__pyx_k_ident), 0, 0, 1, 1},
  {&__pyx_n_s_img, __pyx_k_img, sizeof(__pyx_k_img), 0, 0, 1, 1},
  {&__pyx_n_s_imgProjOptions, __pyx_k_imgProjOptions, sizeof(__pyx_k_imgProjOptions), 0, 0, 1, 1},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_s_info, __pyx_k_info, sizeof(__pyx_k_info), 0, 0, 1, 1},
  {&__pyx_n_s_init, __pyx_k_init, sizeof(__pyx_k_init), 0, 0, 1, 1},
  {&__pyx_n_s_init_dest, __pyx_k_init_dest, sizeof(__pyx_k_init_dest), 0, 0, 1, 1},
  {&__pyx_n_s_init_dest_nodata, __pyx_k_init_dest_nodata, sizeof(__pyx_k_init_dest_nodata), 0, 0, 1, 1},
  {&__pyx_n_s_is_ndarray, __pyx_k_is_ndarray, sizeof(__pyx_k_is_ndarray), 0, 0, 1, 1},
  {&__pyx_n_s_items, __pyx_k_items, sizeof(__pyx_k_items), 0, 0, 1, 1},
  {&__pyx_n_s_key, __pyx_k_key, sizeof(__pyx_k_key), 0, 0, 1, 1},
  {&__pyx_n_s_kwargs, __pyx_k_kwargs, sizeof(__pyx_k_kwargs), 0, 0, 1, 1},
  {&__pyx_n_s_left, __pyx_k_left, sizeof(__pyx_k_left), 0, 0, 1, 1},
  {&__pyx_n_s_log, __pyx_k_log, sizeof(__pyx_k_log), 0, 0, 1, 1},
  {&__pyx_n_s_logging, __pyx_k_logging, sizeof(__pyx_k_logging), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_meta, __pyx_k_meta, sizeof(__pyx_k_meta), 0, 0, 1, 1},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_n_s_name_2, __pyx_k_name_2, sizeof(__pyx_k_name_2), 0, 0, 1, 1},
  {&__pyx_kp_u_ndarray_is_not_C_contiguous, __pyx_k_ndarray_is_not_C_contiguous, sizeof(__pyx_k_ndarray_is_not_C_contiguous), 0, 1, 0, 0},
  {&__pyx_kp_u_ndarray_is_not_Fortran_contiguou, __pyx_k_ndarray_is_not_Fortran_contiguou, sizeof(__pyx_k_ndarray_is_not_Fortran_contiguou), 0, 1, 0, 0},
  {&__pyx_n_s_nearest, __pyx_k_nearest, sizeof(__pyx_k_nearest), 0, 0, 1, 1},
  {&__pyx_n_s_nlines, __pyx_k_nlines, sizeof(__pyx_k_nlines), 0, 0, 1, 1},
  {&__pyx_n_s_nodata, __pyx_k_nodata, sizeof(__pyx_k_nodata), 0, 0, 1, 1},
  {&__pyx_n_s_none, __pyx_k_none, sizeof(__pyx_k_none), 0, 0, 1, 1},
  {&__pyx_n_s_np, __pyx_k_np, sizeof(__pyx_k_np), 0, 0, 1, 1},
  {&__pyx_n_s_npixels, __pyx_k_npixels, sizeof(__pyx_k_npixels), 0, 0, 1, 1},
  {&__pyx_n_s_num_threads, __pyx_k_num_threads, sizeof(__pyx_k_num_threads), 0, 0, 1, 1},
  {&__pyx_n_s_numpy, __pyx_k_numpy, sizeof(__pyx_k_numpy), 0, 0, 1, 1},
  {&__pyx_kp_s_numpy_core_multiarray_failed_to, __pyx_k_numpy_core_multiarray_failed_to, sizeof(__pyx_k_numpy_core_multiarray_failed_to), 0, 0, 1, 0},
  {&__pyx_kp_s_numpy_core_umath_failed_to_impor, __pyx_k_numpy_core_umath_failed_to_impor, sizeof(__pyx_k_numpy_core_umath_failed_to_impor), 0, 0, 1, 0},
  {&__pyx_n_s_oWarper, __pyx_k_oWarper, sizeof(__pyx_k_oWarper), 0, 0, 1, 1},
  {&__pyx_n_s_obj, __pyx_k_obj, sizeof(__pyx_k_obj), 0, 0, 1, 1},
  {&__pyx_n_s_options, __pyx_k_options, sizeof(__pyx_k_options), 0, 0, 1, 1},
  {&__pyx_n_s_osr, __pyx_k_osr, sizeof(__pyx_k_osr), 0, 0, 1, 1},
  {&__pyx_n_s_parse_path, __pyx_k_parse_path, sizeof(__pyx_k_parse_path), 0, 0, 1, 1},
  {&__pyx_n_s_part, __pyx_k_part, sizeof(__pyx_k_part), 0, 0, 1, 1},
  {&__pyx_n_s_pfnTransformer, __pyx_k_pfnTransformer, sizeof(__pyx_k_pfnTransformer), 0, 0, 1, 1},
  {&__pyx_n_s_precision, __pyx_k_precision, sizeof(__pyx_k_precision), 0, 0, 1, 1},
  {&__pyx_n_s_psWOptions, __pyx_k_psWOptions, sizeof(__pyx_k_psWOptions), 0, 0, 1, 1},
  {&__pyx_n_s_pszWarpThread, __pyx_k_pszWarpThread, sizeof(__pyx_k_pszWarpThread), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
  {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
  {&__pyx_n_s_rasterio, __pyx_k_rasterio, sizeof(__pyx_k_rasterio), 0, 0, 1, 1},
  {&__pyx_n_s_rasterio__err, __pyx_k_rasterio__err, sizeof(__pyx_k_rasterio__err), 0, 0, 1, 1},
  {&__pyx_n_s_rasterio__warp, __pyx_k_rasterio__warp, sizeof(__pyx_k_rasterio__warp), 0, 0, 1, 1},
  {&__pyx_n_s_rasterio_control, __pyx_k_rasterio_control, sizeof(__pyx_k_rasterio_control), 0, 0, 1, 1},
  {&__pyx_n_s_rasterio_enums, __pyx_k_rasterio_enums, sizeof(__pyx_k_rasterio_enums), 0, 0, 1, 1},
  {&__pyx_n_s_rasterio_errors, __pyx_k_rasterio_errors, sizeof(__pyx_k_rasterio_errors), 0, 0, 1, 1},
  {&__pyx_n_s_rasterio_transform, __pyx_k_rasterio_transform, sizeof(__pyx_k_rasterio_transform), 0, 0, 1, 1},
  {&__pyx_n_s_rasterio_vfs, __pyx_k_rasterio_vfs, sizeof(__pyx_k_rasterio_vfs), 0, 0, 1, 1},
  {&__pyx_n_s_rdr, __pyx_k_rdr, sizeof(__pyx_k_rdr), 0, 0, 1, 1},
  {&__pyx_n_s_read_crs, __pyx_k_read_crs, sizeof(__pyx_k_read_crs), 0, 0, 1, 1},
  {&__pyx_n_s_read_transform, __pyx_k_read_transform, sizeof(__pyx_k_read_transform), 0, 0, 1, 1},
  {&__pyx_n_s_recursive_round, __pyx_k_recursive_round, sizeof(__pyx_k_recursive_round), 0, 0, 1, 1},
  {&__pyx_n_s_reproject, __pyx_k_reproject, sizeof(__pyx_k_reproject), 0, 0, 1, 1},
  {&__pyx_n_s_resampling, __pyx_k_resampling, sizeof(__pyx_k_resampling), 0, 0, 1, 1},
  {&__pyx_n_s_reshape, __pyx_k_reshape, sizeof(__pyx_k_reshape), 0, 0, 1, 1},
  {&__pyx_n_s_result, __pyx_k_result, sizeof(__pyx_k_result), 0, 0, 1, 1},
  {&__pyx_n_s_retval, __pyx_k_retval, sizeof(__pyx_k_retval), 0, 0, 1, 1},
  {&__pyx_n_s_right, __pyx_k_right, sizeof(__pyx_k_right), 0, 0, 1, 1},
  {&__pyx_n_s_round, __pyx_k_round, sizeof(__pyx_k_round), 0, 0, 1, 1},
  {&__pyx_n_s_row, __pyx_k_row, sizeof(__pyx_k_row), 0, 0, 1, 1},
  {&__pyx_n_s_rows, __pyx_k_rows, sizeof(__pyx_k_rows), 0, 0, 1, 1},
  {&__pyx_n_s_send, __pyx_k_send, sizeof(__pyx_k_send), 0, 0, 1, 1},
  {&__pyx_n_s_shape, __pyx_k_shape, sizeof(__pyx_k_shape), 0, 0, 1, 1},
  {&__pyx_n_s_source, __pyx_k_source, sizeof(__pyx_k_source), 0, 0, 1, 1},
  {&__pyx_n_s_src, __pyx_k_src, sizeof(__pyx_k_src), 0, 0, 1, 1},
  {&__pyx_n_s_src_bidx, __pyx_k_src_bidx, sizeof(__pyx_k_src_bidx), 0, 0, 1, 1},
  {&__pyx_n_s_src_count, __pyx_k_src_count, sizeof(__pyx_k_src_count), 0, 0, 1, 1},
  {&__pyx_n_s_src_crs, __pyx_k_src_crs, sizeof(__pyx_k_src_crs), 0, 0, 1, 1},
  {&__pyx_n_s_src_dataset, __pyx_k_src_dataset, sizeof(__pyx_k_src_dataset), 0, 0, 1, 1},
  {&__pyx_n_s_src_geom, __pyx_k_src_geom, sizeof(__pyx_k_src_geom), 0, 0, 1, 1},
  {&__pyx_n_s_src_nodata, __pyx_k_src_nodata, sizeof(__pyx_k_src_nodata), 0, 0, 1, 1},
  {&__pyx_kp_s_src_nodata_must_be_in_valid_rang, __pyx_k_src_nodata_must_be_in_valid_rang, sizeof(__pyx_k_src_nodata_must_be_in_valid_rang), 0, 0, 1, 0},
  {&__pyx_kp_s_src_nodata_must_be_provided_beca, __pyx_k_src_nodata_must_be_provided_beca, sizeof(__pyx_k_src_nodata_must_be_provided_beca), 0, 0, 1, 0},
  {&__pyx_n_s_src_transform, __pyx_k_src_transform, sizeof(__pyx_k_src_transform), 0, 0, 1, 1},
  {&__pyx_n_s_srcwkt, __pyx_k_srcwkt, sizeof(__pyx_k_srcwkt), 0, 0, 1, 1},
  {&__pyx_n_s_super, __pyx_k_super, sizeof(__pyx_k_super), 0, 0, 1, 1},
  {&__pyx_n_s_tastes_like_gdal, __pyx_k_tastes_like_gdal, sizeof(__pyx_k_tastes_like_gdal), 0, 0, 1, 1},
  {&__pyx_n_s_temp, __pyx_k_temp, sizeof(__pyx_k_temp), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_n_s_throw, __pyx_k_throw, sizeof(__pyx_k_throw), 0, 0, 1, 1},
  {&__pyx_n_s_to_gdal, __pyx_k_to_gdal, sizeof(__pyx_k_to_gdal), 0, 0, 1, 1},
  {&__pyx_n_s_tolerance, __pyx_k_tolerance, sizeof(__pyx_k_tolerance), 0, 0, 1, 1},
  {&__pyx_n_s_top, __pyx_k_top, sizeof(__pyx_k_top), 0, 0, 1, 1},
  {&__pyx_n_s_transform, __pyx_k_transform, sizeof(__pyx_k_transform), 0, 0, 1, 1},
  {&__pyx_n_s_transform_geom, __pyx_k_transform_geom, sizeof(__pyx_k_transform_geom), 0, 0, 1, 1},
  {&__pyx_n_s_udr, __pyx_k_udr, sizeof(__pyx_k_udr), 0, 0, 1, 1},
  {&__pyx_kp_u_unknown_dtype_code_in_numpy_pxd, __pyx_k_unknown_dtype_code_in_numpy_pxd, sizeof(__pyx_k_unknown_dtype_code_in_numpy_pxd), 0, 1, 0, 0},
  {&__pyx_n_s_upper, __pyx_k_upper, sizeof(__pyx_k_upper), 0, 0, 1, 1},
  {&__pyx_kp_s_utf_8, __pyx_k_utf_8, sizeof(__pyx_k_utf_8), 0, 0, 1, 0},
  {&__pyx_n_s_uuid, __pyx_k_uuid, sizeof(__pyx_k_uuid), 0, 0, 1, 1},
  {&__pyx_n_s_uuid4, __pyx_k_uuid4, sizeof(__pyx_k_uuid4), 0, 0, 1, 1},
  {&__pyx_n_s_val, __pyx_k_val, sizeof(__pyx_k_val), 0, 0, 1, 1},
  {&__pyx_n_s_valb, __pyx_k_valb, sizeof(__pyx_k_valb), 0, 0, 1, 1},
  {&__pyx_n_s_vsi_path, __pyx_k_vsi_path, sizeof(__pyx_k_vsi_path), 0, 0, 1, 1},
  {&__pyx_n_s_warp_extras, __pyx_k_warp_extras, sizeof(__pyx_k_warp_extras), 0, 0, 1, 1},
  {&__pyx_n_s_width, __pyx_k_width, sizeof(__pyx_k_width), 0, 0, 1, 1},
  {&__pyx_n_s_wkt, __pyx_k_wkt, sizeof(__pyx_k_wkt), 0, 0, 1, 1},
  {&__pyx_n_s_x, __pyx_k_x, sizeof(__pyx_k_x), 0, 0, 1, 1},
  {&__pyx_n_s_y, __pyx_k_y, sizeof(__pyx_k_y), 0, 0, 1, 1},
  {&__pyx_n_s_z, __pyx_k_z, sizeof(__pyx_k_z), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_round = __Pyx_GetBuiltinName(__pyx_n_s_round); if (!__pyx_builtin_round) __PYX_ERR(0, 36, __pyx_L1_error)
  __pyx_builtin_ValueError = __Pyx_GetBuiltinName(__pyx_n_s_ValueError); if (!__pyx_builtin_ValueError) __PYX_ERR(0, 110, __pyx_L1_error)
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(0, 147, __pyx_L1_error)
  __pyx_builtin_enumerate = __Pyx_GetBuiltinName(__pyx_n_s_enumerate); if (!__pyx_builtin_enumerate) __PYX_ERR(0, 342, __pyx_L1_error)
  __pyx_builtin_super = __Pyx_GetBuiltinName(__pyx_n_s_super); if (!__pyx_builtin_super) __PYX_ERR(0, 637, __pyx_L1_error)
  __pyx_builtin_RuntimeError = __Pyx_GetBuiltinName(__pyx_n_s_RuntimeError); if (!__pyx_builtin_RuntimeError) __PYX_ERR(1, 799, __pyx_L1_error)
  __pyx_builtin_ImportError = __Pyx_GetBuiltinName(__pyx_n_s_ImportError); if (!__pyx_builtin_ImportError) __PYX_ERR(1, 989, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "rasterio/_warp.pyx":65
 * 
 *     # Transform options.
 *     valb = str(antimeridian_offset).encode('utf-8')             # <<<<<<<<<<<<<<
 *     options = CSLSetNameValue(options, "DATELINEOFFSET", <const char *>valb)
 *     if antimeridian_cutting:
 */
  __pyx_tuple_ = PyTuple_Pack(1, __pyx_kp_s_utf_8); if (unlikely(!__pyx_tuple_)) __PYX_ERR(0, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple_);
  __Pyx_GIVEREF(__pyx_tuple_);

  /* "rasterio/_warp.pyx":110
 *     # We require source nodata if destination nodata is passed.
 *     if src_nodata is None and dst_nodata is not None:
 *         raise ValueError("src_nodata must be provided because dst_nodata "             # <<<<<<<<<<<<<<
 *                          "is not None")
 * 
 */
  __pyx_tuple__2 = PyTuple_Pack(1, __pyx_kp_s_src_nodata_must_be_provided_beca); if (unlikely(!__pyx_tuple__2)) __PYX_ERR(0, 110, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2);
  __Pyx_GIVEREF(__pyx_tuple__2);

  /* "rasterio/_warp.pyx":277
 *     if src_nodata is not None:
 *         if not in_dtype_range(src_nodata, source.dtype):
 *             raise ValueError("src_nodata must be in valid range for "             # <<<<<<<<<<<<<<
 *                              "source dtype")
 * 
 */
  __pyx_tuple__4 = PyTuple_Pack(1, __pyx_kp_s_src_nodata_must_be_in_valid_rang); if (unlikely(!__pyx_tuple__4)) __PYX_ERR(0, 277, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__4);
  __Pyx_GIVEREF(__pyx_tuple__4);

  /* "rasterio/_warp.pyx":282
 *     if dst_nodata is not None:
 *         if not in_dtype_range(dst_nodata, destination.dtype):
 *             raise ValueError("dst_nodata must be in valid range for "             # <<<<<<<<<<<<<<
 *                              "destination dtype")
 * 
 */
  __pyx_tuple__5 = PyTuple_Pack(1, __pyx_kp_s_dst_nodata_must_be_in_valid_rang); if (unlikely(!__pyx_tuple__5)) __PYX_ERR(0, 282, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__5);
  __Pyx_GIVEREF(__pyx_tuple__5);

  /* "rasterio/_warp.pyx":291
 *         # Convert 2D single-band arrays to 3D multi-band.
 *         if len(source.shape) == 2:
 *             source = source.reshape(1, *source.shape)             # <<<<<<<<<<<<<<
 *         src_count = source.shape[0]
 *         src_bidx = range(1, src_count + 1)
 */
  __pyx_tuple__6 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__6)) __PYX_ERR(0, 291, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__6);
  __Pyx_GIVEREF(__pyx_tuple__6);

  /* "rasterio/_warp.pyx":305
 *             driver = exc_wrap_pointer(GDALGetDriverByName("MEM"))
 *         except:
 *             raise DriverRegistrationError(             # <<<<<<<<<<<<<<
 *                 "'MEM' driver not found. Check that this call is contained "
 *                 "in a `with rasterio.Env()` or `with rasterio.open()` "
 */
  __pyx_tuple__7 = PyTuple_Pack(1, __pyx_kp_s_MEM_driver_not_found_Check_that); if (unlikely(!__pyx_tuple__7)) __PYX_ERR(0, 305, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__7);
  __Pyx_GIVEREF(__pyx_tuple__7);

  /* "rasterio/_warp.pyx":310
 *                 "block.")
 * 
 *         datasetname = str(uuid.uuid4()).encode('utf-8')             # <<<<<<<<<<<<<<
 *         src_dataset = exc_wrap_pointer(
 *             GDALCreate(driver, <const char *>datasetname, cols, rows,
 */
  __pyx_tuple__8 = PyTuple_Pack(1, __pyx_kp_s_utf_8); if (unlikely(!__pyx_tuple__8)) __PYX_ERR(0, 310, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__8);
  __Pyx_GIVEREF(__pyx_tuple__8);

  /* "rasterio/_warp.pyx":318
 *             src_dataset, "Temporary source dataset for _reproject()")
 * 
 *         log.debug("Created temp source dataset")             # <<<<<<<<<<<<<<
 * 
 *         if src_transform:
 */
  __pyx_tuple__9 = PyTuple_Pack(1, __pyx_kp_s_Created_temp_source_dataset); if (unlikely(!__pyx_tuple__9)) __PYX_ERR(0, 318, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__9);
  __Pyx_GIVEREF(__pyx_tuple__9);

  /* "rasterio/_warp.pyx":343
 *             try:
 *                 for i, obj in enumerate(gcps):
 *                     ident = str(i).encode('utf-8')             # <<<<<<<<<<<<<<
 *                     info = "".encode('utf-8')
 *                     gcplist[i].pszId = ident
 */
  __pyx_tuple__10 = PyTuple_Pack(1, __pyx_kp_s_utf_8); if (unlikely(!__pyx_tuple__10)) __PYX_ERR(0, 343, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__10);
  __Pyx_GIVEREF(__pyx_tuple__10);

  /* "rasterio/_warp.pyx":344
 *                 for i, obj in enumerate(gcps):
 *                     ident = str(i).encode('utf-8')
 *                     info = "".encode('utf-8')             # <<<<<<<<<<<<<<
 *                     gcplist[i].pszId = ident
 *                     gcplist[i].pszInfo = info
 */
  __pyx_tuple__12 = PyTuple_Pack(1, __pyx_kp_s_utf_8); if (unlikely(!__pyx_tuple__12)) __PYX_ERR(0, 344, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__12);
  __Pyx_GIVEREF(__pyx_tuple__12);

  /* "rasterio/_warp.pyx":362
 *         # TODO: handle errors (by retval).
 * 
 *         log.debug("Wrote array to temp source dataset")             # <<<<<<<<<<<<<<
 * 
 *     # If the source is a rasterio MultiBand, no copy necessary.
 */
  __pyx_tuple__13 = PyTuple_Pack(1, __pyx_kp_s_Wrote_array_to_temp_source_datas); if (unlikely(!__pyx_tuple__13)) __PYX_ERR(0, 362, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__13);
  __Pyx_GIVEREF(__pyx_tuple__13);

  /* "rasterio/_warp.pyx":376
 *             src_nodata = rdr.nodata
 *     else:
 *         raise ValueError("Invalid source")             # <<<<<<<<<<<<<<
 * 
 *     # Next, do the same for the destination raster.
 */
  __pyx_tuple__14 = PyTuple_Pack(1, __pyx_kp_s_Invalid_source); if (unlikely(!__pyx_tuple__14)) __PYX_ERR(0, 376, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__14);
  __Pyx_GIVEREF(__pyx_tuple__14);

  /* "rasterio/_warp.pyx":381
 *     if dtypes.is_ndarray(destination):
 *         if len(destination.shape) == 2:
 *             destination = destination.reshape(1, *destination.shape)             # <<<<<<<<<<<<<<
 *             dst_bidx = [1]
 *         else:
 */
  __pyx_tuple__15 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__15)) __PYX_ERR(0, 381, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__15);
  __Pyx_GIVEREF(__pyx_tuple__15);

  /* "rasterio/_warp.pyx":387
 * 
 *         if destination.shape[0] != src_count:
 *             raise ValueError("Destination's shape is invalid")             # <<<<<<<<<<<<<<
 * 
 *         try:
 */
  __pyx_tuple__16 = PyTuple_Pack(1, __pyx_kp_s_Destination_s_shape_is_invalid); if (unlikely(!__pyx_tuple__16)) __PYX_ERR(0, 387, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__16);
  __Pyx_GIVEREF(__pyx_tuple__16);

  /* "rasterio/_warp.pyx":392
 *             driver = exc_wrap_pointer(GDALGetDriverByName("MEM"))
 *         except:
 *             raise DriverRegistrationError(             # <<<<<<<<<<<<<<
 *                 "'MEM' driver not found. Check that this call is contained "
 *                 "in a `with rasterio.Env()` or `with rasterio.open()` "
 */
  __pyx_tuple__17 = PyTuple_Pack(1, __pyx_kp_s_MEM_driver_not_found_Check_that); if (unlikely(!__pyx_tuple__17)) __PYX_ERR(0, 392, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__17);
  __Pyx_GIVEREF(__pyx_tuple__17);

  /* "rasterio/_warp.pyx":399
 *         _, rows, cols = destination.shape
 * 
 *         datasetname = str(uuid.uuid4()).encode('utf-8')             # <<<<<<<<<<<<<<
 *         dst_dataset = exc_wrap_pointer(
 *             GDALCreate(driver, <const char *>datasetname, cols, rows,
 */
  __pyx_tuple__18 = PyTuple_Pack(1, __pyx_kp_s_utf_8); if (unlikely(!__pyx_tuple__18)) __PYX_ERR(0, 399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__18);
  __Pyx_GIVEREF(__pyx_tuple__18);

  /* "rasterio/_warp.pyx":408
 *             dst_dataset, "Temporary destination dataset for _reproject()")
 * 
 *         log.debug("Created temp destination dataset.")             # <<<<<<<<<<<<<<
 * 
 *         for i in range(6):
 */
  __pyx_tuple__19 = PyTuple_Pack(1, __pyx_kp_s_Created_temp_destination_dataset); if (unlikely(!__pyx_tuple__19)) __PYX_ERR(0, 408, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__19);
  __Pyx_GIVEREF(__pyx_tuple__19);

  /* "rasterio/_warp.pyx":414
 * 
 *         if not GDALError.none == GDALSetGeoTransform(dst_dataset, gt):
 *             raise ValueError(             # <<<<<<<<<<<<<<
 *                 "Failed to set transform on temp destination dataset.")
 * 
 */
  __pyx_tuple__20 = PyTuple_Pack(1, __pyx_kp_s_Failed_to_set_transform_on_temp); if (unlikely(!__pyx_tuple__20)) __PYX_ERR(0, 414, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__20);
  __Pyx_GIVEREF(__pyx_tuple__20);

  /* "rasterio/_warp.pyx":433
 *         retval = io_auto(destination, dst_dataset, 1)
 * 
 *         log.debug("Wrote array to temp output dataset")             # <<<<<<<<<<<<<<
 * 
 *         if dst_nodata is None and hasattr(destination, "fill_value"):
 */
  __pyx_tuple__21 = PyTuple_Pack(1, __pyx_kp_s_Wrote_array_to_temp_output_datas); if (unlikely(!__pyx_tuple__21)) __PYX_ERR(0, 433, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__21);
  __Pyx_GIVEREF(__pyx_tuple__21);

  /* "rasterio/_warp.pyx":448
 *             dst_nodata = udr.nodata
 *     else:
 *         raise ValueError("Invalid destination")             # <<<<<<<<<<<<<<
 * 
 *     # Set up GDALCreateGenImgProjTransformer2 keyword arguments.
 */
  __pyx_tuple__22 = PyTuple_Pack(1, __pyx_kp_s_Invalid_destination); if (unlikely(!__pyx_tuple__22)) __PYX_ERR(0, 448, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__22);
  __Pyx_GIVEREF(__pyx_tuple__22);

  /* "rasterio/_warp.pyx":460
 *     # okay.
 *     for key, val in kwargs.items():
 *         key = key.upper().encode('utf-8')             # <<<<<<<<<<<<<<
 *         val = str(val).upper().encode('utf-8')
 *         imgProjOptions = CSLSetNameValue(
 */
  __pyx_tuple__23 = PyTuple_Pack(1, __pyx_kp_s_utf_8); if (unlikely(!__pyx_tuple__23)) __PYX_ERR(0, 460, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__23);
  __Pyx_GIVEREF(__pyx_tuple__23);

  /* "rasterio/_warp.pyx":461
 *     for key, val in kwargs.items():
 *         key = key.upper().encode('utf-8')
 *         val = str(val).upper().encode('utf-8')             # <<<<<<<<<<<<<<
 *         imgProjOptions = CSLSetNameValue(
 *             imgProjOptions, <const char *>key, <const char *>val)
 */
  __pyx_tuple__24 = PyTuple_Pack(1, __pyx_kp_s_utf_8); if (unlikely(!__pyx_tuple__24)) __PYX_ERR(0, 461, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__24);
  __Pyx_GIVEREF(__pyx_tuple__24);

  /* "rasterio/_warp.pyx":475
 *         GDALApproxTransformerOwnsSubtransformer(hTransformArg, 1)
 * 
 *         log.debug("Created transformer and options.")             # <<<<<<<<<<<<<<
 * 
 *     except:
 */
  __pyx_tuple__25 = PyTuple_Pack(1, __pyx_kp_s_Created_transformer_and_options); if (unlikely(!__pyx_tuple__25)) __PYX_ERR(0, 475, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__25);
  __Pyx_GIVEREF(__pyx_tuple__25);

  /* "rasterio/_warp.pyx":487
 *     # warp_extras = psWOptions.papszWarpOptions
 * 
 *     valb = str(num_threads).encode('utf-8')             # <<<<<<<<<<<<<<
 *     warp_extras = CSLSetNameValue(warp_extras, "NUM_THREADS", <const char *>valb)
 * 
 */
  __pyx_tuple__26 = PyTuple_Pack(1, __pyx_kp_s_utf_8); if (unlikely(!__pyx_tuple__26)) __PYX_ERR(0, 487, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__26);
  __Pyx_GIVEREF(__pyx_tuple__26);

  /* "rasterio/_warp.pyx":499
 *     # is fine.
 *     for key, val in kwargs.items():
 *         key = key.upper().encode('utf-8')             # <<<<<<<<<<<<<<
 *         val = str(val).upper().encode('utf-8')
 *         warp_extras = CSLSetNameValue(
 */
  __pyx_tuple__27 = PyTuple_Pack(1, __pyx_kp_s_utf_8); if (unlikely(!__pyx_tuple__27)) __PYX_ERR(0, 499, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__27);
  __Pyx_GIVEREF(__pyx_tuple__27);

  /* "rasterio/_warp.pyx":500
 *     for key, val in kwargs.items():
 *         key = key.upper().encode('utf-8')
 *         val = str(val).upper().encode('utf-8')             # <<<<<<<<<<<<<<
 *         warp_extras = CSLSetNameValue(
 *             warp_extras, <const char *>key, <const char *>val)
 */
  __pyx_tuple__28 = PyTuple_Pack(1, __pyx_kp_s_utf_8); if (unlikely(!__pyx_tuple__28)) __PYX_ERR(0, 500, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__28);
  __Pyx_GIVEREF(__pyx_tuple__28);

  /* "rasterio/_warp.pyx":520
 *         psWOptions.panDstBands[i] = dst_bidx[i]
 * 
 *     log.debug("Set transformer options")             # <<<<<<<<<<<<<<
 * 
 *     # Now that the transformer and warp options are set up, we init
 */
  __pyx_tuple__29 = PyTuple_Pack(1, __pyx_kp_s_Set_transformer_options); if (unlikely(!__pyx_tuple__29)) __PYX_ERR(0, 520, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__29);
  __Pyx_GIVEREF(__pyx_tuple__29);

  /* "rasterio/_warp.pyx":527
 *     try:
 *         exc_wrap_int(oWarper.Initialize(psWOptions))
 *         rows, cols = destination.shape[-2:]             # <<<<<<<<<<<<<<
 * 
 *         log.debug(
 */
  __pyx_slice__30 = PySlice_New(__pyx_int_neg_2, Py_None, Py_None); if (unlikely(!__pyx_slice__30)) __PYX_ERR(0, 527, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__30);
  __Pyx_GIVEREF(__pyx_slice__30);

  /* "rasterio/_warp.pyx":580
 *         transform = transform.to_gdal()
 *     elif any(x is not None for x in (left, bottom, right, top)):
 *         raise ValueError(             # <<<<<<<<<<<<<<
 *             "Some, but not all, bounding box parameters were provided.")
 *     else:
 */
  __pyx_tuple__31 = PyTuple_Pack(1, __pyx_kp_s_Some_but_not_all_bounding_box_pa); if (unlikely(!__pyx_tuple__31)) __PYX_ERR(0, 580, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__31);
  __Pyx_GIVEREF(__pyx_tuple__31);

  /* "rasterio/_warp.pyx":602
 *                     geotransform, &npixels, &nlines, extent, 0))
 * 
 *             log.debug("Created transformer and warp output.")             # <<<<<<<<<<<<<<
 * 
 *         except CPLE_NotSupportedError as err:
 */
  __pyx_tuple__32 = PyTuple_Pack(1, __pyx_kp_s_Created_transformer_and_warp_out); if (unlikely(!__pyx_tuple__32)) __PYX_ERR(0, 602, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__32);
  __Pyx_GIVEREF(__pyx_tuple__32);

  /* "rasterio/_warp.pyx":612
 *                 # "Reprojection failed, err = -14, further errors will be
 *                 # suppressed on the transform object."
 *                 log.info("Encountered points outside of valid dst crs region")             # <<<<<<<<<<<<<<
 *                 pass
 *             else:
 */
  __pyx_tuple__33 = PyTuple_Pack(1, __pyx_kp_s_Encountered_points_outside_of_va); if (unlikely(!__pyx_tuple__33)) __PYX_ERR(0, 612, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__33);
  __Pyx_GIVEREF(__pyx_tuple__33);

  /* "rasterio/_warp.pyx":591
 *     osr = NULL
 * 
 *     with InMemoryRaster(width=width, height=height, transform=transform,             # <<<<<<<<<<<<<<
 *                         gcps=gcps, crs=src_crs) as temp:
 *         try:
 */
  __pyx_tuple__34 = PyTuple_Pack(3, Py_None, Py_None, Py_None); if (unlikely(!__pyx_tuple__34)) __PYX_ERR(0, 591, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__34);
  __Pyx_GIVEREF(__pyx_tuple__34);

  /* "rasterio/_warp.pyx":675
 *             osr = NULL
 * 
 *         log.debug("Exported CRS to WKT.")             # <<<<<<<<<<<<<<
 * 
 *         hds = (<DatasetReaderBase?>self.src_dataset).handle()
 */
  __pyx_tuple__36 = PyTuple_Pack(1, __pyx_kp_s_Exported_CRS_to_WKT); if (unlikely(!__pyx_tuple__36)) __PYX_ERR(0, 675, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__36);
  __Pyx_GIVEREF(__pyx_tuple__36);

  /* "rasterio/_warp.pyx":688
 * 
 *         for key, val in self.warp_extras.items():
 *             key = key.upper().encode('utf-8')             # <<<<<<<<<<<<<<
 *             val = str(val).upper().encode('utf-8')
 *             c_warp_extras = CSLSetNameValue(
 */
  __pyx_tuple__37 = PyTuple_Pack(1, __pyx_kp_s_utf_8); if (unlikely(!__pyx_tuple__37)) __PYX_ERR(0, 688, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__37);
  __Pyx_GIVEREF(__pyx_tuple__37);

  /* "rasterio/_warp.pyx":689
 *         for key, val in self.warp_extras.items():
 *             key = key.upper().encode('utf-8')
 *             val = str(val).upper().encode('utf-8')             # <<<<<<<<<<<<<<
 *             c_warp_extras = CSLSetNameValue(
 *                 c_warp_extras, <const char *>key, <const char *>val)
 */
  __pyx_tuple__38 = PyTuple_Pack(1, __pyx_kp_s_utf_8); if (unlikely(!__pyx_tuple__38)) __PYX_ERR(0, 689, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__38);
  __Pyx_GIVEREF(__pyx_tuple__38);

  /* "rasterio/_warp.pyx":721
 *                             hTransformArg, 1)
 * 
 *                     log.debug("Created transformer and options.")             # <<<<<<<<<<<<<<
 *                     psWOptions.pTransformerArg = hTransformArg
 *                 except Exception:
 */
  __pyx_tuple__39 = PyTuple_Pack(1, __pyx_kp_s_Created_transformer_and_options); if (unlikely(!__pyx_tuple__39)) __PYX_ERR(0, 721, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__39);
  __Pyx_GIVEREF(__pyx_tuple__39);

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":218
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_C_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not C contiguous")             # <<<<<<<<<<<<<<
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 */
  __pyx_tuple__40 = PyTuple_Pack(1, __pyx_kp_u_ndarray_is_not_C_contiguous); if (unlikely(!__pyx_tuple__40)) __PYX_ERR(1, 218, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__40);
  __Pyx_GIVEREF(__pyx_tuple__40);

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":222
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_F_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not Fortran contiguous")             # <<<<<<<<<<<<<<
 * 
 *             info.buf = PyArray_DATA(self)
 */
  __pyx_tuple__41 = PyTuple_Pack(1, __pyx_kp_u_ndarray_is_not_Fortran_contiguou); if (unlikely(!__pyx_tuple__41)) __PYX_ERR(1, 222, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__41);
  __Pyx_GIVEREF(__pyx_tuple__41);

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":259
 *                 if ((descr.byteorder == c'>' and little_endian) or
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")             # <<<<<<<<<<<<<<
 *                 if   t == NPY_BYTE:        f = "b"
 *                 elif t == NPY_UBYTE:       f = "B"
 */
  __pyx_tuple__42 = PyTuple_Pack(1, __pyx_kp_u_Non_native_byte_order_not_suppor); if (unlikely(!__pyx_tuple__42)) __PYX_ERR(1, 259, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__42);
  __Pyx_GIVEREF(__pyx_tuple__42);

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":799
 * 
 *         if (end - f) - <int>(new_offset - offset[0]) < 15:
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")             # <<<<<<<<<<<<<<
 * 
 *         if ((child.byteorder == c'>' and little_endian) or
 */
  __pyx_tuple__43 = PyTuple_Pack(1, __pyx_kp_u_Format_string_allocated_too_shor); if (unlikely(!__pyx_tuple__43)) __PYX_ERR(1, 799, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__43);
  __Pyx_GIVEREF(__pyx_tuple__43);

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":803
 *         if ((child.byteorder == c'>' and little_endian) or
 *             (child.byteorder == c'<' and not little_endian)):
 *             raise ValueError(u"Non-native byte order not supported")             # <<<<<<<<<<<<<<
 *             # One could encode it in the format string and have Cython
 *             # complain instead, BUT: < and > in format strings also imply
 */
  __pyx_tuple__44 = PyTuple_Pack(1, __pyx_kp_u_Non_native_byte_order_not_suppor); if (unlikely(!__pyx_tuple__44)) __PYX_ERR(1, 803, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__44);
  __Pyx_GIVEREF(__pyx_tuple__44);

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":823
 *             t = child.type_num
 *             if end - f < 5:
 *                 raise RuntimeError(u"Format string allocated too short.")             # <<<<<<<<<<<<<<
 * 
 *             # Until ticket #99 is fixed, use integers to avoid warnings
 */
  __pyx_tuple__45 = PyTuple_Pack(1, __pyx_kp_u_Format_string_allocated_too_shor_2); if (unlikely(!__pyx_tuple__45)) __PYX_ERR(1, 823, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__45);
  __Pyx_GIVEREF(__pyx_tuple__45);

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":989
 *         _import_array()
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_umath() except -1:
 */
  __pyx_tuple__46 = PyTuple_Pack(1, __pyx_kp_s_numpy_core_multiarray_failed_to); if (unlikely(!__pyx_tuple__46)) __PYX_ERR(1, 989, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__46);
  __Pyx_GIVEREF(__pyx_tuple__46);

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":995
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_ufunc() except -1:
 */
  __pyx_tuple__47 = PyTuple_Pack(1, __pyx_kp_s_numpy_core_umath_failed_to_impor); if (unlikely(!__pyx_tuple__47)) __PYX_ERR(1, 995, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__47);
  __Pyx_GIVEREF(__pyx_tuple__47);

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":1001
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__48 = PyTuple_Pack(1, __pyx_kp_s_numpy_core_umath_failed_to_impor); if (unlikely(!__pyx_tuple__48)) __PYX_ERR(1, 1001, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__48);
  __Pyx_GIVEREF(__pyx_tuple__48);

  /* "rasterio/_warp.pyx":33
 * 
 * 
 * def recursive_round(val, precision):             # <<<<<<<<<<<<<<
 *     """Recursively round coordinates."""
 *     if isinstance(val, (int, float)):
 */
  __pyx_tuple__49 = PyTuple_Pack(3, __pyx_n_s_val, __pyx_n_s_precision, __pyx_n_s_part); if (unlikely(!__pyx_tuple__49)) __PYX_ERR(0, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__49);
  __Pyx_GIVEREF(__pyx_tuple__49);
  __pyx_codeobj__50 = (PyObject*)__Pyx_PyCode_New(2, 0, 3, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__49, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Users_sean_code_frs_wheel_build, __pyx_n_s_recursive_round, 33, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__50)) __PYX_ERR(0, 33, __pyx_L1_error)

  /* "rasterio/_warp.pyx":41
 * 
 * 
 * def _transform_geom(             # <<<<<<<<<<<<<<
 *         src_crs, dst_crs, geom, antimeridian_cutting, antimeridian_offset,
 *         precision):
 */
  __pyx_tuple__51 = PyTuple_Pack(16, __pyx_n_s_src_crs, __pyx_n_s_dst_crs, __pyx_n_s_geom, __pyx_n_s_antimeridian_cutting, __pyx_n_s_antimeridian_offset, __pyx_n_s_precision, __pyx_n_s_options, __pyx_n_s_src, __pyx_n_s_dst, __pyx_n_s_transform, __pyx_n_s_factory, __pyx_n_s_src_geom, __pyx_n_s_dst_geom, __pyx_n_s_i, __pyx_n_s_valb, __pyx_n_s_result); if (unlikely(!__pyx_tuple__51)) __PYX_ERR(0, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__51);
  __Pyx_GIVEREF(__pyx_tuple__51);
  __pyx_codeobj__52 = (PyObject*)__Pyx_PyCode_New(6, 0, 16, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__51, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Users_sean_code_frs_wheel_build, __pyx_n_s_transform_geom, 41, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__52)) __PYX_ERR(0, 41, __pyx_L1_error)

  /* "rasterio/_warp.pyx":171
 * 
 * 
 * def _reproject(             # <<<<<<<<<<<<<<
 *         source, destination,
 *         src_transform=None,
 */
  __pyx_tuple__54 = PyTuple_Pack(49, __pyx_n_s_source, __pyx_n_s_destination, __pyx_n_s_src_transform, __pyx_n_s_gcps, __pyx_n_s_src_crs, __pyx_n_s_src_nodata, __pyx_n_s_dst_transform, __pyx_n_s_dst_crs, __pyx_n_s_dst_nodata, __pyx_n_s_resampling, __pyx_n_s_init_dest_nodata, __pyx_n_s_num_threads, __pyx_n_s_kwargs, __pyx_n_s_retval, __pyx_n_s_rows, __pyx_n_s_cols, __pyx_n_s_src_count, __pyx_n_s_driver, __pyx_n_s_src_dataset, __pyx_n_s_dst_dataset, __pyx_n_s_GA, __pyx_n_s_gt, __pyx_n_s_srcwkt, __pyx_n_s_dstwkt, __pyx_n_s_osr, __pyx_n_s_warp_extras, __pyx_n_s_pszWarpThread, __pyx_n_s_i, __pyx_n_s_tolerance, __pyx_n_s_gcplist, __pyx_n_s_hTransformArg, __pyx_n_s_pfnTransformer, __pyx_n_s_psWOptions, __pyx_n_s_src_bidx, __pyx_n_s_dtype, __pyx_n_s_datasetname, __pyx_n_s_obj, __pyx_n_s_ident, __pyx_n_s_info, __pyx_n_s_rdr, __pyx_n_s_shape, __pyx_n_s_dst_bidx, __pyx_n_s__53, __pyx_n_s_udr, __pyx_n_s_imgProjOptions, __pyx_n_s_key, __pyx_n_s_val, __pyx_n_s_valb, __pyx_n_s_oWarper); if (unlikely(!__pyx_tuple__54)) __PYX_ERR(0, 171, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__54);
  __Pyx_GIVEREF(__pyx_tuple__54);
  __pyx_codeobj__55 = (PyObject*)__Pyx_PyCode_New(12, 0, 49, 0, CO_VARKEYWORDS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__54, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Users_sean_code_frs_wheel_build, __pyx_n_s_reproject, 171, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__55)) __PYX_ERR(0, 171, __pyx_L1_error)

  /* "rasterio/_warp.pyx":557
 * 
 * 
 * def _calculate_default_transform(src_crs, dst_crs, width, height,             # <<<<<<<<<<<<<<
 *                                  left=None, bottom=None, right=None, top=None,
 *                                  gcps=None, **kwargs):
 */
  __pyx_tuple__56 = PyTuple_Pack(28, __pyx_n_s_src_crs, __pyx_n_s_dst_crs, __pyx_n_s_width, __pyx_n_s_height, __pyx_n_s_left, __pyx_n_s_bottom, __pyx_n_s_right, __pyx_n_s_top, __pyx_n_s_gcps, __pyx_n_s_kwargs, __pyx_n_s_hTransformArg, __pyx_n_s_npixels, __pyx_n_s_nlines, __pyx_n_s_extent, __pyx_n_s_geotransform, __pyx_n_s_osr, __pyx_n_s_wkt, __pyx_n_s_temp, __pyx_n_s_transform, __pyx_n_s_img, __pyx_n_s_err, __pyx_n_s_dst_affine, __pyx_n_s_dst_width, __pyx_n_s_dst_height, __pyx_n_s_genexpr, __pyx_n_s_genexpr, __pyx_n_s_genexpr, __pyx_n_s_i); if (unlikely(!__pyx_tuple__56)) __PYX_ERR(0, 557, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__56);
  __Pyx_GIVEREF(__pyx_tuple__56);
  __pyx_codeobj__57 = (PyObject*)__Pyx_PyCode_New(9, 0, 28, 0, CO_VARKEYWORDS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__56, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Users_sean_code_frs_wheel_build, __pyx_n_s_calculate_default_transform, 557, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__57)) __PYX_ERR(0, 557, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_InitGlobals(void) {
  __pyx_umethod_PyDict_Type_items.type = (PyObject*)&PyDict_Type;
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  __pyx_float_0_125 = PyFloat_FromDouble(0.125); if (unlikely(!__pyx_float_0_125)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1 = PyInt_FromLong(1); if (unlikely(!__pyx_int_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_neg_2 = PyInt_FromLong(-2); if (unlikely(!__pyx_int_neg_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

#if PY_MAJOR_VERSION < 3
PyMODINIT_FUNC init_warp(void); /*proto*/
PyMODINIT_FUNC init_warp(void)
#else
PyMODINIT_FUNC PyInit__warp(void); /*proto*/
PyMODINIT_FUNC PyInit__warp(void)
#endif
{
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannyDeclarations
  #if CYTHON_REFNANNY
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
  if (!__Pyx_RefNanny) {
      PyErr_Clear();
      __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
      if (!__Pyx_RefNanny)
          Py_FatalError("failed to import 'refnanny' module");
  }
  #endif
  __Pyx_RefNannySetupContext("PyMODINIT_FUNC PyInit__warp(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  #ifdef WITH_THREAD /* Python build with threading support? */
  PyEval_InitThreads();
  #endif
  #endif
  /*--- Module creation code ---*/
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("_warp", __pyx_methods, __pyx_k_Raster_and_vector_warping_and_re, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  #if CYTHON_COMPILING_IN_PYPY
  Py_INCREF(__pyx_b);
  #endif
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_rasterio___warp) {
    if (PyObject_SetAttrString(__pyx_m, "__name__", __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "rasterio._warp")) {
      if (unlikely(PyDict_SetItemString(modules, "rasterio._warp", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global init code ---*/
  /*--- Variable export code ---*/
  /*--- Function export code ---*/
  /*--- Type init code ---*/
  __pyx_ptype_8rasterio_3_io_DatasetReaderBase = __Pyx_ImportType("rasterio._io", "DatasetReaderBase", sizeof(struct __pyx_obj_8rasterio_3_io_DatasetReaderBase), 1); if (unlikely(!__pyx_ptype_8rasterio_3_io_DatasetReaderBase)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_vtabptr_8rasterio_3_io_DatasetReaderBase = (struct __pyx_vtabstruct_8rasterio_3_io_DatasetReaderBase*)__Pyx_GetVtable(__pyx_ptype_8rasterio_3_io_DatasetReaderBase->tp_dict); if (unlikely(!__pyx_vtabptr_8rasterio_3_io_DatasetReaderBase)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_vtabptr_8rasterio_5_warp_WarpedVRTReaderBase = &__pyx_vtable_8rasterio_5_warp_WarpedVRTReaderBase;
  __pyx_vtable_8rasterio_5_warp_WarpedVRTReaderBase.__pyx_base = *__pyx_vtabptr_8rasterio_3_io_DatasetReaderBase;
  __pyx_type_8rasterio_5_warp_WarpedVRTReaderBase.tp_base = __pyx_ptype_8rasterio_3_io_DatasetReaderBase;
  if (PyType_Ready(&__pyx_type_8rasterio_5_warp_WarpedVRTReaderBase) < 0) __PYX_ERR(0, 630, __pyx_L1_error)
  __pyx_type_8rasterio_5_warp_WarpedVRTReaderBase.tp_print = 0;
  if (__Pyx_SetVtable(__pyx_type_8rasterio_5_warp_WarpedVRTReaderBase.tp_dict, __pyx_vtabptr_8rasterio_5_warp_WarpedVRTReaderBase) < 0) __PYX_ERR(0, 630, __pyx_L1_error)
  if (PyObject_SetAttrString(__pyx_m, "WarpedVRTReaderBase", (PyObject *)&__pyx_type_8rasterio_5_warp_WarpedVRTReaderBase) < 0) __PYX_ERR(0, 630, __pyx_L1_error)
  __pyx_ptype_8rasterio_5_warp_WarpedVRTReaderBase = &__pyx_type_8rasterio_5_warp_WarpedVRTReaderBase;
  if (PyType_Ready(&__pyx_type_8rasterio_5_warp___pyx_scope_struct___calculate_default_transform) < 0) __PYX_ERR(0, 557, __pyx_L1_error)
  __pyx_type_8rasterio_5_warp___pyx_scope_struct___calculate_default_transform.tp_print = 0;
  __pyx_ptype_8rasterio_5_warp___pyx_scope_struct___calculate_default_transform = &__pyx_type_8rasterio_5_warp___pyx_scope_struct___calculate_default_transform;
  if (PyType_Ready(&__pyx_type_8rasterio_5_warp___pyx_scope_struct_1_genexpr) < 0) __PYX_ERR(0, 576, __pyx_L1_error)
  __pyx_type_8rasterio_5_warp___pyx_scope_struct_1_genexpr.tp_print = 0;
  __pyx_ptype_8rasterio_5_warp___pyx_scope_struct_1_genexpr = &__pyx_type_8rasterio_5_warp___pyx_scope_struct_1_genexpr;
  if (PyType_Ready(&__pyx_type_8rasterio_5_warp___pyx_scope_struct_2_genexpr) < 0) __PYX_ERR(0, 579, __pyx_L1_error)
  __pyx_type_8rasterio_5_warp___pyx_scope_struct_2_genexpr.tp_print = 0;
  __pyx_ptype_8rasterio_5_warp___pyx_scope_struct_2_genexpr = &__pyx_type_8rasterio_5_warp___pyx_scope_struct_2_genexpr;
  /*--- Type import code ---*/
  __pyx_ptype_7cpython_4type_type = __Pyx_ImportType(__Pyx_BUILTIN_MODULE_NAME, "type", 
  #if CYTHON_COMPILING_IN_PYPY
  sizeof(PyTypeObject),
  #else
  sizeof(PyHeapTypeObject),
  #endif
  0); if (unlikely(!__pyx_ptype_7cpython_4type_type)) __PYX_ERR(2, 9, __pyx_L1_error)
  __pyx_ptype_5numpy_dtype = __Pyx_ImportType("numpy", "dtype", sizeof(PyArray_Descr), 0); if (unlikely(!__pyx_ptype_5numpy_dtype)) __PYX_ERR(1, 155, __pyx_L1_error)
  __pyx_ptype_5numpy_flatiter = __Pyx_ImportType("numpy", "flatiter", sizeof(PyArrayIterObject), 0); if (unlikely(!__pyx_ptype_5numpy_flatiter)) __PYX_ERR(1, 168, __pyx_L1_error)
  __pyx_ptype_5numpy_broadcast = __Pyx_ImportType("numpy", "broadcast", sizeof(PyArrayMultiIterObject), 0); if (unlikely(!__pyx_ptype_5numpy_broadcast)) __PYX_ERR(1, 172, __pyx_L1_error)
  __pyx_ptype_5numpy_ndarray = __Pyx_ImportType("numpy", "ndarray", sizeof(PyArrayObject), 0); if (unlikely(!__pyx_ptype_5numpy_ndarray)) __PYX_ERR(1, 181, __pyx_L1_error)
  __pyx_ptype_5numpy_ufunc = __Pyx_ImportType("numpy", "ufunc", sizeof(PyUFuncObject), 0); if (unlikely(!__pyx_ptype_5numpy_ufunc)) __PYX_ERR(1, 861, __pyx_L1_error)
  __pyx_ptype_8rasterio_5_base_DatasetBase = __Pyx_ImportType("rasterio._base", "DatasetBase", sizeof(struct __pyx_obj_8rasterio_5_base_DatasetBase), 1); if (unlikely(!__pyx_ptype_8rasterio_5_base_DatasetBase)) __PYX_ERR(3, 6, __pyx_L1_error)
  __pyx_vtabptr_8rasterio_5_base_DatasetBase = (struct __pyx_vtabstruct_8rasterio_5_base_DatasetBase*)__Pyx_GetVtable(__pyx_ptype_8rasterio_5_base_DatasetBase->tp_dict); if (unlikely(!__pyx_vtabptr_8rasterio_5_base_DatasetBase)) __PYX_ERR(3, 6, __pyx_L1_error)
  __pyx_ptype_8rasterio_3_io_DatasetWriterBase = __Pyx_ImportType("rasterio._io", "DatasetWriterBase", sizeof(struct __pyx_obj_8rasterio_3_io_DatasetWriterBase), 1); if (unlikely(!__pyx_ptype_8rasterio_3_io_DatasetWriterBase)) __PYX_ERR(4, 12, __pyx_L1_error)
  __pyx_vtabptr_8rasterio_3_io_DatasetWriterBase = (struct __pyx_vtabstruct_8rasterio_3_io_DatasetWriterBase*)__Pyx_GetVtable(__pyx_ptype_8rasterio_3_io_DatasetWriterBase->tp_dict); if (unlikely(!__pyx_vtabptr_8rasterio_3_io_DatasetWriterBase)) __PYX_ERR(4, 12, __pyx_L1_error)
  __pyx_ptype_8rasterio_3_io_BufferedDatasetWriterBase = __Pyx_ImportType("rasterio._io", "BufferedDatasetWriterBase", sizeof(struct __pyx_obj_8rasterio_3_io_BufferedDatasetWriterBase), 1); if (unlikely(!__pyx_ptype_8rasterio_3_io_BufferedDatasetWriterBase)) __PYX_ERR(4, 18, __pyx_L1_error)
  __pyx_vtabptr_8rasterio_3_io_BufferedDatasetWriterBase = (struct __pyx_vtabstruct_8rasterio_3_io_BufferedDatasetWriterBase*)__Pyx_GetVtable(__pyx_ptype_8rasterio_3_io_BufferedDatasetWriterBase->tp_dict); if (unlikely(!__pyx_vtabptr_8rasterio_3_io_BufferedDatasetWriterBase)) __PYX_ERR(4, 18, __pyx_L1_error)
  __pyx_ptype_8rasterio_3_io_WarpedVRTReaderBase = __Pyx_ImportType("rasterio._io", "WarpedVRTReaderBase", sizeof(struct __pyx_obj_8rasterio_3_io_WarpedVRTReaderBase), 1); if (unlikely(!__pyx_ptype_8rasterio_3_io_WarpedVRTReaderBase)) __PYX_ERR(4, 22, __pyx_L1_error)
  __pyx_vtabptr_8rasterio_3_io_WarpedVRTReaderBase = (struct __pyx_vtabstruct_8rasterio_3_io_WarpedVRTReaderBase*)__Pyx_GetVtable(__pyx_ptype_8rasterio_3_io_WarpedVRTReaderBase->tp_dict); if (unlikely(!__pyx_vtabptr_8rasterio_3_io_WarpedVRTReaderBase)) __PYX_ERR(4, 22, __pyx_L1_error)
  __pyx_ptype_8rasterio_3_io_InMemoryRaster = __Pyx_ImportType("rasterio._io", "InMemoryRaster", sizeof(struct __pyx_obj_8rasterio_3_io_InMemoryRaster), 1); if (unlikely(!__pyx_ptype_8rasterio_3_io_InMemoryRaster)) __PYX_ERR(4, 29, __pyx_L1_error)
  __pyx_vtabptr_8rasterio_3_io_InMemoryRaster = (struct __pyx_vtabstruct_8rasterio_3_io_InMemoryRaster*)__Pyx_GetVtable(__pyx_ptype_8rasterio_3_io_InMemoryRaster->tp_dict); if (unlikely(!__pyx_vtabptr_8rasterio_3_io_InMemoryRaster)) __PYX_ERR(4, 29, __pyx_L1_error)
  __pyx_ptype_8rasterio_9_features_GeomBuilder = __Pyx_ImportType("rasterio._features", "GeomBuilder", sizeof(struct __pyx_obj_8rasterio_9_features_GeomBuilder), 1); if (unlikely(!__pyx_ptype_8rasterio_9_features_GeomBuilder)) __PYX_ERR(5, 6, __pyx_L1_error)
  __pyx_vtabptr_8rasterio_9_features_GeomBuilder = (struct __pyx_vtabstruct_8rasterio_9_features_GeomBuilder*)__Pyx_GetVtable(__pyx_ptype_8rasterio_9_features_GeomBuilder->tp_dict); if (unlikely(!__pyx_vtabptr_8rasterio_9_features_GeomBuilder)) __PYX_ERR(5, 6, __pyx_L1_error)
  __pyx_ptype_8rasterio_9_features_OGRGeomBuilder = __Pyx_ImportType("rasterio._features", "OGRGeomBuilder", sizeof(struct __pyx_obj_8rasterio_9_features_OGRGeomBuilder), 1); if (unlikely(!__pyx_ptype_8rasterio_9_features_OGRGeomBuilder)) __PYX_ERR(5, 24, __pyx_L1_error)
  __pyx_vtabptr_8rasterio_9_features_OGRGeomBuilder = (struct __pyx_vtabstruct_8rasterio_9_features_OGRGeomBuilder*)__Pyx_GetVtable(__pyx_ptype_8rasterio_9_features_OGRGeomBuilder->tp_dict); if (unlikely(!__pyx_vtabptr_8rasterio_9_features_OGRGeomBuilder)) __PYX_ERR(5, 24, __pyx_L1_error)
  __pyx_ptype_8rasterio_9_features_ShapeIterator = __Pyx_ImportType("rasterio._features", "ShapeIterator", sizeof(struct __pyx_obj_8rasterio_9_features_ShapeIterator), 1); if (unlikely(!__pyx_ptype_8rasterio_9_features_ShapeIterator)) __PYX_ERR(5, 39, __pyx_L1_error)
  /*--- Variable import code ---*/
  /*--- Function import code ---*/
  __pyx_t_1 = __Pyx_ImportModule("rasterio._base"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "get_driver_name", (void (**)(void))&__pyx_f_8rasterio_5_base_get_driver_name, "char const *(GDALDriverH)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "_osr_from_crs", (void (**)(void))&__pyx_f_8rasterio_5_base__osr_from_crs, "OGRSpatialReferenceH (PyObject *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_2 = __Pyx_ImportModule("rasterio._err"); if (!__pyx_t_2) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_2, "exc_wrap_int", (void (**)(void))&__pyx_f_8rasterio_4_err_exc_wrap_int, "int (int)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_2, "exc_wrap_pointer", (void (**)(void))&__pyx_f_8rasterio_4_err_exc_wrap_pointer, "void *(void *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_3 = __Pyx_ImportModule("rasterio._io"); if (!__pyx_t_3) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_3, "in_dtype_range", (void (**)(void))&__pyx_f_8rasterio_3_io_in_dtype_range, "int (PyObject *, PyObject *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_3, "io_auto", (void (**)(void))&__pyx_f_8rasterio_3_io_io_auto, "int (PyObject *, GDALRasterBandH, int, struct __pyx_opt_args_8rasterio_3_io_io_auto *__pyx_optional_args)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "rasterio/_warp.pyx":6
 * include "gdal.pxi"
 * 
 * import logging             # <<<<<<<<<<<<<<
 * import uuid
 * 
 */
  __pyx_t_4 = __Pyx_Import(__pyx_n_s_logging, 0, -1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_logging, __pyx_t_4) < 0) __PYX_ERR(0, 6, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "rasterio/_warp.pyx":7
 * 
 * import logging
 * import uuid             # <<<<<<<<<<<<<<
 * 
 * import numpy as np
 */
  __pyx_t_4 = __Pyx_Import(__pyx_n_s_uuid, 0, -1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_uuid, __pyx_t_4) < 0) __PYX_ERR(0, 7, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "rasterio/_warp.pyx":9
 * import uuid
 * 
 * import numpy as np             # <<<<<<<<<<<<<<
 * 
 * from rasterio._err import (
 */
  __pyx_t_4 = __Pyx_Import(__pyx_n_s_numpy, 0, -1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_np, __pyx_t_4) < 0) __PYX_ERR(0, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "rasterio/_warp.pyx":12
 * 
 * from rasterio._err import (
 *     GDALError, CPLE_IllegalArgError, CPLE_NotSupportedError,             # <<<<<<<<<<<<<<
 *     CPLE_AppDefinedError, CPLE_OpenFailedError)
 * from rasterio import dtypes
 */
  __pyx_t_4 = PyList_New(5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_n_s_GDALError);
  __Pyx_GIVEREF(__pyx_n_s_GDALError);
  PyList_SET_ITEM(__pyx_t_4, 0, __pyx_n_s_GDALError);
  __Pyx_INCREF(__pyx_n_s_CPLE_IllegalArgError);
  __Pyx_GIVEREF(__pyx_n_s_CPLE_IllegalArgError);
  PyList_SET_ITEM(__pyx_t_4, 1, __pyx_n_s_CPLE_IllegalArgError);
  __Pyx_INCREF(__pyx_n_s_CPLE_NotSupportedError);
  __Pyx_GIVEREF(__pyx_n_s_CPLE_NotSupportedError);
  PyList_SET_ITEM(__pyx_t_4, 2, __pyx_n_s_CPLE_NotSupportedError);
  __Pyx_INCREF(__pyx_n_s_CPLE_AppDefinedError);
  __Pyx_GIVEREF(__pyx_n_s_CPLE_AppDefinedError);
  PyList_SET_ITEM(__pyx_t_4, 3, __pyx_n_s_CPLE_AppDefinedError);
  __Pyx_INCREF(__pyx_n_s_CPLE_OpenFailedError);
  __Pyx_GIVEREF(__pyx_n_s_CPLE_OpenFailedError);
  PyList_SET_ITEM(__pyx_t_4, 4, __pyx_n_s_CPLE_OpenFailedError);

  /* "rasterio/_warp.pyx":11
 * import numpy as np
 * 
 * from rasterio._err import (             # <<<<<<<<<<<<<<
 *     GDALError, CPLE_IllegalArgError, CPLE_NotSupportedError,
 *     CPLE_AppDefinedError, CPLE_OpenFailedError)
 */
  __pyx_t_5 = __Pyx_Import(__pyx_n_s_rasterio__err, __pyx_t_4, -1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 11, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_ImportFrom(__pyx_t_5, __pyx_n_s_GDALError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 11, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_GDALError, __pyx_t_4) < 0) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_ImportFrom(__pyx_t_5, __pyx_n_s_CPLE_IllegalArgError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 11, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_CPLE_IllegalArgError, __pyx_t_4) < 0) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_ImportFrom(__pyx_t_5, __pyx_n_s_CPLE_NotSupportedError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 11, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_CPLE_NotSupportedError, __pyx_t_4) < 0) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_ImportFrom(__pyx_t_5, __pyx_n_s_CPLE_AppDefinedError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 11, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_CPLE_AppDefinedError, __pyx_t_4) < 0) __PYX_ERR(0, 13, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_ImportFrom(__pyx_t_5, __pyx_n_s_CPLE_OpenFailedError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 11, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_CPLE_OpenFailedError, __pyx_t_4) < 0) __PYX_ERR(0, 13, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "rasterio/_warp.pyx":14
 *     GDALError, CPLE_IllegalArgError, CPLE_NotSupportedError,
 *     CPLE_AppDefinedError, CPLE_OpenFailedError)
 * from rasterio import dtypes             # <<<<<<<<<<<<<<
 * from rasterio.control import GroundControlPoint
 * from rasterio.enums import Resampling
 */
  __pyx_t_5 = PyList_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 14, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_n_s_dtypes);
  __Pyx_GIVEREF(__pyx_n_s_dtypes);
  PyList_SET_ITEM(__pyx_t_5, 0, __pyx_n_s_dtypes);
  __pyx_t_4 = __Pyx_Import(__pyx_n_s_rasterio, __pyx_t_5, -1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 14, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_ImportFrom(__pyx_t_4, __pyx_n_s_dtypes); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 14, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_dtypes, __pyx_t_5) < 0) __PYX_ERR(0, 14, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "rasterio/_warp.pyx":15
 *     CPLE_AppDefinedError, CPLE_OpenFailedError)
 * from rasterio import dtypes
 * from rasterio.control import GroundControlPoint             # <<<<<<<<<<<<<<
 * from rasterio.enums import Resampling
 * from rasterio.errors import DriverRegistrationError, CRSError, RasterioIOError
 */
  __pyx_t_4 = PyList_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_n_s_GroundControlPoint);
  __Pyx_GIVEREF(__pyx_n_s_GroundControlPoint);
  PyList_SET_ITEM(__pyx_t_4, 0, __pyx_n_s_GroundControlPoint);
  __pyx_t_5 = __Pyx_Import(__pyx_n_s_rasterio_control, __pyx_t_4, -1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_ImportFrom(__pyx_t_5, __pyx_n_s_GroundControlPoint); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_GroundControlPoint, __pyx_t_4) < 0) __PYX_ERR(0, 15, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "rasterio/_warp.pyx":16
 * from rasterio import dtypes
 * from rasterio.control import GroundControlPoint
 * from rasterio.enums import Resampling             # <<<<<<<<<<<<<<
 * from rasterio.errors import DriverRegistrationError, CRSError, RasterioIOError
 * from rasterio.transform import Affine, from_bounds, tastes_like_gdal
 */
  __pyx_t_5 = PyList_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_n_s_Resampling);
  __Pyx_GIVEREF(__pyx_n_s_Resampling);
  PyList_SET_ITEM(__pyx_t_5, 0, __pyx_n_s_Resampling);
  __pyx_t_4 = __Pyx_Import(__pyx_n_s_rasterio_enums, __pyx_t_5, -1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_ImportFrom(__pyx_t_4, __pyx_n_s_Resampling); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Resampling, __pyx_t_5) < 0) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "rasterio/_warp.pyx":17
 * from rasterio.control import GroundControlPoint
 * from rasterio.enums import Resampling
 * from rasterio.errors import DriverRegistrationError, CRSError, RasterioIOError             # <<<<<<<<<<<<<<
 * from rasterio.transform import Affine, from_bounds, tastes_like_gdal
 * from rasterio.vfs import parse_path, vsi_path
 */
  __pyx_t_4 = PyList_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_n_s_DriverRegistrationError);
  __Pyx_GIVEREF(__pyx_n_s_DriverRegistrationError);
  PyList_SET_ITEM(__pyx_t_4, 0, __pyx_n_s_DriverRegistrationError);
  __Pyx_INCREF(__pyx_n_s_CRSError);
  __Pyx_GIVEREF(__pyx_n_s_CRSError);
  PyList_SET_ITEM(__pyx_t_4, 1, __pyx_n_s_CRSError);
  __Pyx_INCREF(__pyx_n_s_RasterioIOError);
  __Pyx_GIVEREF(__pyx_n_s_RasterioIOError);
  PyList_SET_ITEM(__pyx_t_4, 2, __pyx_n_s_RasterioIOError);
  __pyx_t_5 = __Pyx_Import(__pyx_n_s_rasterio_errors, __pyx_t_4, -1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_ImportFrom(__pyx_t_5, __pyx_n_s_DriverRegistrationError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_DriverRegistrationError, __pyx_t_4) < 0) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_ImportFrom(__pyx_t_5, __pyx_n_s_CRSError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_CRSError, __pyx_t_4) < 0) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_ImportFrom(__pyx_t_5, __pyx_n_s_RasterioIOError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_RasterioIOError, __pyx_t_4) < 0) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "rasterio/_warp.pyx":18
 * from rasterio.enums import Resampling
 * from rasterio.errors import DriverRegistrationError, CRSError, RasterioIOError
 * from rasterio.transform import Affine, from_bounds, tastes_like_gdal             # <<<<<<<<<<<<<<
 * from rasterio.vfs import parse_path, vsi_path
 * 
 */
  __pyx_t_5 = PyList_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 18, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_n_s_Affine);
  __Pyx_GIVEREF(__pyx_n_s_Affine);
  PyList_SET_ITEM(__pyx_t_5, 0, __pyx_n_s_Affine);
  __Pyx_INCREF(__pyx_n_s_from_bounds);
  __Pyx_GIVEREF(__pyx_n_s_from_bounds);
  PyList_SET_ITEM(__pyx_t_5, 1, __pyx_n_s_from_bounds);
  __Pyx_INCREF(__pyx_n_s_tastes_like_gdal);
  __Pyx_GIVEREF(__pyx_n_s_tastes_like_gdal);
  PyList_SET_ITEM(__pyx_t_5, 2, __pyx_n_s_tastes_like_gdal);
  __pyx_t_4 = __Pyx_Import(__pyx_n_s_rasterio_transform, __pyx_t_5, -1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 18, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_ImportFrom(__pyx_t_4, __pyx_n_s_Affine); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 18, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Affine, __pyx_t_5) < 0) __PYX_ERR(0, 18, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_ImportFrom(__pyx_t_4, __pyx_n_s_from_bounds); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 18, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_from_bounds, __pyx_t_5) < 0) __PYX_ERR(0, 18, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_ImportFrom(__pyx_t_4, __pyx_n_s_tastes_like_gdal); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 18, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_tastes_like_gdal, __pyx_t_5) < 0) __PYX_ERR(0, 18, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "rasterio/_warp.pyx":19
 * from rasterio.errors import DriverRegistrationError, CRSError, RasterioIOError
 * from rasterio.transform import Affine, from_bounds, tastes_like_gdal
 * from rasterio.vfs import parse_path, vsi_path             # <<<<<<<<<<<<<<
 * 
 * cimport numpy as np
 */
  __pyx_t_4 = PyList_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 19, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_n_s_parse_path);
  __Pyx_GIVEREF(__pyx_n_s_parse_path);
  PyList_SET_ITEM(__pyx_t_4, 0, __pyx_n_s_parse_path);
  __Pyx_INCREF(__pyx_n_s_vsi_path);
  __Pyx_GIVEREF(__pyx_n_s_vsi_path);
  PyList_SET_ITEM(__pyx_t_4, 1, __pyx_n_s_vsi_path);
  __pyx_t_5 = __Pyx_Import(__pyx_n_s_rasterio_vfs, __pyx_t_4, -1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 19, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_ImportFrom(__pyx_t_5, __pyx_n_s_parse_path); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 19, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_parse_path, __pyx_t_4) < 0) __PYX_ERR(0, 19, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_ImportFrom(__pyx_t_5, __pyx_n_s_vsi_path); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 19, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_vsi_path, __pyx_t_4) < 0) __PYX_ERR(0, 19, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "rasterio/_warp.pyx":30
 * 
 * 
 * log = logging.getLogger(__name__)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_logging); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_getLogger); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_name_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_7 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
    }
  }
  if (!__pyx_t_7) {
    __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 30, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_5);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_6)) {
      PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_4};
      __pyx_t_5 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 30, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
      PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_4};
      __pyx_t_5 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 30, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else
    #endif
    {
      __pyx_t_8 = PyTuple_New(1+1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 30, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_7); __pyx_t_7 = NULL;
      __Pyx_GIVEREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_8, 0+1, __pyx_t_4);
      __pyx_t_4 = 0;
      __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_8, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 30, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_log, __pyx_t_5) < 0) __PYX_ERR(0, 30, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "rasterio/_warp.pyx":33
 * 
 * 
 * def recursive_round(val, precision):             # <<<<<<<<<<<<<<
 *     """Recursively round coordinates."""
 *     if isinstance(val, (int, float)):
 */
  __pyx_t_5 = PyCFunction_NewEx(&__pyx_mdef_8rasterio_5_warp_1recursive_round, NULL, __pyx_n_s_rasterio__warp); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_recursive_round, __pyx_t_5) < 0) __PYX_ERR(0, 33, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "rasterio/_warp.pyx":41
 * 
 * 
 * def _transform_geom(             # <<<<<<<<<<<<<<
 *         src_crs, dst_crs, geom, antimeridian_cutting, antimeridian_offset,
 *         precision):
 */
  __pyx_t_5 = PyCFunction_NewEx(&__pyx_mdef_8rasterio_5_warp_3_transform_geom, NULL, __pyx_n_s_rasterio__warp); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_transform_geom, __pyx_t_5) < 0) __PYX_ERR(0, 41, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "rasterio/_warp.pyx":180
 *         dst_crs=None,
 *         dst_nodata=None,
 *         resampling=Resampling.nearest,             # <<<<<<<<<<<<<<
 *         init_dest_nodata=True,
 *         num_threads=1,
 */
  __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_Resampling); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 180, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_nearest); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 180, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_k__3 = __pyx_t_6;
  __Pyx_GIVEREF(__pyx_t_6);
  __pyx_t_6 = 0;

  /* "rasterio/_warp.pyx":171
 * 
 * 
 * def _reproject(             # <<<<<<<<<<<<<<
 *         source, destination,
 *         src_transform=None,
 */
  __pyx_t_6 = PyCFunction_NewEx(&__pyx_mdef_8rasterio_5_warp_5_reproject, NULL, __pyx_n_s_rasterio__warp); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 171, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reproject, __pyx_t_6) < 0) __PYX_ERR(0, 171, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "rasterio/_warp.pyx":557
 * 
 * 
 * def _calculate_default_transform(src_crs, dst_crs, width, height,             # <<<<<<<<<<<<<<
 *                                  left=None, bottom=None, right=None, top=None,
 *                                  gcps=None, **kwargs):
 */
  __pyx_t_6 = PyCFunction_NewEx(&__pyx_mdef_8rasterio_5_warp_7_calculate_default_transform, NULL, __pyx_n_s_rasterio__warp); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 557, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_calculate_default_transform, __pyx_t_6) < 0) __PYX_ERR(0, 557, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "rasterio/_warp.pyx":632
 * cdef class WarpedVRTReaderBase(DatasetReaderBase):
 * 
 *     def __init__(self, src_dataset, dst_crs=None, resampling=Resampling.nearest,             # <<<<<<<<<<<<<<
 *                  tolerance=0.125, src_nodata=None, dst_nodata=None,
 *                  dst_width=None, dst_height=None, dst_transform=None,
 */
  __pyx_t_6 = __Pyx_GetModuleGlobalName(__pyx_n_s_Resampling); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 632, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_nearest); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 632, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_k__35 = __pyx_t_5;
  __Pyx_GIVEREF(__pyx_t_5);
  __pyx_t_5 = 0;

  /* "rasterio/_warp.pyx":1
 * # distutils: language = c++             # <<<<<<<<<<<<<<
 * """Raster and vector warping and reprojection."""
 * 
 */
  __pyx_t_5 = PyDict_New(); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_5) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.pu2Gvd/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":997
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init rasterio._warp", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_DECREF(__pyx_m); __pyx_m = 0;
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init rasterio._warp");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if PY_MAJOR_VERSION < 3
  return;
  #else
  return __pyx_m;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule((char *)modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, (char *)"RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_CheckExact(key)) || likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (PyUnicode_GET_SIZE(**name) != PyUnicode_GET_SIZE(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (PyUnicode_GET_SIZE(**argname) != PyUnicode_GET_SIZE(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = func->ob_type->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* GetModuleGlobalName */
static CYTHON_INLINE PyObject *__Pyx_GetModuleGlobalName(PyObject *name) {
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
    result = PyDict_GetItem(__pyx_d, name);
    if (likely(result)) {
        Py_INCREF(result);
    } else {
#else
    result = PyObject_GetItem(__pyx_d, name);
    if (!result) {
        PyErr_Clear();
#endif
        result = __Pyx_GetBuiltinName(name);
    }
    return result;
}

/* PyFunctionFastCall */
  #if CYTHON_FAST_PYCALL
#include "frameobject.h"
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = PyThreadState_GET();
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = f->f_localsplus;
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, int nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif  // CPython < 3.6
#endif  // CYTHON_FAST_PYCALL

/* PyCFunctionFastCall */
  #if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    return (*((__Pyx_PyCFunctionFast)meth)) (self, args, nargs, NULL);
}
#endif  // CYTHON_FAST_PYCCALL

/* SaveResetException */
  #if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
#endif

/* GetException */
  #if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb) {
#endif
    PyObject *local_type, *local_value, *local_tb;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* PyErrFetchRestore */
    #if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* SwapException */
    #if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = *type;
    tstate->exc_value = *value;
    tstate->exc_traceback = *tb;
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyErr_GetExcInfo(&tmp_type, &tmp_value, &tmp_tb);
    PyErr_SetExcInfo(*type, *value, *tb);
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#endif

/* RaiseException */
    #if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
#if PY_VERSION_HEX >= 0x03030000
    if (cause) {
#else
    if (cause && cause != Py_None) {
#endif
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = PyThreadState_GET();
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* PyIntBinop */
      #if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, CYTHON_UNUSED int inplace) {
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
            x = (long)((unsigned long)a + b);
            if (likely((x^a) >= 0 || (x^b) >= 0))
                return PyInt_FromLong(x);
            return PyLong_Type.tp_as_number->nb_add(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                default: return PyLong_Type.tp_as_number->nb_add(op1, op2);
            }
        }
                x = a + b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla + llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
            double result;
            PyFPE_START_PROTECT("add", return NULL)
            result = ((double)a) + (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceAdd : PyNumber_Add)(op1, op2);
}
#endif

/* PyObjectCallMethO */
      #if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallOneArg */
      #if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, &arg, 1);
    }
#endif
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || PyObject_TypeCheck(func, __pyx_CyFunctionType))) {
#else
    if (likely(PyCFunction_Check(func))) {
#endif
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
#if CYTHON_FAST_PYCCALL
        } else if (PyCFunction_GET_FLAGS(func) & METH_FASTCALL) {
            return __Pyx_PyCFunction_FastCall(func, &arg, 1);
#endif
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* GetItemInt */
        static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (!j) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (wraparound & unlikely(i < 0)) i += PyList_GET_SIZE(o);
    if ((!boundscheck) || likely((0 <= i) & (i < PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (wraparound & unlikely(i < 0)) i += PyTuple_GET_SIZE(o);
    if ((!boundscheck) || likely((0 <= i) & (i < PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely((n >= 0) & (n < PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely((n >= 0) & (n < PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return m->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* PyObjectCallNoArg */
        #if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, NULL, 0);
    }
#endif
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || PyObject_TypeCheck(func, __pyx_CyFunctionType))) {
#else
    if (likely(PyCFunction_Check(func))) {
#endif
        if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
            return __Pyx_PyObject_CallMethO(func, NULL);
        }
    }
    return __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL);
}
#endif

/* RaiseTooManyValuesToUnpack */
          static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* RaiseNeedMoreValuesToUnpack */
          static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* IterFinish */
          static CYTHON_INLINE int __Pyx_IterFinish(void) {
#if CYTHON_FAST_THREAD_STATE
    PyThreadState *tstate = PyThreadState_GET();
    PyObject* exc_type = tstate->curexc_type;
    if (unlikely(exc_type)) {
        if (likely(exc_type == PyExc_StopIteration) || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration)) {
            PyObject *exc_value, *exc_tb;
            exc_value = tstate->curexc_value;
            exc_tb = tstate->curexc_traceback;
            tstate->curexc_type = 0;
            tstate->curexc_value = 0;
            tstate->curexc_traceback = 0;
            Py_DECREF(exc_type);
            Py_XDECREF(exc_value);
            Py_XDECREF(exc_tb);
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#else
    if (unlikely(PyErr_Occurred())) {
        if (likely(PyErr_ExceptionMatches(PyExc_StopIteration))) {
            PyErr_Clear();
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#endif
}

/* UnpackItemEndCheck */
          static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    } else {
        return __Pyx_IterFinish();
    }
    return 0;
}

/* ExtTypeTest */
          static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(PyObject_TypeCheck(obj, type)))
        return 1;
    PyErr_Format(PyExc_TypeError, "Cannot convert %.200s to %.200s",
                 Py_TYPE(obj)->tp_name, type->tp_name);
    return 0;
}

/* UnpackUnboundCMethod */
          static int __Pyx_TryUnpackUnboundCMethod(__Pyx_CachedCFunction* target) {
    PyObject *method;
    method = __Pyx_PyObject_GetAttrStr(target->type, *target->method_name);
    if (unlikely(!method))
        return -1;
    target->method = method;
#if CYTHON_COMPILING_IN_CPYTHON
    #if PY_MAJOR_VERSION >= 3
    if (likely(PyObject_TypeCheck(method, &PyMethodDescr_Type)))
    #endif
    {
        PyMethodDescrObject *descr = (PyMethodDescrObject*) method;
        target->func = descr->d_method->ml_meth;
        target->flag = descr->d_method->ml_flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST);
    }
#endif
    return 0;
}

/* CallUnboundCMethod0 */
          static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self) {
    PyObject *args, *result = NULL;
    if (unlikely(!cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_ASSUME_SAFE_MACROS
    args = PyTuple_New(1);
    if (unlikely(!args)) goto bad;
    Py_INCREF(self);
    PyTuple_SET_ITEM(args, 0, self);
#else
    args = PyTuple_Pack(1, self);
    if (unlikely(!args)) goto bad;
#endif
    result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
    Py_DECREF(args);
bad:
    return result;
}

/* py_dict_items */
          static CYTHON_INLINE PyObject* __Pyx_PyDict_Items(PyObject* d) {
    if (PY_MAJOR_VERSION >= 3)
        return __Pyx_CallUnboundCMethod0(&__pyx_umethod_PyDict_Type_items, d);
    else
        return PyDict_Items(d);
}

/* SliceObject */
          static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(PyObject* obj,
        Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** _py_start, PyObject** _py_stop, PyObject** _py_slice,
        int has_cstart, int has_cstop, CYTHON_UNUSED int wraparound) {
#if CYTHON_USE_TYPE_SLOTS
    PyMappingMethods* mp;
#if PY_MAJOR_VERSION < 3
    PySequenceMethods* ms = Py_TYPE(obj)->tp_as_sequence;
    if (likely(ms && ms->sq_slice)) {
        if (!has_cstart) {
            if (_py_start && (*_py_start != Py_None)) {
                cstart = __Pyx_PyIndex_AsSsize_t(*_py_start);
                if ((cstart == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstart = 0;
        }
        if (!has_cstop) {
            if (_py_stop && (*_py_stop != Py_None)) {
                cstop = __Pyx_PyIndex_AsSsize_t(*_py_stop);
                if ((cstop == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstop = PY_SSIZE_T_MAX;
        }
        if (wraparound && unlikely((cstart < 0) | (cstop < 0)) && likely(ms->sq_length)) {
            Py_ssize_t l = ms->sq_length(obj);
            if (likely(l >= 0)) {
                if (cstop < 0) {
                    cstop += l;
                    if (cstop < 0) cstop = 0;
                }
                if (cstart < 0) {
                    cstart += l;
                    if (cstart < 0) cstart = 0;
                }
            } else {
                if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                    goto bad;
                PyErr_Clear();
            }
        }
        return ms->sq_slice(obj, cstart, cstop);
    }
#endif
    mp = Py_TYPE(obj)->tp_as_mapping;
    if (likely(mp && mp->mp_subscript))
#endif
    {
        PyObject* result;
        PyObject *py_slice, *py_start, *py_stop;
        if (_py_slice) {
            py_slice = *_py_slice;
        } else {
            PyObject* owned_start = NULL;
            PyObject* owned_stop = NULL;
            if (_py_start) {
                py_start = *_py_start;
            } else {
                if (has_cstart) {
                    owned_start = py_start = PyInt_FromSsize_t(cstart);
                    if (unlikely(!py_start)) goto bad;
                } else
                    py_start = Py_None;
            }
            if (_py_stop) {
                py_stop = *_py_stop;
            } else {
                if (has_cstop) {
                    owned_stop = py_stop = PyInt_FromSsize_t(cstop);
                    if (unlikely(!py_stop)) {
                        Py_XDECREF(owned_start);
                        goto bad;
                    }
                } else
                    py_stop = Py_None;
            }
            py_slice = PySlice_New(py_start, py_stop, Py_None);
            Py_XDECREF(owned_start);
            Py_XDECREF(owned_stop);
            if (unlikely(!py_slice)) goto bad;
        }
#if CYTHON_USE_TYPE_SLOTS
        result = mp->mp_subscript(obj, py_slice);
#else
        result = PyObject_GetItem(obj, py_slice);
#endif
        if (!_py_slice) {
            Py_DECREF(py_slice);
        }
        return result;
    }
    PyErr_Format(PyExc_TypeError,
        "'%.200s' object is unsliceable", Py_TYPE(obj)->tp_name);
bad:
    return NULL;
}

/* None */
          static CYTHON_INLINE void __Pyx_RaiseClosureNameError(const char *varname) {
    PyErr_Format(PyExc_NameError, "free variable '%s' referenced before assignment in enclosing scope", varname);
}

/* PyErrExceptionMatches */
          #if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    PyObject *exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
    return PyErr_GivenExceptionMatches(exc_type, err);
}
#endif

/* decode_c_string */
          static CYTHON_INLINE PyObject* __Pyx_decode_c_string(
         const char* cstring, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors)) {
    Py_ssize_t length;
    if (unlikely((start < 0) | (stop < 0))) {
        size_t slen = strlen(cstring);
        if (unlikely(slen > (size_t) PY_SSIZE_T_MAX)) {
            PyErr_SetString(PyExc_OverflowError,
                            "c-string too long to convert to Python");
            return NULL;
        }
        length = (Py_ssize_t) slen;
        if (start < 0) {
            start += length;
            if (start < 0)
                start = 0;
        }
        if (stop < 0)
            stop += length;
    }
    length = stop - start;
    if (unlikely(length <= 0))
        return PyUnicode_FromUnicode(NULL, 0);
    cstring += start;
    if (decode_func) {
        return decode_func(cstring, length, errors);
    } else {
        return PyUnicode_Decode(cstring, length, encoding, errors);
    }
}

/* RaiseNoneIterError */
          static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
}

/* CallNextTpDealloc */
          static void __Pyx_call_next_tp_dealloc(PyObject* obj, destructor current_tp_dealloc) {
    PyTypeObject* type = Py_TYPE(obj);
    while (type && type->tp_dealloc != current_tp_dealloc)
        type = type->tp_base;
    while (type && type->tp_dealloc == current_tp_dealloc)
        type = type->tp_base;
    if (type)
        type->tp_dealloc(obj);
}

/* CallNextTpTraverse */
          static int __Pyx_call_next_tp_traverse(PyObject* obj, visitproc v, void *a, traverseproc current_tp_traverse) {
    PyTypeObject* type = Py_TYPE(obj);
    while (type && type->tp_traverse != current_tp_traverse)
        type = type->tp_base;
    while (type && type->tp_traverse == current_tp_traverse)
        type = type->tp_base;
    if (type && type->tp_traverse)
        return type->tp_traverse(obj, v, a);
    return 0;
}

/* CallNextTpClear */
          static void __Pyx_call_next_tp_clear(PyObject* obj, inquiry current_tp_clear) {
    PyTypeObject* type = Py_TYPE(obj);
    while (type && type->tp_clear != current_tp_clear)
        type = type->tp_base;
    while (type && type->tp_clear == current_tp_clear)
        type = type->tp_base;
    if (type && type->tp_clear)
        type->tp_clear(obj);
}

/* GetVTable */
          static void* __Pyx_GetVtable(PyObject *dict) {
    void* ptr;
    PyObject *ob = PyObject_GetItem(dict, __pyx_n_s_pyx_vtable);
    if (!ob)
        goto bad;
#if PY_VERSION_HEX >= 0x02070000
    ptr = PyCapsule_GetPointer(ob, 0);
#else
    ptr = PyCObject_AsVoidPtr(ob);
#endif
    if (!ptr && !PyErr_Occurred())
        PyErr_SetString(PyExc_RuntimeError, "invalid vtable found for imported type");
    Py_DECREF(ob);
    return ptr;
bad:
    Py_XDECREF(ob);
    return NULL;
}

/* SetVTable */
          static int __Pyx_SetVtable(PyObject *dict, void *vtable) {
#if PY_VERSION_HEX >= 0x02070000
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
#else
    PyObject *ob = PyCObject_FromVoidPtr(vtable, 0);
#endif
    if (!ob)
        goto bad;
    if (PyDict_SetItem(dict, __pyx_n_s_pyx_vtable, ob) < 0)
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* Import */
          static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_VERSION_HEX < 0x03030000
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if (strchr(__Pyx_MODULE_NAME, '.')) {
                #if PY_VERSION_HEX < 0x03030000
                PyObject *py_level = PyInt_FromLong(1);
                if (!py_level)
                    goto bad;
                module = PyObject_CallFunctionObjArgs(py_import,
                    name, global_dict, empty_dict, list, py_level, NULL);
                Py_DECREF(py_level);
                #else
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                #endif
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_VERSION_HEX < 0x03030000
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_VERSION_HEX < 0x03030000
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* ImportFrom */
          static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* CodeObjectCache */
          static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, (size_t)new_max*sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
          #include "compile.h"
#include "frameobject.h"
#include "traceback.h"
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    #if PY_MAJOR_VERSION < 3
    py_srcfile = PyString_FromString(filename);
    #else
    py_srcfile = PyUnicode_FromString(filename);
    #endif
    if (!py_srcfile) goto bad;
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        #else
        py_funcname = PyUnicode_FromString(funcname);
        #endif
    }
    if (!py_funcname) goto bad;
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    Py_DECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    py_code = __pyx_find_code_object(c_line ? c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        PyThreadState_GET(), /*PyThreadState *tstate,*/
        py_code,             /*PyCodeObject *code,*/
        __pyx_d,      /*PyObject *globals,*/
        0                    /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

/* CIntToPy */
          static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
    const int neg_one = (int) -1, const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

/* CIntFromPyVerify */
          #define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntToPy */
          static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
    const long neg_one = (long) -1, const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* Declarations */
          #if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return ::std::complex< float >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return x + y*(__pyx_t_float_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      __pyx_t_float_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
          #if CYTHON_CCOMPLEX
#else
    static CYTHON_INLINE int __Pyx_c_eq_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sum_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_diff_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prod_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        if (b.imag == 0) {
            return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabsf(b.real) >= fabsf(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                float r = b.imag / b.real;
                float s = 1.0 / (b.real + b.imag * r);
                return __pyx_t_float_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            float r = b.real / b.imag;
            float s = 1.0 / (b.imag + b.real * r);
            return __pyx_t_float_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        if (b.imag == 0) {
            return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            float denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_float_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_neg_float(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_float(__pyx_t_float_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conj_float(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE float __Pyx_c_abs_float(__pyx_t_float_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrtf(z.real*z.real + z.imag*z.imag);
          #else
            return hypotf(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_pow_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
            __pyx_t_float_complex z;
            float r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    float denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(a, a);
                    case 3:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(z, a);
                    case 4:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if (b.imag == 0) {
                    z.real = powf(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2f(0, -1);
                }
            } else {
                r = __Pyx_c_abs_float(a);
                theta = atan2f(a.imag, a.real);
            }
            lnr = logf(r);
            z_r = expf(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cosf(z_theta);
            z.imag = z_r * sinf(z_theta);
            return z;
        }
    #endif
#endif

/* Declarations */
          #if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return ::std::complex< double >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return x + y*(__pyx_t_double_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      __pyx_t_double_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
          #if CYTHON_CCOMPLEX
#else
    static CYTHON_INLINE int __Pyx_c_eq_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabs(b.real) >= fabs(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                double r = b.imag / b.real;
                double s = 1.0 / (b.real + b.imag * r);
                return __pyx_t_double_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            double r = b.real / b.imag;
            double s = 1.0 / (b.imag + b.real * r);
            return __pyx_t_double_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            double denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_double_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg_double(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_double(__pyx_t_double_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj_double(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs_double(__pyx_t_double_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrt(z.real*z.real + z.imag*z.imag);
          #else
            return hypot(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
            __pyx_t_double_complex z;
            double r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    double denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(a, a);
                    case 3:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(z, a);
                    case 4:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if (b.imag == 0) {
                    z.real = pow(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2(0, -1);
                }
            } else {
                r = __Pyx_c_abs_double(a);
                theta = atan2(a.imag, a.real);
            }
            lnr = log(r);
            z_r = exp(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cos(z_theta);
            z.imag = z_r * sin(z_theta);
            return z;
        }
    #endif
#endif

/* CIntToPy */
          static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum__NPY_TYPES(enum NPY_TYPES value) {
    const enum NPY_TYPES neg_one = (enum NPY_TYPES) -1, const_zero = (enum NPY_TYPES) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum NPY_TYPES) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(enum NPY_TYPES) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum NPY_TYPES) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum NPY_TYPES) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum NPY_TYPES) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(enum NPY_TYPES),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
          static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
    const int neg_one = (int) -1, const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntFromPy */
          static CYTHON_INLINE GDALDataType __Pyx_PyInt_As_GDALDataType(PyObject *x) {
    const GDALDataType neg_one = (GDALDataType) -1, const_zero = (GDALDataType) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(GDALDataType) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(GDALDataType, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (GDALDataType) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (GDALDataType) 0;
                case  1: __PYX_VERIFY_RETURN_INT(GDALDataType, digit, digits[0])
                case 2:
                    if (8 * sizeof(GDALDataType) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(GDALDataType, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(GDALDataType) >= 2 * PyLong_SHIFT) {
                            return (GDALDataType) (((((GDALDataType)digits[1]) << PyLong_SHIFT) | (GDALDataType)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(GDALDataType) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(GDALDataType, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(GDALDataType) >= 3 * PyLong_SHIFT) {
                            return (GDALDataType) (((((((GDALDataType)digits[2]) << PyLong_SHIFT) | (GDALDataType)digits[1]) << PyLong_SHIFT) | (GDALDataType)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(GDALDataType) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(GDALDataType, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(GDALDataType) >= 4 * PyLong_SHIFT) {
                            return (GDALDataType) (((((((((GDALDataType)digits[3]) << PyLong_SHIFT) | (GDALDataType)digits[2]) << PyLong_SHIFT) | (GDALDataType)digits[1]) << PyLong_SHIFT) | (GDALDataType)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (GDALDataType) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(GDALDataType) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(GDALDataType, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(GDALDataType) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(GDALDataType, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (GDALDataType) 0;
                case -1: __PYX_VERIFY_RETURN_INT(GDALDataType, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(GDALDataType,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(GDALDataType) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(GDALDataType, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(GDALDataType) - 1 > 2 * PyLong_SHIFT) {
                            return (GDALDataType) (((GDALDataType)-1)*(((((GDALDataType)digits[1]) << PyLong_SHIFT) | (GDALDataType)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(GDALDataType) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(GDALDataType, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(GDALDataType) - 1 > 2 * PyLong_SHIFT) {
                            return (GDALDataType) ((((((GDALDataType)digits[1]) << PyLong_SHIFT) | (GDALDataType)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(GDALDataType) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(GDALDataType, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(GDALDataType) - 1 > 3 * PyLong_SHIFT) {
                            return (GDALDataType) (((GDALDataType)-1)*(((((((GDALDataType)digits[2]) << PyLong_SHIFT) | (GDALDataType)digits[1]) << PyLong_SHIFT) | (GDALDataType)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(GDALDataType) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(GDALDataType, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(GDALDataType) - 1 > 3 * PyLong_SHIFT) {
                            return (GDALDataType) ((((((((GDALDataType)digits[2]) << PyLong_SHIFT) | (GDALDataType)digits[1]) << PyLong_SHIFT) | (GDALDataType)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(GDALDataType) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(GDALDataType, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(GDALDataType) - 1 > 4 * PyLong_SHIFT) {
                            return (GDALDataType) (((GDALDataType)-1)*(((((((((GDALDataType)digits[3]) << PyLong_SHIFT) | (GDALDataType)digits[2]) << PyLong_SHIFT) | (GDALDataType)digits[1]) << PyLong_SHIFT) | (GDALDataType)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(GDALDataType) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(GDALDataType, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(GDALDataType) - 1 > 4 * PyLong_SHIFT) {
                            return (GDALDataType) ((((((((((GDALDataType)digits[3]) << PyLong_SHIFT) | (GDALDataType)digits[2]) << PyLong_SHIFT) | (GDALDataType)digits[1]) << PyLong_SHIFT) | (GDALDataType)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(GDALDataType) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(GDALDataType, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(GDALDataType) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(GDALDataType, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            GDALDataType val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (GDALDataType) -1;
        }
    } else {
        GDALDataType val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (GDALDataType) -1;
        val = __Pyx_PyInt_As_GDALDataType(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to GDALDataType");
    return (GDALDataType) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to GDALDataType");
    return (GDALDataType) -1;
}

/* CIntFromPy */
          static CYTHON_INLINE GDALResampleAlg __Pyx_PyInt_As_GDALResampleAlg(PyObject *x) {
    const GDALResampleAlg neg_one = (GDALResampleAlg) -1, const_zero = (GDALResampleAlg) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(GDALResampleAlg) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(GDALResampleAlg, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (GDALResampleAlg) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (GDALResampleAlg) 0;
                case  1: __PYX_VERIFY_RETURN_INT(GDALResampleAlg, digit, digits[0])
                case 2:
                    if (8 * sizeof(GDALResampleAlg) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(GDALResampleAlg, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(GDALResampleAlg) >= 2 * PyLong_SHIFT) {
                            return (GDALResampleAlg) (((((GDALResampleAlg)digits[1]) << PyLong_SHIFT) | (GDALResampleAlg)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(GDALResampleAlg) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(GDALResampleAlg, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(GDALResampleAlg) >= 3 * PyLong_SHIFT) {
                            return (GDALResampleAlg) (((((((GDALResampleAlg)digits[2]) << PyLong_SHIFT) | (GDALResampleAlg)digits[1]) << PyLong_SHIFT) | (GDALResampleAlg)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(GDALResampleAlg) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(GDALResampleAlg, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(GDALResampleAlg) >= 4 * PyLong_SHIFT) {
                            return (GDALResampleAlg) (((((((((GDALResampleAlg)digits[3]) << PyLong_SHIFT) | (GDALResampleAlg)digits[2]) << PyLong_SHIFT) | (GDALResampleAlg)digits[1]) << PyLong_SHIFT) | (GDALResampleAlg)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (GDALResampleAlg) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(GDALResampleAlg) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(GDALResampleAlg, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(GDALResampleAlg) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(GDALResampleAlg, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (GDALResampleAlg) 0;
                case -1: __PYX_VERIFY_RETURN_INT(GDALResampleAlg, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(GDALResampleAlg,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(GDALResampleAlg) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(GDALResampleAlg, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(GDALResampleAlg) - 1 > 2 * PyLong_SHIFT) {
                            return (GDALResampleAlg) (((GDALResampleAlg)-1)*(((((GDALResampleAlg)digits[1]) << PyLong_SHIFT) | (GDALResampleAlg)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(GDALResampleAlg) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(GDALResampleAlg, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(GDALResampleAlg) - 1 > 2 * PyLong_SHIFT) {
                            return (GDALResampleAlg) ((((((GDALResampleAlg)digits[1]) << PyLong_SHIFT) | (GDALResampleAlg)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(GDALResampleAlg) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(GDALResampleAlg, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(GDALResampleAlg) - 1 > 3 * PyLong_SHIFT) {
                            return (GDALResampleAlg) (((GDALResampleAlg)-1)*(((((((GDALResampleAlg)digits[2]) << PyLong_SHIFT) | (GDALResampleAlg)digits[1]) << PyLong_SHIFT) | (GDALResampleAlg)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(GDALResampleAlg) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(GDALResampleAlg, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(GDALResampleAlg) - 1 > 3 * PyLong_SHIFT) {
                            return (GDALResampleAlg) ((((((((GDALResampleAlg)digits[2]) << PyLong_SHIFT) | (GDALResampleAlg)digits[1]) << PyLong_SHIFT) | (GDALResampleAlg)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(GDALResampleAlg) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(GDALResampleAlg, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(GDALResampleAlg) - 1 > 4 * PyLong_SHIFT) {
                            return (GDALResampleAlg) (((GDALResampleAlg)-1)*(((((((((GDALResampleAlg)digits[3]) << PyLong_SHIFT) | (GDALResampleAlg)digits[2]) << PyLong_SHIFT) | (GDALResampleAlg)digits[1]) << PyLong_SHIFT) | (GDALResampleAlg)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(GDALResampleAlg) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(GDALResampleAlg, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(GDALResampleAlg) - 1 > 4 * PyLong_SHIFT) {
                            return (GDALResampleAlg) ((((((((((GDALResampleAlg)digits[3]) << PyLong_SHIFT) | (GDALResampleAlg)digits[2]) << PyLong_SHIFT) | (GDALResampleAlg)digits[1]) << PyLong_SHIFT) | (GDALResampleAlg)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(GDALResampleAlg) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(GDALResampleAlg, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(GDALResampleAlg) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(GDALResampleAlg, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            GDALResampleAlg val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (GDALResampleAlg) -1;
        }
    } else {
        GDALResampleAlg val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (GDALResampleAlg) -1;
        val = __Pyx_PyInt_As_GDALResampleAlg(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to GDALResampleAlg");
    return (GDALResampleAlg) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to GDALResampleAlg");
    return (GDALResampleAlg) -1;
}

/* CIntFromPy */
          static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
    const long neg_one = (long) -1, const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* FetchCommonType */
          static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type) {
    PyObject* fake_module;
    PyTypeObject* cached_type = NULL;
    fake_module = PyImport_AddModule((char*) "_cython_" CYTHON_ABI);
    if (!fake_module) return NULL;
    Py_INCREF(fake_module);
    cached_type = (PyTypeObject*) PyObject_GetAttrString(fake_module, type->tp_name);
    if (cached_type) {
        if (!PyType_Check((PyObject*)cached_type)) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s is not a type object",
                type->tp_name);
            goto bad;
        }
        if (cached_type->tp_basicsize != type->tp_basicsize) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s has the wrong size, try recompiling",
                type->tp_name);
            goto bad;
        }
    } else {
        if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
        PyErr_Clear();
        if (PyType_Ready(type) < 0) goto bad;
        if (PyObject_SetAttrString(fake_module, type->tp_name, (PyObject*) type) < 0)
            goto bad;
        Py_INCREF(type);
        cached_type = type;
    }
done:
    Py_DECREF(fake_module);
    return cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}

/* PyObjectCallMethod1 */
          static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg) {
    PyObject *method, *result = NULL;
    method = __Pyx_PyObject_GetAttrStr(obj, method_name);
    if (unlikely(!method)) goto done;
#if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(method))) {
        PyObject *self = PyMethod_GET_SELF(method);
        if (likely(self)) {
            PyObject *args;
            PyObject *function = PyMethod_GET_FUNCTION(method);
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(function)) {
                PyObject *args[2] = {self, arg};
                result = __Pyx_PyFunction_FastCall(function, args, 2);
                goto done;
            }
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(function)) {
                PyObject *args[2] = {self, arg};
                result = __Pyx_PyCFunction_FastCall(function, args, 2);
                goto done;
            }
            #endif
            args = PyTuple_New(2);
            if (unlikely(!args)) goto done;
            Py_INCREF(self);
            PyTuple_SET_ITEM(args, 0, self);
            Py_INCREF(arg);
            PyTuple_SET_ITEM(args, 1, arg);
            Py_INCREF(function);
            Py_DECREF(method); method = NULL;
            result = __Pyx_PyObject_Call(function, args, NULL);
            Py_DECREF(args);
            Py_DECREF(function);
            return result;
        }
    }
#endif
    result = __Pyx_PyObject_CallOneArg(method, arg);
done:
    Py_XDECREF(method);
    return result;
}

/* CoroutineBase */
          #include <structmember.h>
#include <frameobject.h>
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value);
static PyObject *__Pyx_Coroutine_Close(PyObject *self);
static PyObject *__Pyx_Coroutine_Throw(PyObject *gen, PyObject *args);
#define __Pyx_Coroutine_Undelegate(gen) Py_CLEAR((gen)->yieldfrom)
#if 1 || PY_VERSION_HEX < 0x030300B0
static int __Pyx_PyGen_FetchStopIterationValue(PyObject **pvalue) {
    PyObject *et, *ev, *tb;
    PyObject *value = NULL;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&et, &ev, &tb);
    if (!et) {
        Py_XDECREF(tb);
        Py_XDECREF(ev);
        Py_INCREF(Py_None);
        *pvalue = Py_None;
        return 0;
    }
    if (likely(et == PyExc_StopIteration)) {
        if (!ev) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#if PY_VERSION_HEX >= 0x030300A0
        else if (Py_TYPE(ev) == (PyTypeObject*)PyExc_StopIteration) {
            value = ((PyStopIterationObject *)ev)->value;
            Py_INCREF(value);
            Py_DECREF(ev);
        }
#endif
        else if (unlikely(PyTuple_Check(ev))) {
            if (PyTuple_GET_SIZE(ev) >= 1) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                value = PyTuple_GET_ITEM(ev, 0);
                Py_INCREF(value);
#else
                value = PySequence_ITEM(ev, 0);
#endif
            } else {
                Py_INCREF(Py_None);
                value = Py_None;
            }
            Py_DECREF(ev);
        }
        else if (!PyObject_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration)) {
            value = ev;
        }
        if (likely(value)) {
            Py_XDECREF(tb);
            Py_DECREF(et);
            *pvalue = value;
            return 0;
        }
    } else if (!PyErr_GivenExceptionMatches(et, PyExc_StopIteration)) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    PyErr_NormalizeException(&et, &ev, &tb);
    if (unlikely(!PyObject_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration))) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    Py_XDECREF(tb);
    Py_DECREF(et);
#if PY_VERSION_HEX >= 0x030300A0
    value = ((PyStopIterationObject *)ev)->value;
    Py_INCREF(value);
    Py_DECREF(ev);
#else
    {
        PyObject* args = __Pyx_PyObject_GetAttrStr(ev, __pyx_n_s_args);
        Py_DECREF(ev);
        if (likely(args)) {
            value = PySequence_GetItem(args, 0);
            Py_DECREF(args);
        }
        if (unlikely(!value)) {
            __Pyx_ErrRestore(NULL, NULL, NULL);
            Py_INCREF(Py_None);
            value = Py_None;
        }
    }
#endif
    *pvalue = value;
    return 0;
}
#endif
static CYTHON_INLINE
void __Pyx_Coroutine_ExceptionClear(__pyx_CoroutineObject *self) {
    PyObject *exc_type = self->exc_type;
    PyObject *exc_value = self->exc_value;
    PyObject *exc_traceback = self->exc_traceback;
    self->exc_type = NULL;
    self->exc_value = NULL;
    self->exc_traceback = NULL;
    Py_XDECREF(exc_type);
    Py_XDECREF(exc_value);
    Py_XDECREF(exc_traceback);
}
static CYTHON_INLINE
int __Pyx_Coroutine_CheckRunning(__pyx_CoroutineObject *gen) {
    if (unlikely(gen->is_running)) {
        PyErr_SetString(PyExc_ValueError,
                        "generator already executing");
        return 1;
    }
    return 0;
}
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_SendEx(__pyx_CoroutineObject *self, PyObject *value) {
    PyObject *retval;
    __Pyx_PyThreadState_declare
    assert(!self->is_running);
    if (unlikely(self->resume_label == 0)) {
        if (unlikely(value && value != Py_None)) {
            PyErr_SetString(PyExc_TypeError,
                            "can't send non-None value to a "
                            "just-started generator");
            return NULL;
        }
    }
    if (unlikely(self->resume_label == -1)) {
        PyErr_SetNone(PyExc_StopIteration);
        return NULL;
    }
    __Pyx_PyThreadState_assign
    if (value) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_PYSTON
#else
        if (self->exc_traceback) {
            PyTracebackObject *tb = (PyTracebackObject *) self->exc_traceback;
            PyFrameObject *f = tb->tb_frame;
            Py_XINCREF(__pyx_tstate->frame);
            assert(f->f_back == NULL);
            f->f_back = __pyx_tstate->frame;
        }
#endif
        __Pyx_ExceptionSwap(&self->exc_type, &self->exc_value,
                            &self->exc_traceback);
    } else {
        __Pyx_Coroutine_ExceptionClear(self);
    }
    self->is_running = 1;
    retval = self->body((PyObject *) self, value);
    self->is_running = 0;
    if (retval) {
        __Pyx_ExceptionSwap(&self->exc_type, &self->exc_value,
                            &self->exc_traceback);
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_PYSTON
#else
        if (self->exc_traceback) {
            PyTracebackObject *tb = (PyTracebackObject *) self->exc_traceback;
            PyFrameObject *f = tb->tb_frame;
            Py_CLEAR(f->f_back);
        }
#endif
    } else {
        __Pyx_Coroutine_ExceptionClear(self);
    }
    return retval;
}
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_MethodReturn(PyObject *retval) {
    if (unlikely(!retval && !PyErr_Occurred())) {
        PyErr_SetNone(PyExc_StopIteration);
    }
    return retval;
}
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_FinishDelegation(__pyx_CoroutineObject *gen) {
    PyObject *ret;
    PyObject *val = NULL;
    __Pyx_Coroutine_Undelegate(gen);
    __Pyx_PyGen_FetchStopIterationValue(&val);
    ret = __Pyx_Coroutine_SendEx(gen, val);
    Py_XDECREF(val);
    return ret;
}
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value) {
    PyObject *retval;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(__Pyx_Coroutine_CheckRunning(gen)))
        return NULL;
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_CheckExact(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        {
            if (value == Py_None)
                ret = Py_TYPE(yf)->tp_iternext(yf);
            else
                ret = __Pyx_PyObject_CallMethod1(yf, __pyx_n_s_send, value);
        }
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        retval = __Pyx_Coroutine_FinishDelegation(gen);
    } else {
        retval = __Pyx_Coroutine_SendEx(gen, value);
    }
    return __Pyx_Coroutine_MethodReturn(retval);
}
static int __Pyx_Coroutine_CloseIter(__pyx_CoroutineObject *gen, PyObject *yf) {
    PyObject *retval = NULL;
    int err = 0;
    #ifdef __Pyx_Generator_USED
    if (__Pyx_Generator_CheckExact(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_Coroutine_USED
    if (__Pyx_Coroutine_CheckExact(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    #endif
    {
        PyObject *meth;
        gen->is_running = 1;
        meth = __Pyx_PyObject_GetAttrStr(yf, __pyx_n_s_close);
        if (unlikely(!meth)) {
            if (!PyErr_ExceptionMatches(PyExc_AttributeError)) {
                PyErr_WriteUnraisable(yf);
            }
            PyErr_Clear();
        } else {
            retval = PyObject_CallFunction(meth, NULL);
            Py_DECREF(meth);
            if (!retval)
                err = -1;
        }
        gen->is_running = 0;
    }
    Py_XDECREF(retval);
    return err;
}
static PyObject *__Pyx_Generator_Next(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(__Pyx_Coroutine_CheckRunning(gen)))
        return NULL;
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Generator_Next(yf);
        } else
        #endif
            ret = Py_TYPE(yf)->tp_iternext(yf);
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        return __Pyx_Coroutine_FinishDelegation(gen);
    }
    return __Pyx_Coroutine_SendEx(gen, Py_None);
}
static PyObject *__Pyx_Coroutine_Close(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *retval, *raised_exception;
    PyObject *yf = gen->yieldfrom;
    int err = 0;
    if (unlikely(__Pyx_Coroutine_CheckRunning(gen)))
        return NULL;
    if (yf) {
        Py_INCREF(yf);
        err = __Pyx_Coroutine_CloseIter(gen, yf);
        __Pyx_Coroutine_Undelegate(gen);
        Py_DECREF(yf);
    }
    if (err == 0)
        PyErr_SetNone(PyExc_GeneratorExit);
    retval = __Pyx_Coroutine_SendEx(gen, NULL);
    if (retval) {
        Py_DECREF(retval);
        PyErr_SetString(PyExc_RuntimeError,
                        "generator ignored GeneratorExit");
        return NULL;
    }
    raised_exception = PyErr_Occurred();
    if (!raised_exception
        || raised_exception == PyExc_StopIteration
        || raised_exception == PyExc_GeneratorExit
        || PyErr_GivenExceptionMatches(raised_exception, PyExc_GeneratorExit)
        || PyErr_GivenExceptionMatches(raised_exception, PyExc_StopIteration))
    {
        if (raised_exception) PyErr_Clear();
        Py_INCREF(Py_None);
        return Py_None;
    }
    return NULL;
}
static PyObject *__Pyx_Coroutine_Throw(PyObject *self, PyObject *args) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *typ;
    PyObject *tb = NULL;
    PyObject *val = NULL;
    PyObject *yf = gen->yieldfrom;
    if (!PyArg_UnpackTuple(args, (char *)"throw", 1, 3, &typ, &val, &tb))
        return NULL;
    if (unlikely(__Pyx_Coroutine_CheckRunning(gen)))
        return NULL;
    if (yf) {
        PyObject *ret;
        Py_INCREF(yf);
        if (PyErr_GivenExceptionMatches(typ, PyExc_GeneratorExit)) {
            int err = __Pyx_Coroutine_CloseIter(gen, yf);
            Py_DECREF(yf);
            __Pyx_Coroutine_Undelegate(gen);
            if (err < 0)
                return __Pyx_Coroutine_MethodReturn(__Pyx_Coroutine_SendEx(gen, NULL));
            goto throw_here;
        }
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Coroutine_Throw(yf, args);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_CheckExact(yf)) {
            ret = __Pyx_Coroutine_Throw(yf, args);
        } else
        #endif
        {
            PyObject *meth = __Pyx_PyObject_GetAttrStr(yf, __pyx_n_s_throw);
            if (unlikely(!meth)) {
                Py_DECREF(yf);
                if (!PyErr_ExceptionMatches(PyExc_AttributeError)) {
                    gen->is_running = 0;
                    return NULL;
                }
                PyErr_Clear();
                __Pyx_Coroutine_Undelegate(gen);
                gen->is_running = 0;
                goto throw_here;
            }
            ret = PyObject_CallObject(meth, args);
            Py_DECREF(meth);
        }
        gen->is_running = 0;
        Py_DECREF(yf);
        if (!ret) {
            ret = __Pyx_Coroutine_FinishDelegation(gen);
        }
        return __Pyx_Coroutine_MethodReturn(ret);
    }
throw_here:
    __Pyx_Raise(typ, val, tb, NULL);
    return __Pyx_Coroutine_MethodReturn(__Pyx_Coroutine_SendEx(gen, NULL));
}
static int __Pyx_Coroutine_traverse(PyObject *self, visitproc visit, void *arg) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    Py_VISIT(gen->closure);
    Py_VISIT(gen->classobj);
    Py_VISIT(gen->yieldfrom);
    Py_VISIT(gen->exc_type);
    Py_VISIT(gen->exc_value);
    Py_VISIT(gen->exc_traceback);
    return 0;
}
static int __Pyx_Coroutine_clear(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    Py_CLEAR(gen->closure);
    Py_CLEAR(gen->classobj);
    Py_CLEAR(gen->yieldfrom);
    Py_CLEAR(gen->exc_type);
    Py_CLEAR(gen->exc_value);
    Py_CLEAR(gen->exc_traceback);
    Py_CLEAR(gen->gi_name);
    Py_CLEAR(gen->gi_qualname);
    return 0;
}
static void __Pyx_Coroutine_dealloc(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject_GC_UnTrack(gen);
    if (gen->gi_weakreflist != NULL)
        PyObject_ClearWeakRefs(self);
    if (gen->resume_label > 0) {
        PyObject_GC_Track(self);
#if PY_VERSION_HEX >= 0x030400a1
        if (PyObject_CallFinalizerFromDealloc(self))
#else
        Py_TYPE(gen)->tp_del(self);
        if (self->ob_refcnt > 0)
#endif
        {
            return;
        }
        PyObject_GC_UnTrack(self);
    }
    __Pyx_Coroutine_clear(self);
    PyObject_GC_Del(gen);
}
static void __Pyx_Coroutine_del(PyObject *self) {
    PyObject *res;
    PyObject *error_type, *error_value, *error_traceback;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    __Pyx_PyThreadState_declare
    if (gen->resume_label <= 0)
        return ;
#if PY_VERSION_HEX < 0x030400a1
    assert(self->ob_refcnt == 0);
    self->ob_refcnt = 1;
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&error_type, &error_value, &error_traceback);
    res = __Pyx_Coroutine_Close(self);
    if (res == NULL)
        PyErr_WriteUnraisable(self);
    else
        Py_DECREF(res);
    __Pyx_ErrRestore(error_type, error_value, error_traceback);
#if PY_VERSION_HEX < 0x030400a1
    assert(self->ob_refcnt > 0);
    if (--self->ob_refcnt == 0) {
        return;
    }
    {
        Py_ssize_t refcnt = self->ob_refcnt;
        _Py_NewReference(self);
        self->ob_refcnt = refcnt;
    }
#if CYTHON_COMPILING_IN_CPYTHON
    assert(PyType_IS_GC(self->ob_type) &&
           _Py_AS_GC(self)->gc.gc_refs != _PyGC_REFS_UNTRACKED);
    _Py_DEC_REFTOTAL;
#endif
#ifdef COUNT_ALLOCS
    --Py_TYPE(self)->tp_frees;
    --Py_TYPE(self)->tp_allocs;
#endif
#endif
}
static PyObject *
__Pyx_Coroutine_get_name(__pyx_CoroutineObject *self)
{
    PyObject *name = self->gi_name;
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_name(__pyx_CoroutineObject *self, PyObject *value)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value))) {
#else
    if (unlikely(value == NULL || !PyString_Check(value))) {
#endif
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    tmp = self->gi_name;
    Py_INCREF(value);
    self->gi_name = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_Coroutine_get_qualname(__pyx_CoroutineObject *self)
{
    PyObject *name = self->gi_qualname;
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_qualname(__pyx_CoroutineObject *self, PyObject *value)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value))) {
#else
    if (unlikely(value == NULL || !PyString_Check(value))) {
#endif
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    tmp = self->gi_qualname;
    Py_INCREF(value);
    self->gi_qualname = value;
    Py_XDECREF(tmp);
    return 0;
}
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
            PyTypeObject* type, __pyx_coroutine_body_t body, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    __pyx_CoroutineObject *gen = PyObject_GC_New(__pyx_CoroutineObject, type);
    if (gen == NULL)
        return NULL;
    gen->body = body;
    gen->closure = closure;
    Py_XINCREF(closure);
    gen->is_running = 0;
    gen->resume_label = 0;
    gen->classobj = NULL;
    gen->yieldfrom = NULL;
    gen->exc_type = NULL;
    gen->exc_value = NULL;
    gen->exc_traceback = NULL;
    gen->gi_weakreflist = NULL;
    Py_XINCREF(qualname);
    gen->gi_qualname = qualname;
    Py_XINCREF(name);
    gen->gi_name = name;
    Py_XINCREF(module_name);
    gen->gi_modulename = module_name;
    PyObject_GC_Track(gen);
    return gen;
}

/* PatchModuleWithCoroutine */
              static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    int result;
    PyObject *globals, *result_obj;
    globals = PyDict_New();  if (unlikely(!globals)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_coroutine_type",
    #ifdef __Pyx_Coroutine_USED
        (PyObject*)__pyx_CoroutineType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_generator_type",
    #ifdef __Pyx_Generator_USED
        (PyObject*)__pyx_GeneratorType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "_module", module) < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "__builtins__", __pyx_b) < 0)) goto ignore;
    result_obj = PyRun_String(py_code, Py_file_input, globals, globals);
    if (unlikely(!result_obj)) goto ignore;
    Py_DECREF(result_obj);
    Py_DECREF(globals);
    return module;
ignore:
    Py_XDECREF(globals);
    PyErr_WriteUnraisable(module);
    if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, "Cython module failed to patch module with custom type", 1) < 0)) {
        Py_DECREF(module);
        module = NULL;
    }
#else
    py_code++;
#endif
    return module;
}

/* PatchGeneratorABC */
              #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
static PyObject* __Pyx_patch_abc_module(PyObject *module);
static PyObject* __Pyx_patch_abc_module(PyObject *module) {
    module = __Pyx_Coroutine_patch_module(
        module, ""
"if _cython_generator_type is not None:\n"
"    try: Generator = _module.Generator\n"
"    except AttributeError: pass\n"
"    else: Generator.register(_cython_generator_type)\n"
"if _cython_coroutine_type is not None:\n"
"    try: Coroutine = _module.Coroutine\n"
"    except AttributeError: pass\n"
"    else: Coroutine.register(_cython_coroutine_type)\n"
    );
    return module;
}
#endif
static int __Pyx_patch_abc(void) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    static int abc_patched = 0;
    if (!abc_patched) {
        PyObject *module;
        module = PyImport_ImportModule((PY_VERSION_HEX >= 0x03030000) ? "collections.abc" : "collections");
        if (!module) {
            PyErr_WriteUnraisable(NULL);
            if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning,
                    ((PY_VERSION_HEX >= 0x03030000) ?
                        "Cython module failed to register with collections.abc module" :
                        "Cython module failed to register with collections module"), 1) < 0)) {
                return -1;
            }
        } else {
            module = __Pyx_patch_abc_module(module);
            abc_patched = 1;
            if (unlikely(!module))
                return -1;
            Py_DECREF(module);
        }
        module = PyImport_ImportModule("backports_abc");
        if (module) {
            module = __Pyx_patch_abc_module(module);
            Py_XDECREF(module);
        }
        if (!module) {
            PyErr_Clear();
        }
    }
#else
    if (0) __Pyx_Coroutine_patch_module(NULL, NULL);
#endif
    return 0;
}

/* Generator */
              static PyMethodDef __pyx_Generator_methods[] = {
    {"send", (PyCFunction) __Pyx_Coroutine_Send, METH_O,
     (char*) PyDoc_STR("send(arg) -> send 'arg' into generator,\nreturn next yielded value or raise StopIteration.")},
    {"throw", (PyCFunction) __Pyx_Coroutine_Throw, METH_VARARGS,
     (char*) PyDoc_STR("throw(typ[,val[,tb]]) -> raise exception in generator,\nreturn next yielded value or raise StopIteration.")},
    {"close", (PyCFunction) __Pyx_Coroutine_Close, METH_NOARGS,
     (char*) PyDoc_STR("close() -> raise GeneratorExit inside generator.")},
    {0, 0, 0, 0}
};
static PyMemberDef __pyx_Generator_memberlist[] = {
    {(char *) "gi_running", T_BOOL, offsetof(__pyx_CoroutineObject, is_running), READONLY, NULL},
    {(char*) "gi_yieldfrom", T_OBJECT, offsetof(__pyx_CoroutineObject, yieldfrom), READONLY,
     (char*) PyDoc_STR("object being iterated by 'yield from', or None")},
    {0, 0, 0, 0, 0}
};
static PyGetSetDef __pyx_Generator_getsets[] = {
    {(char *) "__name__", (getter)__Pyx_Coroutine_get_name, (setter)__Pyx_Coroutine_set_name,
     (char*) PyDoc_STR("name of the generator"), 0},
    {(char *) "__qualname__", (getter)__Pyx_Coroutine_get_qualname, (setter)__Pyx_Coroutine_set_qualname,
     (char*) PyDoc_STR("qualified name of the generator"), 0},
    {0, 0, 0, 0, 0}
};
static PyTypeObject __pyx_GeneratorType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    "generator",
    sizeof(__pyx_CoroutineObject),
    0,
    (destructor) __Pyx_Coroutine_dealloc,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_HAVE_FINALIZE,
    0,
    (traverseproc) __Pyx_Coroutine_traverse,
    0,
    0,
    offsetof(__pyx_CoroutineObject, gi_weakreflist),
    0,
    (iternextfunc) __Pyx_Generator_Next,
    __pyx_Generator_methods,
    __pyx_Generator_memberlist,
    __pyx_Generator_getsets,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if PY_VERSION_HEX >= 0x030400a1
    0,
#else
    __Pyx_Coroutine_del,
#endif
    0,
#if PY_VERSION_HEX >= 0x030400a1
    __Pyx_Coroutine_del,
#endif
};
static int __pyx_Generator_init(void) {
    __pyx_GeneratorType_type.tp_getattro = PyObject_GenericGetAttr;
    __pyx_GeneratorType_type.tp_iter = PyObject_SelfIter;
    __pyx_GeneratorType = __Pyx_FetchCommonType(&__pyx_GeneratorType_type);
    if (unlikely(!__pyx_GeneratorType)) {
        return -1;
    }
    return 0;
}

/* CheckBinaryVersion */
              static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* ModuleImport */
              #ifndef __PYX_HAVE_RT_ImportModule
#define __PYX_HAVE_RT_ImportModule
static PyObject *__Pyx_ImportModule(const char *name) {
    PyObject *py_name = 0;
    PyObject *py_module = 0;
    py_name = __Pyx_PyIdentifier_FromString(name);
    if (!py_name)
        goto bad;
    py_module = PyImport_Import(py_name);
    Py_DECREF(py_name);
    return py_module;
bad:
    Py_XDECREF(py_name);
    return 0;
}
#endif

/* TypeImport */
              #ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(const char *module_name, const char *class_name,
    size_t size, int strict)
{
    PyObject *py_module = 0;
    PyObject *result = 0;
    PyObject *py_name = 0;
    char warning[200];
    Py_ssize_t basicsize;
#ifdef Py_LIMITED_API
    PyObject *py_basicsize;
#endif
    py_module = __Pyx_ImportModule(module_name);
    if (!py_module)
        goto bad;
    py_name = __Pyx_PyIdentifier_FromString(class_name);
    if (!py_name)
        goto bad;
    result = PyObject_GetAttr(py_module, py_name);
    Py_DECREF(py_name);
    py_name = 0;
    Py_DECREF(py_module);
    py_module = 0;
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#ifndef Py_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if (!strict && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. Expected %zd, got %zd",
            module_name, class_name, basicsize, size);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    else if ((size_t)basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s has the wrong size, try recompiling. Expected %zd, got %zd",
            module_name, class_name, basicsize, size);
        goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(py_module);
    Py_XDECREF(result);
    return NULL;
}
#endif

/* FunctionImport */
              #ifndef __PYX_HAVE_RT_ImportFunction
#define __PYX_HAVE_RT_ImportFunction
static int __Pyx_ImportFunction(PyObject *module, const char *funcname, void (**f)(void), const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    d = PyObject_GetAttrString(module, (char *)"__pyx_capi__");
    if (!d)
        goto bad;
    cobj = PyDict_GetItemString(d, funcname);
    if (!cobj) {
        PyErr_Format(PyExc_ImportError,
            "%.200s does not export expected C function %.200s",
                PyModule_GetName(module), funcname);
        goto bad;
    }
#if PY_VERSION_HEX >= 0x02070000
    if (!PyCapsule_IsValid(cobj, sig)) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, PyCapsule_GetName(cobj));
        goto bad;
    }
    tmp.p = PyCapsule_GetPointer(cobj, sig);
#else
    {const char *desc, *s1, *s2;
    desc = (const char *)PyCObject_GetDesc(cobj);
    if (!desc)
        goto bad;
    s1 = desc; s2 = sig;
    while (*s1 != '\0' && *s1 == *s2) { s1++; s2++; }
    if (*s1 != *s2) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, desc);
        goto bad;
    }
    tmp.p = PyCObject_AsVoidPtr(cobj);}
#endif
    *f = tmp.fp;
    if (!(*f))
        goto bad;
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(d);
    return -1;
}
#endif

/* InitStrings */
              static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
static CYTHON_INLINE char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if CYTHON_COMPILING_IN_CPYTHON && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
#if PY_VERSION_HEX < 0x03030000
        char* defenc_c;
        PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
        if (!defenc) return NULL;
        defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
        {
            char* end = defenc_c + PyBytes_GET_SIZE(defenc);
            char* c;
            for (c = defenc_c; c < end; c++) {
                if ((unsigned char) (*c) >= 128) {
                    PyUnicode_AsASCIIString(o);
                    return NULL;
                }
            }
        }
#endif
        *length = PyBytes_GET_SIZE(defenc);
        return defenc_c;
#else
        if (__Pyx_PyUnicode_READY(o) == -1) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
        if (PyUnicode_IS_ASCII(o)) {
            *length = PyUnicode_GET_LENGTH(o);
            return PyUnicode_AsUTF8(o);
        } else {
            PyUnicode_AsASCIIString(o);
            return NULL;
        }
#else
        return PyUnicode_AsUTF8AndSize(o, length);
#endif
#endif
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (PyInt_Check(x) || PyLong_Check(x))
#else
  if (PyLong_Check(x))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = PyNumber_Int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = PyNumber_Long(x);
  }
  #else
  if (m && m->nb_int) {
    name = "int";
    res = PyNumber_Long(x);
  }
  #endif
#else
  res = PyNumber_Int(x);
#endif
  if (res) {
#if PY_MAJOR_VERSION < 3
    if (!PyInt_Check(res) && !PyLong_Check(res)) {
#else
    if (!PyLong_Check(res)) {
#endif
      PyErr_Format(PyExc_TypeError,
                   "__%.4s__ returned non-%.4s (type %.200s)",
                   name, name, Py_TYPE(res)->tp_name);
      Py_DECREF(res);
      return NULL;
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(x);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
